/*------------------------------------------------------------\
|                                                             |
| Tool    :                     c4                            |
|                                                             |
| File    :                    main.c                         |
|                                                             |
| Date    :                   11.09.99                        |
|                                                             |
| Author  :               Alain Greiner                       |
|                                                             |
\------------------------------------------------------------*/


/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>


#include "C4_topologie.h"
#include "C4_check.h"
#include "C4_lofig.h"
#include "C4_layout.h"

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/

void 
C4Usage ()
{
  fprintf (stderr, "\t\tC4 [Options] Input_name [Output_name]\n\n");
  fprintf (stdout, "\t\tOptions : -V Verbose mode on\n");
  fprintf (stdout, "\t\tOptions : -P Output buffer X4  \n");
  fprintf (stdout, "\t\tOptions : -C Saves caracterised befig  \n");
  fprintf (stdout, "\t\tOptions : -N Saves logig    \n");
  fprintf (stdout, "\t\tOptions : -L Saves phgig    \n");
  fprintf (stdout, "\n");

  exit (1);
}

/*------------------------------------------------------------\
|                                                             |
|       La fonction C4_main est la fonction principale        |
|                                                             |
|	Elle prend en entree un pointeur de befig et          |
|	le nom souhaite pour le(s) fichier(s) de sortie       |
|	et des flags definissant le traitement demamde        |
|	- FlagNetlist : on veut la netlist                    |
|	- Flaglayout  : on veut le layout                     |
|	- FlagNetlist : on veut un VHDL caraterise            |
|	- FlagPower   : on veut un buffer x4                  |
|	Elle renvoie 0 en cas de succes.                      |
|	Elle renvoie -1 si le modele VHDL fourni en entree    |
|	ne respecte pas les contraintes C4.                   |
|	Elle renvoie -2 si l'expression booleenne est trop    |
|	complexe pour etre synthetisee.                       |
\------------------------------------------------------------*/

int	
C4_main(befig_list *BEFIG,
	char *OutputFileName,
	int FlagNetlist,
	int FlagLayout,
	int FlagCarac,
	int FlagPower,
	int FlagVerbose)
{
  phfig_list	*PHFIG;
  lofig_list	*LOFIG;
  C4_fig_route	*C4FIG;

  int		buffer;

  mbkenv ();
  autenv ();
  ablenv ();
  bddenv ();

// verification de la befig 
// si buffer = 0 la cellule ne respecte pas les conditions de synthese
// si buffer = 1 la cellule est inverseuse
// si buffer = 2 elle est non inverseuse

  buffer = C4_check(BEFIG);
  if(buffer == 0) return(-1);

  if (FlagVerbose)
    {
      fprintf (stdout, "befig %s successfully checked\n", BEFIG->NAME);
    }

// construction de la lofig et ecriture sur disque eventuelle

  LOFIG = C4_lofig(BEFIG,buffer);
  if(FlagNetlist) 
	{
	LOFIG->NAME = namealloc (OutputFileName);
	savelofig(LOFIG);
	}

  if (FlagVerbose)
    {
      fprintf (stdout, "lofig %s successfully build\n", LOFIG->NAME);
    }

// construction de la c4fig 
// si la figure n'est pas routable, le pointeur C4FIG est nul

  C4FIG = C4_topologie(LOFIG, BEFIG);
  if(C4FIG == (C4_fig_route *) 0) return (-2);

  if (FlagVerbose)
    {
      fprintf (stdout, "c4fig %s successfully build\n", C4FIG->NOM);
      C4_aff_fig(C4FIG);
    }

// construction phfig et ecriture sur disque eventuelle

  if (FlagLayout)
	{
	if ((buffer==1) && (FlagPower == 0)) PHFIG = C4_layout(0,C4FIG);
	if ((buffer==1) && (FlagPower != 0)) PHFIG = C4_layout(1,C4FIG);
	if ((buffer==2) && (FlagPower == 0)) PHFIG = C4_layout(2,C4FIG);
	if ((buffer==2) && (FlagPower != 0)) PHFIG = C4_layout(3,C4FIG);
	PHFIG->NAME = namealloc (OutputFileName);
	savephfig(PHFIG);
	}
  
// caracterisation befig et ecriture sur disque

  if (FlagCarac)
    {

/*
      C4_carac(BEFIG,LOFIG);
*/	
      BEFIG->NAME = namealloc (OutputFileName);
      vhdlsavebefig (BEFIG, 0);
      if (FlagVerbose)
	{
	  fprintf (stdout, "characterized befig %s saved\n", BEFIG->NAME);
	}
    }

  return (0);
}

/*--------------------------------------------------------------\
|                                                             	|
|                             main                            	|
|                                                             	|
|	La fonction main() analyse les parametres de la		|
|	ligne de commande, charge la befig,                     |
|       et affiche les messages d'erreurs			|
|                                                             	|
\--------------------------------------------------------------*/

int 
main (argc, argv)

     int argc;
     char *argv[];

{
  char *InputFileName;
  char *OutputFileName;
  int Number;
  int Index;
  int Status;
  char Option;
  befig_list *BEFIG;
  int FlagVerbose = 0;
  int FlagPower =0;
  int FlagCarac = 0;
  int FlagNetlist = 0;
  int FlagLayout = 0;

  InputFileName = (char *) 0;
  OutputFileName = (char *) 0;

  if (argc < 2)
    C4Usage ();

  for (Number = 1; Number < argc; Number++)
    {
      if (argv[Number][0] == '-')
	{
	  for (Index = 1; argv[Number][Index] != '\0'; Index++)
	    {
	      Option = argv[Number][Index];

	      switch (Option)
		{
		case 'C':
                    FlagCarac = 1;
                    break;
                case 'V':
                    FlagVerbose =1;
		  break;
                case 'P':
                    FlagPower =1;     
                    break;
                case 'N':
                    FlagNetlist =1;     
                    break;
                case 'L':
                    FlagLayout =1;     
                    break;
		default:
		    C4Usage ();
		}
	    }
	}
      else if (InputFileName == (char *) 0)
	InputFileName = argv[Number];
      else if (OutputFileName == (char *) 0)
	OutputFileName = argv[Number];
      else
	C4Usage ();
    }

  if (InputFileName == (char *) 0)
    C4Usage ();
  if (OutputFileName == (char *) 0)
    OutputFileName = InputFileName;

// chargement de la befig et appel de C4_main

  BEFIG = vhdlloadbefig ((befig_list *) 0, InputFileName, 3);
  if (FlagVerbose)
	fprintf (stdout, "befig %s sucessfully loaded\n", BEFIG->NAME);
  Status = C4_main(BEFIG, OutputFileName,
		FlagNetlist, FlagLayout, FlagCarac,
		FlagPower, FlagVerbose);
  if(Status == -1) { 
  	fprintf(stdout,
	"C4 error : the VHDL model of %s is not synthesisable by C4\n",
		InputFileName);
  	fprintf(stdout,
	"The model must respect the following constraints :\n");
  	fprintf(stdout,
	"- no internal signal\n");
  	fprintf(stdout,
	"- no register\n");
  	fprintf(stdout,
	"- no bus\n");
  	fprintf(stdout,
	"- only one output signal\n");
  	fprintf(stdout,
	"- only OR and AND operators\n");
  	fprintf(stdout,
	"- only one optional NOT to define the output polarity\n");
  	fprintf(stdout,
	"- no multiple instanciation of input variables\n");
	}
  if(Status == -2) { 
  	fprintf(stdout,
	"C4 error : the boolean expression of %s\n", InputFileName);
  	fprintf(stdout,
	"is to complex to be routed by C4 in the SXLIB format\n");
	}
}

