/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    File        :    placer_create_grid.c

    Date        :    27-09-1994
    Version     :    1.0

    Auteur      :    BEAURIN oliver

    Description :    Create the symbolic grid for placer

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>

#include MUT_H
#include MLO_H
#include MPH_H

#include "dpr.h"
#include "placer_structure.h"
#include "placer_utils.h"

/*---------------------------------------------------------\
                     create_alim_list_from_connectors
\---------------------------------------------------------*/
alim *create_alim_list_from_connectors( ptphcon )
phcon_list *ptphcon;
{
    alim       *ptalim;

    ptalim = NULL;
    for( ; ptphcon; ptphcon = ptphcon->NEXT )
    {
        if( ptphcon->ORIENT == WEST )
        {
            if( isvdd( ptphcon->NAME ) )
                ptalim = addalim( ptalim, TYPE_VDD, ptphcon->YCON );
            else if( isvss( ptphcon->NAME ) )
                ptalim = addalim( ptalim, TYPE_VSS, ptphcon->YCON );
        }
    }

    return ptalim;

} /* end de create_alim_list_from_connectors */

/*---------------------------------------------------------\
                     create_alim_list_two_levels
\---------------------------------------------------------*/
alim *create_alim_list_two_levels( modelname )
char *modelname;
{
    alim       *ptalim;
    phcon_list *ptphcon;
    phfig_list *ptphfig;
    phfig_list *ptmainphfig;
    phins_list *ptphins;

    ptmainphfig = getphfig( modelname, 'A' );

    ptalim = NULL;
    for( ptphins = ptmainphfig->PHINS; ptphins; ptphins=ptphins->NEXT )
    {
        ptphfig = getphfig( ptphins->FIGNAME, 'P' );

        for( ptphcon = ptphfig->PHCON; ptphcon; ptphcon = ptphcon->NEXT )
        {
            if( ptphcon->ORIENT == WEST )
            {
                if( isvdd( ptphcon->NAME ) )
                {
                    ptalim = addalim( ptalim, TYPE_VDD,
                                      ptphins->YINS - ptmainphfig->YAB1
                                    + ( ( ptphins->TRANSF == SYM_Y )
                                      ? ( ptphfig->YAB2 - ptphcon->YCON )
                                      : ( ptphcon->YCON - ptphfig->YAB1 ) ) );
                }
                else if( isvss( ptphcon->NAME ) )
                {
                    ptalim = addalim( ptalim, TYPE_VSS,
                                      ptphins->YINS - ptmainphfig->YAB1
                                    + ( ( ptphins->TRANSF == SYM_Y )
                                      ? ( ptphfig->YAB2 - ptphcon->YCON )
                                      : ( ptphcon->YCON - ptphfig->YAB1 ) ) );
                }
            }
        }
    }

    return ptalim;
} /* end de create_alim_list_two_levels */

/*---------------------------------------------------------\
                     create_slice_list
\---------------------------------------------------------*/
slice *create_slice_list( ptnextslice, modelname, insname, mainindex, placed_bloc )
slice *ptnextslice;
char *modelname;
char *insname;
long mainindex;
char placed_bloc;
{
    phfig_list *ptphfig;
    alim       *ptalim;
    alim       *ptalim1;
    alim       *ptalim2;
    alim       *ptalimmax;
    alim       *ptalimprevmax;
    alim       *ptalimsorted;
    slice      *ptslice;
    long        ymax;
    char        first_slice;

    ptphfig = getphfig( modelname, 'P' );
    if( placed_bloc )
    {
        ptalim = create_alim_list_two_levels( modelname );
    }
    else
    {
        ptalim = create_alim_list_from_connectors( ptphfig->PHCON );
    }

    if( ptalim == NULL )
    {
        fprintf( stderr, "*** dpp error *** : no alim connectors in model %s\n", ptphfig->NAME );
        exit( 1 );
    }

    ptalimsorted = NULL;
    while( ptalim )
    {
        ymax          = ptalim->Y;
        ptalimmax     = ptalim;
        ptalimprevmax = NULL;

        for( ptalim1 = ptalim->NEXT, ptalim2 = ptalim; ptalim1; ptalim2 = ptalim1, ptalim1 = ptalim1->NEXT )
        {
            if( ( ptalim1->Y == ymax ) )
            {
/*                 fprintf( stderr, "*** dpp warning *** : Two alim connectors on the same Y in model %s\n", ptphfig->NAME ); */
                ptalim2->NEXT = ptalim1->NEXT;
                mbkfree( (void *)ptalim1 );
                ptalim1 = ptalim2;
                continue;
            }
            if( ptalim1->Y > ymax )
            {
                ptalimmax     = ptalim1;
                ptalimprevmax = ptalim2;
                ymax          = ptalim1->Y;
            }
        }

        ptalimsorted = addalim( ptalimsorted, ptalimmax->TYPE, ptalimmax->Y );
        if( ptalimprevmax )
        {
            ptalimprevmax->NEXT = ptalimmax->NEXT;
            mbkfree( (void *)ptalimmax );
        }
        else
        {
            ptalim = ptalimmax->NEXT;
            mbkfree( (void *)ptalimmax );
        }
    }

    ptslice = ptnextslice;

    first_slice = TRUE;
    for( ptalim1 = ptalimsorted; ptalim1->NEXT; ptalim1 = ptalim1->NEXT, mainindex++ )
    {
        ptslice = addslice( ptslice,
                            first_slice ? modelname : NULL,
                            insname,
                            mainindex,
                            ( ptphfig->XAB2 - ptphfig->XAB1 ) / SCALE_X,
                            ( ptalim1->NEXT->Y - ptalim1->Y ) / SCALE_X,
                            ptalim1->TYPE, ptalim1->NEXT->TYPE );
       first_slice = FALSE;
    }

    freealim( ptalimsorted );

    return( ptslice );

} /* end of create_slice_list */

/*---------------------------------------------------------\
                     create_column_one_level
\---------------------------------------------------------*/
column *create_column_one_level( ptnextcolumn, modelname, insname, placed_bloc )
column   *ptnextcolumn;
char     *modelname;
char     *insname;
char      placed_bloc;
{
    column     *ptcolumn;

    ptcolumn = addcolumn( ptnextcolumn,
                          create_slice_list( (slice *)NULL, modelname, insname, 0L, placed_bloc ),
                          modelname,
                          insname,
                          placed_bloc ? COL_PLACED : COL_ROUTED );

    return( ptcolumn );

} /* end of create_column_one_level */

/*---------------------------------------------------------\
                     create_column_two_level
\---------------------------------------------------------*/
void create_column_two_level( ptcolumn, modelname )
column   *ptcolumn;
char     *modelname;
{
    instance   *ptinstance;
    instance   *ptinstance1;
    instance   *ptinstance2;
    instance   *ptinstancemax;
    instance   *ptinstanceprevmax;
    instance   *ptinstancesorted;
    slice      *ptslice;
    lofig_list *ptlofig;
    loins_list *ptloins;
    long        indexmax;

    ptlofig = getlofig( modelname, 'A' );

    ptinstance = NULL;

    if( !ptlofig->LOINS )
    {
        fprintf( stderr, "*** dpp error *** : two level hierarchy not respected in model '%s' (no instances)\n", modelname );
        exit( 1 );
    }

    for( ptloins = ptlofig->LOINS; ptloins; ptloins = ptloins->NEXT )
    {
        ptinstance = addinstance( ptinstance, ptloins->FIGNAME, ptloins->INSNAME, getindex( ptloins->INSNAME ) );
    }

    ptinstancesorted = NULL;
    while( ptinstance )
    {
        indexmax          = ptinstance->INDEX;
        ptinstancemax     = ptinstance;
        ptinstanceprevmax = NULL;

        for( ptinstance1 = ptinstance->NEXT, ptinstance2 = ptinstance;
             ptinstance1;
             ptinstance2 = ptinstance1, ptinstance1 = ptinstance1->NEXT )
        {
            if( ( ptinstance1->INDEX == indexmax ) )
            {
                fprintf( stderr,
                         "*** dpp error *** : Two instances on the same INDEX in model %s\n",
                         modelname );
                exit( 1 );
            }
            if( ptinstance1->INDEX < indexmax )
            {
                ptinstancemax     = ptinstance1;
                ptinstanceprevmax = ptinstance2;
                indexmax          = ptinstance1->INDEX;
            }
        }

        ptinstancesorted = addinstance( ptinstancesorted,
                                        ptinstancemax->MODELNAME,
                                        ptinstancemax->INSNAME,
                                        ptinstancemax->INDEX );
        if( ptinstanceprevmax )
        {
            ptinstanceprevmax->NEXT = ptinstancemax->NEXT;
            mbkfree( (void *)ptinstancemax );
        }
        else
        {
            ptinstance = ptinstancemax->NEXT;
            mbkfree( (void *)ptinstancemax );
        }
    }

    ptslice = NULL;
    for( ptinstance1 = ptinstancesorted; ptinstance1; ptinstance1 = ptinstance1->NEXT )
    {
        ptslice   = create_slice_list( ptslice,
                                       ptinstance1->MODELNAME,
                                       ptinstance1->INSNAME,
                                       ptinstance1->INDEX,
                                       FALSE );
    }

    freeinstance( ptinstancesorted );

    ptcolumn->SLICES = ptslice;
    ptcolumn->TYPE   = COL_COLUMN;

} /* end of create_column_two_level */

/*---------------------------------------------------------\
                     create_slice_list_for_glue
\---------------------------------------------------------*/
slice *create_slice_list_for_glue( modelname, insname, nbslices )
char *modelname;
char *insname;
long  nbslices;
{
    long        slicei;
    slice      *ptslice = NULL;
    phfig_list *pt_model;
    long        height;
    long        width;

    pt_model = getphfig( modelname, 'I' );

    width = ( pt_model->XAB2 - pt_model->XAB1 ) / SCALE_X;

    if( pt_model->PHINS == NULL )
    {
        fprintf( stderr, "*** dpp error *** : instance %s (%s) have no instance.\n", insname, modelname );
        exit( 1 );
    }

    pt_model = getphfig( pt_model->PHINS->FIGNAME, 'P' );
	height = ( pt_model->YAB2 - pt_model->YAB1 ) / SCALE_X;

    for( slicei = 0; slicei < nbslices; slicei++ )
    {
         ptslice = addslice( ptslice,
                             ( slicei == 0 ) ? modelname : NULL,
                             insname,
                             slicei,
                             width,
                             height,
                             slicei & 1 ? TYPE_VDD : TYPE_VSS,
                             slicei & 1 ? TYPE_VSS : TYPE_VDD );
    }

    return ptslice;

} /* end of create_slice_list_for_glue */
