/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    File        :   placer_main.c

    Date        :   30-06-94
    Version     :   1

    Auteur      :   BEAURIN oliver

    Description :   Placer

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>

#include MUT_H
#include MLO_H
#include MPH_H
#include APR_H

#include GFP_H

#include "dpr.h"
#include "placer_structure.h"
#include "placer_utils.h"
#include "placer_decoup.h"
#include "placer_create_grid.h"
#include "placer_create_con.h"
#include "placer_dprfile.h"
#include "strecher_phcut.h"
#include "strecher_bics.h"

/*---------------------------------------------------------\
                     defines
\---------------------------------------------------------*/
#define WIDTH_OF_GLUE_PTYPE 0xbaffe
#define CREATED_CON         0xcafe

/*---------------------------------------------------------\
                     locals vars
\---------------------------------------------------------*/
static chain_list *model_already_done = NULL;
static phcut_list *the_streches = NULL;

/*---------------------------------------------------------\
                     parse_strech_param
\---------------------------------------------------------*/
int parse_strech_param( arg )
char *arg;
{
    char *p;
    char *p_y;
    char *p_delta;
    long  y;
    long  delta;
    char  y_done = FALSE;
    char  save;
    phcut_list *cut;

    for( p = p_delta = p_y = arg; ; p++ )
    {
        if( ( *p < '0' ) || ( *p > '9' ) )
        {
            if( ( p == p_y ) || ( p == p_delta ) )
            {
                return FALSE;
            }
            if( *p == ':' )
            {
                if( y_done )
                {
                    return FALSE;
                }
                y_done = TRUE;
                p_delta = p;
            }
            else if(( *p == ',' )||( *p == '\0' ))
            {
                if( ! y_done )
                {
                    return FALSE;
                }
                save = *p;
                *p = *p_delta = '\0';
                y = atol( p_y );
                delta = atol( ++p_delta );

                for( cut = the_streches; cut; cut = cut->NEXT )
                {
                    if( cut->YCUT == y )
                    {
                        return FALSE;
                    }
                }
                the_streches = addophcut( the_streches, y, delta );
                if( save == '\0' )
                {
                    return TRUE;
                }
                y_done = FALSE;
                *p = save;
                p_y = p_delta = p + 1;
            }
            else
            {
                return FALSE;
            }
        }
    }
} /* end of parse_strech_param */

/*---------------------------------------------------------\
                    found_nb_slices
\---------------------------------------------------------*/
long found_nb_slices ( ptplacement )
placement *ptplacement;
{
    column  *ptcolumn;
    slice   *ptslice;
    long     nbslicesmax;
    long     nbslices;

    if( !ptplacement->COLUMNS )
    {
        return 0L;
    }

    nbslicesmax = -1L;

    for( ptcolumn = ptplacement->COLUMNS;
         ptcolumn;
         ptcolumn = ptcolumn->NEXT )
    {
        for( ptslice = ptcolumn->SLICES, nbslices = -1L;
             ptslice;
             ptslice = ptslice->NEXT)
        {
            if( ptslice->INDEX > nbslices )
                nbslices = ptslice->INDEX;
        }
 
        if(nbslices > nbslicesmax)
            nbslicesmax = nbslices;
    }

    return nbslicesmax + 1;

} /* end of found_nb_slices */

/*---------------------------------------------------------\
                    found_nb_cols
\---------------------------------------------------------*/
long found_nb_cols( ptplacement )
placement *ptplacement;
{
    column  *ptcolumn;
    long     nbcols;

    if( !ptplacement->COLUMNS )
    {
        return 0L;
    }

    for( ptcolumn = ptplacement->COLUMNS, nbcols = 0L;
         ptcolumn;
         ptcolumn = ptcolumn->NEXT, nbcols++ )
	{
		ptcolumn->INDEX = nbcols;
	}
    
    return nbcols;

} /* end of found_nb_cols */

/*---------------------------------------------------------\
                    make_slices_array
\---------------------------------------------------------*/
void make_slices_array ( ptplacement )
placement  *ptplacement;
{
    long slicei;
    long nbslices;

    nbslices = ptplacement->NB_SLICES;

    ptplacement->SLICESHEIGHTS = (long * )mbkalloc( ( nbslices + 1 ) * sizeof( long ) );
    ptplacement->YSLICE = (long * )mbkalloc( ( nbslices + 1 ) * sizeof( long ) );
    ptplacement->TYPES = (char * )mbkalloc( ( nbslices + 1 ) * sizeof( char ) );

    for( slicei = 0; slicei <= nbslices;  slicei++ )
    {
        ptplacement->SLICESHEIGHTS[ slicei ] = -1;
        ptplacement->TYPES[ slicei ] = -1;
    }

} /* end of make_slices_array */

/*---------------------------------------------------------\
                     check_column_slice_heights
\---------------------------------------------------------*/
void check_column_slice_heights( ptplacement, ptcolumn )
placement  *ptplacement;
column     *ptcolumn;
{
    slice   *ptslice;
    long     colwidth;

    colwidth = 0;

    for( ptslice = ptcolumn->SLICES;
         ptslice;
         ptslice = ptslice->NEXT)
    {
        if( ptslice->WIDTH > colwidth )
            colwidth = ptslice->WIDTH;

        if( ptplacement->SLICESHEIGHTS[ ptslice->INDEX ] == -1 )
        {
            ptplacement->SLICESHEIGHTS[ ptslice->INDEX ] = ptslice->HEIGHT;
        }
        else
        {
            if( ptplacement->SLICESHEIGHTS[ ptslice->INDEX ] != ptslice->HEIGHT )
            {
                fprintf( stderr, "*** dpp error *** : All slice don't have the same height.\n" );
                exit( 1 );
            }
        }
    }
    ptcolumn->WIDTH = colwidth; 

} /* end of check_column_slice_heights */

/*---------------------------------------------------------\
                    check_slices_heights
\---------------------------------------------------------*/
void check_slices_heights ( ptplacement )
placement  *ptplacement;
{
    column  *ptcolumn;

    if( !ptplacement->COLUMNS )
    {
        return;
    }

    for( ptcolumn = ptplacement->COLUMNS;
         ptcolumn;
         ptcolumn = ptcolumn->NEXT )
    {
        check_column_slice_heights( ptplacement, ptcolumn );
    }
    
} /* end of check_slices_heights */

/*---------------------------------------------------------\
                    check_alims
\---------------------------------------------------------*/
void check_alims( ptplacement )
placement  *ptplacement;
{
    column  *ptcolumn;
    slice   *ptslice;
    char     type;

    if( !ptplacement->COLUMNS )
    {
        return;
    }

    for( ptcolumn = ptplacement->COLUMNS;
         ptcolumn;
         ptcolumn = ptcolumn->NEXT )
    {
        for( ptslice = ptcolumn->SLICES;
             ptslice;
             ptslice = ptslice->NEXT)
        {
            if( ( ptslice->MODEL_NAME ) && ( ptslice->INDEX & 1 ) )
            {
                type = ptslice->TYPE_UP;
            }
            else
            {
                type = ptslice->TYPE_DOWN;
            }
            if( ptplacement->TYPES[ ptslice->INDEX ] == -1 )
            {
                ptplacement->TYPES[ ptslice->INDEX ] = type;
            }
            else
            {
                if( ptplacement->TYPES[ ptslice->INDEX ] != type )
                {
                    fprintf( stderr, "*** dpp error *** : Alims correspondances error in %s slice : %ld.\n", ptcolumn->INSTANCE_NAME, ptslice->INDEX );
                    exit( 1 );
                }
            }
        }
    }
    
} /* end of check_alims */

/*---------------------------------------------------------\
                    fill_yslice_array
\---------------------------------------------------------*/
void fill_yslice_array ( ptplacement )
placement  *ptplacement;
{
    long slicei;
    long yslice;
    long nbslices;

    nbslices = ptplacement->NB_SLICES;

    yslice = 0L;
    for( slicei = 0; slicei <= nbslices; slicei ++ )
    {
        ptplacement->YSLICE[ slicei ] = yslice;
        if( ptplacement->SLICESHEIGHTS[ slicei ] != -1L )
            yslice += ptplacement->SLICESHEIGHTS[ slicei ];
    }

} /* end of fill_yslice_array */

/*---------------------------------------------------------\
                    in_already_done
\---------------------------------------------------------*/
char in_already_done ( name )
char *name;
{
    chain_list *ptchain;

    for( ptchain = model_already_done; ptchain; ptchain = ptchain->NEXT )
    {
        if( (char *)ptchain->DATA == name )
            return TRUE;
    }

    return FALSE;

} /* end of in_already_done */

/*---------------------------------------------------------\
                    make_cols
\---------------------------------------------------------*/
void make_cols( ptplacement, ptcolumn )
placement  *ptplacement;
column     *ptcolumn;
{
    phfig_list *ptphcol;
    slice      *ptslice;
    long        nbslices;

    nbslices = ptplacement->NB_SLICES;

    if( !ptcolumn )
    {
        fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (make_cols 1), i can't continue...\n" );
        exit( 1 );
    }

    if( ptcolumn->TYPE != COL_COLUMN )
    {
        fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (make_cols 2), i can't continue...\n" );
        exit( 1 );
    }

    if( in_already_done( ptcolumn->MODEL_NAME ) )
    {
        if( verbose_mode == VERBOSE_ALL )
        {
            fprintf( stdout, "        model already create.\n" );
            fflush( stdout );
        }
        return;
    }

    delphfig( ptcolumn->MODEL_NAME );
    ptphcol = addphfig( ptcolumn->MODEL_NAME );

    model_already_done = addchain( model_already_done, ptcolumn->MODEL_NAME );

    defab( ptphcol,
           0L, 0L,
           ptcolumn->WIDTH * SCALE_X,
           ptplacement->YSLICE[ nbslices ] * SCALE_X );
    
    for( ptslice = ptcolumn->SLICES;
         ptslice;
         ptslice = ptslice->NEXT)
    {
        if( ptslice->MODEL_NAME )
        {
            addphins( ptphcol,
                      ptslice->MODEL_NAME,
                      ptslice->INSTANCE_NAME,
                      ( ptslice->INDEX & 1 ) ? SYM_Y : NOSYM,
                      0L,
                      ptplacement->YSLICE[ ptslice->INDEX ] * SCALE_X );
        }
    }

    savephfig( ptphcol );

} /* end of make_cols */

/*---------------------------------------------------------\
                     free_alim_con
\---------------------------------------------------------*/
void free_alim_con( phfigname )
char       *phfigname;
{
    phfig_list *ptphfig;
    phcon_list *ptphcon;

    ptphfig = getphfig( phfigname, 'A' );

    for( ptphcon = ptphfig->PHCON; ptphcon; ptphcon = ptphcon->NEXT )
    {
        if( isvdd( ptphcon->NAME ) || isvss( ptphcon->NAME ) )
        {
            delphcon( ptphfig, ptphcon );
        }
    }

} /* end of free_alim_con */

/*---------------------------------------------------------\
                     make_alim_con
\---------------------------------------------------------*/
void make_alim_con( ptplacement, phfigname )
placement  *ptplacement;
char       *phfigname;
{
    phfig_list *ptphfig;
    phcon_list *ptphcon;
    long slicei;
    long nbslices;

    ptphfig = getphfig( phfigname, 'A' );

    for( ptphcon = ptphfig->PHCON; ptphcon; ptphcon = ptphcon->NEXT )
    {
        if( isvdd( ptphcon->NAME ) || isvss( ptphcon->NAME ) )
        {
            return;
        }
    }

    nbslices = ptplacement->NB_SLICES;
    for( slicei = 0; slicei <= nbslices; slicei ++ )
    {
        addphcon( ptphfig, EAST,
                  ( slicei & 1 ) ? VDD : VSS,
                  ptphfig->XAB2,
                  ptplacement->YSLICE[ slicei ] * SCALE_X,
                  DPR_Var.LAYER_ALIM,
                  ( slicei & 1 ) ? 6L * SCALE_X : 8L * SCALE_X );
        addphcon( ptphfig, WEST,
                  ( slicei & 1 ) ? VDD : VSS,
                  ptphfig->XAB1,
                  ptplacement->YSLICE[ slicei ] * SCALE_X,
                  DPR_Var.LAYER_ALIM,
                  ( slicei & 1 ) ? 6L * SCALE_X : 8L * SCALE_X );
    }

} /* end of make_alim_con */

/*---------------------------------------------------------\
                     make_strech
\---------------------------------------------------------*/
void make_strech( ptplacement, ptmainlofig, strech_name )
placement  *ptplacement;
lofig_list *ptmainlofig;
char       *strech_name;
{
    para_strech_list *head_param = NULL;
    para_strech_list *param;
    column     *ptcolumn;
    phcut_list *cut;
    lofig_list *ptlofig;
    phfig_list *ptphfig;
    loins_list *ptloins;
    chain_list *done = NULL;
    chain_list *ptchain;
    char       *new_name;

    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> Convert slice to lambda.\n" );
    }

    for( cut = the_streches; cut; cut=cut->NEXT )
    {

        if( cut->YCUT >= ptplacement->NB_SLICES )
        {
            fprintf( stderr, "*** dpp warning *** : Can't strech slice %ld because datapath have only %ld slices.\n",
                     cut->YCUT, ptplacement->NB_SLICES );
            continue;
        }

        param = MkParaStrech(1);

        param->ymin  = ptplacement->YSLICE[cut->YCUT];
        param->ymax  = ptplacement->YSLICE[cut->YCUT + 1];
        param->delta = cut->PITCHNB;

        param->NEXT = head_param;
        head_param  = param;

        ptplacement->SLICESHEIGHTS[cut->YCUT] += ( cut->PITCHNB * 5 );
    }

    for( ptcolumn = ptplacement->COLUMNS;
         ptcolumn;
         ptcolumn = ptcolumn->NEXT )
    {
        if( ptcolumn->TYPE == COL_EMPTY )
        {
            continue;
        }

        for( ptchain = done; ptchain; ptchain = ptchain->NEXT )
        {
            if( (char *)ptchain->DATA ==  ptcolumn->MODEL_NAME )
            {
                break;
            }
        }

        if( ptchain == NULL )
        {
            if( verbose_mode == VERBOSE_ALL )
            {
                fprintf( stdout, "     -> Streching model %s of instance %s.\n",
                                 ptcolumn->MODEL_NAME, ptcolumn->INSTANCE_NAME );
            }
            ptlofig = getlofig( ptcolumn->MODEL_NAME, 'P' );
            done = addchain( done, ptcolumn->MODEL_NAME );

            for( ptchain = ptmainlofig->MODELCHAIN; ptchain; ptchain = ptchain->NEXT )
            {
                if( (char *)ptchain->DATA ==  ptcolumn->MODEL_NAME )
                {
                    break;
                }
            }
            if( ptchain == NULL )
            {
                fprintf( stderr, "*** dpr error *** : Unable to found model %s in lofig %s.\n",
                                 ptcolumn->MODEL_NAME, ptmainlofig->NAME );
                exit( 1 );
            }

            make_alim_con( ptplacement, ptcolumn->MODEL_NAME );

            if( new_name = strecher ( ptcolumn->MODEL_NAME, head_param ) )
            {
                if( ptcolumn->TYPE != COL_ROUTED )
                {
                    free_alim_con( ptcolumn->MODEL_NAME );
                    free_alim_con( new_name );

                    ptphfig = getphfig( new_name, 'A' );
                    while( delphseg( ptphfig, ptphfig->PHSEG ) );
                    savephfig( ptphfig );
                }
                ptcolumn->MODEL_NAME = new_name;
            }
            else if( verbose_mode == VERBOSE_ALL )
            {
                fprintf( stdout, "        No slice of model %s must be streched.\n",
                                 ptcolumn->MODEL_NAME );
            }


            done = addchain( done, ptcolumn->MODEL_NAME );
            ptchain->DATA = ptlofig->NAME = ptcolumn->MODEL_NAME;
            ptchain = done->NEXT;
            done->NEXT = done->NEXT->NEXT;
            ptchain->NEXT = done;
            done = ptchain;

            if( new_name )
            {
                savelofig( ptlofig );
            }
        }
        else
        {
            if( !ptchain->NEXT )
            {
                fprintf( stderr, "*** dpr bug *** : i am sorry but you find a bug (make_strech 1), i can't continue...\n" );
                exit( 1 );
            }
            if( verbose_mode == VERBOSE_ALL )
            {
                fprintf( stdout, "     -> Model %s of instance %s already done.\n",
                                 ptcolumn->MODEL_NAME, ptcolumn->INSTANCE_NAME );
            }
            ptcolumn->MODEL_NAME = (char *)(ptchain->NEXT->DATA);
        }

        for( ptloins = ptmainlofig->LOINS; ptloins; ptloins=ptloins->NEXT )
        {
            if( ptloins->INSNAME == ptcolumn->INSTANCE_NAME )
            {
                ptloins->FIGNAME = ptcolumn->MODEL_NAME;
                break;
            }
        }
        if( ptloins == NULL )
        {
            fprintf( stderr, "*** dpr error *** : Unable to found instance %s in model %s.\n",
                             ptcolumn->INSTANCE_NAME, ptmainlofig->NAME );
            exit( 1 );
        }
    }
    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> Calculating new slices heigths.\n" );
    }
    fill_yslice_array( ptplacement );

    ptplacement->MODEL_NAME = ptmainlofig->NAME = namealloc( strech_name );
    if( verbose_mode == VERBOSE_ALL )
    {
        fprintf( stdout, "     -> Saving lofig %s.\n", ptmainlofig->NAME );
    }
    savelofig( ptmainlofig );
} /* end of make_strech */

/*---------------------------------------------------------\
                    make_phfig_placement
\---------------------------------------------------------*/
void make_phfig_placement ( ptplacement, ptlofig )
placement  *ptplacement;
lofig_list *ptlofig;
{
    locon_list *ptlocon;
    locon_list *ptloconi;
    chain_list *ptchain;
    phfig_list *ptphfig;
    column     *ptcolumn;
    long        X;
    long        Index = -1;
    con_NS     *ptconNS;
    char        found_NS_con;
    con_EW     *ptconEW;
    con_EW     *conEW_min;
    con_EW     *conEW_max;
    column     *conEW_colmin;
    column     *conEW_colmax;
    long        conEW_minX;
    long        conEW_maxX;
    long        nbslices;
    long        slicei;
    long        track;
    ptype_list *ptype;
    lofig_ab   *defab_struct;
	ulocon_place_list *locon_place;

    nbslices = ptplacement->NB_SLICES;
 
    if( !ptplacement->COLUMNS )
    {
        return;
    }

    ptphfig = addphfig( ptplacement->MODEL_NAME );

    X     = 0L;

    for( ptcolumn = ptplacement->COLUMNS;
         ptcolumn;
         ptcolumn = ptcolumn->NEXT )
    {
        Index = ptcolumn->INDEX;

        if( ptcolumn->TYPE == COL_EMPTY )
        {
            continue;
        }

        ptcolumn->X = X * SCALE_X;

        addphins( ptphfig,
                  ptcolumn->MODEL_NAME,
                  ptcolumn->INSTANCE_NAME,
                  NOSYM,
                  X * SCALE_X,
                  0L );

        X = X + ptcolumn->WIDTH;
    }

    ptype = getptype( ptlofig->USER, (long)C_GfpUT_UserLoFigAB );
    if( ptype != NULL )
    {
        defab_struct = (lofig_ab *)( ptype->DATA );

        defab( ptphfig,
               0L, defab_struct->DY1 * SCALE_X,
               X * SCALE_X,
               ( ptplacement->YSLICE[ nbslices ] + defab_struct->DY2 ) * SCALE_X );
    }
    else
    {
        defab( ptphfig,
               0L, 0L,
               X * SCALE_X,
               ptplacement->YSLICE[ nbslices ] * SCALE_X );
    }


    for( ptlocon=ptlofig->LOCON; ptlocon; ptlocon=ptlocon->NEXT )
    {
        if( isvdd( ptlocon->NAME ) || isvss( ptlocon->NAME ) )
        {
            continue;
        }


        ptchain = (chain_list *)( getptype( ptlocon->SIG->USER, (long)LOFIGCHAIN)->DATA );

        found_NS_con = FALSE;
        conEW_minX   = Index + 1;
        conEW_maxX   = -1L;
        conEW_min    = NULL;
        conEW_max    = NULL;
        conEW_colmin = NULL;
        conEW_colmax = NULL;

        while( ptchain )
        {
            ptloconi = (locon_list *)( ptchain->DATA );
            ptchain  = ptchain->NEXT;

            if( ptloconi->TYPE != INTERNAL )
            {
                continue;
            }

            if( isvdd( ptloconi->NAME ) || isvss( ptloconi->NAME ) )
            {
                continue;
            }

            ptcolumn = (column *)getptype( ( (loins_list *)ptloconi->ROOT)->USER, (long)COLUMN_PTYPE )->DATA;
            if( ( ptconNS = getcon_NS( ptcolumn->CON_NS, ptloconi->NAME ) ) != NULL )
            {
                addphcon( ptphfig,
                          ptconNS->TYPE,
                          ptlocon->NAME,
                          ptconNS->POSX + ptcolumn->X,
                          ( ptconNS->TYPE == NORTH ) ? ptplacement->YSLICE[ nbslices ] * SCALE_X : 0L,
                          ALU1,
                          ptconNS->WIDTH );

                ptype = getptype( ptlocon->USER, (long)C_GfpUT_UserLoConPlace );
                if( ptype != NULL )
                {
                    fprintf( stderr, "*** dpp warning *** : unable to move control connector %s\n", ptlocon->NAME );
                }
                found_NS_con = TRUE;
            }
            else if( ( ptconEW = getcon_EW( ptcolumn->CON_EW, ptloconi->NAME ) ) != NULL )
            {
                if( ptcolumn->INDEX <= conEW_minX )
                {
                    conEW_minX   = ptcolumn->INDEX;
                    conEW_colmin = ptcolumn;
                    conEW_min    = ptconEW;
                }
                if( ptcolumn->INDEX >= conEW_maxX )
                {
                    conEW_maxX   = ptcolumn->INDEX;
                    conEW_colmax = ptcolumn;
                    conEW_max    = ptconEW;
                }
            }
            else
            {
                fprintf( stderr, "*** dpp error *** : No correspondances between logical and physical view of model %s.\n",
                         ptcolumn->MODEL_NAME );
                exit( 1 );
            }
        }

        if( !found_NS_con )
        {
            if( ( conEW_max == NULL ) || ( conEW_min == NULL ) )
            {
                fprintf( stderr,
                         "*** dpp error *** : Unable to found physical connector for logical connector : %s\n",
                         ptlocon->NAME );
                exit( 1 );
            }

            ptype = getptype( ptlocon->USER, (long)C_GfpUT_UserLoConPlace );
            if( ptype != NULL )
            {
                locon_place = (ulocon_place_list *)( ptype->DATA );

                while( locon_place )
                {
                    if( ( locon_place->SIDE == C_UtFDP_SideTypeNorth )
                      ||( locon_place->SIDE == C_UtFDP_SideTypeSouth ) )
                    {
                        addphcon( ptphfig,
                                  ( locon_place->SIDE == C_UtFDP_SideTypeSouth ) ? SOUTH : NORTH,
                                  ptlocon->NAME,
                                  0L,
                                  ( locon_place->SIDE == C_UtFDP_SideTypeNorth ) ? ptplacement->YSLICE[ nbslices ] * SCALE_X
                                                                                 : 0L,
                                  ALU1,
                                  SCALE_X );
                    }
                    else
                    {
                        if( ( locon_place->SLICE >= ptplacement->NB_SLICES )
                          ||( locon_place->SLICE < 0 ) )
                        {
                            slicei = -1L;
                        }
                        else
                        {
                            slicei = locon_place->SLICE;
                        }
                        if( slicei < 0 )
                        {
                            slicei = YtoSlice( ptplacement, conEW_colmin, conEW_min );
                        }
                        if( ( locon_place->TRACK > ( ( ptplacement->SLICESHEIGHTS[ slicei ] - 15L ) / 5L ) )
                          ||( locon_place->TRACK < 0L ) )
                        {
                            track = -1L;
                        }
                        else
                        {
                            track = locon_place->TRACK;
                        }
                        if( verbose_mode == VERBOSE_ALL )
                        {
                            if ( track == -1 )
                            {
                                fprintf( stdout,
                                         "     Moving connector %s on %s side, slice %ld, default track.\n",
                                         ptlocon->NAME,
                                         ( locon_place->SIDE == C_UtFDP_SideTypeEast ) ? "EAST" : "WEST",
                                         slicei );
                                fflush( stdout );
                            }
                            else
                            {
                                fprintf( stdout,
                                         "     Moving connector %s on %s side, slice %ld, track %ld.\n",
                                         ptlocon->NAME,
                                         ( locon_place->SIDE == C_UtFDP_SideTypeEast ) ? "EAST" : "WEST",
                                         slicei, track );
                                fflush( stdout );
                            }
                        }
                        if( track < 0 )
                        {
                            track = 0L;
                        }
                        else
                        {
                            track = 5L * track + ( ( slicei & 1 ) ? 7L : 8L );
                        }
                        addphcon( ptphfig,
                                  ( locon_place->SIDE == C_UtFDP_SideTypeEast ) ? EAST : WEST,
                                  ptlocon->NAME,
                                  ( locon_place->SIDE == C_UtFDP_SideTypeEast ) ? X * SCALE_X : 0L,
                                  ( ptplacement->YSLICE[ slicei ] + track ) * SCALE_X,
                                  ALU1,
                                  2 * SCALE_X );
                    }
                    locon_place = locon_place->NEXT;
                }
            } else if( conEW_minX < ( Index - conEW_maxX - 1 ) )
            {
                addphcon( ptphfig,
                          WEST,
                          ptlocon->NAME,
                          0L,
                          ptplacement->YSLICE[ YtoSlice( ptplacement, conEW_colmin, conEW_min ) ] * SCALE_X,
                          ALU2,
                          2 * SCALE_X );
            }
            else
            {
                addphcon( ptphfig,
                          EAST,
                          ptlocon->NAME,
                          X * SCALE_X,
                          ptplacement->YSLICE[ YtoSlice( ptplacement, conEW_colmax, conEW_max ) ] * SCALE_X,
                          ALU2,
                          2 * SCALE_X );
            }
        }
    }

    PlCon_MakePlacement( ptphfig->NAME );

    savephfig( ptphfig );

} /* end of make_phfig_placement */

/*---------------------------------------------------------\
                     create_column
\---------------------------------------------------------*/
void create_column( ptplacement )
placement  *ptplacement;
{
    column *ptcolumn;
    slice  *ptslice;
    long    nbslices;

    nbslices = ptplacement->NB_SLICES;

    for( ptcolumn = ptplacement->COLUMNS; ptcolumn; ptcolumn=ptcolumn->NEXT )
    {
        if( ptcolumn->TYPE != COL_GLUE )
            continue;


        if( islogic_column( ptcolumn->MODEL_NAME, nbslices ) )
        {
            if( verbose_mode == VERBOSE_ALL )
            {
                fprintf( stdout,
                        "     -> Indexed column     : '%s'.\n",
                         ptcolumn->INSTANCE_NAME );
                fflush( stdout );
            }
            create_column_two_level( ptcolumn, ptcolumn->MODEL_NAME );

            for( ptslice = ptcolumn->SLICES; ptslice; ptslice = ptslice->NEXT )
            {
                if( ptslice->INDEX >= nbslices )
                {
                    fprintf( stderr, "*** dpp error *** : logic column have more slice than datapath.\n" );
                    exit( 1 );
                }
            }

            check_column_slice_heights( ptplacement, ptcolumn );
            fill_yslice_array( ptplacement );
            make_cols( ptplacement, ptcolumn ); 
            create_con_list_for_column( ptcolumn );
        }
    }
    
} /* end of create_column */

/*---------------------------------------------------------\
                     create_glue_logic_bloc
\---------------------------------------------------------*/
void create_glue_logic_bloc( ptplacement, nb_ite, nb_cells )
placement  *ptplacement;
long        nb_ite;
long        nb_cells;
{
    phfig_list   *ptphfig;
    column       *ptcolumn;
    column       *ptcolumnnext;
    PlaceConList *ptplacecon;
    long          nbcolumns;
    long          nbslices;

    nbslices = ptplacement->NB_SLICES;

    for( ptcolumn = ptplacement->COLUMNS; ptcolumn; ptcolumn=ptcolumn->NEXT )
    {
        if( ptcolumn->TYPE != COL_GLUE )
            continue;


		if( verbose_mode == VERBOSE_ALL )
        {
			fprintf( stdout,
					"     -> Logic Bloc  : '%s'.\n",
					 ptcolumn->INSTANCE_NAME );
            fflush( stdout );
        }

        if( !in_already_done( ptcolumn->MODEL_NAME ) )
        {
            model_already_done = addchain( model_already_done, ptcolumn->MODEL_NAME );

            ptplacecon = create_con_for_glue_placing( ptplacement, ptcolumn );
            nbcolumns  = glue_logic_placer( ptcolumn->MODEL_NAME, nbslices, nb_cells, nb_ite, NULL, ptplacecon );

            ptphfig       = getphfig( ptcolumn->MODEL_NAME, 'A' );
            ptphfig->USER = addptype( ptphfig->USER, WIDTH_OF_GLUE_PTYPE, (void *)nbcolumns );
        }
        else
        {
            if( verbose_mode == VERBOSE_ALL )
            {
                fprintf( stdout, "        Model already placed.\n" );
                fflush( stdout );
            }

            ptphfig   = getphfig( ptcolumn->MODEL_NAME, 'A' );
            nbcolumns = (long)( getptype( ptphfig->USER, (long)WIDTH_OF_GLUE_PTYPE)->DATA );
        }

        nbcolumns--;

        for( ptcolumnnext = ptcolumn->NEXT; ptcolumnnext; ptcolumnnext=ptcolumnnext->NEXT )
        {
            ptcolumnnext->INDEX += nbcolumns;
        }

        for( ; nbcolumns != 0; nbcolumns-- )
        {
            ptcolumn->NEXT        = addcolumn( ptcolumn->NEXT, NULL, NULL, NULL, COL_EMPTY );
            ptcolumn->NEXT->INDEX = ptcolumn->INDEX + nbcolumns;
        }

        ptcolumn->SLICES = create_slice_list_for_glue( ptcolumn->MODEL_NAME,
                                                       ptcolumn->INSTANCE_NAME,
                                                       nbslices );

		if( verbose_mode == VERBOSE_ALL )
        {
			fprintf( stdout, "\n" );
            fflush( stdout );
        }

        check_column_slice_heights( ptplacement, ptcolumn );
        fill_yslice_array( ptplacement );
        create_con_list_for_placed_bloc( ptcolumn );
    }
    
} /* end of create_glue_logic_bloc */

/*---------------------------------------------------------\
                     placer
\---------------------------------------------------------*/
void placer( lofig_name, strech_name, dp_width, nb_ite, nb_cells )
char *lofig_name;
char *strech_name;
long  dp_width;
long  nb_ite;
long  nb_cells;
{
    lofig_list *ptlofig;
    loins_list *ptloins;
    placement  *ptplacement;
    long        nbslices;
    long        nbcols;
    long        operator_type;

    fprintf( stdout, " * Placement step.\n" );
    if( verbose_mode )
    {
        fprintf( stdout, "   o Getting netlist.\n" );
        fflush( stdout );
    }
    ptlofig = getlofig( lofig_name, 'A' );
    lofigchain( ptlofig );

    if( verbose_mode )
    {
        fprintf( stdout, "   o Getting dpr file.\n" );
        fflush( stdout );
    }
    PlCon_Init();
    UtFDP_LoadPlacement( lofig_name );

    ptlofig = getlofig( lofig_name, 'A' );
    if( verbose_mode )
    {
        fprintf( stdout, "   o Building symbolic grid.\n" );
        fflush( stdout );
    }

    ptplacement = addplacement( lofig_name );

    for( ptloins = ptlofig->LOINS; ptloins; ptloins = ptloins->NEXT )
    {
        operator_type = UtHir_GetModelType( ptloins->FIGNAME );

        if( operator_type == C_UtHir_TypeOperatorRouted )
        {
            if( verbose_mode == VERBOSE_ALL )
            {
                fprintf( stdout,
                         "     -> Routed bloc : '%s' (%s).\n",
                         ptloins->INSNAME, ptloins->FIGNAME );
                fflush( stdout );
            }

            ptplacement->COLUMNS = create_column_one_level(
                                          ptplacement->COLUMNS,
                                          ptloins->FIGNAME,
                                          ptloins->INSNAME,
                                          FALSE );

            create_con_list_for_routed_bloc( ptplacement->COLUMNS );
        }
        else if( operator_type == C_UtHir_TypeOperatorPlaced )
        {
            if( verbose_mode == VERBOSE_ALL )
            {
                fprintf( stdout,
                        "     -> Placed bloc : '%s' (%s).\n",
                         ptloins->INSNAME, ptloins->FIGNAME );
                fflush( stdout );
            }

            ptplacement->COLUMNS = create_column_one_level(
                                          ptplacement->COLUMNS,
                                          ptloins->FIGNAME,
                                          ptloins->INSNAME,
                                          TRUE );
            create_con_list_for_placed_bloc( ptplacement->COLUMNS );
        }
        else if( operator_type == C_UtHir_TypeOperatorUser )
        {
            if( verbose_mode == VERBOSE_ALL )
            {
                fprintf( stdout,
                         "     -> Glue bloc   : '%s' (%s).\n",
                         ptloins->INSNAME, ptloins->FIGNAME );
                fflush( stdout );
            }

            ptplacement->COLUMNS = addcolumn( ptplacement->COLUMNS,
                                              NULL,
                                              ptloins->FIGNAME,
                                              ptloins->INSNAME,
                                              COL_GLUE );
        }
        else
        {
            fprintf( stderr,
                     "*** dpp error *** : two level hierarchy not respected in %s for instance %s(%s)\n",
                     ptlofig->NAME, ptloins->INSNAME, ptloins->FIGNAME );
            exit( 1 );
        }

        ptloins->USER = addptype( ptloins->USER, COLUMN_PTYPE, (void *)ptplacement->COLUMNS );
		ptplacement->COLUMNS->LOINS = ptloins;
    }

    nbslices = found_nb_slices( ptplacement );

    if( ( nbslices == 0 ) && ( dp_width == 0 ) )
    {
        fprintf( stderr,
                 "*** dpp error *** : unable to found DP width, please use the -w option.\n" );
        exit( 1 );
    }

    ptplacement->NB_SLICES = (nbslices > dp_width ) ? nbslices : dp_width;

    if( verbose_mode )
    {
        fprintf( stdout, "   o Ckecking slices heights.\n" );
        fflush( stdout );
    }

    make_slices_array( ptplacement );
    check_slices_heights( ptplacement ); /* and fill col width field */
    fill_yslice_array( ptplacement );

    nbcols = found_nb_cols( ptplacement );
    if( verbose_mode )
    {
        fprintf( stdout, "\n" );
        fprintf( stdout, "   o Nb slices : %d.\n", (int)ptplacement->NB_SLICES );
        fprintf( stdout, "   o Nb cols   : %d.\n", (int)nbcols );
        fprintf( stdout, "\n" );
        fprintf( stdout, "   o Creating column.\n" );
        fflush( stdout );
    }
    create_column( ptplacement );

    if( verbose_mode )
    {
        fprintf( stdout, "   o Creating glue logic blocs.\n" );
        fflush( stdout );
    }
    create_glue_logic_bloc( ptplacement, nb_ite, nb_cells );

    if( verbose_mode )
    {
        fprintf( stdout, "   o Checking alims correspondances.\n" );
        fflush( stdout );
    }
    check_alims( ptplacement );

    if( strech_name )
    {
        if( UtTec_V.Type == C_UtTec_TypeAsGa3 )
        {
            fprintf(stderr, "*** dpp error *** : strecher not available in AsGa technologie.\n" );
            exit( 1 );
        }
        if( verbose_mode )
        {
            fprintf( stdout, "   o Streching.\n" );
            fflush( stdout );
        }
        /*\ ATTENTION : a partir du momment ou on est passe dans la fonction de strech
                        la structure de donnees n'est plus completement coherente. en
                        effet la hauteur des slices dans le tableau SLICESHEIGHT et dans
                        le champ de la structure slice n'est plus identique. Prendre
                        le tableau SLICESHEIGHT comme valeur correcte
        \*/
        make_strech( ptplacement, ptlofig, strech_name );
    }

    if( verbose_mode )
    {
        fprintf( stdout, "   o Creating global phfig view.\n" );
        fflush( stdout );
    }
    make_phfig_placement( ptplacement, ptlofig );

    if( verbose_mode )
    {
        fprintf( stdout, "   o Freeing symbolic grid.\n" );
        fflush( stdout );
    }
    freeplacement( ptplacement );
    freechain( model_already_done );
    model_already_done = NULL;

/* penser a liberer les lofigch, ptlofigchain ET les ptypes columns */

    fprintf( stdout, " * Placement done.\n" );
    fflush( stdout );

    return;
} /* end of placer */
