/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    File        :    placer_structure.h

    Date        :   26-09-94
    Version     :   1

    Auteur      :   BEAURIN oliver

\*********************************************************************/

#ifndef PLACER_STRUCTURE
#define PLACER_STRUCTURE

/*---------------------------------------------------------\
                     defines
\---------------------------------------------------------*/
#define COLUMN_PTYPE 0x1234321

/*---------------------------------------------------------\
                     structures
\---------------------------------------------------------*/
struct struct_con_NS
    {
        struct struct_con_NS *NEXT;
        char                 *NAME;
        char                  TYPE;
        long                  POSX;
        long                  WIDTH;
    };
struct struct_con_EW
    {
        struct struct_con_EW *NEXT;
        char                 *NAME;
        long                  YPOS;
    };
struct struct_instance
    {
        struct struct_instance *NEXT;
        char                   *MODELNAME;
        char                   *INSNAME;
        long                    INDEX;
    };
struct struct_alim
    {
        struct struct_alim   *NEXT;
        char                  TYPE;
        long                  Y;
    };
struct struct_slice
    {
        struct struct_slice  *NEXT;
        char                 *INSTANCE_NAME;
        char                 *MODEL_NAME;
        long                  HEIGHT;
        long                  WIDTH;
        long                  INDEX;
        char                  TYPE_DOWN;
        char                  TYPE_UP;
    };
struct struct_column
    {
        struct struct_column *NEXT;
        struct struct_slice  *SLICES;
        struct struct_con_NS *CON_NS;
        struct struct_con_EW *CON_EW;
        char                 *INSTANCE_NAME;
        char                 *MODEL_NAME;
        long                  WIDTH;
        long                  X;
        long                  INDEX;
        char                  TYPE;
        loins_list           *LOINS;
    };
struct struct_placement
    {
        struct struct_column *COLUMNS;
        long                  NB_SLICES;
        long                 *SLICESHEIGHTS;
        long                 *YSLICE;
        char                 *TYPES;
        char                 *MODEL_NAME;
    };

typedef struct struct_con_EW    con_EW;
typedef struct struct_con_NS    con_NS;
typedef struct struct_instance  instance;
typedef struct struct_alim      alim;
typedef struct struct_slice     slice;
typedef struct struct_column    column;
typedef struct struct_placement placement;

/*---------------------------------------------------------\
                     defines
\---------------------------------------------------------*/
#define COL_ROUTED 0
#define COL_PLACED 1
#define COL_GLUE   2
#define COL_COLUMN 3
#define COL_EMPTY  4

#define TYPE_VDD   0
#define TYPE_VSS   1

/*---------------------------------------------------------\
                     prototypes
\---------------------------------------------------------*/
#ifdef __STDC__

extern void        display_placement( placement * );
extern con_NS     *add_con_NS       ( con_NS *, char *, char, long, long );
extern void        free_con_NS_list ( con_NS * );
extern con_EW     *add_con_EW       ( con_EW *, char *, long );
extern void        free_con_EW_list ( con_EW * );
extern con_NS     *getcon_NS        ( con_NS *, char * );
extern con_EW     *getcon_EW        ( con_EW *, char * );
extern instance   *addinstance      ( instance *, char *, char *, long );
extern void        freeinstance     ( instance * );
extern alim       *addalim          ( alim *, char, long );
extern void        freealim         ( alim * );
extern slice      *addslice         ( slice *, char *, char *, long, long, long, char, char );
extern void        freeslicelist    ( slice * );
extern column     *addcolumn        ( column *, slice *, char *, char *, char );
extern void        freecolumnlist   ( column * );
extern placement  *addplacement     ( char * );
extern void        freeplacement    ( placement * );

#else

extern void        display_placement( );
extern con_NS     *add_con_NS       ( );
extern void        free_con_NS_list ( );
extern con_EW     *add_con_EW       ( );
extern void        free_con_EW_list ( );
extern con_NS     *getcon_NS        ( );
extern con_EW     *getcon_EW        ( );
extern instance   *addinstance      ( );
extern void        freeinstance     ( );
extern alim       *addalim          ( );
extern void        freealim         ( );
extern slice      *addslice         ( );
extern void        freeslicelist    ( );
extern column     *addcolumn        ( );
extern void        freecolumnlist   ( );
extern placement  *addplacement     ( );
extern void        freeplacement    ( );

#endif

#endif /* PLACER_STRUCTURE */
