/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    File        :    placer_utils.c

    Date        :   27-09-94
    Version     :   1

    Auteur      :   BEAURIN oliver

    Description :   Utilities for placement step.

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>

#include MUT_H
#include MLO_H
#include MPH_H

#include "dpr.h"
#include "placer_structure.h"

/*---------------------------------------------------------\
                     freelofigchain
\---------------------------------------------------------*/
void freelofigchain( pt_lofig )
lofig_list *pt_lofig;
{
    losig_list *pt_losig;
    ptype_list *ptype;

    for (pt_losig = pt_lofig->LOSIG; pt_losig != NULL; pt_losig = pt_losig->NEXT)
    {
        ptype = getptype( pt_losig->USER, (long)LOFIGCHAIN );
        freechain( (chain_list *)ptype->DATA );

        ptype->DATA = (void *)NULL;
        pt_losig->USER = delptype( pt_losig->USER, (long)LOFIGCHAIN );
    }

} /* fin de freelofigchain */

/*---------------------------------------------------------\
                     nbcells_of_lofig
\---------------------------------------------------------*/
long nbcells_of_lofig( modelname )
char *modelname;
{
    lofig_list *ptlofig;
    loins_list *ptloins;
    long        nbcells;

    ptlofig = getlofig( modelname, 'A' );

    for( nbcells = 0, ptloins = ptlofig->LOINS;
         ptloins;
         nbcells++, ptloins = ptloins->NEXT );

    return nbcells;

} /* end of nbcells_of_lofig */

/*---------------------------------------------------------\
                    YtoSlice
\---------------------------------------------------------*/
long YtoSlice( ptplacement, ptcol, ptcon )
placement *ptplacement;
column    *ptcol;
con_EW    *ptcon;
{
    slice *ptslice;
    long   base_slice;
    long   y;
    long   nbslices;

    nbslices = ptplacement->NB_SLICES;

    for( ptslice = ptcol->SLICES; ptslice; ptslice = ptslice->NEXT )
    {
		/* PATCH JP: ptcon->INSNAME devient ptcon->NAME.
         * if( ( ptslice->INSTANCE_NAME == ptcon->NAME )
         *   &&( ptslice->MODEL_NAME != NULL ) )
		 */
		if ( ptslice->MODEL_NAME != NULL )
        {
            base_slice = ptslice->INDEX;
            y          = ptcon->YPOS + ptplacement->YSLICE[ base_slice ];

            if( y >= ptplacement->YSLICE[ nbslices ] )
            {
                fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (YtoSlice 1), i can't continue...\n" );
                exit( 1 );
            }

            while( y >= ptplacement->YSLICE[ base_slice + 1] )
            {
                base_slice ++;
            }

            return( base_slice );
        }
    }

    fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (YtoSlice 2), i can't continue...\n" );
    exit( 1 );
} /* end of YtoSlice */

/*---------------------------------------------------------\
                     getonemorelevelconname
\---------------------------------------------------------*/
char *getonemorelevelconname( conname, ptlofig, insname )
char       *conname;
lofig_list *ptlofig;
char       *insname;
{
    loins_list *ptloins;
    locon_list *ptlocon;
    chain_list *ptchain;

    for( ptloins=ptlofig->LOINS; ptloins; ptloins = ptloins->NEXT )
    {
        if( ptloins->INSNAME == insname )
        {
            break;
        }
    }

    if( ptloins == NULL )
    {
        fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (getonemorelevelconname 1), i can't continue...\n" );
        exit( 1 );
    }

    for( ptlocon=ptloins->LOCON; ptlocon; ptlocon=ptlocon->NEXT )
    {
        if( ptlocon->NAME == conname )
        {
            break;
        }
    }

    if( ptlocon == NULL )
    {
        fprintf( stderr, "*** dpp error *** : No connector %s in model %s of %s.\n",
                 conname, ptloins->FIGNAME, ptlofig->NAME );
        exit( 1 );
    }

    ptchain = (chain_list *)( getptype( ptlocon->SIG->USER, (long)LOFIGCHAIN)->DATA );

    if( ptchain == NULL )
    {
        fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (getonemorelevelconname 3), i can't continue...\n" );
        exit( 1 );
    }

    do
    {
        ptlocon = (locon_list *)( ptchain->DATA );
        ptchain = ptchain->NEXT;
    } while( ( ptlocon->TYPE == INTERNAL ) && ( ptchain ) );

    if( ( ptchain == NULL ) && ( ptlocon->TYPE == INTERNAL ) )
    {
        return NULL;
    }

    return ptlocon->NAME;

} /* end of getonemorelevelconname */

/*---------------------------------------------------------\
                     get_con_name_from_ref_name
\---------------------------------------------------------*/
char *get_con_name_from_ref_name( refname )
char *refname;
{
    char *con_name;
    char *return_name;
    char *p;
    long  len;

    for( p = refname + strlen( refname ) - 1; (*p != '_') && (p != refname) && (isdigit( *p )); p-- );

    if( *p == '_' )
    {
        len = (long)p - (long)refname;

        con_name = mbkalloc( len + 1 );
        strncpy( con_name, refname, (int)len );
        con_name[ len] = '\0';

        return_name = namealloc( con_name );

        mbkfree( con_name );

        return return_name;
    }

    fprintf( stderr, "*** dpp error *** : reference name %s is not indexed !\n", refname );
    exit( 1 );

} /* end of get_con_name_from_ref_name */

/*---------------------------------------------------------\
                     getindex
\---------------------------------------------------------*/
long getindex( name )
char *name;
{
    char *p;

    for( p = name + strlen( name ) - 1; (*p != '_') && (p != name) && (isdigit( *p )); p-- );

    if( *p == '_' )
    {
        return( atol( p + 1 ) );
    }

    fprintf( stderr, "*** dpp error *** : instance name %s is not indexed !\n", name );
    exit( 1 );

} /* end of getindex */

/*---------------------------------------------------------\
                     isindexed
\---------------------------------------------------------*/
int isindexed( name )
char *name;
{
    char *p;

    for( p = name + strlen( name ) - 1; (*p != '_') && (p != name) && (isdigit( *p )); p-- );

    return( *p == '_' );

} /* end of isindexed */

/*---------------------------------------------------------\
                     islogic_column
\---------------------------------------------------------*/
char islogic_column( modelname, nbslices )
char *modelname;
long  nbslices;
{
    lofig_list *ptlofig;
    loins_list *ptloins;
    long        min_index = -1L;
    long        max_index = -1L;
    long       *index_array;
    long        index;

    ptlofig = getlofig( modelname, 'A' );

    for( ptloins = ptlofig->LOINS; ptloins; ptloins=ptloins->NEXT )
    {
        if( !isindexed( ptloins->INSNAME ) )
        {
            return FALSE;
        }
        index = getindex( ptloins->INSNAME );

        if( ( index < 0 ) || ( index >= nbslices ) )
        {
            return FALSE;
        }

        if( min_index < 0 )
        {
            min_index = index;
            max_index = index;
        }
        else
        {
            if( index < min_index )
                min_index = index;
            if( index > max_index )
                max_index = index;
        }
    }

    if( min_index < 0 )
    {
        return FALSE;
    }

    if( min_index == max_index )
    {
        return TRUE;
    }

    index_array = (long *)mbkalloc( ( max_index - min_index + 1 ) * sizeof( long ) );
    for( index = 0; index <= ( max_index - min_index ); index ++ )
    {
        index_array[ index ] = -1L;
    }

    for( ptloins = ptlofig->LOINS; ptloins; ptloins=ptloins->NEXT )
    {
        index = getindex( ptloins->INSNAME );
        if( index_array[ index - min_index ] != -1L )
        {
            mbkfree( index_array );
            return FALSE;
        }

        index_array[ index - min_index ] = 0L;
    }

    mbkfree( index_array );

    return TRUE;
} /* end of islogic_column */

