/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/********************************************************************
*                                                                   *
* Laboratoire MASI CAO-VLSI, UPMC, Copyright 1991 1992 1993 1994    *
*                                                                   *
* Software support Email : cao-vlsi@masi.ibp.fr                     *
*                                                                   *
* Authors : Zouina AKTOUF &                                         *
*           El Arabi RHOMARI &                                      *
*           Jose MARTINS DOS SANTOS                                 *
*                                                                   *
* Supervision & Modifications : Lotfi BEN AMMAR                     *
*                                                                   *
* Restructuration : Olivier BEAURIN                                 *
*                                                                   *
********************************************************************/
/*****************************************************************************
*                  Fichier: router_dprlin.c                                  *
*                                                                            *
*   version : 3.0  date    : 19/03/92                                        *
*                                                                            *
*   version : 3.1  date    : 27/03/94                                        *
*                                                                            *
*****************************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include MLO_H
#include MPH_H

#include GFP_H

#include "dpr.h"
#include "router_dpr.h"
#include "router_util.h"

/*---------------------------------------------------------\
                     widthalim
\---------------------------------------------------------*/
long            widthalim (ptcpos)
chain_list     *ptcpos;
{
	return (((POSSIBILITY *) ptcpos->DATA)->WIDTH);
}

/*---------------------------------------------------------\
                     inserttrack
\---------------------------------------------------------*/
void            inserttrack (ptgabarit, ptcpos)
GABARIT        *ptgabarit;
chain_list     *ptcpos;

{
	TRACK          *pttrack = NULL;
	TRACK          *pttrack_save = NULL;
	POSSIBILITY    *ptpos = NULL;

	for (; ptcpos; ptcpos = ptcpos->NEXT)
	{
		ptpos = (POSSIBILITY *) ptcpos->DATA;
		pttrack_save = NULL;
		for (pttrack = ptgabarit->TRACK; pttrack; pttrack = pttrack->NEXT)
		{
			if (pttrack->Y == ptpos->Y)
			{
				pttrack_save = pttrack;
				break;
			}
			else if (pttrack->Y > ptpos->Y)
				break;
			else
				pttrack_save = pttrack;
		}
		if (pttrack_save)
		{
			if (pttrack_save->Y != ptpos->Y)
			{
				ptgabarit->NBTRACK += 1;
				pttrack_save->NEXT = addtrack (pttrack_save->NEXT, ptpos->Y);
			}
		}
		else
		{
			ptgabarit->NBTRACK += 1;
			ptgabarit->TRACK = addtrack (ptgabarit->TRACK, ptpos->Y);
		}
	}
}

/*---------------------------------------------------------\
			 con_in_gabarit
\---------------------------------------------------------*/
int             con_in_gabarit (ptcon, ptgabarit)
CONNECTOR      *ptcon;
GABARIT        *ptgabarit;

{
	chain_list     *ptcpos = NULL;
	POSSIBILITY    *ptpos = NULL;
	TRACK          *ptstrack = NULL;
	TRACK          *pttrack = NULL;

	if ((ptcon->MINPOS <= ptgabarit->TRACK->Y) ||
	    (ptcon->MAXPOS >= ptgabarit->TRACK->Y + ptgabarit->HEIGHT))
		return FALSE;
	for (ptcpos = ptcon->POS; ptcpos; ptcpos = ptcpos->NEXT)
	{
		ptpos = (POSSIBILITY *) ptcpos->DATA;
		ptstrack = NULL;
		for (pttrack = ptgabarit->TRACK; pttrack; pttrack = pttrack->NEXT)
		{
			if (ptpos->Y == pttrack->Y)
				break;
			else if (pttrack->Y > ptpos->Y)
				break;
		}
		if (pttrack)
		{
			if (pttrack->Y != ptpos->Y)
				return FALSE;
		}
		else if ((ptgabarit->TRACK->Y + ptgabarit->HEIGHT) != ptpos->Y)
        {
			return FALSE;
        }
	}
	return TRUE;
}

/*---------------------------------------------------------\
                     valider_gabarit
\---------------------------------------------------------*/
GABARIT        *valider_gabarit (ptgabarit, height)
GABARIT        *ptgabarit;
long            height;

{
	/* a finir: provisoirement, on cable le bit slice 10 pistes */
	int             i = 0;
	long            y = 0;
    long            nbtrack;
	TRACK          *pttrack = NULL;

	switch(UtTec_V.Type)
	{
		case C_UtTec_TypeAsGa3:
            nbtrack = (height - 12 * SCALE_X) / ( 5 * SCALE_X ) + 2;
			break;
		case C_UtTec_TypeCMOS2:
		case C_UtTec_TypeUnknow:
		default:
            nbtrack = (height - 14 * SCALE_X) / ( 5 * SCALE_X ) + 2;
			break;
	}

	if (isvdd (ptgabarit->ALIM1))
	{
		ptgabarit->NBTRACK = nbtrack;
		ptgabarit->TRACK->NEXT = NULL;
		pttrack = ptgabarit->TRACK;
		/*
		 * PATCH AsGa.
		 */
		y = pttrack->Y + DPR_Var.TOP_TRACK * SCALE_X;
		for (i = 1; i < nbtrack; i++)
		{
			pttrack->NEXT = addtrack (NULL, y);
			pttrack = pttrack->NEXT;
			y = y + 5 * SCALE_X;
		}
	}
	else
	{
		if (isvss (ptgabarit->ALIM1))
		{
			ptgabarit->NBTRACK = nbtrack;
			ptgabarit->TRACK->NEXT = NULL;
			pttrack = ptgabarit->TRACK;
			/*
			 * PATCH AsGa : Premiere piste a 6 lambdas du haut de l'AB.
			 */
			y = pttrack->Y + DPR_Var.BOT_TRACK * SCALE_X;
			for (i = 1; i < nbtrack; i++)
			{
				pttrack->NEXT = addtrack (NULL, y);
				pttrack = pttrack->NEXT;
				y = y + 5 * SCALE_X;
			}
		}
		else
		{
            fflush( stdout );
			fprintf (stderr, "BUG: First slice Alim must be VDD* or VSS*\n");
			exit (1);
		}
	}
	return ptgabarit;
}

/*---------------------------------------------------------\
                     initgabarit
\---------------------------------------------------------*/
GABARIT        *initgabarit (ptname, walim1, ylin, nextalim, ptcheadcon, ptcheadallow)
char           *ptname;
long            walim1;
long            ylin;
char           *nextalim;
chain_list     *ptcheadcon;
chain_list     *ptcheadallow;

{
	chain_list     *ptccon = NULL;
	chain_list     *ptcallow = NULL;
	GABARIT        *ptgabarit = NULL;
	CONNECTOR      *ptcon = NULL;
	SEGMENT        *ptallow = NULL;
	TRACK          *pttrack = NULL;
	long            yalim2;

	ptgabarit = (GABARIT *) mbkalloc (sizeof (GABARIT));
	ptgabarit->ALIM1 = ptname;
	ptgabarit->WALIM1 = walim1;
	ptgabarit->NBTRACK = 1;
	ptgabarit->TRACK = addtrack (NULL, ylin);

	pttrack = ptgabarit->TRACK;

	for (ptccon = ptcheadcon; ptccon; ptccon = ptccon->NEXT)
	{
		ptcon = (CONNECTOR *) ptccon->DATA;
		if (!ptcon)
		{
            fflush( stdout );
			fprintf (stderr, "Error: Unable to initialize the gabarit, not valid slice!\n");
			exit (1);
		}
		if (!strncmp (ptcon->NAME, nextalim, 3))
		{
			yalim2 = ptcon->MAXPOS;
			ptgabarit->HEIGHT = ptcon->MAXPOS - ylin;
			ptgabarit->WALIM2 = widthalim (ptcon->POS);
			break;
		}
	}
	if (!ptccon)
	{
        fflush( stdout );
		fprintf (stderr, "Error: VDD and VSS alims must be available in all the instances\n");
		exit (1);
	}

	/*
	 * creer les pistes horizontales du gabarit a partir des
	 * transparences
	 */
	for (ptcallow = ptcheadallow; ptcallow; ptcallow = ptcallow->NEXT)
	{
		ptallow = (SEGMENT *) ptcallow->DATA;
		if ((ptallow) && (ptallow->Y1 > ylin))
			break;
	}
	if (ptcallow)
	{
		for (; ptcallow->DATA; ptcallow = ptcallow->NEXT)
		{
			ptallow = (SEGMENT *) ptcallow->DATA;
			if (ptallow->Y1 > yalim2)
				break;
			else
			{
				if (pttrack->Y != ptallow->Y1)	/* pour eviter les
								 * redondances */
				{
					pttrack->NEXT = addtrack (NULL, ptallow->Y1);
					pttrack = pttrack->NEXT;
					ptgabarit->NBTRACK += 1;
				}
			}
		}
	}

	/*
	 * rajouter la detection des pistes suplementaires du aux connecteurs
	 * fixes d'orientation NORD SUD et de Layer ALU2
	 */

	return (ptgabarit);
}

/*---------------------------------------------------------\
                     addlin
\---------------------------------------------------------*/
chain_list     *addlin (ptchead, ylin, width, index, ptgab)
chain_list     *ptchead;
long            ylin;
long            width;
long            index;
GABARIT        *ptgab;

{
	LINE           *ptlin = NULL;

	ptlin = (LINE *) mbkalloc (sizeof (LINE));
	ptlin->X = 0;
	ptlin->Y = ylin;
	ptlin->WIDTH = width;
	ptlin->HEIGHT = ptgab->HEIGHT;
	ptlin->INDEX = index;
	ptlin->GAB = ptgab;
	ptlin->GRID = NULL;
	ptlin->SIG = NULL;
	ptlin->CON = NULL;
	ptlin->INS = NULL;
	ptlin->N_WELL = NULL;
	ptlin->ALLOW = NULL;

	return addchain (ptchead, (void *) ptlin);
}

/*---------------------------------------------------------\
                     insertlin
\---------------------------------------------------------*/
chain_list     *insertlin (ptccon, ylin, width, indexlin)
chain_list     *ptccon;
long            ylin;
long            width;
long            indexlin;

{
	chain_list     *ptclin = NULL;
	chain_list     *ptccol = NULL;
	chain_list     *ptcallow = NULL;
	chain_list     *ptccon_lin = NULL;
	chain_list     *ptccell_lin = NULL;
	chain_list     *ptcsig_lin = NULL;
	COLUMN         *ptcol = NULL;
	LINE           *ptlin = NULL;
	CONNECTOR      *ptcon = NULL;
	SIGNAL         *ptsig = NULL;
	GABARIT        *ptgabarit = NULL;
    INSTANCE       *ptins = NULL;
	char           *nextalim = NULL;
	long            minlin;
	long            walim1;
	long            ynextalim;

	ptcon = (CONNECTOR *) ptccon->DATA;
	if (isvdd(ptcon->NAME))
		nextalim = namealloc("vss");
	else if (isvss (ptcon->NAME))
		nextalim = namealloc("vdd");
	else
	{
        fflush( stdout );
		fprintf (stderr,"BUG: The instance must begin with an Alim terminal, %s\n", ptcon->NAME);
		exit (1);
	}
	walim1 = widthalim (ptcon->POS);
	ptins = (INSTANCE *) ptcon->INS->DATA;
	ptcallow = ptins->ALLOW;
	ptgabarit = initgabarit (ptcon->NAME, walim1, ylin, nextalim, ptccon, ptcallow);
	ptgabarit = valider_gabarit (ptgabarit, ptgabarit->HEIGHT);
	ptccell_lin = NULL;
	ptccon_lin = NULL;
	ptcsig_lin = NULL;
	ptclin = addlin (NULL, ylin, width, indexlin, ptgabarit);
	for (ptccol = ptcon->COL; ptccol->NEXT; ptccol = ptccol->NEXT->NEXT)
	{
		ptcol = (COLUMN *) ptccol->DATA;
		if (!ptcol->USER)
			continue;
		ptcon = (CONNECTOR *) ptcol->USER->DATA;
		if (ptcon->MINPOS == ylin)
		{
			if (!strncmp (ptcon->NAME, ptgabarit->ALIM1, 3))
			{
				ptcol->USER = ptcol->USER->NEXT;
				ptccell_lin = addchain (ptccell_lin, (void *) ptcon->INS->DATA);
			}
			else
			{
                fflush( stdout );
				fprintf (stderr,"Error: Bad alternance of the alim terminals!!\n");
				exit (1);
			}
		}
		else
			/* c.a.d ptcon->MINPOS > ylin */
			continue;
		for (; ptcol->USER; ptcol->USER = ptcol->USER->NEXT)
		{
			ptcon = (CONNECTOR *) ptcol->USER->DATA;
			if (!ptcon)
			{
                fflush( stdout );
				fprintf (stderr,"Error: VDD and VSS alims must delimt all the slices!\n");
				exit (1);
			}
			if (!strncmp (ptcon->NAME, nextalim, 3))
			{
				ynextalim = ptcon->MAXPOS;
				if (!(ptcol->USER->NEXT->DATA))
					if ((ptcol->USER = ptcol->USER->NEXT->NEXT))
					{
						ptcon = (CONNECTOR *) ptcol->USER->DATA;
						if ((strncmp (ptcon->NAME, nextalim, 3)) && (ptcon->MINPOS == ynextalim))
						{
                            fflush( stdout );
							fprintf (stderr, "Error: Bad alternance of the alim terminals!\n");
							exit (1);
						}
					}
				break;
			}
			if (!con_in_gabarit (ptcon, ptgabarit))
			{
                fflush( stdout );
				fprintf (stderr, "*** dpr error *** : Terminal %s does not fit to the slice gabarit(instance %s of model %s)\n",
                         ptcon->NAME,
                         ((INSTANCE *)(ptcon->INS->DATA))->INSNAME,
                         ((INSTANCE *)(ptcon->INS->DATA))->FIGNAME);
				exit (1);
			}
			else
			{
				ptccon_lin = addchain (ptccon_lin, (void *) ptcon);
				ptcon->LIN = ptclin;
				ptsig = (SIGNAL *) ptcon->SIG->DATA;
				if (!existsig (ptcsig_lin, ptsig))
				{
					minlin = ptsig->MIN_LIN;
					ptsig->MIN_LIN = dpr_min (ptsig->MIN_LIN, indexlin);
					ptsig->HEIGHT = ((minlin == MAXINT) ? 0 : dpr_max (minlin + ptsig->HEIGHT, indexlin)
							 - ptsig->MIN_LIN);
					ptcsig_lin = addchain (ptcsig_lin, (void *) ptsig);
					ptsig->LIN = addchain (ptsig->LIN, (void *) ptclin->DATA);
				}
			}
		}
	}
	ptlin = (LINE *) ptclin->DATA;
	ptlin->CON = ptccon_lin;
	ptlin->SIG = ptcsig_lin;
	ptlin->INS = ptccell_lin;

	return (ptclin);
}

/*---------------------------------------------------------\
                     dprlin
\---------------------------------------------------------*/
void            dprlin (ptroot)
FIGURE         *ptroot;

{
	chain_list     *ptclin = NULL;
	chain_list     *ptccol = NULL;
	chain_list     *ptccon = NULL;
	chain_list     *ptccon1 = NULL;
	GABARIT        *ptgabarit;
	TRACK          *pttrack = NULL;
	CONNECTOR      *ptcon = NULL;
	CONNECTOR      *ptcon1 = NULL;
	POSSIBILITY    *ptpos = NULL;
	SIGNAL         *ptsig = NULL;
	LINE           *ptlin = NULL;
	COLUMN         *ptcol = NULL;
	long            ylin_pere;
	long            width = 0;
	long            minlin = MAXINT;
	int             indexlin = 1;
	int             detectlin = TRUE;
	chain_list     *ptcconbug = NULL;
	CONNECTOR      *ptconbug = NULL;
	chain_list     *ptcins = NULL;
	INSTANCE       *ptins = NULL;

	ptroot->LIN = addchain (ptroot->LIN, (void *) NULL);
	ptclin = ptroot->LIN;
	while (detectlin)
	{
		ptccon = NULL;
		ylin_pere = MAXLONG;
		for (ptccol = ptroot->COL->NEXT->NEXT; ptccol->NEXT; ptccol = ptccol->NEXT->NEXT)
		{
			ptcol = (COLUMN *) ptccol->DATA;
			if (ptcol->USER)
			{
				ptcon = (CONNECTOR *) ptcol->USER->DATA;
				if ((ptcon->TYPE != EXTERNAL) && (ptcon->MINPOS < ylin_pere))
				{
					ylin_pere = ptcon->MINPOS;
					ptccon = ptcol->USER;
				}
			}
		}
		if (ptccon)
		{
			/*
			 * Avoir la hauteur du canal horizontal, si existe,
			 * et chercher le(s) gabarit(s) correspondant(s)
			 */
			ptclin->NEXT = insertlin (ptccon, ylin_pere, width, indexlin);
			ptclin = ptclin->NEXT;
			ptlin = (LINE *) ptclin->DATA;
			ptgabarit = ptlin->GAB;

			/*
			 * mettre les connecteurs DF place de la figure sur
			 * le gabarit de la ligne
			 */
			for (ptccon1 = ptroot->CON; ptccon1; ptccon1 = ptccon1->NEXT)
			{
				ptcon1 = (CONNECTOR *) ptccon1->DATA;
				/*
				 * dans sigconpere on cree un connecteur par
				 * possibilite: pas besoin de for
				 */
				ptpos = (POSSIBILITY *) ptcon1->POS->DATA;
				if ((ptcon1->COL) && (!ptcon1->LIN) &&
				    ((ptpos->ACCES == EAST) || (ptpos->ACCES == WEST)))
				{
					if ((ptpos->Y >= ptlin->Y) && (ptpos->Y < ptlin->Y + ptlin->HEIGHT))
					{
						if (!con_in_gabarit (ptcon1, ptgabarit))	/* c.a.d flottant */
						{
							pttrack = ptgabarit->TRACK->NEXT;
							ptpos->Y = pttrack->Y;
							ptcon1->MINPOS = dpr_min (ptcon1->MINPOS, pttrack->Y);
							ptcon1->MAXPOS = dpr_max (ptcon1->MAXPOS, pttrack->Y);
							for (pttrack = pttrack->NEXT; pttrack; pttrack = pttrack->NEXT)
							{
								ptcon1->POS->NEXT = addpos (ptcon1->POS->NEXT, ptpos->X,
											    pttrack->Y, ptpos->LAYER,
											    ptpos->ACCES, ptpos->TYPE,
											    ptpos->TRACK, ptpos->WIDTH,
								    ptcon1);
								ptcon1->NBPOSS = ptcon1->NBPOSS + 1;
								ptcon1->MINPOS = dpr_min (ptcon1->MINPOS, pttrack->Y);
								ptcon1->MAXPOS = dpr_max (ptcon1->MAXPOS, pttrack->Y);
							}
						}
						else
							/* c.a.d fixe */
						{
							ptcon1->MINPOS = ptpos->Y;
							ptcon1->MAXPOS = ptpos->Y;
						}
						ptlin->CON = addchain (ptlin->CON, (void *) ptcon1);
						ptcon1->LIN = ptclin;
						ptsig = (SIGNAL *) ptcon1->SIG->DATA;
						if (!existsig (ptlin->SIG, ptsig))
						{
							minlin = ptsig->MIN_LIN;
							ptsig->MIN_LIN = dpr_min (ptsig->MIN_LIN, indexlin);
							ptsig->HEIGHT = ((minlin == MAXINT) ? 0 :
									 dpr_max (minlin + ptsig->HEIGHT,
										  indexlin) - ptsig->MIN_LIN);
							ptlin->SIG = addchain (ptlin->SIG, (void *) ptsig);
							ptsig->LIN = addchain (ptsig->LIN, (void *) ptlin);
						}
					}
				}
			}
			indexlin++;
		}
		else
			detectlin = FALSE;
	}
	ptroot->LIN = ptroot->LIN->NEXT;

}
