/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    File        :   router_phwork.c

    Date        :   02 04 94
    Version     :   1

    Auteur      :   BEAURIN oliver

    Description :   All functions which add ph elememt.

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include MUT_H
#include MLO_H
#include MPH_H

#include "router_dpr.h"

/*---------------------------------------------------------\
                     addvia
\---------------------------------------------------------*/
 chain_list *addvia(ptchead,x,y,type)
 chain_list *ptchead;
 long        x,y;
 char        type;

 {
 VIA *ptnewvia;

 ptnewvia = (VIA *)mbkalloc(sizeof(VIA));

 ptnewvia->X    = x;
 ptnewvia->Y    = y;
 ptnewvia->TYPE = type;

 return addchain(ptchead,(void *)ptnewvia);

 }

/*---------------------------------------------------------\
                     add_via_in_op
\---------------------------------------------------------*/
void            add_via_in_op (ptcol, x, y)
COLUMN         *ptcol;
long            x;
long            y;
{
    VIA            *ptvia = NULL;
    chain_list     *ptcvia;
    chain_list     *ptcvia_save = NULL;

    for (ptcvia = ptcol->VIA; ptcvia; ptcvia = ptcvia->NEXT)
    {
        ptvia = (VIA *) ptcvia->DATA;
        if (ptvia->Y > y)
            break;
        else if ((ptvia->X == x) && (ptvia->Y == y))
        {
            ptcvia_save = ptcvia;
            break;
        }
        else
            ptcvia_save = ptcvia;
    }
    if (ptcvia_save)
    {
        if ((ptvia->X != x) || (ptvia->Y != y))
            ptcvia_save->NEXT = addvia (ptcvia_save->NEXT, x, y, (char *) CONT_VIA);
    }
    else
        ptcol->VIA = addvia (ptcol->VIA, x, y, (char *) CONT_VIA);
}

/*---------------------------------------------------------\
                     seg_inclu

fonction seg_inclu() : retourne TRUE si l'un des deux segments
contient l'autre sinon FALSE.
\---------------------------------------------------------*/
int             seg_inclu (seg1_y1, seg1_y2, seg2_y1, seg2_y2)
long            seg1_y1;
long            seg1_y2;
long            seg2_y1;
long            seg2_y2;

{
    if (seg1_y1 <= seg2_y1)
    {
        if (seg1_y2 < seg2_y1)
        {
            return FALSE;
            /* +-+  */
            /* |2|  */
            /* +-+  */
            /* +-+      */
            /* |1|      */
            /* +-+      */
        }
        else
        {
            return TRUE;
            /* +-+         |     +-+ | +-+ +-+ | +-+     */
            /* |2|         | +-+ |2| | | | |2| | | | +-+ */
            /* +-+ +-+     | |1| +-+ | |1| +-+ | |1| |2| */
            /*     |1|     | | |     | | |     | | | +-+ */
            /*     +-+     | +-+     | +-+     | +-+     */
        }
    }
    else
    {
        if (seg1_y1 <= seg2_y2)
        {
            return TRUE;
            /* +-+     | +-+     | +-+     | +-+     */
            /* |1|     | |1|     | | |     | | | +-+ */
            /* +-+ +-+ | | | +-+ | |1| +-+ | |1| |2| */
            /*     |2| | +-+ |2| | | | |2| | | | +-+ */
            /*     +-+ |     +-+ | +-+ +-+ | +-+     */
        }
        else
        {
            return FALSE;
            /* +-+      */
            /* |1|      */
            /* +-+      */
            /* +-+  */
            /* |2|  */
            /* +-+  */
        }
    }
}

/*---------------------------------------------------------\
                     addseg
\---------------------------------------------------------*/
 chain_list *addseg(ptchead,ptname,x1,y1,x2,y2,width,type,layer)
 chain_list *ptchead;
 char       *ptname;
 long        x1, y1, x2, y2;
 long        width;
 char        type, layer;
 
 {
 SEGMENT   *ptseg=NULL;

 ptseg = (SEGMENT *)mbkalloc(sizeof(SEGMENT));
 
 ptseg->NAME = ptname ;
 ptseg->X1 = x1 ;
 ptseg->X2 = x2 ;
 ptseg->Y1 = y1 ;
 ptseg->Y2 = y2 ;
 ptseg->WIDTH = width ;
 ptseg->TYPE = type ;
 ptseg->LAYER = layer ;

 return (addchain(ptchead,(void *)ptseg));
 
 }

/*---------------------------------------------------------\
                     insertseg

fonction insertseg() :insertion d'un segment dans la
colonne. avec trie et unification des segments.
\---------------------------------------------------------*/
void            insertseg (ptcol, x1, y1, x2, y2, layer, width, name)
COLUMN         *ptcol;
long            x1, x2;
long            y1, y2;
char            layer;
int             width;
char           *name;
{
    chain_list     *ptcseg_save = NULL;
    chain_list     *ptcseg = NULL;
    SEGMENT        *ptseg = NULL;
    int             inclu = 0;

    for (ptcseg = ptcol->SEG; ptcseg; ptcseg = ptcseg->NEXT)
    {
        ptseg = (SEGMENT *) ptcseg->DATA;
        if (ptseg->Y1 > y1)
            break;
        else if (ptseg->Y1 == y1)
            if (inclu = seg_inclu (ptseg->X1, ptseg->X2, x1, x2))
            {
                ptcseg_save = ptcseg;
                break;
            }
            else
                ptcseg_save = ptcseg;
        else
            ptcseg_save = ptcseg;
    }
    if (ptcseg_save)
    {
        if (inclu)
        {
            ptseg->X1 = dpr_min (ptseg->X1, x1);
            ptseg->X2 = dpr_max (ptseg->X2, x2);
        }
        else
            ptcseg_save->NEXT = addseg (ptcseg_save->NEXT, name, x1, y1, x2, y2, width, 'H', layer);
    }
    else
        ptcol->SEG = addseg (ptcol->SEG, name, x1, y1, x2, y2, width, 'H', layer);
}

