/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_intv.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/


/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  <stdlib.h>
#include  <math.h>
#include  MPH_H
#include  MUT_H

#include  "strecher_support.h"
#include  "strecher_intv.h"

/*---------------------------------------------------------\
                     defines
\---------------------------------------------------------*/
#define    F_INTV_EX_INF       0
#define    F_INTV_EX_SUP       0
#define    F_INTV_EX_TOTAL    (F_INTV_EX_INF | F_INTV_EX_SUP)
#define    F_INTV_IC_INF       1
#define    F_INTV_IC_SUP       2
#define    F_INTV_IC_TOTAL    (F_INTV_IC_INF | F_INTV_IC_SUP)

/*---------------------------------------------------------\
                     addintv
\---------------------------------------------------------*/
intv_list      *addintv(pthead, inf, sup)
intv_list      *pthead;
long            inf, sup;
{
    intv_list      *pintv;

    /* Allocation de l'intervalle */
    pintv = (intv_list *) mbkalloc(sizeof(intv_list));

    if( inf > sup )
    {
        fprintf(stderr,"*** strecher bugs *** : in addintv inf > sup(%ld,%ld).\n",inf,sup);
        exit( 1 );
    }

    /* Initialisation des champs. */
    pintv->INF = inf;
    pintv->SUP = sup;
    pintv->NEXT = pthead;

    return (pintv);
}

/*---------------------------------------------------------\
                     delintv
\---------------------------------------------------------*/
intv_list      *delintv(pthead, inf, sup)
intv_list      *pthead;
long            inf, sup;
{
    intv_list     **ppintv, *pHEAD, *pDEL;


    pHEAD = pthead;
    for (ppintv = &pHEAD; (*ppintv) != (intv_list *) NULL; ppintv = &(*ppintv)->NEXT)
        if (((*ppintv)->INF == inf) && ((*ppintv)->SUP == sup))
        {
            pDEL = (*ppintv);

            (*ppintv) = (*ppintv)->NEXT;
            mbkfree((void *) pDEL);

            return (pHEAD);
        }

    return (pHEAD);
}


/*---------------------------------------------------------\
                     free_ls_intv
\---------------------------------------------------------*/
void            free_ls_intv(pthead)
intv_list      *pthead;
{
    intv_list      *pintv;


    /* Boucle de destruction directe de la chaine. */
    for (pintv = pthead; pintv != (intv_list *) NULL;
         pintv = pintv->NEXT)
        mbkfree(pintv);
} /* end of free_ls_intv */

/*---------------------------------------------------------\
                     modify_ls_intv
\---------------------------------------------------------*/
intv_list      *modify_ls_intv(pthead, inf, sup)
intv_list      *pthead;
long            inf, sup;
{

#define  EX_INF     0
#define  EX_SUP     0
#define  EX_TOTAL   0
#define  IC_INF     1
#define  IC_SUP     2
#define  IC_TOTAL   3

    intv_list     **ppintv, *pHEAD, *pDEL;
    long            inclusion;

    pHEAD = pthead;
    for (ppintv = &pHEAD; (*ppintv) != (intv_list *) NULL;
         ppintv = &(*ppintv)->NEXT)
    {
        /* Test de non-interraction : cf DOC. */
LOOP:   if ((*ppintv)->SUP <= inf)
            continue;
        if ((*ppintv)->INF >= sup)
            break;

        inclusion = ((((*ppintv)->INF < sup) && ((*ppintv)->SUP > sup)) ? IC_SUP : EX_SUP)
                  + ((((*ppintv)->INF < inf) && ((*ppintv)->SUP > inf)) ? IC_INF : EX_INF);

        switch (inclusion)
        {
        case IC_TOTAL:
            {
                (*ppintv)->NEXT = addintv((*ppintv)->NEXT, sup, (*ppintv)->SUP);
                (*ppintv)->SUP = inf;
            }
            return (pHEAD);
        case IC_SUP:
            (*ppintv)->INF = sup;
            return (pHEAD);
        case IC_INF:
            (*ppintv)->SUP = inf;
            break;
        case EX_TOTAL:
            {
                pDEL = (*ppintv);
                (*ppintv) = (*ppintv)->NEXT;

                mbkfree((void *) pDEL);
                if ((*ppintv) == (intv_list *) NULL)
                    return (pHEAD);
            }
            goto LOOP;
        }
    }

    return (pHEAD);

#undef  EX_INF
#undef  EX_SUP
#undef  EX_TOTAL
#undef  IC_INF
#undef  IC_SUP
#undef  IC_TOTAL
}

/*---------------------------------------------------------\
                     merge_ls_intv
\---------------------------------------------------------*/
intv_list      *merge_ls_intv(pthead1, pthead2)
intv_list      *pthead1;
intv_list      *pthead2;
{
    intv_list     **ppintv1, *pthead, *pintv2, *pDEL;
    long            iex_INF, iex_SUP, inclusion;

    pthead = pthead1;
    ppintv1 = &pthead;

    pintv2 = pthead2;

    /* Au moin deux intervalles dans la liste a prendre en compte. */
    if ((pintv2 == (intv_list *) NULL) || (pintv2->NEXT == (intv_list *) NULL))
    {
        return (pthead);
    }
    else
    {
        iex_INF = pintv2->SUP;
        iex_SUP = pintv2->NEXT->INF;

        pintv2 = pintv2->NEXT;
    }

    /* Boucle sur les intervalles interdits de la liste a fusionner. */
    while (TRUE)
    {
        /* Passage a l'intervalle interdit suivant. */
        if (iex_SUP <= (*ppintv1)->INF)
        {
            if (pintv2->NEXT == (intv_list *) NULL)
                return (pthead);

            iex_INF = pintv2->SUP;
            iex_SUP = pintv2->NEXT->INF;

            pintv2 = pintv2->NEXT;
        }

        /* Sauter les intervalles dont la borne SUPERIEURE est inferieure a la borne INFERIEURE de l'interdit. */
        while ((*ppintv1)->SUP <= iex_INF)
            if ((*ppintv1)->NEXT == (intv_list *) NULL)
                return (pthead);
            else
                ppintv1 = &(*ppintv1)->NEXT;

        /* Calcul du type d'inclusion. */
        inclusion = (((*ppintv1)->INF < iex_INF) && ((*ppintv1)->SUP > iex_INF) ? F_INTV_IC_INF : F_INTV_EX_INF)
                  + (((*ppintv1)->INF < iex_SUP) && ((*ppintv1)->SUP > iex_SUP) ? F_INTV_IC_SUP : F_INTV_EX_SUP);

        /* Traitement specifique a chaque type d'inclusion. */
        switch (inclusion)
        {
        case F_INTV_IC_TOTAL:
            (*ppintv1)->NEXT = addintv((*ppintv1)->NEXT, iex_SUP, (*ppintv1)->SUP);
            (*ppintv1)->SUP = iex_INF;

            /* Intervalle suivant a retailler. */
            if ((*ppintv1)->NEXT == (intv_list *) NULL)
                return (pthead);
            else
                ppintv1 = &(*ppintv1)->NEXT;

            break;
        case F_INTV_IC_SUP:
            (*ppintv1)->INF = iex_SUP;
            break;
        case F_INTV_IC_INF:
            (*ppintv1)->SUP = iex_INF;

            /* Intervalle suivant a retailler. */
            if ((*ppintv1)->NEXT == (intv_list *) NULL)
                return (pthead);
            else
                ppintv1 = &(*ppintv1)->NEXT;

            break;
        case F_INTV_EX_TOTAL:
            pDEL = *ppintv1;
            ppintv1 = &(*ppintv1)->NEXT;

            mbkfree((void *) pDEL);
            if (*ppintv1 == (intv_list *) NULL)
                return (pthead);

            break;
        }
    }
} /* end of merge_ls_intv */

/*---------------------------------------------------------\
                     trunc_ls_intv
\---------------------------------------------------------*/
intv_list      *trunc_ls_intv(pthead, inf, sup)
intv_list      *pthead;
long            inf, sup;
{

#define  EX_INF     0
#define  EX_SUP     0
#define  EX_TOTAL   0
#define  IC_INF     1
#define  IC_SUP     2
#define  IC_TOTAL   3

    intv_list     **ppintv, *pHEAD, *pDEL;
    long            inclusion;

    pHEAD = pthead;
    for (ppintv = &pHEAD; (*ppintv) != (intv_list *) NULL;
         ppintv = &(*ppintv)->NEXT)
    {

        /* Type d'inclusion des bornes "inf" et "sup" dans l'intervalle courant. */
LOOP:   inclusion = ((((*ppintv)->INF < sup) && ((*ppintv)->SUP >= sup)) ? IC_SUP : EX_SUP)
                  + ((((*ppintv)->INF <= inf) && ((*ppintv)->SUP > inf)) ? IC_INF : EX_INF);

        switch (inclusion)
        {
        case IC_TOTAL:
            (*ppintv)->SUP = sup;
            (*ppintv)->INF = inf;
            break;
        case IC_SUP:
            (*ppintv)->SUP = sup;
            break;
        case IC_INF:
            (*ppintv)->INF = inf;
            break;
        case EX_TOTAL:
            {                   /* Destruction de l'intervalle. */
                pDEL = (*ppintv);
                (*ppintv) = (*ppintv)->NEXT;

                mbkfree((void *) pDEL);
                /* Test fin de boucle. */
                if ((*ppintv) == (intv_list *) NULL)
                    return (pHEAD);
            }
            goto LOOP;
        }
    }

    return (pHEAD);

#undef  EX_INF
#undef  EX_SUP
#undef  EX_TOTAL
#undef  IC_INF
#undef  IC_SUP
#undef  IC_TOTAL
}

/*---------------------------------------------------------\
                     debug_intv
\---------------------------------------------------------*/
void            debug_intv(n,pthead)
int             n;
intv_list      *pthead;
{
    intv_list      *pintv;

    if(!mbkgetenv("DEBUG_STRECHER"))
        return;

    if(atoi(mbkgetenv("DEBUG_STRECHER"))<n)
        return;

    for (pintv = pthead; pintv != (intv_list *) NULL ; pintv = pintv->NEXT)
    {
        fprintf(stdout,"(%ld,%ld) ",pintv->INF,pintv->SUP);
    }
    fprintf(stdout,"\n");
} /* end of debug_intv */

/*---------------------------------------------------------\
                     in_ls_intv
\---------------------------------------------------------*/
long            in_ls_intv(pthead, inf, sup)
intv_list      *pthead;
long            inf, sup;
{
    intv_list      *pintv;

    for (pintv = pthead; (pintv != (intv_list *) NULL) && (pintv->INF <= sup); pintv = pintv->NEXT)
        if ((pintv->INF <= inf) && (pintv->SUP >= sup))
            return (TRUE);

    return (FALSE);
}

/*---------------------------------------------------------\
                     ls_intvtoseg
\---------------------------------------------------------*/
void            ls_intvtoseg(pthead, ptfig, layer, width, yseg)
intv_list      *pthead;
phfig_list     *ptfig;
char            layer;
long            width, yseg;
{
    intv_list      *pintv;


    for (pintv = pthead; pintv != (intv_list *) NULL;
         pintv = pintv->NEXT)
        addphseg(ptfig, layer, width,
                 lambda_ceil(pintv->INF), yseg,
                 lambda_floor(pintv->SUP), yseg,
                 (char *) NULL);
}
