/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_supply.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  <string.h>
#include  MPH_H
#include  MUT_H

#include  "strecher_support.h"
#include  "strecher_supply.h"

/*---------------------------------------------------------\
                     addosupply
\---------------------------------------------------------*/
static supply_list *addosupply(pthead, ysupply, supply_type)
supply_list    *pthead;
long            ysupply;
char            supply_type;
{
    supply_list   **ppsupply, *pHEAD, *pNEW;


    pHEAD = pthead;
    for (ppsupply = &pHEAD; *ppsupply != (supply_list *) NULL;
         ppsupply = &(*ppsupply)->NEXT)
    {
        if ((*ppsupply)->YSUPPLY == ysupply)
            return (pHEAD);
        if ((*ppsupply)->YSUPPLY > ysupply)
            break;
    }

    pNEW = (supply_list *) mbkalloc(sizeof(supply_list));

    pNEW->YSUPPLY = ysupply;
    pNEW->TYPE = supply_type;

    pNEW->NEXT = *ppsupply;
    *ppsupply = pNEW;

    return (pHEAD);
} /* end of addosupply */

/*---------------------------------------------------------\
                     is_out_of_supply
\---------------------------------------------------------*/
int is_out_of_supply( pthead, min, max )
supply_list    *pthead;
long            min;
long            max;
{
    supply_list    *psupply;

    if( pthead == NULL )
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher warning *** : no alim connectors found.\n" );
        return FALSE;
    }

    if( max <= pthead->YSUPPLY )
    {
        return TRUE;
    }

    for (psupply = pthead; psupply->NEXT ; psupply = psupply->NEXT);

    if( min >= psupply->YSUPPLY )
    {
        return TRUE;
    }

    return FALSE;

} /* end of is_out_of_supply */

/*---------------------------------------------------------\
                     getnsupply
\---------------------------------------------------------*/
supply_list    *getnsupply(pthead, ycut)
supply_list    *pthead;
long            ycut;
{
    supply_list    *psupply;

    /* Moin de deux elements dans la liste. */
    if ((pthead == (supply_list *) NULL) || (pthead->NEXT == (supply_list *) NULL))
        return ((supply_list *) NULL);

    for (psupply = pthead; psupply->NEXT != (supply_list *) NULL;
         psupply = psupply->NEXT)
        if ((psupply->YSUPPLY < ycut) && (psupply->NEXT->YSUPPLY >= ycut))
            return (psupply);

    return ((supply_list *) NULL);
}

/*---------------------------------------------------------\
                     index_calc
\---------------------------------------------------------*/
long            index_calc(ptslice, mult, yref)
supply_list    *ptslice;
long            mult;
long            yref;
{
    long            NB_through, NO_through;


    /* Nombre de transparences du slice. */
    NB_through = (ptslice->NEXT->YSUPPLY - ptslice->YSUPPLY
                  - SX_WIDTH_VDD / 2
                  - SX_WIDTH_VSS / 2
                  + SX_WIDTH_TALU2
        ) / SX_PITCH_TALU2 - 1;

    if (ptslice->TYPE == C_SUPPLY_VDD)
        NO_through = (yref - ptslice->YSUPPLY
                      - SX_WIDTH_VDD / 2
                      + SX_WIDTH_TALU2 / 2
            ) / SX_PITCH_TALU2 - 1;
    else
        NO_through = (ptslice->NEXT->YSUPPLY - yref
                      - SX_WIDTH_VDD / 2
                      + SX_WIDTH_TALU2 / 2
            ) / SX_PITCH_TALU2 - 1;

    return (NO_through + mult * NB_through);
} /* end of index_calc */

/*---------------------------------------------------------\
                     dup_ls_supply
\---------------------------------------------------------*/
void           *dup_ls_supply(a_ptfig_src, a_ptfig_dest)
phfig_list     *a_ptfig_src;
phfig_list     *a_ptfig_dest;
{
#define  access_supply  ( (supply_list*)(ptuser->DATA) )

    ptype_list     *ptuser;
    supply_list    *psupply, *phead, *pEND = NULL;

    /* Recuperation de la structure d'origine. */
    if ((ptuser = get_ls_data(a_ptfig_src, C_SUPP_USUPPLY, FALSE))
                == (ptype_list *) NULL)
        return ((void *) NULL);

    if (ptuser->DATA == (void *) NULL)
        return ((void *) NULL);

    /* Duplication de la liste. */
    phead = (supply_list *) NULL;

    for (psupply = access_supply; psupply != (supply_list *) NULL;
         psupply = psupply->NEXT)
    {
        if (psupply == access_supply)
        {
            /* Creation de la tete de liste. */
            pEND = (supply_list *) mbkalloc(sizeof(supply_list));
            phead = pEND;
        }
        else
        {
            /* Addition d'un element en fin de liste. */
            pEND->NEXT = (supply_list *) mbkalloc(sizeof(supply_list));
            pEND = pEND->NEXT;
        }

        pEND->YSUPPLY = psupply->YSUPPLY;
        pEND->TYPE = psupply->TYPE;

        pEND->NEXT = (supply_list *) NULL;
    }

    return ((void *) phead);

#undef  access_supply
} /* end of dup_ls_supply */

/*---------------------------------------------------------\
                     extract_ls_supply
\---------------------------------------------------------*/
void            extract_ls_supply(ptfig)
phfig_list     *ptfig;
{
    ptype_list     *ptuser;
    phcon_list     *pcon;
    supply_list    *ptsupply;


    ptuser = get_ls_data(ptfig, C_SUPP_USUPPLY, F_SUPP_UCREAT);
    ptsupply = (supply_list *) ptuser->DATA;

    if( ptsupply )
    {
        return;
    }

    for (pcon = ptfig->PHCON; pcon != (phcon_list *) NULL;
         pcon = pcon->NEXT)
    {
        if ((pcon->LAYER != ALU2) || ((pcon->ORIENT != WEST) && (pcon->ORIENT != EAST)))
            continue;

        if (isvdd(pcon->NAME) && (pcon->WIDTH == SX_WIDTH_VDD))
            ptsupply = addosupply(ptsupply, pcon->YCON, C_SUPPLY_VDD);

        if (isvss(pcon->NAME) && (pcon->WIDTH == SX_WIDTH_VSS))
            ptsupply = addosupply(ptsupply, pcon->YCON, C_SUPPLY_VSS);
    }

    ptuser->DATA = (void *) ptsupply;
}
