/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: SharedGN.c,v 1.5 2000/03/08 17:43:33 jpc Exp $       |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                   | |%
 * %| |     F p G e n   P a c k a g e   1.10                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, MASI, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author    :            Jean-Paul CHAPUT                           | |%
 * %| |  E-mail    :        cao-vlsi@masi.ibp.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Module    :        "SharedGN.c"                                   | |%
 * %| | ****************************************************************** | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */


# include  <stdio.h>
# include  <string.h>
# include  <time.h>

# include  <genlib.h>

# include  "GenericGN.h"
# include  "SharedGN.h"


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |                    Variables Declarations                      | |%
 *   %| \----------------------------------------------------------------/ |%
 *   %|                                                                    |%
 *   %| All the following variables are declared `extern' in "SharedGN.h". |%
 */

       GV_t GV;


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                        Internal Constants                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */



/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |               Internal Structures and Variables                    | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |                       Macros Definitions                       | |%
 *   %| \----------------------------------------------------------------/ |%
 */

/* Macro for various index computings. */
# define  mc_IndexToSlice(index)      \
            (((index) >= GV.aWidthDP) \
                ?  (index)            \
                : ((index) + GV.aSlice))
# define  mc_SliceToPower(slice)      \
            (((slice) + GV.aWidthDP % 2) / 2)
# define  mc_SliceToPowerType(slice)  \
            (((slice) + GV.aWidthDP) % 2)
# define  mc_SliceToPowerName(slice)  \
            ((mc_SliceToPowerType(slice)) ? "vdd" : "vss")
# define  mc_IndexToPower(index)      \
            (mc_SliceToPower(mc_IndexToSlice((index))))
# define  mc_IndexToPowerName(index)  \
            (mc_SliceToPowerName(mc_IndexToSlice((index))))
# define  mc_SliceToBit(slice)        \
            ((GV.aMSB0) ? (GV.aWidthDP - 1 - slice) : (slice))
# define  mc_IndexToBit(index)        \
            (mc_SliceToBit(mc_IndexToSlice((index))))

/* Macro for standard terminals names. */
# define  mc_VHDL_WIDE(msb_i,lsb_i)  \
            (UtBas_PrintF("(%ld %s %ld)",                \
                          mc_IndexToBit(msb_i),          \
                          (GV.aMSB0) ? "to" : "downto",  \
                          mc_IndexToBit(lsb_i)))
# define  mc_VHDL_FULL_WIDE()        \
            (mc_VHDL_WIDE(GV.aWidthOp - 1, 0))
# define  mc_VHDL_MEMM_WIDE(msb_i,lsb_i) \
            (UtBas_PrintF("(%ld downto %ld)", msb_i, lsb_i))
# define  mc_VHDL_FULL_BUS(conname)  \
            (UtBas_PrintF("%s%s", (conname), mc_VHDL_FULL_WIDE()))
# define  mc_POWER_NAME(slice)  \
            (GENLIB_NAME("%s_%d", mc_SliceToPowerName(slice), \
                           mc_SliceToPower(slice)))
# define  mc_VDDE(conname,index)  \
            ((GV.aPower)                            \
              ? UtBas_PrintF("%s => vdd"   , (conname))  \
              : UtBas_PrintF("%s => vdd_%d", (conname),  \
                                             mc_IndexToPower(index)))
# define  mc_VSSE(conname,index)  \
            ((GV.aPower)                            \
              ? UtBas_PrintF("%s => vss"   , (conname))  \
              : UtBas_PrintF("%s => vss_%d", (conname),  \
                                             mc_IndexToPower(index)))




/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                  Internal Functions Declarations                   | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

static void  Init          __FP(( void ));
static long  FunctionCheck __FP(( void ));

static char *OperName      __FP(( void ));
static void  HeaderVHDL    __FP(( void ));
static void  SignalVHDL    __FP(( void ));
static void  DataFlowBool  __FP(( void ));
static void  DataFlowMask  __FP(( void ));
static void  DataFlowBuse  __FP(( void ));
static void  DataFlowMux   __FP(( void ));
static void  DataFlowComp  __FP(( void ));
static void  DataFlowMem   __FP(( void ));
static void  EntityVHDL    __FP(( void ));
static void  ArchiVHDL     __FP(( void ));
static void  CreateVHDL    __FP(( void ));


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                     Functions Definitions                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "Init()"                                             | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Perform all the initialisation needed to use the boolean functions     |%
 * %| generator. This function must be used only one time, at the first call |%
 * %| of the 'Dispatch' function. The initialisations performed are listed   |%
 * %| below :                                                                |%
 * %|   1. Load the error data base whith module specific errors.            |%
 * %|   2. Select the functions that will generate structural view and the   |%
 * %| physical view.                                                         |%
 * %|                                                                        |%
 */

static void  Init()
{
  /* Attribution of a module identifier. */
  GV.ModuleId = GfpUT_GetModuleId("GfpGN", F_GfpUT_ModuleAdd);
  /* Load error and warning DB. */
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_WIDTH, "[%s]:Invalid bus width:%ld.");
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_SLICE,
    "[%s]:Invalid LSB starting slice:%ld." );
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_DRIVE,
    "[%s]:invalid output drive:%ld." );
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_NOVIEW, "[%s]:No view to generate.");
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_OPENVBE,
    "[%s]:Unable to create file:\"%s\"." );
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_FUNCTION,
    "[%s]:Unknow function code:%ld." );
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_WidthDP,
    "Invalid data-path wide : %d (%d < WideDP < %d)." );
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_WidthOp,
    "[%s]:Invalid operator wide : %d (%d < WideOp < %d)." );
  GfpUT_AddError( GV.ModuleId, IE_GfpGN_SliceEven,
    "[%s]:This operator does not support odd starting slice : %d." );
  GfpUT_AddWarng( GV.ModuleId, IW_GfpGN_PowerMerge,
    "[%s]:Layout power supplies of model \"%s\" are not merged." );
  /* Set the target library dependant functions. */
  switch(UtLib_V.Type) {
    case C_UtLib_TargetScLib:
      GV.pFunctionArgs      = ScGN_FunctionArgs;
      GV.pFunctionCells     = ScGN_FunctionCells;
      GV.pHaveBuff          = ScGN_HaveBuff;
      GV.pCreateNetlist     = ScGN_CreateNetlist;
      GV.pCreateNetlistComp = ScGN_CreateNetlistComp;
      GV.pCreateLayout      = ScGN_CreateLayout;
      GV.pCreateLayoutComp  = ScGN_CreateLayoutComp;
      break;
    case C_UtLib_TargetFpLib:
    default:
      GV.pFunctionArgs      = FpGN_FunctionArgs;
      GV.pFunctionCells     = FpGN_FunctionCells;
      GV.pHaveBuff          = FpGN_HaveBuff;
      GV.pCreateNetlist     = FpGN_CreateNetlist;
      GV.pCreateNetlistComp = FpGN_CreateNetlistComp;
      GV.pCreateLayout      = FpGN_CreateLayout;
      GV.pCreateLayoutComp  = FpGN_CreateLayoutComp;
      break;
  }
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "ShGN_llog2()"                                       | |%
 * %| \--------------------------------------------------------------------/ |%
 */

extern long  ShGN_llog2( value )
  long  value;
{ long  bit; for(bit = 1; (value >> bit); bit++); return(bit - 1); }


/*  Function  :  'FunctionCheck'                                            */
/* ------------------------------------------------------------------------ */
/*  Check if the value given as argument have a legal value. That is,
 *  a value corresponding to one of a boolean function.
 */

static long  FunctionCheck()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

        return( TRUE );

    default:

        return( FALSE );
  }
}


/*  Function  :  'ShGN_FunctionDrive'                                       */
/* ------------------------------------------------------------------------ */
/*  Return the maximal drive value associated with each boolean function.
 *  When no choice is possible the value is "1".
 */

extern long  ShGN_FunctionDrive()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:       return( C_GfpGN_DriveInv   );
    case C_GfpGN_FunctionBuff:      return( C_GfpGN_DriveBuff  );
    case C_GfpGN_FunctionNand2:     return( C_GfpGN_DriveNand2 );
    case C_GfpGN_FunctionNand3:     return( C_GfpGN_DriveNand3 );
    case C_GfpGN_FunctionAnd2:      return( C_GfpGN_DriveAnd2  );
    case C_GfpGN_FunctionAnd3:      return( C_GfpGN_DriveAnd3  );
    case C_GfpGN_FunctionNor2:      return( C_GfpGN_DriveNor2  );
    case C_GfpGN_FunctionNor3:      return( C_GfpGN_DriveNor3  );
    case C_GfpGN_FunctionOr2:       return( C_GfpGN_DriveOr2   );
    case C_GfpGN_FunctionOr3:       return( C_GfpGN_DriveOr3   );
    case C_GfpGN_FunctionXnor2:     return( C_GfpGN_DriveXnor2 );
    case C_GfpGN_FunctionXor2:      return( C_GfpGN_DriveXor2  );
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
    default:

      return(0);
  }
}


/*  Function  :  'ShGN_FunctionStr'                                         */
/* ------------------------------------------------------------------------ */
/*  Return the string used as root for building the model name.             */

extern char *ShGN_FunctionStr()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:       return( "inv"       );
    case C_GfpGN_FunctionBuff:      return( "buff"      );
    case C_GfpGN_FunctionNand2:     return( "nand2"     );
    case C_GfpGN_FunctionNand3:     return( "nand3"     );
    case C_GfpGN_FunctionAnd2:      return( "and2"      );
    case C_GfpGN_FunctionAnd3:      return( "and3"      );
    case C_GfpGN_FunctionNor2:      return( "nor2"      );
    case C_GfpGN_FunctionNor3:      return( "nor3"      );
    case C_GfpGN_FunctionOr2:       return( "or2"       );
    case C_GfpGN_FunctionOr3:       return( "or3"       );
    case C_GfpGN_FunctionXnor2:     return( "xnor2"     );
    case C_GfpGN_FunctionXor2:      return( "xor2"      );
    case C_GfpGN_FunctionNand2mask: return( "nand2mask" );
    case C_GfpGN_FunctionNor2mask:  return( "nor2mask"  );
    case C_GfpGN_FunctionXnor2mask: return( "xnor2mask" );
    case C_GfpGN_FunctionConst:     return( "const"     );
    case C_GfpGN_FunctionNbuse:     return( "nbuse"     );
    case C_GfpGN_FunctionBuse:      return( "buse"      );
    case C_GfpGN_FunctionNmux2cs:   return( "nmux2cs"   );
    case C_GfpGN_FunctionMux2cs:    return( "mux2cs"    );
    case C_GfpGN_FunctionMux3cd:    return( "mux3cd"    );
    case C_GfpGN_FunctionMux3cs:    return( "mux3cs"    );
    case C_GfpGN_FunctionMux4cs:    return( "mux4cs"    );
    case C_GfpGN_FunctionMux5cs:    return( "mux5cs"    );
    case C_GfpGN_FunctionMux6cs:    return( "mux6cs"    );
    case C_GfpGN_FunctionAdd2s:     return( "add2s"     );
    case C_GfpGN_FunctionNul:       return( "nul"       );
    case C_GfpGN_FunctionNpl1:      return( "npl1"      );
    case C_GfpGN_FunctionPl1:       return( "pl1"       );
    case C_GfpGN_FunctionPmsnw:     return( "pmsnw"     );
    case C_GfpGN_FunctionPmsnwq:    return( "pmsnwq"    );
    case C_GfpGN_FunctionPms:       return( "pms"       );
    case C_GfpGN_FunctionPmsq:      return( "pmsq"      );
    case C_GfpGN_FunctionPmst:      return( "pmst"      );
    case C_GfpGN_FunctionPmstq:     return( "pmstq"     );
    case C_GfpGN_FunctionPdffnw:    return( "pdffnw"    );
    case C_GfpGN_FunctionPdffnwq:   return( "pdffnwq"   );
    case C_GfpGN_FunctionPdff:      return( "pdff"      );
    case C_GfpGN_FunctionPdffr:     return( "pdffr"     );
    case C_GfpGN_FunctionPdffq:     return( "pdffq"     );
    case C_GfpGN_FunctionPdfft:     return( "pdfft"     );
    case C_GfpGN_FunctionPdffut:    return( "pdffut"    );
    case C_GfpGN_FunctionPdffrt:    return( "pdffrt"    );
    case C_GfpGN_FunctionPdfftq:    return( "pdfftq"    );
    default:                        return( "errorgen"  );
  }
}


/*  Function  :  'OperName'                                                 */
/* ------------------------------------------------------------------------ */
/*  Generate the complete model name of the operator to create. The fonction
 *  use all the significant arguments to build this name. It return a
 *  pointer to an internal static string.
 */

static char *OperName()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      return( UtBas_PrintF( "%s_%ldx%ldx%ldx%ld%c_cl",
                                 ShGN_FunctionStr(),
                                 GV.aWidthDP,
                                 GV.aWidthOp,
                                 GV.aSlice,
                                 GV.aDrive,
                               ((GV.aMSB0) ? 'm' : 'l')) );

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:

      return( UtBas_PrintF( "%s%s_%ldx%ldx%ld%c_cl",
                            ShGN_FunctionStr(),
                            GfpUT_XlToStr(GV.aXlMask,
                                         GV.aWidthDP,
                                         C_GfpUT_BaseHexadecimal),
                            GV.aWidthDP,
                            GV.aWidthOp,
                            GV.aSlice,
                          ((GV.aMSB0) ? 'm' : 'l')) );

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdfftq:

      return( UtBas_PrintF( "%s_%ldx%ldx%ld%c_cl",
                            ShGN_FunctionStr(),
                            GV.aWidthDP,
                            GV.aWidthOp,
                            GV.aSlice,
                          ((GV.aMSB0) ? 'm' : 'l')) );

    default:

      return( UtBas_PrintF( "errormodel" ) );
  }
}


/*  Function  :  'HeaderVHDL'                                               */
/* ------------------------------------------------------------------------ */
/*  Write comment header in the file.                                       */

static void  HeaderVHDL()
{
  char *String;
  long  Time, Length;


  /* Print the standart distribution header. */
  fprintf(GV.PtFile,
"-- ###-------------------------------------------------------------###\n"
          );
  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );
  fprintf( GV.PtFile,
"-- #          Alliance CAD System 3.2b                                #\n"
          );
  fprintf( GV.PtFile,
"-- #             FpGen Package 1.15                                  #\n"
          );
  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );
  fprintf( GV.PtFile,
"-- #  Copyright(c) 93-97, MASI, CAO-VLSI Team                        #\n"
          );
  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );
  fprintf( GV.PtFile,
"-- #  Author  :              Jean-Paul CHAPUT                        #\n"
          );
  fprintf( GV.PtFile,
"-- #  E-mail  : alliance-support@asim.lip6.fr                        #\n"
          );
  fprintf( GV.PtFile,
"-- # *************************************************************** #\n"
          );


  /* Print the generator name. */
  String = UtBas_PrintF( "\"%sg\"", ShGN_FunctionStr() );
  fprintf( GV.PtFile, "-- #  Generator   : %-47s  #\n", String );


  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );


  /* Print name of the generated file. */
  String = UtBas_PrintF( "\"%s.vbe\".", GV.aModelName );
  fprintf( GV.PtFile, "-- #  File  :  %-52s  #\n", String );


  fprintf( GV.PtFile,
"-- # --------------------------------------------------------------- #\n"
          );


  /* Print the date of generation. */
  Time = time((time_t*)NULL);
  String = UtBas_PrintF( "%s", asctime( localtime( &Time ) ) );
  Length = strlen( String );
  String[ Length - 1 ] = (char)0;
  fprintf(GV.PtFile, "-- #            Date : %-44s  #\n", String);


  /* Print the call used to generate the model. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      String = UtBas_PrintF(
                 "\"%sg %ld %ld %ld %ld%s -vhdl \\",
                 ShGN_FunctionStr(),
                 GV.aWidthDP,
                 GV.aWidthOp,
                 GV.aSlice,
                 GV.aDrive,
                (GV.aMSB0) ? " -msb0"  : "" );
      fprintf( GV.PtFile, "-- #  Generator Call : %-44s  #\n", String );

      String = UtBas_PrintF( "-o %s\"", GV.aModelName );
      fprintf( GV.PtFile, "-- # %62s  #\n", String );

      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:

      String = UtBas_PrintF(
                 "\"%sg %ld %ld %ld \\",
                 ShGN_FunctionStr(),
                 GV.aWidthDP,
                 GV.aWidthOp,
                 GV.aSlice );
      fprintf( GV.PtFile, "-- #  Generator Call : %-44s  #\n", String );

      String = UtBas_PrintF(
                 "0x%s \\", GfpUT_XlToStr( GV.aXlMask,
                                           GV.aWidthOp,
                                           C_GfpUT_BaseHexadecimal ) );
      fprintf( GV.PtFile, "-- # %62s  #\n", String );

      String = UtBas_PrintF(
                 "%s -vhdl -o %s",
                (GV.aMSB0) ? " -msb0"  : "",
                 GV.aModelName );
      fprintf( GV.PtFile, "-- # %62s  #\n", String );

      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdfftq:

      String = UtBas_PrintF(
                 "\"%sg %ld %ld %ld %s -vhdl \\",
                 ShGN_FunctionStr(),
                 GV.aWidthDP,
                 GV.aWidthOp,
                 GV.aSlice,
                (GV.aMSB0) ? " -msb0"  : "");
      fprintf( GV.PtFile, "-- #  Generator Call : %-44s  #\n", String);

      String = UtBas_PrintF( "-o %s\"", GV.aModelName );
      fprintf( GV.PtFile, "-- # %62s  #\n", String );

      break;
  }


  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );
  fprintf( GV.PtFile,
"-- ###-------------------------------------------------------------###\n"
          );
}


/*  Function  :  'EntityVHDL'                                               */
/* ------------------------------------------------------------------------ */
/*  Create the ENTITY description of the operator. Essentially PORT part.
 */

static void  EntityVHDL()
{
  char *BusWide;
  long  Slice, fHaveBuff;


  /* Compute the VHDL full bus wide string. */
  BusWide = mc_VHDL_FULL_WIDE();


  fprintf( GV.PtFile, "\nENTITY %s IS\n\tPORT(\n", GV.aModelName );

  /* Controls Terminals. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionConst:

      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "msk" );
      break;

    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "sel" );
      break;

    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:

      fprintf( GV.PtFile,
               "%14s : in    BIT_VECTOR(1 downto 0);\n", "sel" );
      break;

    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile,
               "%14s : in    BIT_VECTOR(2 downto 0);\n", "sel" );
      break;

    case C_GfpGN_FunctionAdd2s:

      fprintf( GV.PtFile, "%14s :   out BIT;\n", "cout" );
      fprintf( GV.PtFile, "%14s :   out BIT;\n", "over" );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "ck" );
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "scin" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "test" );

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "wen" );

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "ckm" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "cks" );
      break;

    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "scin" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "test" );

    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "wen" );

    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "ck" );
      break;

    case C_GfpGN_FunctionPdffrt:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "scin" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "test" );

    case C_GfpGN_FunctionPdffr:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "reset" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "wen" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "ck" );
      break;
  }

  /* Terminals of input data buses. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionConst:

      break;

    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i5", BusWide );

    case C_GfpGN_FunctionMux5cs:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i4", BusWide );

    case C_GfpGN_FunctionMux4cs:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i3", BusWide );

    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i2", BusWide );

    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionAdd2s:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i1", BusWide );

    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionNul:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i0", BusWide );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "d", BusWide );
      break;
  }

  /* Terminals of output data buses. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionOr2:

      fprintf( GV.PtFile, "%14s :   out BIT_VECTOR%s;\n",  "o", BusWide );

      if ( UtTec_Family() == C_UtTec_FamilyCMOS )
      {
        fprintf( GV.PtFile, "%14s : inout BIT_VECTOR%s;\n", "no", BusWide );
      }

      break;

    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:

      fprintf( GV.PtFile, "%14s :   out BIT_VECTOR%s;\n", "o", BusWide );
      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:

      fprintf( GV.PtFile,
               "%14s :   out MUX_VECTOR%s BUS;\n", "o", BusWide );
      break;

    case C_GfpGN_FunctionNul:

      fprintf( GV.PtFile, "%14s :   out BIT;\n", "o" );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "%14s : inout BIT_VECTOR%s;\n", "q", BusWide );
      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:

      fprintf( GV.PtFile, "%14s : inout BIT_VECTOR%s;\n",  "q", BusWide );
      fprintf( GV.PtFile, "%14s :   out BIT_VECTOR%s;\n", "nq", BusWide );
      break;
  }


  /* Select the buffer model name. */
  /* (NULL if no need of a buffer) */
  fHaveBuff = GV.pHaveBuff();

  /* Power supplies terminals. */
  if (GV.aPower)
  {
    fprintf(GV.PtFile, "%14s : in    BIT;\n", "vdd");
    fprintf(GV.PtFile, "%14s : in    BIT\n" , "vss");
  }
  else
  {
    for(Slice = 0 ;
        Slice < GV.aWidthDP + ((fHaveBuff) ? 2 : 0) ;
        Slice++)
      fprintf(GV.PtFile, "%14s : in    BIT;\n", mc_POWER_NAME(Slice));

    /* No semi-colon at the end of the last connector.
     * This explain why we separate the last iteration of the loop.
     */
    fprintf(GV.PtFile, "%14s : in    BIT\n", mc_POWER_NAME(Slice));
  }


  /* *** End of interface description *** */
  fprintf(GV.PtFile, "\t);\nEND %s;\n\n\n", GV.aModelName);
}


/*  Function  :  'SignalVHDL'                                               */
/* ------------------------------------------------------------------------ */
/*  Internal signals declaration.                                           */

static void  SignalVHDL()
{
  char *BusWide, *MemWide;


  BusWide = mc_VHDL_FULL_WIDE();
  MemWide = mc_VHDL_MEMM_WIDE( GV.aSlice + GV.aWidthOp - 1,
                               GV.aSlice );


  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionNul:

      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile,
               "\tSIGNAL %20s : BIT_VECTOR%s;\n", "otmp", BusWide );
      break;

    case C_GfpGN_FunctionAdd2s:

      {
        char *CryBusWide;

        CryBusWide = UtBas_PrintF(
                       "(%ld downto %ld)",
                       GV.aWidthOp + GV.aSlice,
                       GV.aSlice );
        fprintf( GV.PtFile,
                 "\tSIGNAL %20s : BIT_VECTOR%s;\n", "cry", CryBusWide);
      }
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "wmd" );
      fprintf( GV.PtFile,
               "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "memm", MemWide );
      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:

      fprintf( GV.PtFile,
               "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "memm", MemWide );

    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:

      fprintf( GV.PtFile,
               "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "mems", MemWide );
      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "wmd");
      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "ws" );
      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT_VECTOR%s;\n", "dwen", MemWide);
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      fprintf(GV.PtFile, "\tSIGNAL %20s : BIT;\n"         , "wmd"  );
      fprintf(GV.PtFile, "\tSIGNAL %20s : BIT;\n"         , "ws"   );
      fprintf(GV.PtFile, "\tSIGNAL %20s : BIT_VECTOR%s;\n", "dwen" , MemWide);
      fprintf(GV.PtFile, "\tSIGNAL %20s : BIT_VECTOR%s;\n", "dtest", MemWide);
      fprintf(GV.PtFile,
        "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "memm", MemWide );
      fprintf(GV.PtFile,
        "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "mems", MemWide );

      break;
  }
  fprintf( GV.PtFile, "\n" );
}


/*  Function  :  'DataFlowBool'                                   */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for booleans operators.                   */

static void  DataFlowBool()
{
	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionInv   :
			fprintf(GV.PtFile, "\to <= not i0;\n\n");
			break;
		case C_GfpGN_FunctionBuff  :
			fprintf(GV.PtFile, "\to <= i0;\n\n");
			break;
		case C_GfpGN_FunctionNand2 :
			fprintf(GV.PtFile, "\to <= not (i0 and i1);\n\n");
			break;
		case C_GfpGN_FunctionNand3 :
			fprintf(GV.PtFile, "\to <= not (i0 and i1 and i2);\n\n");
			break;
		case C_GfpGN_FunctionAnd2  :
			fprintf(GV.PtFile, "\t o <=      i0 and i1 ;\n"  );

            if ( UtTec_Family() == C_UtTec_FamilyCMOS )
            {
			  fprintf(GV.PtFile, "\tno <= not (i0 and i1);\n\n");
			}

			break;
		case C_GfpGN_FunctionAnd3  :
			fprintf(GV.PtFile, "\t o <=      i0 and i1 and i2 ;\n"  );

            if ( UtTec_Family() == C_UtTec_FamilyCMOS )
            {
			  fprintf(GV.PtFile, "\tno <= not (i0 and i1 and i2);\n\n");
			}

			break;
		case C_GfpGN_FunctionNor2  :
			fprintf(GV.PtFile, "\to <= not (i0 or i1);\n\n");
			break;
		case C_GfpGN_FunctionNor3  :
			fprintf(GV.PtFile, "\to <= not (i0 or i1 or i2);\n\n");
			break;
		case C_GfpGN_FunctionOr2   :
			fprintf(GV.PtFile, "\t o <=      i0 or i1 ;\n"  );

            if ( UtTec_Family() == C_UtTec_FamilyCMOS )
            {
			  fprintf(GV.PtFile, "\tno <= not (i0 or i1);\n\n");
			}

			break;
		case C_GfpGN_FunctionOr3   :
			fprintf(GV.PtFile, "\t o <=      i0 or i1 or i2 ;\n"  );

            if ( UtTec_Family() == C_UtTec_FamilyCMOS )
            {
			  fprintf(GV.PtFile, "\tno <= not (i0 or i1 or i2);\n\n");
			}

			break;
		case C_GfpGN_FunctionXnor2 :
			fprintf(GV.PtFile, "\to <= not (i0 xor i1);\n\n");
			break;
		case C_GfpGN_FunctionXor2  :
			fprintf(GV.PtFile, "\to <= i0 xor i1;\n\n");
			break;
	}
}


/*  Function  :  'DataFlowMask'                                   */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for controllable mask operators.          */

static void  DataFlowMask()
{
	fprintf(GV.PtFile, "\tWITH  msk  SELECT\n");
	fprintf(GV.PtFile, "\t\totmp <=  i0          WHEN \'0\',\n");

	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionNand2mask:
			fprintf(GV.PtFile,
					"\t\t\t\t(i0 and %s) WHEN \'1\';\n",
					  GfpUT_XlToStrVHDL(GV.aXlMask ,
									   GV.aWidthOp));
			break;
		case C_GfpGN_FunctionNor2mask:
			fprintf(GV.PtFile,
					"\t\t\t\t(i0  or %s) WHEN \'1\';\n",
					 GfpUT_XlToStrVHDL(GV.aXlMask ,
									   GV.aWidthOp));
			break;
		case C_GfpGN_FunctionXnor2mask:
			fprintf(GV.PtFile,
					"\t\t\t\t(i0 xor %s) WHEN \'1\';\n",
					 GfpUT_XlToStrVHDL(GV.aXlMask ,
									   GV.aWidthOp));
			break;
	}

	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionNand2mask:
		case C_GfpGN_FunctionNor2mask:
		case C_GfpGN_FunctionXnor2mask:
			fprintf(GV.PtFile, "\n\t\to <= not otmp;\n\n");
			break;
		default:
			fprintf(GV.PtFile, "\n\to <= otmp;\n\n");
			break;
	}
}


/*  Function  :  'DataFlowConst'                                  */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for constants operators.                  */

static void  DataFlowConst()
{
	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionConst:
			fprintf(GV.PtFile, "\to <= %s;\n\n",
				  GfpUT_XlToStrVHDL(GV.aXlMask ,
								   GV.aWidthOp));
			break;
	}
}


/*  Function  :  'DataFlowBuse'                                   */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for tristate (bus) operators.             */

static void  DataFlowBuse()
{
	fprintf(GV.PtFile, "\tlwts:BLOCK(sel = '1')\n");
	fprintf(GV.PtFile, "\tBEGIN\n"               );

	/* Specific Data-Flow equations. */
	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionNbuse:
			fprintf(GV.PtFile, "\t\to <= GUARDED not i0;\n");
			break;
		case C_GfpGN_FunctionBuse:
			fprintf(GV.PtFile, "\t\to <= GUARDED i0;\n");
			break;
	}

	fprintf(GV.PtFile, "\tEND BLOCK lwts;\n\n");
}


/*  Function  :  'DataFlowMux'                                              */
/* ------------------------------------------------------------------------ */
/*  Data-Flow equations for multiplexer operators.                          */

static void  DataFlowMux()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:

      fprintf( GV.PtFile,     "\tWITH  sel  SELECT\n" );
      fprintf( GV.PtFile,     "\t\totmp <= "          );
      fprintf( GV.PtFile,         "i0  WHEN \'0\',\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN \'1\';\n" );
      break;

    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:

      fprintf( GV.PtFile,       "\tWITH  sel  SELECT\n" );
      fprintf( GV.PtFile,     "\t\totmp <= "            );
      fprintf( GV.PtFile,         "i0  WHEN  \"00\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN  \"01\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti2  WHEN  \"10\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"11\";\n" );
      break;

    case C_GfpGN_FunctionMux4cs:

      fprintf( GV.PtFile,       "\tWITH  sel  SELECT\n" );
      fprintf( GV.PtFile,     "\t\totmp <= "            );
      fprintf( GV.PtFile,         "i0  WHEN  \"00\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN  \"01\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti2  WHEN  \"10\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti3  WHEN  \"11\";\n" );
      break;

    case C_GfpGN_FunctionMux5cs:

      fprintf( GV.PtFile,       "\tWITH  sel  SELECT\n"  );
      fprintf( GV.PtFile,     "\t\totmp <= "             );
      fprintf( GV.PtFile,         "i0  WHEN  \"000\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN  \"001\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti2  WHEN  \"010\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti3  WHEN  \"011\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti4  WHEN  \"100\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"101\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"110\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"111\";\n" );
      break;

    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile,       "\tWITH  sel  SELECT\n"  );
      fprintf( GV.PtFile,     "\t\totmp <= "             );
      fprintf( GV.PtFile,         "i0  WHEN  \"000\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN  \"001\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti2  WHEN  \"010\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti3  WHEN  \"011\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti4  WHEN  \"100\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"110\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti5  WHEN  \"101\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"111\";\n" );
      break;
  }

  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNmux2cs:

      fprintf( GV.PtFile, "\t\to <= not otmp;\n\n" );
      break;

    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile, "\t\to <= otmp;\n\n" );
      break;
  }
}


/*  Function  :  'DataFlowAdd'                                    */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for adder operators.                      */

static void  DataFlowAdd()
{
	char  CryBusWide1[IC_GfpGN_BusWideMaxLen];
	char  CryBusWide2[IC_GfpGN_BusWideMaxLen];

	sprintf(CryBusWide1,
			"(%ld downto %ld)", GV.aWidthOp
							   +GV.aSlice  -1,
							    GV.aSlice    );
	sprintf(CryBusWide2,
			"(%ld downto %ld)", GV.aWidthOp
							   +GV.aSlice    ,
							    GV.aSlice  +1);

	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionAdd2s:
			fprintf(GV.PtFile, "\tcry(0)           <= \'0\';\n");
			fprintf(GV.PtFile, "\tcry%s <= (i0 and i1)\n",
												   CryBusWide2);
			fprintf(GV.PtFile, "\t\t\t\t\t or (i0 and cry%s)\n",
														 CryBusWide1);
			fprintf(GV.PtFile, "\t\t\t\t\t or (i1 and cry%s);\n\n",
														    CryBusWide1);
			fprintf(GV.PtFile, "\to        <= ");
			fprintf(GV.PtFile, "i0 xor i1 xor cry%s;\n\n", CryBusWide1);
			fprintf(GV.PtFile, "\tcout     <= cry(%ld);\n",
											   GV.aWidthOp
											  +GV.aSlice  );
			fprintf(GV.PtFile, "\tover     <= cry(%ld)",
											   GV.aWidthOp
										      +GV.aSlice  );
			fprintf(GV.PtFile,          " xor cry(%ld);\n\n",
											   GV.aWidthOp-1
										      +GV.aSlice    );
			break;
	}
}


/*  Function  :  'DataFlowComp'                                   */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for comparators operators.                */

static void  DataFlowComp()
{
	long *XlNul, XlIndex;

	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionNul:
			XlNul = GfpUT_XlAlloc();

			for(XlIndex=0; XlIndex<C_GfpUT_XLSIZE; XlIndex++)
				GfpUT_XlSetLong(XlNul, XlIndex, 0L);

			fprintf(GV.PtFile,
					"\to <= \'1\' WHEN (i0 = %s) ELSE \'0\';\n\n",
						GfpUT_XlToStrVHDL(XlNul,GV.aWidthOp));

			GfpUT_XlFree(XlNul);
			break;
	}
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "DataFlowMem()"                                      | |%
 * %| \--------------------------------------------------------------------/ |%
 */

static void  DataFlowMem()
{
  long  MemFlag;
  char *MemMast, *MemSlav, *Invert, *Direct, *sDin;

  MemFlag = 0L;
  sDin     = "dwen";
  /* Select the name of master and slave registers. */
  switch(GV.aFunction) {
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
      MemFlag |= IF_GfpGN_MemMaster;
      MemMast  = "memm";
      MemSlav  = "memm";
      sDin     = "d";
      break;
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
      MemFlag |= IF_GfpGN_MemScan;
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
      MemFlag |=  IF_GfpGN_MemMaster | IF_GfpGN_MemSlave;
      MemMast  = "memm";
      MemSlav  = "mems";
      break;
  }
  /* Set the reset flag. */
  switch(GV.aFunction) {
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfftq:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdffut:
      break;
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdffr:
      MemFlag |=  IF_GfpGN_MemReset;
      break;
  }
  /* Select the polarity for slave and outputs. */
  switch(GV.aFunction) {
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
      Invert = "not";
      Direct = "   ";
      break;
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfftq:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
      Invert = "   ";
      Direct = "not";
      break;
  }

  /* Equations for the master/slave control signal. */
  switch(GV.aFunction) {
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
      fprintf(GV.PtFile, "   wmd <=     ck;\n");
      break;
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
      fprintf(GV.PtFile, "   wmd <=     ckm;\n");
      fprintf(GV.PtFile, "   ws  <=     cks;\n");
      break;
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdfftq:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdffr:
      fprintf(GV.PtFile, "   wmd <=     ck;\n");
      fprintf(GV.PtFile, "   ws  <= not ck;\n");
      break;
    case C_GfpGN_FunctionPdffut:
      fprintf(GV.PtFile, "   wmd <= not      ck ;\n");
      fprintf(GV.PtFile, "   ws  <= not (not ck);\n");
      break;
  }
  fprintf(GV.PtFile, "\n");
  /* Equations for the mux control signals. */
  switch(GV.aFunction) {
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
      break;
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
      fprintf(GV.PtFile, "   WITH test SELECT\n" );
      fprintf(GV.PtFile, "     dtest <= dwen WHEN \'0\',\n");
      fprintf(GV.PtFile, "     q%s & scin ",
        mc_VHDL_WIDE(GV.aWidthOp - 2, 0));
      fprintf(GV.PtFile,                    "WHEN \'1\';\n");
      sDin = "dtest";
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
      fprintf(GV.PtFile, "   WITH wen SELECT\n" );
      fprintf(GV.PtFile, "     dwen  <=    d WHEN \'1\',\n");
      fprintf(GV.PtFile, "                 q WHEN \'0\';\n");
      break;
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
      break;
  }
  fprintf(GV.PtFile, "\n");

  /* Write in slave register (MS and DFF only). */
  if (   (GV.aFunction != C_GfpGN_FunctionPdffnw)
      && (GV.aFunction != C_GfpGN_FunctionPdffnwq)) {
    if (MemFlag & IF_GfpGN_MemReset) {
      long *XlReset, i;

      XlReset = GfpUT_XlAlloc();
      for(i = 0; i <= GV.aWidthOp / 32; i++)
        GfpUT_XlSetLong(XlReset, i, 0x00000000);
      fprintf(GV.PtFile, "   lreset:BLOCK(reset = \'1\') BEGIN ");
      fprintf(GV.PtFile, "%s <= GUARDED %s %s; ", MemMast, Invert, 
              GfpUT_XlToStrVHDL( XlReset,  GV.aWidthOp));
      fprintf(GV.PtFile, "END BLOCK lreset;\n");
      GfpUT_XlFree(XlReset);
    }
    /* Write in master registers. */
    fprintf( GV.PtFile, "\tlwmd:BLOCK(wmd = \'1\') BEGIN " );
    fprintf( GV.PtFile, "%s <= GUARDED %s %s;    ", MemMast, Invert, sDin );
    fprintf( GV.PtFile, "END BLOCK lwmd;\n" );
    /* Write in slave register (MS and DFF only). */
    if (MemFlag & IF_GfpGN_MemSlave) {
      fprintf( GV.PtFile, "\tlws :BLOCK(ws = \'1\') BEGIN " );
      fprintf( GV.PtFile, "%s <= GUARDED     %s; ", MemSlav, MemMast );
      fprintf( GV.PtFile, "END BLOCK lws;\n" );
    }
    fprintf( GV.PtFile, "\n" );
  }
  else
  {
    fprintf( GV.PtFile, "\tlck:BLOCK(((not ck) and not(ck\'STABLE)) = \'1\')\n");
    fprintf( GV.PtFile, "\tBEGIN\n" );
    fprintf( GV.PtFile, "\t\t%s <= GUARDED %s d;\n ", Invert, MemSlav );
    fprintf( GV.PtFile, "\tEND BLOCK lck;\n" );
    fprintf( GV.PtFile, "\n" );
  }
  /* Affect the outputs. */
  switch(GV.aFunction) {
    case C_GfpGN_FunctionNpl1:
      fprintf( GV.PtFile, "\tq <= not %s;\n", MemSlav );
      break;
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
      fprintf( GV.PtFile, "\tnq <= %s %s;\n", Direct, MemSlav );
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfftq:
      fprintf( GV.PtFile, "\t q <= %s %s;\n", Invert, MemSlav );
      fprintf(GV.PtFile, "\n");
      break;
  }
  fprintf(GV.PtFile, "\n");
}



/*  Function  :  'ArchiVHDL'                                               */
/* ----------------------------------------------------------------------- */
/*  Create the ARCHITECTURE description of the operator. This description
 *  use the VHDL data-flow subset.
 */

static void  ArchiVHDL()
{
  long  Slice, fHaveBuff;


  /* Begin of behavioral view. */
  fprintf(GV.PtFile,
      "ARCHITECTURE VBE OF %s IS\n\n",
                     GV.aModelName);

  /* Declaration of internals signals. */
  SignalVHDL();

  fprintf(GV.PtFile, "BEGIN\n\n");


  /* Data-Flow equations. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      DataFlowBool();
      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

      DataFlowMask();
      break;

    case C_GfpGN_FunctionConst:

      DataFlowConst();
      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:

      DataFlowBuse();
      break;

    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      DataFlowMux();
      break;

    case C_GfpGN_FunctionAdd2s:

      DataFlowAdd();
      break;

    case C_GfpGN_FunctionNul:

      DataFlowComp();
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      DataFlowMem();
      break;
  }


  /* Check power supplies connections. */
  fprintf(GV.PtFile, "\t-- ****************>");
  fprintf(GV.PtFile, " Supplies Check "      );
  fprintf(GV.PtFile, "<****************\n\n" );

  /* Select the buffer model name. */
  /* (NULL if no need of a buffer) */
  fHaveBuff = GV.pHaveBuff();

  /* Power supplies terminals. */
  if (GV.aPower)
  {
    fprintf(GV.PtFile, "\tASSERT(vdd = '1')\n");
    fprintf(GV.PtFile,
            "\t\tREPORT \"Power supply is missing on vdd of model %s.\"\n",
             GV.aModelName);
    fprintf(GV.PtFile, "\t\tSEVERITY WARNING;\n\n");

    fprintf(GV.PtFile, "\tASSERT(vss = '0')\n");
    fprintf(GV.PtFile,
            "\t\tREPORT \"Power supply is missing on vss of model %s.\"\n",
             GV.aModelName);
    fprintf(GV.PtFile, "\t\tSEVERITY WARNING;\n\n");
  }
  else
    for(Slice = 0 ;
        Slice < GV.aWidthDP + ((fHaveBuff) ? 2 : 0) ;
        Slice++)
    {
      fprintf(GV.PtFile, "\tASSERT(%s = '%c')\n",
              mc_POWER_NAME(Slice),
              mc_SliceToPowerType(Slice) ?   '1' :   '0');
      fprintf(GV.PtFile,
              "\t\tREPORT \"Power supply is missing on %s of model %s.\"\n",
               mc_POWER_NAME(Slice),
               GV.aModelName);
      fprintf(GV.PtFile, "\t\tSEVERITY WARNING;\n\n");
    }


  /* End of behavioral view. */
  fprintf(GV.PtFile, "END VBE;\n");
}


/*  Function  :  'CreateVHDL'                                     */
/* -------------------------------------------------------------- */
/*  Generation of the VHDL behavioral view.                       */

static void  CreateVHDL()
{
	/* Creating the file holding behavioral view. */
	if ((GV.PtFile=mbkfopen(GV.aModelName,
											  "vbe",
										 WRITE_TEXT)) == (FILE*)NULL)
		GfpUT_PrintError(GV.ModuleId,
					   IE_GfpGN_OPENVBE,
						  ShGN_FunctionStr(),
					 GV.aModelName);

	/* Write an identification header. */
	HeaderVHDL();

	/* Write the VHDL file. */
	EntityVHDL();
	ArchiVHDL();

	/* Close the file holding the behavioral view. */
	fclose(GV.PtFile);
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "GfpGN_GenericDispatch()"                            | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Interface of the generic operator generator. The argument list change  |%
 * %| acording to the asked operator. The complete list of functions         |%
 * %| "GenericDispatch" can emulate is given in the "GenericGN.h" include    |%
 * %| file.                                                                  |%
 */

extern char *GfpGN_GenericDispatch __FPV(( long aFunction, ... ))
{
  static    long  FirstCall = FALSE;
         va_list  ArgList;
            long  aVHDL;
            long  aNetlist;
            long  aLayout;
            long  aIcon;
            long  aPower;

  /* Do specific initialisations at first generator call. */
  if (!FirstCall) { Init(); FirstCall = TRUE; }
  /* Set the mark for all the strings allocated at this call. */
  GV.PrintFMark = UtBas_PrintFMark();

  __ANSI_C( va_start( ArgList, aFunction ); )
    __KR_C( va_start( ArgList ); )
  /* Get the three first arguments.           */
  /* (not depending of the emulated function) */
  __ANSI_C( GV.aFunction  = aFunction; )
    __KR_C( GV.aFunction  = va_arg(ArgList, long ); )
            GV.aModelName = va_arg(ArgList, char*);
  /* Check the asked function. */
  if (!FunctionCheck())
    GfpUT_PrintError(
      GV.ModuleId, IE_GfpGN_FUNCTION, ShGN_FunctionStr(), GV.aFunction );
  /* Get the aguments to read, acording to the function. */
  GV.pFunctionArgs( ArgList );
  if (GV.aViews == FALSE)
    GfpUT_PrintError( GV.ModuleId, IE_GfpGN_NOVIEW, ShGN_FunctionStr() ); 
  /* Init the "LeafCellIndex". */
  GV.pFunctionCells();
  /* If necessary, generate the model name. */
  if (GV.aModelName == (char*)NULL) GV.aModelName = OperName();
  /* All the needed arguments have been read. */
  va_end( ArgList );

  /* Generation of behaviorial view. */
  if (GV.aViews & IF_GfpGN_ViewVHDL) CreateVHDL();
  /* Generation of netlist view. */
  switch(UtTec_Family()) {
    case C_UtTec_FamilyAsGa: break;
    case C_UtTec_FamilyCMOS:
    default:
      if (GV.aViews & IF_GfpGN_ViewNetlist) {
        switch(GV.aFunction) {
          case C_GfpGN_FunctionNul: GV.pCreateNetlistComp(); break;
          default:                  GV.pCreateNetlist();     break;
        }
      }
      break;
  }
  /* Generation of layout view. */
  if (GV.aViews & IF_GfpGN_ViewLayout)
    switch(GV.aFunction) {
      case C_GfpGN_FunctionNul: GV.pCreateLayoutComp(); break;
      default:                  GV.pCreateLayout();     break;
    }
  /* Generation of icon.                */
  /* if (GV.aViews & IF_GfpGN_ViewIcon) */
  /* Free all string allocated. */
  UtBas_PrintFFree( GV.PrintFMark );
  /* Return name of the generated model. */
  return( GV.aModelName );
}
