/*
 * $Author: ac $
 * $Id: bflchrono.c,v 1.1 1999/10/21 12:51:49 ac Exp $
 * $Date: 1999/10/21 12:51:49 $
 */
/*
 * $Log: bflchrono.c,v $
 * Revision 1.1  1999/10/21 12:51:49  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include BFL_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

bflchrono_t *
bflchrononew(void)
{
    bflchrono_t *chrono;

    if ((chrono = (bflchrono_t *) malloc(sizeof(bflchrono_t))) == NULL)
	bflsyserror(1, "bflchrononew");
    bflchronoreset(chrono);
    bflchronostart(chrono);
    return chrono;
}

void
bflchronofree(bflchrono_t * chrono)
{
    free(chrono);
}

/*
 */
void
bflchronoreset(bflchrono_t * chrono)
{
    struct rusage rusage;
    
    getrusage(RUSAGE_SELF, &rusage);
    chrono->stop.ru_utime = chrono->lap.ru_utime = chrono->start.ru_utime = rusage.ru_utime;
    chrono->stop.ru_stime = chrono->lap.ru_stime = chrono->start.ru_stime = rusage.ru_stime;
}

void
bflchronostart(bflchrono_t * chrono)
{
    struct rusage rusage;

    getrusage(RUSAGE_SELF, &rusage);

    if (chrono->stop.ru_utime.tv_sec != 0) {
	bfltime_t diff;

	diff.ru_utime.tv_sec  = rusage.ru_utime.tv_sec  - chrono->stop.ru_utime.tv_sec ;
	diff.ru_utime.tv_usec = rusage.ru_utime.tv_usec - chrono->stop.ru_utime.tv_usec;
	diff.ru_stime.tv_sec  = rusage.ru_stime.tv_sec  - chrono->stop.ru_stime.tv_sec ;
	diff.ru_stime.tv_usec = rusage.ru_stime.tv_usec - chrono->stop.ru_stime.tv_usec;

	chrono->start.ru_utime.tv_sec  += diff.ru_utime.tv_sec ;
	chrono->start.ru_utime.tv_usec += diff.ru_utime.tv_usec;
	chrono->start.ru_stime.tv_sec  += diff.ru_stime.tv_sec ;
	chrono->start.ru_stime.tv_usec += diff.ru_stime.tv_usec;
 
	chrono->stop.ru_utime.tv_sec  = 0;
	chrono->stop.ru_utime.tv_usec = 0;
	chrono->stop.ru_stime.tv_sec  = 0;
	chrono->stop.ru_stime.tv_usec = 0;

	chrono->lap.ru_utime.tv_sec  += diff.ru_utime.tv_sec ;
	chrono->lap.ru_utime.tv_usec += diff.ru_utime.tv_usec;
	chrono->lap.ru_stime.tv_sec  += diff.ru_stime.tv_sec ;
	chrono->lap.ru_stime.tv_usec += diff.ru_stime.tv_usec;
    } else {
	chrono->lap.ru_utime.tv_sec  = chrono->start.ru_utime.tv_sec  = rusage.ru_utime.tv_sec ;
	chrono->lap.ru_utime.tv_usec = chrono->start.ru_utime.tv_usec = rusage.ru_utime.tv_usec;
	chrono->lap.ru_stime.tv_sec  = chrono->start.ru_stime.tv_sec  = rusage.ru_stime.tv_sec ;
	chrono->lap.ru_stime.tv_usec = chrono->start.ru_stime.tv_usec = rusage.ru_stime.tv_usec;
    }
}

void
bflchronostop( bflchrono_t *chrono )
{
    struct rusage rusage;
    
    if (chrono->stop.ru_utime.tv_sec != 0) {
	getrusage(RUSAGE_SELF, &rusage);

	chrono->stop.ru_utime.tv_sec  = rusage.ru_utime.tv_sec ;
	chrono->stop.ru_utime.tv_usec = rusage.ru_utime.tv_usec;
	chrono->stop.ru_stime.tv_sec  = rusage.ru_stime.tv_sec ;
	chrono->stop.ru_stime.tv_usec = rusage.ru_stime.tv_usec;
    }
}

bfltime_t
bflchronoelapsed( bflchrono_t *chrono )
{
    struct rusage rusage;
    bfltime_t time;

    if (chrono->stop.ru_utime.tv_sec != 0) {
	time.ru_utime.tv_sec  = chrono->stop.ru_utime.tv_sec ; 
	time.ru_utime.tv_usec = chrono->stop.ru_utime.tv_usec; 
        time.ru_stime.tv_sec  = chrono->stop.ru_stime.tv_sec ; 
        time.ru_stime.tv_usec = chrono->stop.ru_stime.tv_usec; 
    } else {
	getrusage(RUSAGE_SELF, &rusage);

	time.ru_utime.tv_sec  = rusage.ru_utime.tv_sec ;
	time.ru_utime.tv_usec = rusage.ru_utime.tv_usec;
	time.ru_stime.tv_sec  = rusage.ru_stime.tv_sec ;
	time.ru_stime.tv_usec = rusage.ru_stime.tv_usec;
    }

    time.ru_utime.tv_sec  -= chrono->start.ru_utime.tv_sec ;
    time.ru_utime.tv_usec -= chrono->start.ru_utime.tv_usec;
    time.ru_stime.tv_sec  -= chrono->start.ru_stime.tv_sec ;
    time.ru_stime.tv_usec -= chrono->start.ru_stime.tv_usec;

    return time;
}

bfltime_t
bflchronolap( bflchrono_t *chrono)
{
    struct rusage rusage;
    bfltime_t diff;

    if (chrono->stop.ru_utime.tv_sec != 0) {
	diff.ru_utime.tv_sec  = chrono->stop.ru_utime.tv_sec ; 
	diff.ru_utime.tv_usec = chrono->stop.ru_utime.tv_usec; 
        diff.ru_stime.tv_sec  = chrono->stop.ru_stime.tv_sec ; 
        diff.ru_stime.tv_usec = chrono->stop.ru_stime.tv_usec; 
    } else {
	getrusage(RUSAGE_SELF, &rusage);

	diff.ru_utime.tv_sec  = rusage.ru_utime.tv_sec ;
	diff.ru_utime.tv_usec = rusage.ru_utime.tv_usec;
	diff.ru_stime.tv_sec  = rusage.ru_stime.tv_sec ;
	diff.ru_stime.tv_usec = rusage.ru_stime.tv_usec;
    }

    diff.ru_utime.tv_sec  -= chrono->lap.ru_utime.tv_sec ;
    diff.ru_utime.tv_usec -= chrono->lap.ru_utime.tv_usec;
    diff.ru_stime.tv_sec  -= chrono->lap.ru_stime.tv_sec ;
    diff.ru_stime.tv_usec -= chrono->lap.ru_stime.tv_usec;

    if (chrono->stop.ru_utime.tv_sec != 0) {
	chrono->lap.ru_utime.tv_sec  = rusage.ru_utime.tv_sec ;
	chrono->lap.ru_utime.tv_usec = rusage.ru_utime.tv_usec;
	chrono->lap.ru_stime.tv_sec  = rusage.ru_stime.tv_sec ;
	chrono->lap.ru_stime.tv_usec = rusage.ru_stime.tv_usec;
    }

    return diff;
}

#define TIME(tv)	((double)(tv).tv_sec + (double)(tv).tv_usec / 1000000.0)

char *
bflchronostring( bfltime_t time )
{
    static char str[BUFSIZ];
    
    sprintf(str, "(user := %.2fs, sys := %.2fs)", TIME(time.ru_utime), TIME(time.ru_stime));
    return str;
}
