/*
 * $Author: ac $
 * $Id: bfllist.c,v 1.2 1999/10/21 12:51:51 ac Exp $
 * $Date: 1999/10/21 12:51:51 $
 */
/*
 * $Log: bfllist.c,v $
 * Revision 1.2  1999/10/21 12:51:51  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:37  alliance
 * fpmap import
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include BFL_H
/* #include "bfllist.h"
 * #include "bflmem.h"
 * #include "bflerr.h"
 */
/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

#define BFL_LIST_SWAP(list1, list2) { \
	bfllistnode_t *__tmp_node = list1->node; \
	bflsize_t __tmp_length = list1->length; \
	list1->node = list2->node; \
	list1->length = list2->.length; \
	list2->node = __tmp_node; \
	list2->length = __tmp_length; \
}

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

static void
bfllisttransfer __P((bfllistnode_t *, bfllistnode_t *, bfllistnode_t *));

/*----------------------------------------------------------------------*
static void
bfllistsplice __P((bfllist_t *, bfllistnode_t *, bfllist_t *, bfllistnode_t *));
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

/*
 * deplace la sous-liste (first, last) a la position pos.
 * (last pointe un apres le dernier element a deplacer.
 */
static void
bfllisttransfer(pos, first, last)
    bfllistnode_t *pos;
    bfllistnode_t *first;
    bfllistnode_t *last;
{
    bfllistnode_t *tmp;

    last->prev->next = pos;
    first->prev->next = last;
    pos->prev->next = first;
    tmp = pos->prev;
    pos->prev = last->prev;
    last->prev = first->prev;
    first->prev = tmp;
}

/*----------------------------------------------------------------------*
static void
bfllistsplice(list1, pos, list2, i)
	bfllist_t *list1;
	bfllistnode_t *pos;
	bfllist_t *list2;
	bfllistnode_t *i;
{
	bfllistnode_t *j = i;
	if (pos == i || pos == bfllistnext(j))
		return;
	bfllisttransfer(pos, i, j);
	++list1->length;
	--list2->length;
}
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

bfllist_t *
bfllistnew()
{
    bfllist_t *list;

    list = (bfllist_t *) bflmemalloc(sizeof(bfllist_t));
    list->length = 0;
    list->node = (bfllistnode_t *) bflmemalloc(sizeof(bfllistnode_t));
    list->node->prev = list->node;
    list->node->next = list->node;
    return list;
}

void
bfllistfree(list)
    bfllist_t *list;
{
    bfllistclean(list);
    bflmemfree(list->node, sizeof(bfllistnode_t));
    bflmemfree(list, sizeof(bfllist_t));
}

bfllist_t *
bfllistcopy(list)
    bfllist_t *list;
{
    bfllist_t *copy = bfllistnew();
    bfllistnode_t *first = bfllistbegin(list);
    bfllistnode_t *last = bfllistend(list);

    while (first != last) {
	bfllistpushback(copy, bfllistdata(first));
	first = bfllistnext(first);
    }
    return copy;
}

void
bfllistclean(list)
    bfllist_t *list;
{
    bfllistnode_t *first = bfllistbegin(list);
    bfllistnode_t *last = bfllistend(list);

    while (first != last) {
	bfllistnode_t *next = bfllistnext(first);

	bfllisterase(list, first);
	first = next;
    }
}

bfllistnode_t *
bfllistinsert(list, pos, data)
    bfllist_t *list;
    bfllistnode_t *pos;
    bflptr_t data;
{
    bfllistnode_t *node = (bfllistnode_t *) bflmemalloc(sizeof(bfllistnode_t));

    node->data = data;
    node->next = pos;
    node->prev = pos->prev;
    pos->prev->next = node;
    pos->prev = node;
    list->length++;
    return node;
}

void
bfllisterase(list, pos)
    bfllist_t *list;
    bfllistnode_t *pos;
{
    pos->prev->next = pos->next;
    pos->next->prev = pos->prev;
    bflmemfree(pos, sizeof(bfllistnode_t));
    list->length--;
}

bfllistnode_t *
bfllistfind(bfllist_t *list, bflcompar_fn compar, bflptr_t data)
{
    bfllistnode_t *first = bfllistbegin(list);
    bfllistnode_t *last = bfllistend(list);

    while (first != last) {
	if (compar(bfllistdata(first), data) == 0)
	    return first;
	first = bfllistnext(first);;
    }
    return (bfllistnode_t *) 0;
}

bfllistnode_t *
bfllistfindptr(bfllist_t *list, bflptr_t data)
{
    bfllistnode_t *first = bfllistbegin(list);
    bfllistnode_t *last = bfllistend(list);

    while (first != last) {
	if (bfllistdata(first) == data)
	    return first;
	first = bfllistnext(first);;
    }
    return (bfllistnode_t *) 0;
}

bfllistnode_t *
bfllistnth(list, index)
    bfllist_t *list;
    bflsize_t index;
{
    bfllistnode_t *first;

    if (index > bfllistsize(list))
	bflerror(1, "bfllistnth: index out of range");

    first = bfllistbegin(list);
    while (index-- != 0)
	first = bfllistnext(first);
    return first;
}

void
bfllistremove(bfllist_t *list, bflcompar_fn compar, bflptr_t data)
{
    bfllistnode_t *first = bfllistbegin(list);
    bfllistnode_t *last = bfllistend(list);

    while (first != last) {
	bfllistnode_t *next = bfllistnext(first);

	if (compar(bfllistdata(first), data) == 0)
	    bfllisterase(list, first);
	first = next;
    }
}

void
bfllistremoveptr(bfllist_t *list, bflptr_t data)
{
    bfllistnode_t *first = bfllistbegin(list);
    bfllistnode_t *last = bfllistend(list);

    while (first != last) {
	bfllistnode_t *next = bfllistnext(first);

	if (bfllistdata(first) == data)
	    bfllisterase(list, first);
	first = next;
    }
}

void
bfllistreverse(list)
    bfllist_t *list;
{
    bfllistnode_t *first;
    bfllistnode_t *last;

    if (bfllistsize(list) < 2)
	return;
    first = bfllistnext(bfllistbegin(list));
    last = bfllistend(list);
    while (first != last) {
	bfllistnode_t *next = bfllistnext(first);

	bfllisttransfer(bfllistbegin(list), first, next);
	first = next;
    }
}

void
bfllistmerge(list1, list2, compar)
    bfllist_t *list1;
    bfllist_t *list2;
    bflcompar_fn compar;
{
    bfllistnode_t *first1 = bfllistbegin(list1);
    bfllistnode_t *last1 = bfllistend(list1);
    bfllistnode_t *first2 = bfllistbegin(list2);
    bfllistnode_t *last2 = bfllistend(list2);

    while (first1 != last1 && first2 != last2) {
	if (compar(first2->data, first1->data) < 0) {
	    bfllistnode_t *next = bfllistnext(first2);

	    bfllisttransfer(first1, first2, next);
	    first2 = next;
	} else {
	    first1 = bfllistnext(first1);
	}
    }
    if (first2 != last2)
	bfllisttransfer(last1, first2, last2);
    list1->length += list2->length;
    list2->length = 0;
}

void
bfllistsort(list, compar)
    bfllist_t *list;
    bflcompar_fn compar;
{
    bfllist_t *sorted;
    bfllist_t *elem;
    bfllistnode_t *first;
    bfllistnode_t *last;

    if (bfllistsize(list) < 2)
	return;

    sorted = bfllistnew();
    elem = bfllistnew();

    first = bfllistbegin(list);
    last = bfllistend(list);
    while (first != last) {
	bfllistnode_t *next = bfllistnext(first);

	bfllisttransfer(bfllistbegin(elem), first, next);
	bfllistmerge(sorted, elem, compar);
	first = next;
    }

    bfllisttransfer(bfllistbegin(list), bfllistbegin(sorted), bfllistend(sorted));

    bfllistfree(sorted);
    bfllistfree(elem);
}

/*
 * void
 * bfllistsort(list, compar)
 * bfllist_t *list;
 * bflcompar_fn compar;
 * {
 * bfllist_t *carry = bfllistnew();
 * bfllist_t counter[64];
 * int i, fill = 0;
 * 
 * if (bfllistsize(list) < 2)
 * return;
 * 
 * while (!bfllistempty(list)) {
 * bfllistsplice(carry, bfllistbegin(carry), list, bfllistbegin(list));
 * i = 0;
 * while(i < fill && !bfllistempty(&counter[i])) {
 * bfllistmerge(&counter[i], carry, compar);
 * BFL_LIST_SWAP(carry, &counter[i++]);
 * }
 * BFL_LIST_SWAP(carry, &counter[i]);
 * if (i == fill)
 * fill++;
 * } 
 * 
 * for (i = 1; i < fill; i++)
 * bfllistmerge(&counter[i], &counter[i-1], compar);
 * BFL_LIST_SWAP(list, &counter[fill-1]);
 * 
 * bfllistfree(carry);
 * }
 */
