/*
 * $Author: ac $
 * $Id: bflmem.c,v 1.2 1999/10/21 12:51:52 ac Exp $
 * $Date: 1999/10/21 12:51:52 $
 */
/*
 * $Log: bflmem.c,v $
 * Revision 1.2  1999/10/21 12:51:52  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:38  alliance
 * fpmap import
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include BFL_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

#define	__MEM_ALIGN		8
#define __MEM_MAX_BYTES		128
#define __MEM_NFREELISTS	(__MEM_MAX_BYTES / __MEM_ALIGN)
#define __MEM_NOBJS		32

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

typedef union __mem_obj {
    union __mem_obj *next;
} __mem_obj_t;

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

static __mem_obj_t *free_list[__MEM_NFREELISTS];

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

#define ROUND_UP(size)		(((size) + __MEM_ALIGN-1) & ~(__MEM_ALIGN-1))
#define FREELIST_INDEX(size)	(((size) + __MEM_ALIGN-1) / __MEM_ALIGN-1)

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

static bflptr_t
mem_refill(bflsize_t size)
{
    bflptr_t chunk = bflcalloc(__MEM_NOBJS, size);
    __mem_obj_t **my_free_list;
    __mem_obj_t *result;
    __mem_obj_t *current_obj, *next_obj;
    int i;

    my_free_list = free_list + FREELIST_INDEX(size);

/* Build free list in chunk */
    result = (__mem_obj_t *) chunk;
    *my_free_list = next_obj = (__mem_obj_t *) (chunk + size);
    for (i = 1;; i++) {
	current_obj = next_obj;
	next_obj = (__mem_obj_t *) ((bflptr_t) next_obj + size);
	if (__MEM_NOBJS - 1 == i) {
	    current_obj->next = 0;
	    break;
	} else {
	    current_obj->next = next_obj;
	}
    }
    return (result);
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

bflptr_t
bflmemalloc(bflsize_t size)
{
    __mem_obj_t **my_free_list;
    __mem_obj_t *obj;

    if (size > (size_t) __MEM_MAX_BYTES)
	return bflalloc(size);

    my_free_list = free_list + FREELIST_INDEX(size);
    obj = *my_free_list;

    if (obj == 0)
	obj = mem_refill(ROUND_UP(size));
    else
	*my_free_list = obj->next;

    return obj;
}

bflptr_t
bflmemresize(bflptr_t ptr, bflsize_t old_size, bflsize_t new_size)
{
    bflptr_t res;

    if (old_size > __MEM_MAX_BYTES && new_size > __MEM_MAX_BYTES)
	return bflrealloc(ptr, new_size);

    if (ROUND_UP(old_size) == ROUND_UP(new_size))
	return ptr;

    res = bflmemalloc(new_size);
    memcpy(res, ptr, new_size > old_size ? old_size : new_size);
    bflmemfree(ptr, old_size);
    return res;
}

void
bflmemfree(bflptr_t ptr, bflsize_t size)
{
    __mem_obj_t *obj = (__mem_obj_t *) ptr;
    __mem_obj_t **my_free_list;

    if (size > __MEM_MAX_BYTES) {
	bflfree(ptr);
	return;
    }
    my_free_list = free_list + FREELIST_INDEX(size);
    obj->next = *my_free_list;
    *my_free_list = obj;
}

bflptr_t
bflalloc(bflsize_t size)
{
    bflptr_t ptr;

    if ((ptr = malloc(size)) == NULL)
	bflsyserror(1, NULL);
    return ptr;
}

bflptr_t
bflcalloc(bflsize_t nmemb, bflsize_t size)
{
    bflptr_t ptr;

    if ((ptr = calloc(nmemb, size)) == NULL)
	bflsyserror(1, NULL);
    return ptr;
}

bflptr_t
bflrealloc(bflptr_t ptr, bflsize_t size)
{
    if ((ptr = realloc(ptr, size)) == NULL)
	bflsyserror(1, NULL);
    return ptr;
}

void
bflfree(bflptr_t ptr)
{
    free(ptr);
}
