/*
 * $Author: ac $
 * $Id: fpm_mapp.c,v 1.2 1999/10/21 12:52:05 ac Exp $
 * $Date: 1999/10/21 12:52:05 $
 */
/*
 * $Log: fpm_mapp.c,v $
 * Revision 1.2  1999/10/21 12:52:05  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:41  alliance
 * fpmap import
 *
 * Revision 1.3  1998/01/26 15:29:10  ac
 * Bug fix in the mapping function.
 *
 * Revision 1.2  1998/01/06 11:56:42  ac
 * Added .lax parameter file support.
 *
 * Revision 1.1  1997/11/14 11:49:46  ac
 * Initial revision
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include  MUT_H
#include  BFL_H
#include  AUT_H
#include  ABL_H
#include  BDD_H
#include  ABE_H
#include "fpm_network.h"
#include "fpm_user.h"
#include "fpm_mapp.h"

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/* # define DEBUG */

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

static char *
fpm_aux_name()
{
    static int	NumAux;
    char	Buffer[BUFSIZ];

    do {
	sprintf( Buffer, "auxil%d", NumAux++);
    } while ( namefind( Buffer ) != 0 );

    return namealloc( Buffer );
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

/* Cree un signal `Name' avec `Node'. Ajoute une sortie `Name', avec le bdd
 *  `Node'. Ajoute une entree, avec le meme nom. Retourne le bdd de l'entree.
 *  Le BddSystem associe est modifie (Tous ce qui pointait sur Node utilise
 *  maintenant la nouvelle entree.
 */
bddnode *
fpm_mapp_node( bddnode *Node, ablexpr **Expr )
{
	befig_list     *Figure;
	fpm_network_t  *Network;
	char           *Name;
	fpm_expr_t     *FpmExpr;
	fpm_atom_t     *FpmAtom;
	bddnode        *BddNode;
	char          **NameIn;
	long            NameInSize;      
	long            NumberNameIn;
	bddindex       *IndexIn;
	long            IndexInSize;      
	bddindex        Index;

	Figure  = FpmUser->BEFIG;
	Network = FpmUser->NETWORK;
	
	Name = fpm_aux_name();
	*Expr = simpablexpr( *Expr );
# ifdef DEBUG
    printf( "    --> %s = ", Name );
    viewablexpr( *Expr, ABL_VIEW_VHDL );
    putchar( '\n' );
# endif

	/* Node devient une variable terminale, le vieux Node est dans
	 * BddNode !!
	 */
	BddNode = addbddvarauxsingle( Network->BDD_SYSTEM, Node );

	FpmExpr           = fpm_expr_alloc();
	FpmExpr->NAME     = Name;
	FpmExpr->BDD      = BddNode;
	FpmExpr->BDD_ATOM = NULL;
	FpmExpr->USER     = NULL;

	bflmapinsert( Network->MAP_EXPR, FpmExpr->NAME, FpmExpr );

	FpmAtom           = fpm_atom_alloc();
	FpmAtom->NAME     = Name;
	FpmAtom->BDD      = Node;
	FpmAtom->BDD_EXPR = (chain_list *)0;
	FpmAtom->USER     = (ptype_list *)0;
	
	bflmapinsert( Network->MAP_ATOM, FpmAtom->NAME, FpmAtom );
	
	NumberNameIn = bflmapsize( Network->MAP_ATOM ) - 1;
	NameIn       = Network->NAME_IN;
	NameInSize   = Network->NAME_IN_SIZE;

	if ( NumberNameIn >= NameInSize ) {
		NameInSize <<= 1;
		NameIn = resizebddnamein( NameIn, NameInSize, NameInSize );

		Network->NAME_IN      = NameIn;
		Network->NAME_IN_SIZE = NameInSize;
	}

	NameIn[ NumberNameIn ] = FpmAtom->NAME;

	IndexIn     = Network->INDEX_IN;
	IndexInSize = Network->INDEX_IN_SIZE;
	Index       = FpmAtom->BDD->INDEX - BDD_INDEX_MIN;

	if ( Index >= IndexInSize ) {
		IndexIn = resizebddindexin( IndexIn, IndexInSize, Index << 1 );

		Network->INDEX_IN      = IndexIn;
		Network->INDEX_IN_SIZE = (long)( Index ) << 1;
	}

	IndexIn[ Index ] = (bddindex)NumberNameIn;

	fpm_network_depend( Network );
	
	Figure->BEAUX = beh_addbeaux( Figure->BEAUX, Name, *Expr, BddNode );
	*Expr = createablatom( Name );
	return Node;
}

void
fpm_mapp()
{
	befig_list	*Figure;
	fpm_network_t	*Network;
	beaux_list	*BehAux;
	beout_list	*BehOut;
	bebus_list	*BehBus;
	bebux_list	*BehBux;
	bereg_list	*BehReg;
	binode_list	*BiNode;

	Figure  = FpmUser->BEFIG;
	Network = FpmUser->NETWORK;

	setbddlocalsystem(Network->BDD_SYSTEM);
	garbagebddsystem(BddLocalSystem);

	(*FpmUser->ARCHI_CLEAR) (0);

	printf("\n * Running the Fpga mapper on `%s' ...\n", FpmUser->FILE_VHDL);

	/* Mappe les sorties...
	 */
	for (BehAux = Figure->BEAUX; BehAux != NULL; BehAux = BehAux->NEXT) {
		if (BehAux->NODE != NULL)
			BehAux->ABL = (*FpmUser->ARCHI_MAPP) (BehAux->NODE, 0);
	}

	for (BehOut = Figure->BEOUT; BehOut != NULL; BehOut = BehOut->NEXT) {
		if (BehOut->NODE != NULL)
			BehOut->ABL = (*FpmUser->ARCHI_MAPP) (BehOut->NODE, 1);
	}

	for (BehBus = Figure->BEBUS; BehBus != NULL; BehBus = BehBus->NEXT) {
		BiNode = BehBus->BINODE;
		/* Fixme: */
	}

	for (BehBux = Figure->BEBUX; BehBux != NULL; BehBux = BehBux->NEXT) {
		BiNode = BehBux->BINODE;
		/* Fixme: */
	}

	for (BehReg = Figure->BEREG; BehReg != NULL; BehReg = BehReg->NEXT) {
		BiNode = BehReg->BINODE;
		/* Fixme: */
	}

	garbagebddsystem(BddLocalSystem);

	(*FpmUser->ARCHI_REPORT) ();
}
