/*
 * $Author: ac $
 * $Id: xnf_drive.c,v 1.3 1999/10/21 12:52:15 ac Exp $
 * $Date: 1999/10/21 12:52:15 $
 */
/*
 * $Log: xnf_drive.c,v $
 * Revision 1.3  1999/10/21 12:52:15  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.2  1999/05/18 15:39:12  ac
 * Bug fix in the xnf driver.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:42  alliance
 * fpmap import
 *
 */

/* Be careful... There's a side effect when ones call xnf_ibuf: A new
 * signal is created and connected to the 'O' pin of the ibuf. Any subsequent
 * utilisation of this name will result in a subtitution with the created
 * signal... That's not very clear ? Have a look at:
 * 	xnf.c:xnf_ibuf
 * 	xnf.c:xnf_rename
 * 	xnf.c:xnf_name
 */
#include <assert.h>
#include <stdio.h>
#include  MUT_H
#include  BFL_H
#include  AUT_H
#include  ABL_H
#include  ABE_H
#include  BUT_H
#include  XNF_H


static void
xnf_input( File, BehFigure, IobMode, Name )
    FILE	*File;
    befig_list	*BehFigure;
    int		 IobMode;
    char	*Name;
{
    if ( IobMode ) {
        xnf_ext( File, Name, 'I' );
	xnf_ibuf( File, Name, 0 );
    }
}

static void
xnf_output( File, BehFigure, IobMode, Name )
    FILE	*File;
    befig_list	*BehFigure;
    int		 IobMode;
    char	*Name;
{
    beout_list	*BehOut;
    ablexpr	*OutExpr;
    char	*PadName;
    char	*OutName;

    if ( ( BehOut = but_findbeout( BehFigure, Name ) ) == NULL )
	bflerror( 1, "xnf_drive: can't find output `%s'", Name );

    PadName = BehOut->NAME;
    OutExpr = BehOut->ABL;
    
    if ( IobMode )  {
	if ( ABL_ATOM( OutExpr ) ) {
	    OutName = ABL_ATOM_VALUE( OutExpr );
	} else {
	    OutName = xnf_signal_name();
	    but_substname( BehFigure, PadName, OutName );
	}

	xnf_ext( File, PadName, 'O' );
	xnf_obuf( File, PadName, OutName, 0 );
	xnf_eqn( File, OutExpr, OutName );
    } else {
	xnf_eqn( File, OutExpr, PadName );
    }
    
    /*
    if ( IobMode ) xnf_obuf( File, OutName, OutName, 0 );
    xnf_ext( File, OutName, 'O' );

    if ( !ABL_ATOM( OutExpr ) )
    */
}

static void
xnf_busout( File, BehFigure, IobMode, Name )
    FILE	*File;
    befig_list	*BehFigure;
    int		 IobMode;
    char	*Name;
{
    bflerror( 1, "xnf_drive: bus are stil unsupported by the driver" );
}

static void
xnf_businout( File, BehFigure, IobMode, Name )
    FILE	*File;
    befig_list	*BehFigure;
    int		 IobMode;
    char	*Name;
{
    bflerror( 1, "xnf_drive: bus are stil unsupported by the driver" );
}

static void
xnf_register( File, BehFigure, BehReg )
    FILE	*File;
    befig_list	*BehFigure;
    bereg_list	*BehReg;
{
    ablexpr    *ValExpr;
    ablexpr    *CndExpr;
    ptype_list *ClkPtr;
    ptype_list *ClkType;
    char       *ValName;
    char       *CndName;
    char       *ClkName;
    
    if ( BehReg->BIABL->NEXT != 0 )
        bflerror( 1, "xnf_drive: multiple drivers on register `%s'", BehReg->NAME );

    ValExpr = BehReg->BIABL->VALABL;
    CndExpr = BehReg->BIABL->CNDABL;
    ClkPtr = getptype( BehReg->BIABL->USER, BUT_CLOCK_NAME );
    ClkType = getptype( BehReg->BIABL->USER, BUT_CLOCK_TYPE );
    
    if ( ClkType->DATA != BUT_FF_RE && ClkType->DATA != BUT_FF_FE )
	bflerror( 1, "xnf_drive: invalid type for register `%s' (%d)", BehReg->NAME, ClkType->DATA );
    	
    assert( ABL_ATOM( ValExpr ) );
    assert( ( CndExpr == 0 ) || ABL_ATOM( CndExpr ) );
    assert( ClkPtr != 0 );

    ValName = ABL_ATOM_VALUE( ValExpr );
    CndName = ( CndExpr != 0 ) ? ABL_ATOM_VALUE( CndExpr ) : 0;
    ClkName = ( char * )ClkPtr->DATA;
    
    xnf_dff( File, ValName, CndName, ClkName, BehReg->NAME, 0 );
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

void
xnf_drive( befig_list *BehFigure, int IobMode )
{
    FILE       *File;
    bepor_list *BehPor;
    beaux_list *BehAux;
    bereg_list *BehReg;
    bebux_list *BehBux;

    if ( ( File = mbkfopen( BehFigure->NAME, "xnf", "w" ) ) == NULL )
	bflsyserror( 1, "can't open File `%s.xnf'", BehFigure->NAME );

    printf( "   - writing `%s.xnf'...\n", BehFigure->NAME );

    xnf_prolog( File, "generated by xnf driver" );
    
    for( BehPor = BehFigure->BEPOR; BehPor != 0; BehPor = BehPor->NEXT ) {
	switch ( BehPor->DIRECTION ) {
	
	case 'I':
	    xnf_input( File, BehFigure, IobMode, BehPor->NAME );
	    break;
	
	case 'O':
	    xnf_output( File, BehFigure, IobMode, BehPor->NAME );
	    break;
	
	case 'B':
	    /*
	     * pas la peine de se prendre la tete, c'est une specificite
	     * Alliance. On drive donc un obuf simple.
	     */
	    xnf_output( File, BehFigure, IobMode, BehPor->NAME );
	    break;
	
	case 'Z':
	    xnf_busout( File, BehFigure, IobMode, BehPor->NAME );
	    break;
	    
	case 'T':
	    xnf_businout( File, BehFigure, IobMode, BehPor->NAME );
	    break;
	    
	default:
	    bflerror( 1, "xnf_drive: unknown port direction for port `%s'", BehPor->NAME );
	}
    }
    
    for ( BehReg = BehFigure->BEREG; BehReg != 0; BehReg = BehReg->NEXT )
	xnf_register( File, BehFigure, BehReg );

    for ( BehAux = BehFigure->BEAUX; BehAux != 0; BehAux = BehAux->NEXT )
	    xnf_eqn( File, BehAux->ABL, BehAux->NAME );

    for ( BehBux = BehFigure->BEBUX; BehBux != 0; BehBux = BehBux->NEXT )
	bflerror( 1, "xnf_drive: bus are stil unsupported by the driver" );

    xnf_epilog( File );
    fclose( File );
}
