/*
   ### -------------------------------------------------- ### 
   $Author: hcl $
   $Date: 2001/09/26 14:27:51 $
   $Log: ocrWRoutingUtil.c,v $
   Revision 1.1  2001/09/26 14:27:51  hcl
   premier commit....

   Revision 1.3  2000/02/26 00:24:09  root
   2 points OK

   Revision 1.2  2000/01/25 15:49:00  root
   *** empty log message ***

   ### -------------------------------------------------- ###
 */

#include <stdlib.h>
#include MUT_H
#include MLO_H
#include MLU_H
#include MPH_H
#include MPU_H
#include "ocr.h"
#include "ocrUtil.h"
#include "ocrWRoutingDataBase.h"
#include "ocrWindow.h"
#include "mbk_tree.h"
#include "ocrWRoutingUtil.h"

#define LAYER_PARITY_MASK  (ocrNaturalInt)1

static char *res_id =
  "$Id: ocrWRoutingUtil.c,v 1.1 2001/09/26 14:27:51 hcl Exp $";

ocrWSegment *HEAD_OCRWSEGMENT = NULL;	/* segment buffer head */

/** 
 * Fonction de comparaison entre Entier
 * pour les fonctions sur les arbres RBTREE
**/

static int
deltaCompare (void *first, void *second)
{

  return ((ocrNaturalInt) first == (ocrNaturalInt) second) ?
    0 : (((ocrNaturalInt) first > (ocrNaturalInt) second) ? 1 : -1);
}

/** 
 * Cration de la Grille de Routage
 * Rservation de la place mmoire ncessaire
**/

ocrWRoutingGrid *
createWGrid (ocrNaturalInt size_h,
             ocrNaturalInt size_v, ocrNaturalInt nb_of_layers)
{
  ocrWRoutingGrid *pt =
    (ocrWRoutingGrid *) mbkalloc (sizeof (ocrWRoutingGrid));

  pt->SIZE_H = size_h;
  pt->SIZE_V = size_v;
  pt->NB_OF_LAYERS = nb_of_layers;
  pt->DATA =
    (ocrWSegment **) mbkalloc (size_h * size_v * nb_of_layers *
                               sizeof (ocrWSegment *));
  return pt;
}

/** 
 * initialisation de la grille compltement libre
 * de contraintes
**/
void
initWGrid (ocrWRoutingGrid * i_pGrid, ocrRoutingParameters * i_pParam)
{
  ocrNaturalInt l_uLayer, i, j;
  ocrWSegment *l_pSegment;

  // cration de segments libres de longueur MAX pour chaque noeud
  for (l_uLayer = 0; l_uLayer < i_pGrid->NB_OF_LAYERS; l_uLayer++)
    if (getDirection (i_pParam, l_uLayer) == ocrHorizontal)
      for (i = 0; i < i_pGrid->SIZE_V; i++)
        {
          l_pSegment = createWSegment (i, l_uLayer, 0, i_pGrid->SIZE_H - 1,
                                       WSEGMENT_FREE);
          for (j = l_pSegment->P_MIN; j <= l_pSegment->P_MAX; j++)
            setWGrid (i_pGrid, l_pSegment, j, i, l_uLayer);
        }
    else                        // vertical

      for (j = 0; j < i_pGrid->SIZE_H; j++)
        {
          l_pSegment = createWSegment (j, l_uLayer, 0, i_pGrid->SIZE_V - 1,
                                       WSEGMENT_FREE);
          for (i = l_pSegment->P_MIN; i <= l_pSegment->P_MAX; i++)
            setWGrid (i_pGrid, l_pSegment, j, i, l_uLayer);
        }

  // contraintes NORTH ALU 2
  i = i_pGrid->SIZE_V - 1;
  l_uLayer = 0;
  for (j = 0; j < i_pGrid->SIZE_H; j++)
    {
      l_pSegment = createWSegment (i, l_uLayer, j, j, WSEGMENT_FREE);
      setWGrid (i_pGrid, l_pSegment, j, i, l_uLayer);
    }

  // contraintes NORTH en ALU 4
  if (i_pGrid->NB_OF_LAYERS >= 3)
    {
      l_pSegment = getWSegment (i_pGrid, 0, 0, 2);
      l_pSegment->SIGNAL_INDEX = WSEGMENT_OBSTACLE;
    }

  // contraintes SOUTH en ALU 4
  if (i_pGrid->NB_OF_LAYERS >= 3)
    {
      l_pSegment = getWSegment (i_pGrid, 0, i_pGrid->SIZE_V - 1, 2);
      l_pSegment->SIGNAL_INDEX = WSEGMENT_OBSTACLE;
    }

  // contraintes EAST
  l_pSegment = getWSegment (i_pGrid, i_pGrid->SIZE_H - 1, 0, 1);
  l_pSegment->SIGNAL_INDEX = WSEGMENT_OBSTACLE;
}

/**
 * Destruction de la Grille de Routage 
**/

void
freeWGrid (ocrWRoutingGrid * i_pGrid)
{
/*
   ocrNaturalInt l_uLayer, i, j;
   ocrWSegment *l_pSegment;

   // destruction des segments.
   for (l_uLayer = 0; l_uLayer < i_pGrid->NB_OF_LAYERS; l_uLayer++)
   for (i = 0; i < i_pGrid->SIZE_V; i++)
   for (j = 0; j < i_pGrid->SIZE_H; j++)
   {
   l_pSegment = getWSegment (i_pGrid, i, j, l_uLayer);
   if (l_pSegment)
   {
   freeWSegment (l_pSegment);
   setWGrid (i_pGrid, NULL, i, j, l_uLayer);
   }
   }
 */
  mbkfree (i_pGrid->DATA);
  mbkfree (i_pGrid);
}

/** 
 * Cration d'un segment pour un noeud de la Grille de Routage
**/

ocrWSegment *
createWSegment (ocrNaturalInt offset,
                ocrNaturalInt layer,
                ocrNaturalInt p_min,
                ocrNaturalInt p_max, ocrNaturalInt signal_index)
{
  ocrWSegment *pt;
  register int i;

  if (HEAD_OCRWSEGMENT == NULL)
    {
      pt = (ocrWSegment *) mbkalloc (BUFSIZE * sizeof (ocrWSegment));
      HEAD_OCRWSEGMENT = pt;
      for (i = 1; i < BUFSIZE; i++)
        {
          (ocrWSegment *) pt->LABELS_LEFT_TOP = pt + 1;
          pt++;
        }
      (ocrWSegment *) pt->LABELS_LEFT_TOP = NULL;
    }

  pt = HEAD_OCRWSEGMENT;
  HEAD_OCRWSEGMENT = (ocrWSegment *) HEAD_OCRWSEGMENT->LABELS_LEFT_TOP;
  pt->OFFSET = offset;
  pt->LAYER = layer;
  pt->P_MIN = p_min;
  pt->P_MAX = p_max;
  // Ajout
  pt->TAG = 0;
  pt->RB = 0;
  pt->LT = 0;
  pt->SIGNAL_INDEX = signal_index;
  pt->LABELS_LEFT_TOP = RBTREENULL;
  pt->LABELS_RIGHT_BOTTOM = RBTREENULL;
  return pt;
}

/** 
 * Supprime un segment
**/

void
freeWSegment (ocrWSegment * segment)
{

  (ocrWSegment *) segment->LABELS_LEFT_TOP = HEAD_OCRWSEGMENT;
  HEAD_OCRWSEGMENT = segment;
}

/** 
 * Retourne le segment de la grille aux coord (x,y,z) 
**/

ocrWSegment *
getWSegment (ocrWRoutingGrid * grid,
             ocrNaturalInt x, ocrNaturalInt y, ocrNaturalInt z)
{

  return (grid->DATA[z +
                     y * grid->NB_OF_LAYERS +
                     x * grid->SIZE_V * grid->NB_OF_LAYERS]);
}

ocrWSegment *
getWSegmentCV (ocrWRoutingGrid * grid, ocrVirtualConnector * i_pVirCon)
{

  return (grid->DATA[i_pVirCon->Z +
                     i_pVirCon->Y /*/ 500 */  * grid->NB_OF_LAYERS + i_pVirCon->X	/* / 500 */
                     * grid->SIZE_V * grid->NB_OF_LAYERS]);
}


/** 
 * Mise  jour de la Grille de Routage
 * Association d'un segment a un noeud de la Grille
**/

void
setWGrid (ocrWRoutingGrid * grid,
          ocrWSegment * segment,
          ocrNaturalInt x, ocrNaturalInt y, ocrNaturalInt z)
{
  grid->DATA[z +
             y * grid->NB_OF_LAYERS +
             x * grid->SIZE_V * grid->NB_OF_LAYERS] = segment;
}

/** **/

int
isLabeled (ocrWSegment * segment)
{

  return !((segment->LABELS_LEFT_TOP == RBTREENULL) &&
           (segment->LABELS_RIGHT_BOTTOM == RBTREENULL));
}

/** 
 * Est-ce que le segment est occup
**/

int
isObstructed (ocrWSegment * segment)
{

  return (segment->SIGNAL_INDEX != WSEGMENT_FREE);
}

/** 
 * Renvoie la direction du segment 
**/

ocrRoutingDirection
getWSegDirection (ocrRoutingParameters * i_param, ocrWSegment * i_segment)
{
  if (i_segment == NULL)
    puts ("ERROR getWSegDirection");
  fflush (stdout);

  return ((i_segment->LAYER & LAYER_PARITY_MASK) ?
          (ocrRoutingDirection) (!(i_param->EVEN_LAYERS_DIRECTION))
          : i_param->EVEN_LAYERS_DIRECTION);
}

/** 
 * Renvoie la coordonne X du segment 
**/

ocrNaturalInt
getWSegXCoord (ocrRoutingParameters * i_param,
               ocrWSegment * i_segment, ocrNaturalInt i_point)
{
  return ((getWSegDirection (i_param, i_segment) == ocrHorizontal) ?
          i_point : i_segment->OFFSET);
}

/** 
 * Renvoie la coordonne Y du segment 
**/

ocrNaturalInt
getWSegYCoord (ocrRoutingParameters * i_param,
               ocrWSegment * i_segment, ocrNaturalInt i_point)
{
// Y dpend de la direction du LAYER
  return ((getWSegDirection (i_param, i_segment) == ocrHorizontal) ?
          i_segment->OFFSET : i_point);
}

/** 
 * Cration d'un segment physique reprsentant le signal index
 * entre first_point et second_point dans le LAYER considr 
 * (ici segment->LAYER)
 *
 * et mise  jour de la Grille car
 * Les noeuds aligns avec ce nouveau segment sont tronqus
 *
 * Il est suppos que :
 * first_point < second_point
**/

ocrWSegment *
splitWSegment (ocrRoutingParameters * param,
               ocrWRoutingGrid * grid,
               ocrWSegment * segment,
               ocrNaturalInt first_point,
               ocrNaturalInt second_point, ocrNaturalInt index)
{
  ocrNaturalInt seg1_length = first_point - segment->P_MIN;
  ocrNaturalInt seg2_length = second_point - first_point + 1;
  ocrNaturalInt seg3_length = segment->P_MAX - second_point;
  ocrWSegment *new_segment1;
  ocrWSegment *new_segment2;
  ocrNaturalInt i;

  if (seg1_length > seg3_length)
    {
      if (seg1_length > seg2_length)
        {
          new_segment1 = createWSegment (segment->OFFSET,
                                         segment->LAYER,
                                         first_point, second_point, index);
          if (getWSegDirection (param, new_segment1) == ocrHorizontal)
            for (i = first_point; i <= second_point; i++)
              setWGrid (grid, new_segment1, i, new_segment1->OFFSET,
                        new_segment1->LAYER);
          else
            for (i = first_point; i <= second_point; i++)
              setWGrid (grid, new_segment1, new_segment1->OFFSET, i,
                        new_segment1->LAYER);
          if (seg3_length)
            {
              new_segment2 = createWSegment (segment->OFFSET,
                                             segment->LAYER,
                                             second_point + 1,
                                             segment->P_MAX,
                                             segment->SIGNAL_INDEX);
              if (getWSegDirection (param, new_segment2) == ocrHorizontal)
                for (i = second_point + 1; i <= segment->P_MAX; i++)
                  setWGrid (grid, new_segment2, i, new_segment2->OFFSET,
                            new_segment2->LAYER);
              else
                for (i = second_point + 1; i <= segment->P_MAX; i++)
                  setWGrid (grid, new_segment2, new_segment2->OFFSET, i,
                            new_segment2->LAYER);
            }
          segment->P_MAX = first_point - 1;
          return new_segment1;
        }
    }
  else
    {
      if (seg3_length > seg2_length)
        {                       //here

          new_segment1 = createWSegment (segment->OFFSET,
                                         segment->LAYER,
                                         first_point, second_point, index);
          if (getWSegDirection (param, new_segment1) == ocrHorizontal)
            for (i = first_point; i <= second_point; i++)
              setWGrid (grid, new_segment1, i, new_segment1->OFFSET,
                        new_segment1->LAYER);
          else
            for (i = first_point; i <= second_point; i++)
              setWGrid (grid, new_segment1, new_segment1->OFFSET, i,
                        new_segment1->LAYER);
          if (seg1_length)
            {
              new_segment2 = createWSegment (segment->OFFSET,
                                             segment->LAYER,
                                             segment->P_MIN,
                                             first_point - 1,
                                             segment->SIGNAL_INDEX);
              if (getWSegDirection (param, new_segment2) == ocrHorizontal)
                for (i = segment->P_MIN; i <= first_point - 1; i++)
                  setWGrid (grid, new_segment2, i, new_segment2->OFFSET,
                            new_segment2->LAYER);
              else
                for (i = segment->P_MIN; i <= first_point - 1; i++)
                  setWGrid (grid, new_segment2, new_segment2->OFFSET, i,
                            new_segment2->LAYER);
            }
          segment->P_MIN = second_point + 1;
          return new_segment1;
        }
    }
  if (seg1_length)
    {
      new_segment1 = createWSegment (segment->OFFSET,
                                     segment->LAYER,
                                     segment->P_MIN,
                                     first_point - 1, segment->SIGNAL_INDEX);
      if (getWSegDirection (param, new_segment1) == ocrHorizontal)
        for (i = segment->P_MIN; i <= first_point - 1; i++)
          setWGrid (grid, new_segment1, i, new_segment1->OFFSET,
                    new_segment1->LAYER);
      else
        for (i = segment->P_MIN; i <= first_point - 1; i++)
          setWGrid (grid, new_segment1, new_segment1->OFFSET, i,
                    new_segment1->LAYER);
    }
  if (seg3_length)
    {
      new_segment2 = createWSegment (segment->OFFSET,
                                     segment->LAYER,
                                     second_point + 1,
                                     segment->P_MAX, segment->SIGNAL_INDEX);
      if (getWSegDirection (param, new_segment2) == ocrHorizontal)
        for (i = second_point + 1; i <= segment->P_MAX; i++)
          setWGrid (grid, new_segment2, i, new_segment2->OFFSET,
                    new_segment2->LAYER);
      else
        for (i = second_point + 1; i <= segment->P_MAX; i++)
          setWGrid (grid, new_segment2, new_segment2->OFFSET, i,
                    new_segment2->LAYER);
    }
  segment->P_MIN = first_point;
  segment->P_MAX = second_point;
  segment->SIGNAL_INDEX = index;
  return segment;
}

/** get L1 distance (manhattan distance) in a segment **/

ocrNaturalInt
manhattan1Distance (ocrRoutingParameters * param,
                    ocrWSegment * segment, ocrNaturalInt p1, ocrNaturalInt p2)
{

  return ((getWSegDirection (param, segment) == ocrHorizontal) ?
          param->PITCH_H * ABSDIFF (p1, p2)
          : param->PITCH_V * ABSDIFF (p1, p2));
}

/** 
 * Get induced delta in a segment 
**/

ocrNaturalInt
inducedDelta (ocrRoutingParameters * param,
              ocrWSegment * segment,
              ocrNaturalInt xtarget,
              ocrNaturalInt ytarget, rbtree * node, ocrNaturalInt point)
{
//printf("-> noeud %ld point = %ld point(node) = %ld\n",
  //   DELTA (node),point,POINT (node));

  return (DELTA (node) +
          manhattan1Distance (param,
                              segment,
                              POINT (node),
                              point) -
          manhattan2Distance (param,
                              getWSegXCoord (param, segment, POINT (node)),
                              getWSegYCoord (param, segment, POINT (node)),
                              xtarget,
                              ytarget) +
          manhattan2Distance (param,
                              getWSegXCoord (param, segment, point),
                              getWSegYCoord (param, segment, point),
                              xtarget, ytarget));
}

/** **/

rbtree *
getLabel (ocrRoutingParameters * param,
          ocrNaturalInt xtarget,
          ocrNaturalInt ytarget,
          ocrWSegment * segment, ocrNaturalInt delta, ocrNaturalInt point)
{
  rbtree *label_LT;
  rbtree *label_RB;
  rbtree *previous;
  rbtree *next;
  ocrNaturalInt coord_target =
    (getWSegDirection (param, segment) == ocrHorizontal) ? xtarget : ytarget;

  if (point <= coord_target)
    {
//    puts("label");
      label_LT = getrbtree2 (segment->LABELS_LEFT_TOP,
                             (void *) delta, &previous, &next, deltaCompare);

      if (previous == RBTREENULL)
        {
          if (label_LT == RBTREENULL)
            {
//              puts("label0");
              return RBTREENULL;
            }
          else
            {
//              puts("label1");
              if (inducedDelta (param,
                                segment,
                                xtarget, ytarget, label_LT, point) != delta)
                return RBTREENULL;
            }
        }
      else
        {
          if (label_LT == RBTREENULL)
            {
//              puts("label2");
              if (inducedDelta (param,
                                segment,
                                xtarget, ytarget, previous, point) != delta)
                return RBTREENULL;
              label_LT = previous;
            }
          else
            {
//              puts("label3");
              if ((inducedDelta (param,
                                 segment,
                                 xtarget,
                                 ytarget,
                                 label_LT,
                                 point) != delta) ||
                  (inducedDelta (param,
                                 segment,
                                 xtarget, ytarget, previous, point) < delta))
                return RBTREENULL;
              return label_LT;
            }
        }

      if ((POINT (label_LT) == coord_target) &&
          ((label_RB = getmostrbtree (segment->LABELS_RIGHT_BOTTOM)) !=
           RBTREENULL) && (DELTA (label_RB) == DELTA (label_LT)))
        {

//              puts("RB");
          return label_RB;
        }
      else
        {
//              puts("LT");
          return label_LT;
        }

    }
  else
    {
      label_RB = getrbtree2 (segment->LABELS_RIGHT_BOTTOM,
                             (void *) delta, &previous, &next, deltaCompare);

      if (previous == RBTREENULL)
        {
          if (label_RB == RBTREENULL)
            {
//              puts("label10");
              return RBTREENULL;
            }
          else
            {
//              puts("label11");
              if (inducedDelta (param,
                                segment,
                                xtarget, ytarget, label_RB, point) != delta)
                return RBTREENULL;
            }
        }
      else
        {
          if (label_RB == RBTREENULL)
            {
//              puts("label12");

              if (inducedDelta (param,
                                segment,
                                xtarget, ytarget, previous, point) != delta)
                return RBTREENULL;
              label_RB = previous;
            }
          else
            {
//              puts("label13");
              if ((inducedDelta (param,
                                 segment,
                                 xtarget,
                                 ytarget,
                                 label_RB,
                                 point) != delta) ||
                  (inducedDelta (param,
                                 segment,
                                 xtarget, ytarget, previous, point) < delta))
                return RBTREENULL;
              return label_RB;
            }
        }

      if ((POINT (label_RB) == coord_target + 1) &&
          ((label_LT = getmostrbtree (segment->LABELS_LEFT_TOP)) !=
           RBTREENULL) && (DELTA (label_LT) < DELTA (label_RB)))
        return label_LT;
      else
        return label_RB;

    }
}

/** **/

ocrNaturalInt
getNextPlane (ocrWRoutingGrid * grid,
              ocrNaturalInt ztarget, ocrNaturalInt plane)
{
  ocrInt symplane;

  if (plane == ztarget)
    // retourne le plus proche entre 0 et NB_OF_LAYERS - 1
    return ((ABSDIFF (grid->NB_OF_LAYERS - 1, ztarget) > ABSDIFF (0, ztarget))
            ? grid->NB_OF_LAYERS - 1 : 0);

  symplane = (plane > ztarget) ?
    ztarget - (ABSDIFF (plane, ztarget) - 1)
    : ztarget + ABSDIFF (plane, ztarget);
  if (symplane < 0)
    return (plane - 1);
  else
    {
      if (symplane >= grid->NB_OF_LAYERS)
        return (plane + 1);
      else
        return symplane;
    }
}
