
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: cox_package.c						*/
/* date		: Jun  1 1994						*/
/* version	: v1.00							*/
/* author	: DUNOYER Julien					*/
/* description	: This file contains the set of probability dedicated	*/
/*                functions to link with generic symbolic simulator 	*/
/*	cox_addvalue, cox_freevalue, cox_cpyvalue  , cox_updvalue,	*/
/*	cox_cmpvalue, cox_evalvalue, cox_resolvalue,			*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#include MUT_H
#include LOG_H
#include "beh.h"
#include "bhl.h"
/* LUDO
#include BST_H
*/
#include CST_H
#include GSS_H
#include COX_H
#include PRB_H
#include "cox_colorset.h"

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */

static char         *SPG_INSNAME;

/* ###--------------------------------------------------------------### */
/* function     : cox_createatom                                        */
/* description  : create SHARED terminals (atoms) abl.			*/
/* called func. : addchain, gss_chktab, gss_addtab			*/
/* ###--------------------------------------------------------------### */
  
static struct chain *cox_createatom(terminal)

char              *terminal ;

  {
  struct chain *res_abl ;

    /* ###-------------------------------------### */
    /* ### search a previously record terminal ### */
    /* ###-------------------------------------### */

  res_abl = (struct chain*) gss_chktab (COX_DICTIONNARY, terminal,
                                        COX_TERMABL_DEF, GSS_PNT0FIELD);

  if(! res_abl)
    {
    res_abl = addchain (NULL, (void *)namealloc(terminal));
    gss_addtab (COX_DICTIONNARY, terminal,
                COX_TERMABL_DEF, GSS_PNT0FIELD, res_abl);
    }

  return (res_abl);
  }

/* ###--------------------------------------------------------------### */
/* function     : cox_createexpr                                        */
/* description  : create SHARED operators abl.				*/
/* called func. : addchain, gss_chktab, gss_addtab			*/
/* ###--------------------------------------------------------------### */
  
static struct chain *cox_createexpr(operator)

short              operator ;

  {
  struct chain *res_abl ;

    /* ###-------------------------------------### */
    /* ### search a previously record operator ### */
    /* ###-------------------------------------### */

  res_abl = (struct chain*)
            gss_chktab (COX_DICTIONNARY, operator,
                        COX_OPERABL_DEF, GSS_PNT0FIELD);

  if(! res_abl)
    {
    res_abl = addchain (NULL, operator);
    gss_addtab (COX_DICTIONNARY, operator,
                COX_OPERABL_DEF, GSS_PNT0FIELD, res_abl);
    }

  return (addchain (NULL, (void *)res_abl));
  }

/* ###--------------------------------------------------------------### */
/* function     : cox_freabl                                            */
/* description  : free recursively an abl keeping atoms only		*/
/* called func. : delchain, freechain, cox_freebl                       */
/* ###--------------------------------------------------------------### */
  
static void cox_freabl(sce_abl)

struct chain *sce_abl ;

{
struct chain *tmp_abl ;

    /* ###-------------------------------------### */
    /* ###   dont free TERMINALS & OPERATORS   ### */
    /* ###     because they are SHARED ABL.    ### */
    /* ###-------------------------------------### */
if (sce_abl)
  {
  if(! ATOM(sce_abl))
    {
    tmp_abl= sce_abl;
    delchain(CAR(sce_abl),CAR(sce_abl));
    while((sce_abl = CDR(sce_abl)))
      cox_freabl(CAR(sce_abl));
    freechain(tmp_abl);
    }
  }
}


/* ###--------------------------------------------------------------### */
/* function     : cox_dupabl                                            */
/* description  : duplicate recursively an abl 				*/
/* called func. : ATOM,CDR,VALUE_ATOM,cox_createatom, cox_createexpr    */
/* ###--------------------------------------------------------------### */
  
static struct chain *cox_dupabl(sce_abl)

struct chain *sce_abl ;

{

struct chain *res_abl;

if(ATOM(sce_abl))
  {
  res_abl = cox_createatom (VALUE_ATOM(sce_abl));
  }

else 
  {
  res_abl = cox_createexpr (OPER(sce_abl));

  while (sce_abl = CDR(sce_abl))
    {
    CDR(res_abl) =
       addchain(CDR(res_abl),(void *)cox_dupabl(CAR(sce_abl)));
    }
  }
return (res_abl);
}


/* ###--------------------------------------------------------------### */
/* function     : cox_createsuperabl 					*/
/* description  : Recursively Construct a superabl for an abl		*/
/*		  and with a color list for recursive reference.	*/
/* called func  : cox_createsupabl,gss_chktab,gss_addtab,gss_chktab	*/
/* ###--------------------------------------------------------------### */

static struct chain *cox_createsuperabl(sce_abl,extracted_col,instance)

struct chain      *sce_abl;
struct coxcol     *extracted_col;
struct lkdins     *instance;

  {
  struct chain  *spg_abl = NULL;
  struct chain  *tmp_abl = NULL;
  struct coxcol *propagate_set = NULL;

    /* ###-------------------------------------### */
    /* ### if the abl is a terminal,           ### */
    /* ### explode it in it's basic abl        ### */
    /* ### and then continue the traversal     ### */
    /* ### algorithme with this new abl.       ### */
    /* ###-------------------------------------### */

/*----
  if((ATOM (sce_abl)) && (strcmp (VALUE_ATOM (sce_abl), "'0'"))
                      && (strcmp (VALUE_ATOM (sce_abl), "'1'"))
                      && (strcmp (VALUE_ATOM (sce_abl), "'d'")))
----*/

  if(ATOM (sce_abl))
    {
    if (strcmp (VALUE_ATOM (sce_abl), "'d'") &&
        strcmp (VALUE_ATOM (sce_abl), "'1'") &&
        strcmp (VALUE_ATOM (sce_abl), "'0'"))
      {
    /* ###-------------------------------------### */
    /* ### => find the propagated colors set.  ### */
    /* ###-------------------------------------### */

      propagate_set = ((struct coxval *) gss_getvalue (instance,
                      searchInputCct(instance->BEFIG->CIRCUI,
                                     VALUE_ATOM(sce_abl))))->PROPAGATE;

     /* ###--------------------------------------------------### */
     /* ### if extracted colors set and propagated colors    ### */
     /* ###    set are not disjunctives                      ### */
     /* ###          => dupplicate the terminal abl,         ### */
     /* ###             & continue the traversal algorithme. ### */
     /* ### else     => dupplicate the terminal abl.         ### */
     /* ###--------------------------------------------------### */

      if(coc_GetInSetElement (propagate_set , extracted_col->FUNDAMENTAL) != NULL)
        {
        spg_abl =(struct chain*) gss_chktab (COX_DICTIONNARY,
                                             VALUE_ATOM (sce_abl),
                                             COX_BASICABL_DEF,
                                             GSS_PNT0FIELD);

        spg_abl = cox_createsuperabl (cox_dupabl(spg_abl),extracted_col,instance);
        }
      else
        {
        spg_abl = sce_abl;
        }
      }
    else
      {
      spg_abl = sce_abl;
      }
    }

   /* ###--------------------------------------### */
   /* ### if the abl is an expression,         ### */ 
   /* ### duplicate the expression and create  ### */ 
   /* ### the superabl of it's operands with   ### */ 
   /* ### the same list of color for reference.### */ 
   /* ###--------------------------------------### */

  else
    {
    tmp_abl = sce_abl;

    spg_abl = cox_createexpr (OPER(tmp_abl));

    while (tmp_abl = CDR(tmp_abl))
      {
      CDR(spg_abl) = addchain(CDR(spg_abl),
              (void*)cox_createsuperabl(CAR(tmp_abl), extracted_col, instance));
      }
    }

  return (spg_abl);
  }


/* ###--------------------------------------------------------------### */
/* function	: cox_bddevalvalue					*/
/* description	: computes a BDD's value (proba) 			*/
/* called func.	: cox_toolbug						*/
/* ###--------------------------------------------------------------### */

static float cox_bddevalvalue (instance , bdd_expression)

struct lkdins *instance;
struct node   *bdd_expression;
  {
  short              indexbdd;
  float              value;
  float              lowvalue = 0.0;
  float              highvalue = 0.0;

  if (bdd_expression == zero)
    {
    return (0.0);
    }
  else if (bdd_expression == one)
    {
    return (1.0);
    }

  indexbdd = bdd_expression->index ;

  value = ((struct coxval *)(gss_getvalue(instance, indexbdd)))->PROBABACK;

  if ( value != 1.0)
    {
    lowvalue = cox_bddevalvalue(instance, bdd_expression->low);

    lowvalue = (1.0 - value) * lowvalue;
    }

  if ( value != 0.0)
    {
    highvalue = cox_bddevalvalue(instance, bdd_expression->high);

    highvalue = value * highvalue;
    }

  value = lowvalue + highvalue;

  return(value);
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_BackCoeffCorrel					*/
/* description	: computes a BDD's coeff correl depending on node's val	*/
/* called func.	: cox_toolbug						*/
/* ###--------------------------------------------------------------### */

static float cox_BackCoeffCorrel (bdd_expression, instance, indexvar)

struct node   *bdd_expression;
struct lkdins *instance;
short          indexvar;
  {
  short              indexbdd;
  float              value;
  float              lowvalue = 0.0;
  float              highvalue = 0.0;


    indexbdd = bdd_expression->index ;

    if ( indexbdd > indexvar )
      {
      value = ((struct coxval *)(gss_getvalue(instance, indexbdd)))->PROBABACK;

      if ( value != 1.0)
        {
        lowvalue = cox_BackCoeffCorrel(bdd_expression->low, instance, indexvar);

        lowvalue = (1.0 - value) * lowvalue;
        }

      if ( value != 0.0)
        {
        highvalue = cox_BackCoeffCorrel(bdd_expression->high,instance,indexvar);

        highvalue = value * highvalue;
        }

      value = lowvalue + highvalue;
      }
/* ici on devrait utiliser la meme eval proba que prb,                      */
/* mais il y a un dico, bref, c'est pas cool, on recopie le code, sans dico */

    else if ( indexbdd == indexvar )
      {
      lowvalue  = cox_bddevalvalue(instance, bdd_expression->low);

      highvalue = cox_bddevalvalue(instance, bdd_expression->high);

      value = highvalue - lowvalue;
      }
    else
      {
      value = 0.0;
      }
  return(value);
  }


/* ###--------------------------------------------------------------### */
/* function	: cox_BackSensibility					*/
/* description	: computes a BDD's sensibility depending on node's val	*/
/* called func.	: cox_toolbug						*/
/* ###--------------------------------------------------------------### */

static float cox_BackSensibility (bdd_expression, instance, indexvar1, indexvar2)

struct node   *bdd_expression;
struct lkdins *instance;
short          indexvar1;
short          indexvar2;
  {
  short              indexbdd;
  float              value;
  float              lowvalue = 0.0;
  float              highvalue = 0.0;


    indexbdd = bdd_expression->index ;

    if ( indexbdd > indexvar1 )
      {
      value = ((struct coxval *)(gss_getvalue(instance, indexbdd)))->PROBABACK;

      if ( value != 1.0)
        {
        lowvalue = cox_BackSensibility(bdd_expression->low, instance,
                                       indexvar1, indexvar2);

        lowvalue = (1.0 - value) * lowvalue;
        }

      if ( value != 0.0)
        {
        highvalue = cox_BackSensibility(bdd_expression->high, instance,
                                       indexvar1, indexvar2);

        highvalue = value * highvalue;
        }

      value = lowvalue + highvalue;
      }

    else if ( indexbdd == indexvar1 )
      {
      lowvalue  = cox_BackCoeffCorrel(bdd_expression->low, instance, indexvar2);

      highvalue = cox_BackCoeffCorrel(bdd_expression->high,instance, indexvar2);
  
      value = highvalue - lowvalue;
      }
    else
      {
      value = 0.0;
      }
  return(value);
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_addvalue						*/
/* description	: addition function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void *cox_addvalue ()

  {
  struct coxval *value;

  value  = (struct coxval *)mbkalloc (sizeof (struct coxval));
  value->PROPAGATE = NULL;
  value->GENERATE  = NULL;
  value->EXTRACTED = NULL;
  value->PROBABACK = 0.5 ;
  value->EVALUATED = COX_NOT_EVALUATED;

  return(value);
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_freevalue						*/
/* description	: liberation function for values			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void cox_freevalue (value)

  struct coxval *value;

  {
  if ( value != NULL )
    {
    coc_FreeSet ( value->PROPAGATE );
    coc_FreeSet ( value->GENERATE );
    coc_FreeSet ( value->EXTRACTED );
    mbkfree (value);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_cmpvalue						*/
/* description	: comparaison function for values			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int cox_cmpvalue (value1,value2)

  struct coxval *value1;
  struct coxval *value2;

  {
  if ( COX_EVALUATION & COX_LEVELIZED_EVALUATION )
    {
    if ( value1->EVALUATED != COX_NOT_EVALUATED )
      {
      return(0);
      }
    else
      {
      if ( ( value2->EVALUATED == COX_NOT_EVALUATED )
        && ( value2->GENERATE == NULL ) )
        {
        return(0);
        }
      else
        {
        return(1);
        }
      }
    }
  else
    {
    if (coc_SetCompare(value1->PROPAGATE, value2->PROPAGATE) == 0)
      {
      if (coc_SetCompare(value1->GENERATE, value2->GENERATE) == 0)
        {
        if (coc_SetCompare(value1->EXTRACTED, value2->EXTRACTED) == 0)
          {
          return (0);
          }
        else
          return (1);
        }
      else
        return (1);
      }
    else
      return (1);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_cpyvalue						*/
/* description	: copy function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int cox_cpyvalue (value1,value2)

  struct coxval *value1;
  struct coxval *value2;

  {
  int errflg = 0;

  coc_FreeSet (value1->PROPAGATE);
  coc_FreeSet (value1->GENERATE );
  coc_FreeSet (value1->EXTRACTED);

  value1->PROPAGATE = coc_CopySet (value2->PROPAGATE);
  value1->GENERATE  = coc_CopySet (value2->GENERATE);
  value1->EXTRACTED = coc_CopySet (value2->EXTRACTED);
  value1->PROBABACK = value2->PROBABACK;

  if ( COX_EVALUATION & COX_LEVELIZED_EVALUATION )
    {
    value1->EVALUATED = value2->EVALUATED;
    }

  return(errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_updvalue						*/
/* description	: copy function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int cox_updvalue (value1,value2)

  struct coxval *value1;
  struct coxval *value2;

  {
  int errflg = 0;

  coc_FreeSet (value1->PROPAGATE);
  coc_FreeSet (value1->EXTRACTED);

  value1->PROPAGATE = coc_CopySet (value2->PROPAGATE);
  value1->EXTRACTED = coc_CopySet (value2->EXTRACTED);

  if ( COX_EVALUATION & COX_LEVELIZED_EVALUATION )
    {
    value1->EVALUATED = value2->EVALUATED;
    }

  if (( value1->GENERATE == NULL ) && ( value2->GENERATE != NULL ))
    {
    value1->GENERATE  = coc_CopySet (value2->GENERATE);
    }

  if (( value1->PROBABACK == 0.5 ) && ( value2->PROBABACK != 0.5 ))
    {
    value1->PROBABACK = value2->PROBABACK;
    }

  return(errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_resolvalue					*/
/* description	: resolution function for typed signals			*/
/*		  return flag:0 ok, 1 error				*/
/*		  return default value 0				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int cox_resolvalue (projected_bivalue,
                    resolution_function_type,
                    current_value,
                    result_value)

  struct prjbvl  *projected_bivalue;
  char		  resolution_function_type;
  struct coxval  *current_value;
  struct coxval  *result_value;

  {
  int             errflg = 0;
  struct prjbvl  *tmp_bivalue;

  if ( COX_EVALUATION & COX_LEVELIZED_EVALUATION )
    {
    for ( tmp_bivalue = projected_bivalue;
          tmp_bivalue != NULL ;
          tmp_bivalue = tmp_bivalue->NEXT)
      {
      if ( (((struct coxval *)(tmp_bivalue->CNDVAL))->EVALUATED
                                                      == COX_NOT_EVALUATED) ||
           (((struct coxval *)(tmp_bivalue->DRVVAL))->EVALUATED
                                                      == COX_NOT_EVALUATED) )
        {
        return(errflg);
        }
      }
    result_value->EVALUATED = COX_EVALUATED;
    }

  switch (resolution_function_type)
    {
    case 'R' :
      result_value->PROPAGATE = NULL;
      result_value->EXTRACTED = NULL;
      result_value->GENERATE  = current_value->GENERATE;

      break;

    case 'W' :
      result_value->PROPAGATE = NULL;
      result_value->EXTRACTED = NULL;
      result_value->GENERATE  = current_value->GENERATE;

      break;

    case 'M' :
      result_value->PROPAGATE = NULL;
      result_value->EXTRACTED = NULL;
      result_value->GENERATE  = current_value->GENERATE;

      break;

    default :
      cox_toolbug (22, "cox_resolvalue", NULL, resolution_function_type);
      break;
    }

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_evalvalue						*/
/* description	: computes a BDD's value depending on its node's value	*/
/* 		  and on its expression (here the variable list).	*/
/* called func.	: vhu_toolbug						*/
/* ###--------------------------------------------------------------### */

int cox_evalvalue (instance , expression, target_type, value)

struct lkdins  *instance;
void           *expression;
unsigned char   target_type;
struct coxval  *value;
  {
  int                errflg = 0;
  void              *index;
  struct chain      *var_list;
  struct coxval     *value_locale;
  struct coxcol     *ext_localset    = NULL;
  struct coxcol     *pro_localset    = NULL;
  struct coxcol     *ext_colorset    = NULL;
  struct coxcol     *pro_colorset    = NULL;
  struct coxcol     *prev_propagate  = NULL;
  unsigned int       param           = 1   ;
  float              fparam          = 1.0 ;
  unsigned int       var_nbr         = 0   ;
  unsigned int       i               = 0   ;
/* 3 lignes Ajoute par Juju pour sa these */
  static      NSupgates = 0;
  static      NGates    = 0;
  static      Vtyp      = 0;
  static      Vmax      = 0;
  struct chain     *var1 = NULL;
  struct chain     *var2 = NULL;
  unsigned int      lvl1 = 0   ;
  unsigned int      lvl2 = 0   ;

  if ( COX_EVALUATION & COX_LEVELIZED_EVALUATION )
    {
    for ( var_list = PRB_SUPERVARLIST(expression);
          var_list != NULL;
          var_list = var_list->NEXT )
      {
      value_locale = (struct coxval *)
                     gss_getvalue(instance, var_list->DATA);

      if (value_locale->EVALUATED == COX_NOT_EVALUATED)
        {
        return(errflg);
        }
      }

    value->EVALUATED = COX_EVALUATED;
    }

  switch ( target_type )
    {
    case GSS_SIGDEF :
    case GSS_AUXDEF :
      {
      if (COX_HEURISTIC & COX_DECREM_HEURISTIC) 
        {
        coc_FreeSet (value->EXTRACTED);
        coc_FreeSet (value->PROPAGATE);
        value->EXTRACTED = NULL;
        value->PROPAGATE = NULL;

        var_nbr = 0;
        for ( var_list = PRB_SUPERVARLIST(expression);
              var_list != NULL; var_list = var_list->NEXT )
          {
          value_locale = (struct coxval *) gss_getvalue(instance,
                                                        var_list->DATA);

          pro_localset = coc_CreateUnionSet (value_locale->GENERATE,
                                             value_locale->PROPAGATE);

          ext_localset = coc_CreateIntersectionSet (value->PROPAGATE,
                                                    pro_localset);

          ext_colorset = coc_CreateUnionSet (value->EXTRACTED,
                                             ext_localset);

          pro_colorset = coc_CreateUnionSet (value->PROPAGATE,
                                             pro_localset);

          coc_FreeSet (pro_localset);
          coc_FreeSet (ext_localset);
          coc_FreeSet (value->EXTRACTED);
          coc_FreeSet (value->PROPAGATE);

          value->EXTRACTED = ext_colorset;
          value->PROPAGATE = pro_colorset;
          var_nbr ++;
          }

        for  (ext_colorset = value->EXTRACTED;
              ext_colorset != NULL; ext_colorset = ext_colorset->NEXT)
          {
          PRB_SUPERABL(expression) =
                 cox_createsuperabl(PRB_SUPERABL(expression),
                                     ext_colorset, instance);

          PRB_SUPERFLAG(expression) |= DEF_PRB_SUPERGATE;
          }

        PRB_SUPERBDD(expression) =
               beh_abl2bdd( instance->BEFIG, PRB_SUPERABL(expression) );

        PRB_SUPERVARLIST(expression) =
               beh_abl2varlist( instance->BEFIG, PRB_SUPERABL(expression) );

/* Lignes Ajoutees Pour reevaluer les couleurs propagees apres construction */
/*   de la SUpergate.. Meme chose pour dunoyer et roethig heuristics        */

        coc_FreeSet (value->PROPAGATE);
        value->PROPAGATE = NULL;
        var_nbr = 0;
        for ( var_list = PRB_SUPERVARLIST(expression);
              var_list != NULL; var_list = var_list->NEXT )
          {
          value_locale = (struct coxval *) gss_getvalue(instance,
                                                        var_list->DATA);

          pro_localset = coc_CreateUnionSet (value_locale->GENERATE,
                                             value_locale->PROPAGATE);

          pro_colorset = coc_CreateUnionSet (value->PROPAGATE,
                                             pro_localset);

          coc_FreeSet (pro_localset);
          coc_FreeSet (value->PROPAGATE);

          value->PROPAGATE = pro_colorset;
          var_nbr ++;
          }

        if (COX_HEURISTIC & COX_PEDRAM_HEURISTIC) 
          {
          param = 1 ;
          }
        else if (COX_HEURISTIC & COX_SETH_HEURISTIC) 
          {
          if ( var_nbr > 1 ) param = 1;
          else               param = 0;
          }
        else
          {
          param = 0 ;
          }

        coc_DecremPropagate( value->PROPAGATE, param );

        value->PROPAGATE = coc_Levelize( value->PROPAGATE );
        }

      else if (COX_HEURISTIC & COX_DIVIDE_HEURISTIC) 
        {
        coc_FreeSet (value->EXTRACTED);
        coc_FreeSet (value->PROPAGATE);
        value->EXTRACTED = NULL;
        value->PROPAGATE = NULL;

        var_nbr = 0;
        for ( var_list = PRB_SUPERVARLIST(expression);
              var_list != NULL; var_list = var_list->NEXT )
          {
          value_locale = (struct coxval *) gss_getvalue(instance,
                                                        var_list->DATA);

          pro_localset = coc_CreateUnionSet (value_locale->GENERATE,
                                             value_locale->PROPAGATE);

          ext_localset = coc_CreateIntersectionSet (value->PROPAGATE,
                                                    pro_localset);

          ext_colorset = coc_CreateUnionSet (value->EXTRACTED,
                                             ext_localset);

          pro_colorset = coc_CreateUnionSet (value->PROPAGATE,
                                             pro_localset);

          coc_FreeSet (pro_localset);
          coc_FreeSet (ext_localset);
          coc_FreeSet (value->EXTRACTED);
          coc_FreeSet (value->PROPAGATE);

          value->EXTRACTED = ext_colorset;
          value->PROPAGATE = pro_colorset;
          var_nbr ++;
          }

        for  (ext_colorset = value->EXTRACTED;
              ext_colorset != NULL; ext_colorset = ext_colorset->NEXT)
          {
          PRB_SUPERABL(expression) =
                 cox_createsuperabl(PRB_SUPERABL(expression),
                                     ext_colorset, instance);

          PRB_SUPERFLAG(expression) |= DEF_PRB_SUPERGATE;
          }

        PRB_SUPERBDD(expression) =
               beh_abl2bdd( instance->BEFIG, PRB_SUPERABL(expression) );

        PRB_SUPERVARLIST(expression) =
               beh_abl2varlist( instance->BEFIG, PRB_SUPERABL(expression) );

/* Lignes Ajoutees Pour reevaluer les couleurs propagees apres construction */
/*   de la SUpergate.. Meme chose pour dunoyer et roethig heuristics        */

        coc_FreeSet (value->PROPAGATE);
        value->PROPAGATE = NULL;
        var_nbr = 0;
        for ( var_list = PRB_SUPERVARLIST(expression);
              var_list != NULL; var_list = var_list->NEXT )
          {
          value_locale = (struct coxval *) gss_getvalue(instance,
                                                        var_list->DATA);

          pro_localset = coc_CreateUnionSet (value_locale->GENERATE,
                                             value_locale->PROPAGATE);

          pro_colorset = coc_CreateUnionSet (value->PROPAGATE,
                                             pro_localset);

          coc_FreeSet (pro_localset);
          coc_FreeSet (value->PROPAGATE);

          value->PROPAGATE = pro_colorset;
          var_nbr ++;
          }


        if (COX_HEURISTIC & COX_ROETHIG_HEURISTIC) 
          {
          param = var_nbr;
          }
        else if (COX_HEURISTIC & COX_DUNOYER_HEURISTIC) 
          {
          param = 1;
          for (i = 1; (i < var_nbr) && (i < 30); i++)
            param = 2 * param;
          }
        else
          {
          param = 1 ;
          }

        coc_DividePropagate( value->PROPAGATE, param );

        value->PROPAGATE = coc_Levelize( value->PROPAGATE );
        }

      else if (COX_HEURISTIC & COX_BACKPACK_HEURISTIC) 
        {
/* Modifs pour l'introduction de la sensibilite et du coeff de correl  */
/* 1. Dans la recherche des couleurs extraites : on les prend toutes,  */
/*    sans modifier les niveaux.                                       */
/* 2. Pour chaque couleur extraite, on recherche le couple d'entrees   */
/*    avec les couleurs de niveau max                                  */
/* 3. Pour cette couleur, le BDD de F, le couple de variables trouvees */
/*    on calcule la sensibilite (le facteur flottant fparam)           */
/* 4. On modifie en consequence la liste des couleurs extraites pour   */
/*    ne conserver qu'une partie des couleurs                          */

        coc_FreeSet (value->EXTRACTED);
        coc_FreeSet (value->PROPAGATE);
        value->EXTRACTED = NULL;
        value->PROPAGATE = NULL;

        for ( var_list = PRB_SUPERVARLIST(expression);
              var_list != NULL; var_list = var_list->NEXT )
          {
          value_locale = (struct coxval *) gss_getvalue(instance,
                                                        var_list->DATA);

          pro_localset = coc_CreateUnionSet (value_locale->GENERATE,
                                             value_locale->PROPAGATE);

          ext_localset = coc_CreateIntersectionSet (value->PROPAGATE,
                                                    pro_localset);

          ext_colorset = coc_CreateUnionSet (value->EXTRACTED,
                                             ext_localset);

          pro_colorset = coc_CreateUnionSet (value->PROPAGATE,
                                             pro_localset);

          coc_FreeSet (pro_localset);
          coc_FreeSet (ext_localset);
          coc_FreeSet (value->EXTRACTED);
          coc_FreeSet (value->PROPAGATE);

          value->EXTRACTED = ext_colorset;
          value->PROPAGATE = pro_colorset;
          }

/* Etape 2, on cherche 2 variables pour chaque couleur extraite        */
        for  (ext_colorset = value->EXTRACTED;
              ext_colorset != NULL; ext_colorset = ext_colorset->NEXT)
          {
          var1 = NULL;
          var2 = NULL;
          lvl1 = 0;
          lvl2 = 0;
          for ( var_list = PRB_SUPERVARLIST(expression);
                var_list != NULL; var_list = var_list->NEXT )
            {
            value_locale = (struct coxval *) gss_getvalue(instance,
                                                          var_list->DATA);

            for  (pro_colorset = value_locale->PROPAGATE;
                  pro_colorset != NULL; pro_colorset = pro_colorset->NEXT)
              {
              if ( COC_ELEMENT_COMPARE(pro_colorset->FUNDAMENTAL,
                                       ext_colorset->FUNDAMENTAL) == 0 )
                {
                if (pro_colorset->LEVEL > lvl1)
                  {
                  var2 = var1;
                  lvl2 = lvl1;
                  var1 = var_list;
                  lvl1 = pro_colorset->LEVEL;
                  }
                else if (pro_colorset->LEVEL > lvl2)
                  {
                  var2 = var_list;
                  lvl2 = pro_colorset->LEVEL;
                  }
                break;
                }
              }

            if ( value_locale->GENERATE != NULL )
              {
              if ( COC_ELEMENT_COMPARE( value_locale->GENERATE->FUNDAMENTAL,
                                        ext_colorset->FUNDAMENTAL) == 0 )
                {
                var2 = var1;
                lvl2 = lvl1;
                var1 = var_list;
                lvl1 =  value_locale->GENERATE->LEVEL;
                }
              }
            }


/* Etape 3, on calcule la sensibilite                                  */

          if (var1->DATA > var2->DATA)
            {
/* printf("Entree Sensib pour indexvar1= %d, indexvar2= %d | ", var1->DATA, var2->DATA); */
            fparam = cox_BackSensibility( PRB_SUPERBDD(expression),
                                                       instance,
                                                       var1->DATA,
                                                       var2->DATA);
            }
          else
            {
/* printf("Entree Sensib pour indexvar1= %d, indexvar2= %d | ", var2->DATA, var1->DATA); */
            fparam = cox_BackSensibility( PRB_SUPERBDD(expression),
                                          instance,
                                          var2->DATA,
                                          var1->DATA);
            }
/* printf("EXTRACTION : fparam vaut %-4.4f pour index %d et %d\n", fparam, var1->DATA, var2->DATA); */
          coc_BackExtracted (ext_colorset, fparam, lvl1, lvl2);
          }

/* Etape 4, on vire les couleurs inutiles                              */

        value->EXTRACTED = coc_Levelize( value->EXTRACTED );

        for  (ext_colorset = value->EXTRACTED;
              ext_colorset != NULL; ext_colorset = ext_colorset->NEXT)
          {
          PRB_SUPERABL(expression) =
                 cox_createsuperabl(PRB_SUPERABL(expression),
                                     ext_colorset, instance);

          PRB_SUPERFLAG(expression) |= DEF_PRB_SUPERGATE;
          }

        PRB_SUPERBDD(expression) =
               beh_abl2bdd( instance->BEFIG, PRB_SUPERABL(expression) );

        PRB_SUPERVARLIST(expression) =
               beh_abl2varlist( instance->BEFIG, PRB_SUPERABL(expression) );

/* Attention, la version update_setlevel  ne marchant pas, on utilise */
/* le procede plus conventionnel de recalcul du PROPAGATE             */
/* ce qui est commente correspond a updatesetlevel                    */


        coc_FreeSet (value->PROPAGATE);
        value->PROPAGATE = NULL;

        var_nbr = 0;
        for ( var_list = PRB_SUPERVARLIST(expression);
              var_list != NULL; var_list = var_list->NEXT )
          {
          var_nbr ++;
          }

        if (var_nbr < 31)
          {
        for ( var_list = PRB_SUPERVARLIST(expression);
              var_list != NULL; var_list = var_list->NEXT )
          {
          value_locale = (struct coxval *) gss_getvalue(instance,
                                                        var_list->DATA);

          pro_localset = coc_CreateUnionSet (value_locale->GENERATE,
                                             value_locale->PROPAGATE);

          if (pro_localset != NULL)
            {
            fparam = cox_BackCoeffCorrel (PRB_SUPERBDD(expression),
                                          instance, var_list->DATA);

/* printf("PROPAGATION : fparam vaut %-4.4f pour index %d\n", fparam, var_list->DATA); */

            coc_BackPropagate (pro_localset, fparam);
            }

          pro_colorset = coc_CreateUnionSet (value->PROPAGATE,
                                             pro_localset);

          coc_FreeSet (pro_localset);
          coc_FreeSet (value->PROPAGATE);

          value->PROPAGATE = pro_colorset;
          }

          }

        value->PROPAGATE = coc_Levelize( value->PROPAGATE );          
        }

      else if (COX_HEURISTIC == COX_NO_HEURISTIC) 
        {
        coc_FreeSet (value->EXTRACTED);
        coc_FreeSet (value->PROPAGATE);
        value->EXTRACTED = NULL;
        value->PROPAGATE = NULL;

        var_nbr = 0;
        for ( var_list = PRB_SUPERVARLIST(expression);
              var_list != NULL; var_list = var_list->NEXT )
          {
          value_locale = (struct coxval *) gss_getvalue(instance,
                                                        var_list->DATA);

          pro_localset = coc_CreateUnionSet (value_locale->GENERATE,
                                             value_locale->PROPAGATE);

          ext_localset = coc_CreateIntersectionSet (value->PROPAGATE,
                                                    pro_localset);

          ext_colorset = coc_CreateUnionSet (value->EXTRACTED,
                                             ext_localset);

          pro_colorset = coc_CreateUnionSet (value->PROPAGATE,
                                             pro_localset);

          coc_FreeSet (pro_localset);
          coc_FreeSet (ext_localset);
          coc_FreeSet (value->EXTRACTED);
          coc_FreeSet (value->PROPAGATE);

          value->EXTRACTED = ext_colorset;
          value->PROPAGATE = pro_colorset;
          var_nbr ++;
          }

/* 7 lignes ajoutees par Juju pour sa these               */
/* pour tester un truc (doit etre laisse en commentaire   */
/*----------
        var_nbr = 0;
        for  (ext_colorset = value->EXTRACTED;
              ext_colorset != NULL; ext_colorset = ext_colorset->NEXT)
          {
          var_nbr++;
          }
if (var_nbr <= 24)
------------*/

        for  (ext_colorset = value->EXTRACTED;
              ext_colorset != NULL; ext_colorset = ext_colorset->NEXT)
          {

          PRB_SUPERABL(expression) =
                 cox_createsuperabl(PRB_SUPERABL(expression),
                                     ext_colorset, instance);

          PRB_SUPERFLAG(expression) |= DEF_PRB_SUPERGATE;
          }

        PRB_SUPERVARLIST(expression) =
               beh_abl2varlist( instance->BEFIG, PRB_SUPERABL(expression) );

/* lignes commentees par Juju pour sa these               */
/* Quand il teste le temps d'execution de la construction */
/* des Supergates sans le developpement des BDDs          */

        PRB_SUPERBDD(expression) =
               beh_abl2bdd( instance->BEFIG, PRB_SUPERABL(expression) );

        }

/* 21 lignes Ajoute par Juju pour sa these */
      var_nbr = 0;
      for ( var_list = PRB_SUPERVARLIST(expression);
            var_list != NULL; var_list = var_list->NEXT )
        {
        var_nbr ++;
        }

      NGates ++;
      if ( value->EXTRACTED != NULL )
        {
        NSupgates ++;
        Vtyp += var_nbr;

        if (var_nbr > Vmax)
          {
          Vmax = var_nbr;
          }
        printf ("%d\t", var_nbr);
        printf ("%d\t", NSupgates);
        printf ("%d\t", NGates);
        printf ("%-4.4f\t", (float)Vtyp / (float)NSupgates);
        printf ("%d\t", Vtyp);
        printf ("%d\n", Vmax);
        }
      break;
      }

    case GSS_REGDEF | GSS_CNDDEF:
    case GSS_REGDEF | GSS_DRVDEF:
    case GSS_BUSDEF | GSS_CNDDEF:
    case GSS_BUSDEF | GSS_DRVDEF:
    case GSS_BUXDEF | GSS_CNDDEF:
    case GSS_BUXDEF | GSS_DRVDEF:
    case GSS_MSGDEF :
      break;

    default :
      cox_toolbug (23,"cox_evalvalue",instance->INSNAME, target_type);
      break;
 
    }
  return(errflg);
  }
