
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: cox_simulator.c					*/
/* date		: Mar  1 1994						*/
/* version	: v1.00							*/
/* author	: DUNOYER Julien					*/
/* description	: This file contains the high level function :		*/
/*		  cox_simulator						*/
/*		  this  function do a symbolic simulation to extract	*/
/*		  correlation in a circuit.				*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <sys/time.h>

#include MUT_H
#include GSS_H
#include COX_H

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cox_simulator						*/
/* description	: symbolic simulation function for correlation		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */


struct dct_entry **cox_simulator  (pt_lofig, pt_befig)

  struct lofig      *pt_lofig;
  struct befig      *pt_befig;

  {
  struct gssseq     *pt_gssseq;
  struct lkdfig     *pt_lkdfig;
  struct funpack    *cox_package;
  char               cox_errflg = 0;
  int                cur_error;


  gss_init ();

  GSS_DELAY_MODE = GSS_ZERO_DELAY;

  cox_package = gss_addfunpack( cox_addvalue,
                                cox_freevalue,
                                cox_cmpvalue,
                                cox_cpyvalue,
                                cox_updvalue,
                                cox_evalvalue,
                                cox_resolvalue,
                                NULL,
                                NULL,
                                NULL);

  /* put the LOCONs of each instance  in the same order as the BEPOR.*/

  gss_resord (pt_lofig, pt_befig);

  /* restore the mode of ext. ports when the declared mode is LINKAGE*/

  gss_restor (pt_lofig, pt_befig);

  if (cox_errflg != 0)
    {
    cox_errflg = cox_error (300, NULL);
    exit (1);
    }

        /* ###------------------------------------------------------### */
        /*    Link previous structures to create the simulator's data   */
        /* base structures                                              */
        /* ###------------------------------------------------------### */

  cox_message (1, NULL, 0);

  pt_lkdfig   = gss_linker (pt_lofig, cox_package);

        /* ###------------------------------------------------------### */
        /*    Running the sequence description loader                   */
        /* ###------------------------------------------------------### */

  cox_message (2, NULL, 0);

  pt_gssseq  = cox_loadsequence (pt_lofig, pt_befig);

  cox_message (3, NULL, 0);

  gss_lnkgssseq (pt_gssseq);

  /* ###------------------------------------------------------------### */
  /*    Simulating the current figure                                   */
  /* ###------------------------------------------------------------### */

        /* ###------------------------------------------------------### */
        /*    Reaching the stability.                                   */
        /* ###------------------------------------------------------### */

  cox_message (4, NULL, 0);

  gss_circuitinit(pt_lkdfig, COX_INIT_VALUE, cox_package);

  gss_forcesequence (pt_gssseq, cox_package);

  gss_circuitstable(pt_lkdfig, cox_package);

  gss_frelkdfig (pt_lkdfig, cox_package);
  }
