# include MUT_H
# include CST_H



/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_AddSetElt.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_AddSetElement					*/
/* description	: Add a pointer element in a pointer set.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

chain_list  *cst_AddSetElement (set,element)

chain_list    *set     ;
void          *element ;

  {
  chain_list  *set_pnt = NULL;
  chain_list  *sav_pnt = NULL;

  CST_STATUS = CST_SUCCEED;

  if ( set == NULL )
    {
    set = addchain ( set, element );
    return( set );
    }

  if ( CST_ELEMENT_COMPARE(element,set->DATA) == 0 )
    {
    return( set );
    }

  if ( CST_ELEMENT_COMPARE(element,set->DATA) < 0 )
    {
    set = addchain ( set, element );
    return( set );
    }

  sav_pnt = set;
  set_pnt = set->NEXT;

  while ( set_pnt != NULL )
    {
    if ( CST_ELEMENT_COMPARE(element,set_pnt->DATA) == 0 )
      {
      return( set );
      }

    if ( CST_ELEMENT_COMPARE(element,set_pnt->DATA) < 0 )
      {
      set_pnt = addchain ( set_pnt, element );
      sav_pnt->NEXT = set_pnt;
      return( set );
      }

    sav_pnt = set_pnt;
    set_pnt = set_pnt->NEXT;
    }

  set_pnt = addchain ( set_pnt, element );
  sav_pnt->NEXT = set_pnt;
  return (set);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_CrtCompSet.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_CreateComplementSet				*/
/* description	: Create The Complement between two pointers sets.	*/
/*                This mean set1 - set2.				*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

chain_list  *cst_CreateComplementSet (set1, set2)

chain_list    *set1     ;
chain_list    *set2     ;

  {
  chain_list  *set_pnt  = NULL;

  CST_STATUS = CST_SUCCEED;

  while ((set1 != NULL) && (set2 != NULL))
    {
    if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) < 0 )
      {
      set_pnt = addchain (set_pnt, set1->DATA);
      set1 = set1->NEXT;
      }
    else
      {
      if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) == 0 )
        {
        set1 = set1->NEXT;
        set2 = set2->NEXT;
        }
      else
        {
        cst_error(5, "cst_CreateComplementSet", NULL, 0);
        set2 = set2->NEXT;
        }
      }
    }

  while (set1 != NULL)
    {
    set_pnt = addchain (set_pnt, set1->DATA);
    set1 = set1->NEXT;
    }

  if (set2 != NULL)
    {
    cst_error(5, "cst_CreateComplementSet", NULL, 0);
    }

  set_pnt = reverse (set_pnt);
  return (set_pnt);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_CrtInterSet.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_CreateIntersectionSet				*/
/* description	: Create The Intersection between two pointers sets.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

chain_list  *cst_CreateIntersectionSet (set1, set2)

chain_list    *set1     ;
chain_list    *set2     ;

  {
  chain_list  *set_pnt  = NULL;

  CST_STATUS = CST_SUCCEED;

  while ((set1 != NULL) && ( set2 != NULL))
    {
    if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) == 0 )
      {
      set_pnt = addchain (set_pnt, set1->DATA);
      set1 = set1->NEXT;
      set2 = set2->NEXT;
      }
    else
      {
      if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) < 0 )
        {
        set1 = set1->NEXT;
        }
      else
        {
        set2 = set2->NEXT;
        }
      }
    }

  set_pnt = reverse( set_pnt);

  return (set_pnt);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_CrtUnionSet.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_CreateUnionSet					*/
/* description	: Create The Union between two pointers sets.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

chain_list  *cst_CreateUnionSet (set1, set2)

chain_list    *set1     ;
chain_list    *set2     ;

  {
  chain_list  *set_pnt  = NULL;

  CST_STATUS = CST_SUCCEED;

  while ((set1 != NULL) && ( set2 != NULL))
    {
    if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) == 0 )
      {
      set_pnt = addchain (set_pnt, set1->DATA);
      set1 = set1->NEXT;
      set2 = set2->NEXT;
      }
    else
      {
      if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) < 0 )
        {
        set_pnt = addchain (set_pnt, set1->DATA);
        set1 = set1->NEXT;
        }
      else
        {
        set_pnt = addchain (set_pnt, set2->DATA);
        set2 = set2->NEXT;
        }
      }
    }

  while (set1 != NULL)
    {
    set_pnt = addchain (set_pnt, set1->DATA);
    set1 = set1->NEXT;
    }

  while (set2 != NULL)
    {
    set_pnt = addchain (set_pnt, set2->DATA);
    set2 = set2->NEXT;
    }

  set_pnt = reverse(set_pnt);

  return (set_pnt);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_DisjunctSet.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_DisjunctiveSet					*/
/* description	: check The Disjunction between two pointers sets.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

int cst_DisjunctiveSet (set1, set2)

chain_list    *set1     ;
chain_list    *set2     ;

  {
  CST_STATUS = CST_SUCCEED;

  while ((set1 != NULL) && ( set2 != NULL))
    {
    if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) == 0 )
      {
      return (0);
      }
    else
      {
      if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) < 0 )
        {
        set1 = set1->NEXT;
        }
      else
        {
        set2 = set2->NEXT;
        }
      }
    }

  return (1);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_GetFirstElt.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_GetFirstElement					*/
/* description	: Get the first pointer element in a pointer set.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void  *cst_GetFirstElement (set)

chain_list    *set     ;

  {
  void  *element = NULL;

  CST_STATUS = CST_SUCCEED;

  if ( set == NULL )
    {
    CST_STATUS = CST_FAILED;
    }
  else
    {
    element = set->DATA;
    }

  return ( element );
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_GetLastElt.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_GetLastElement					*/
/* description	: Get the last pointer element in a pointer set.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void  *cst_GetLastElement (set)

chain_list    *set     ;

  {
  void  *element = NULL;

  CST_STATUS = CST_SUCCEED;

  if ( set == NULL )
    {
    CST_STATUS = CST_FAILED;
    }
  else
    {
    while (set->NEXT != NULL)
      {
      set = set->NEXT;
      }
    element = set->DATA;
    }

  return ( element );
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_GetNextElt.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_GetNextElement					*/
/* description	: Get the next pointer element in a pointer set.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void  *cst_GetNextElement (set, current)

chain_list    *set     ;
void          *current ;

  {
  void  *element = NULL;

  CST_STATUS = CST_SUCCEED;

  if ( set == NULL )
    {
    cst_error(2,"cst_GetNextElement",NULL,0);
    return ( element );
    }

  while (set != NULL)
    {
    if ( CST_ELEMENT_COMPARE(current,set->DATA) < 0 )
      {
      element = set->DATA;
      break;
      }
    set = set->NEXT;
    }

  if (element == NULL)
    {
    CST_STATUS = CST_FAILED;
    }

  return ( element );
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_GetPrevElt.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_GetPreviousElement				*/
/* description	: Get the previous pointer element in a pointer set.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void  *cst_GetPreviousElement (set, current)

chain_list    *set     ;
void          *current ;

  {
  void        *element = NULL;
  chain_list  *sav_pnt = NULL;

  CST_STATUS = CST_SUCCEED;

  if ( set == NULL )
    {
    cst_error(2,"cst_GetPreviousElement",NULL,0);
    }

  while (set != NULL)
    {
    if ( CST_ELEMENT_COMPARE(current,set->DATA) <= 0 )
      {
      break;
      }
    sav_pnt = set;
    set = set->NEXT;
    }

  if (sav_pnt == NULL)
    {
    CST_STATUS = CST_FAILED;
    }
  else
    {
    element = sav_pnt->DATA;
    }

  return ( element );
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_IsInSetElt.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_IsInSetElement					*/
/* description	: Is a pointer element in a pointer set.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

int cst_IsInSetElement (set,element)

chain_list    *set     ;
void          *element ;

  {
  chain_list  *set_pnt = NULL;

  CST_STATUS = CST_SUCCEED;

  if ( set == NULL )
    {
    cst_error(2,"cst_IsInSetElement",NULL,0);
    return( 0 );
    }

  set_pnt = set;

  while ( set_pnt != NULL )
    {
    if ( CST_ELEMENT_COMPARE(element,set_pnt->DATA) == 0 )
      {
      return( 1 );
      }

    if ( CST_ELEMENT_COMPARE(element,set_pnt->DATA) < 0 )
      {
      return( 0 );
      }

    set_pnt = set_pnt->NEXT;
    }

  return ( 0 );
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_RemSetElt.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_RemSetElement					*/
/* description	: Remove a pointer element in a pointer set.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

chain_list  *cst_RemSetElement (set,element)

chain_list    *set     ;
void          *element ;

  {
  chain_list  *set_pnt = NULL;

  CST_STATUS = CST_SUCCEED;

  if ( set == NULL )
    {
    cst_error(2,"cst_RemSetElement",NULL,0);
    return( set );
    }

  set_pnt = set;

  while ( set_pnt != NULL )
    {
    if ( CST_ELEMENT_COMPARE(element,set_pnt->DATA) == 0 )
      {
      set = delchain (set, set_pnt);
      return( set );
      }

    if ( CST_ELEMENT_COMPARE(element,set_pnt->DATA) < 0 )
      {
      cst_error(4,"cst_RemSetElement",NULL,0);
      return( set );
      }

    set_pnt = set_pnt->NEXT;
    }

  cst_error(4,"cst_RemSetElement",NULL,0);
  return (set);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cst_CmpSet.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_SetCompare					*/
/* description	: compare two sets.					*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

int  cst_SetCompare (set1, set2)

chain_list    *set1     ;
chain_list    *set2     ;

  {
  CST_STATUS = CST_SUCCEED;

  while ((set1 != NULL) && ( set2 != NULL))
    {
    if ( CST_ELEMENT_COMPARE(set1->DATA, set2->DATA) == 0 )
      {
      set1 = set1->NEXT;
      set2 = set2->NEXT;
      }
    else
      return(1);
    }

  if ((set1 != NULL) || (set2 != NULL))
      return(1);

  return (0);
  }

/* ###--------------------------------------------------------------### */
/*									*/
/* file		: cst_error.c						*/
/* date		: Dec 24 1994						*/
/* version	: v000							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: system level functions				*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: cst_error						*/
/* description	: print an error message on standard error output	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void cst_error (code, str1, str2, dat1)

int   code;
char *str1;
char *str2;
int   dat1;

  {
  CST_ERRCNT++;
  CST_ERRCOD = code;
  CST_STATUS = CST_ERROR;

  if (((CST_MSGLVL & CST_MSG__ERR) != 0) && ((CST_ERRCNT < CST_MAXERR)))
    {
    fprintf (stderr, "Cst Error %d in func. `%s`: ", code, str1);

    switch (code)
      {

      default:
        fprintf (stderr,"unknown error: %d\n", code);
        break;
      }
    }
  if (CST_ERRCNT == CST_MAXERR)
    fprintf (stderr ,"Too many errors : cant explain them further more"); 
  }

/* ###--------------------------------------------------------------### */
/*									*/
/* file		: cst_globals.c						*/
/* date		: Dec 17 1994						*/
/* version	: v000							*/
/* authors	: DUNOYER Julien & ABDALLAH Nizar			*/
/* description	: global variables					*/
/*									*/
/* ###--------------------------------------------------------------### */


unsigned long  CST_STATUS  = CST_SUCCEED;	/* function call status	*/

unsigned long  CST_MSGLVL  = CST_MSG__ERR;	/* message level	*/

unsigned int   CST_ERRCOD  =  0;		/* error code		*/
unsigned int   CST_ERRCNT  =  0;		/* error count		*/
unsigned int   CST_MAXERR  = 10;		/* maximum error count	*/
