
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: gss_lspec.c						*/
/* date		: Feb 13 1995						*/
/* version	: v1.00							*/
/* author	: DUNOYER Julien					*/
/* contents	: This file contains the set of functions used by the	*/
/*		  linker to build data structures, and to check the	*/
/*		  consistency of the description for a generic 		*/
/*		  symbolic simulation :					*/
/*									*/
/*		  gss_addlkdfig , gss_addlkdins  , gss_addsimsig ,	*/
/*		  gss_addbussig , gss_addwrbsig  , gss_addwrireg ,	*/
/*		  gss_addwriaux , gss_addwribux  , gss_addprjbvl ,	*/
/*		  gss_linker    , gss_restor     , gss_resord    ,	*/
/*		  gss_addredlst , gss_makgst     ,			*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include MLO_H
#include "beh.h"
#include GSS_H

static void gss_fresimsig ();
static void gss_frebussig ();
static void gss_frelkdins ();

/* ###--------------------------------------------------------------### */
/* function	: gss_addprjbvl						*/
/* description	: Used to add an projected bi-value to an wrbsig	*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct prjbvl *gss_addprjbvl (lastprjbvl, usr_package)

struct prjbvl *lastprjbvl;
struct funpack *usr_package;

  {
  struct prjbvl *ptprjbvl;

  ptprjbvl = (struct prjbvl *)mbkalloc (sizeof(struct prjbvl));

  ptprjbvl->CNDVAL = (void *) gss_addvalue (usr_package) ;

  ptprjbvl->DRVVAL = (void *) gss_addvalue (usr_package) ;

  ptprjbvl->NEXT   = lastprjbvl;

  return (ptprjbvl);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_freprjbvl						*/
/* description	: free a list of SIMSIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_freprjbvl (listprjbvl, usr_package)

struct prjbvl *listprjbvl;
struct funpack *usr_package;

  {
  struct prjbvl *ptprjbvl;

  while (listprjbvl != NULL)
    {
    gss_freevalue (listprjbvl->CNDVAL, usr_package);
    gss_freevalue (listprjbvl->DRVVAL, usr_package);

    ptprjbvl   = listprjbvl;
    listprjbvl = listprjbvl->NEXT;
    mbkfree (ptprjbvl);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addlkdfig						*/
/* description	: prepare an empty LKDFIG structure			*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct lkdfig *gss_addlkdfig (ptlofig)

struct lofig *ptlofig;

  {
  struct lkdfig *ptlkdfig;

  ptlkdfig         = (struct lkdfig *) mbkalloc (sizeof(struct lkdfig));
  ptlkdfig->NEXT   = NULL;
  ptlkdfig->NAME   = ptlofig->NAME;
  ptlkdfig->LKDINS = NULL;
  ptlkdfig->SIMSIG = NULL;
  ptlkdfig->BUSSIG = NULL;

  return (ptlkdfig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frelkdfig						*/
/* description	: free a list of LKDFIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void gss_frelkdfig (listlkdfig, usr_package)

struct lkdfig *listlkdfig;
struct funpack *usr_package;

  {
  struct lkdfig *ptlkdfig;

  while (listlkdfig != NULL)
    {
    gss_frelkdins (listlkdfig->LKDINS, usr_package);
    gss_fresimsig (listlkdfig->SIMSIG, usr_package);
    gss_frebussig (listlkdfig->BUSSIG, usr_package);

    ptlkdfig   = listlkdfig;
    listlkdfig = listlkdfig->NEXT;
    mbkfree (ptlkdfig);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_adddepsig						*/
/* description	: Used to add an structure depsig 			*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct depsig *gss_adddepsig (lastdepsig, signal, type)

struct depsig  *lastdepsig;
void           *signal;
unsigned char   type;
  {
  struct depsig *ptdepsig;

  ptdepsig          = (struct depsig *)mbkalloc (sizeof(struct depsig));

  ptdepsig->SIGNAL  = signal;
  ptdepsig->TYPE    = type;

  ptdepsig->NEXT    = lastdepsig;

  return(ptdepsig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_fredepsig						*/
/* description	: Used to free an structure depsig 			*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static void gss_fredepsig (listdepsig)

struct depsig  *listdepsig;
  {
  struct depsig *ptdepsig;

  while (listdepsig != NULL)
    {
    ptdepsig   = listdepsig;
    listdepsig = listdepsig->NEXT;
    mbkfree (ptdepsig);
    }
  }
  
/* ###--------------------------------------------------------------### */
/* function	: gss_addsimsig						*/
/* description	: Used to add an structure simsig to an lkdfig		*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct simsig *gss_addsimsig (lastsimsig, ptlosig, usr_package)

struct simsig  *lastsimsig;
struct losig   *ptlosig;
struct funpack *usr_package;

  {
  struct simsig *ptsimsig;

  ptsimsig            = (struct simsig *)mbkalloc (sizeof(struct simsig));

  ptsimsig->DEPSIG    = NULL;

  ptsimsig->HISTORY   = gss_addevthis (NULL, 0, 0) ;

  ptsimsig->CURVAL    = (void *) gss_addvalue (usr_package) ;

  ptsimsig->PRJVAL    = (void *) gss_addvalue (usr_package) ;

  ptsimsig->NEXT      = lastsimsig;

  return (ptsimsig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_fresimsig						*/
/* description	: free a list of SIMSIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_fresimsig (listsimsig, usr_package)

struct simsig *listsimsig;
struct funpack *usr_package;

  {
  struct simsig *ptsimsig;

  while (listsimsig != NULL)
    {
    gss_fredepsig (listsimsig->DEPSIG);
    gss_freevthis (listsimsig->HISTORY);
    gss_freevalue (listsimsig->CURVAL, usr_package);
    gss_freevalue (listsimsig->PRJVAL, usr_package);

    ptsimsig   = listsimsig;
    listsimsig = listsimsig->NEXT;
    mbkfree (ptsimsig);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addbussig						*/
/* description	: Used to add an structure bussig to an lkdfig		*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct bussig *gss_addbussig (lastbussig, ptlosig, usr_package)

struct bussig *lastbussig;
struct losig  *ptlosig;
struct funpack *usr_package;

  {
  struct bussig *ptbussig;

  ptbussig            = (struct bussig *)mbkalloc (sizeof(struct bussig));

  ptbussig->DEPSIG    = NULL;

  ptbussig->HISTORY   = gss_addevthis (NULL, 0, 0);

  ptbussig->CURVAL    = (void *) gss_addvalue (usr_package) ;

  ptbussig->PRJVAL    = (void *) gss_addvalue (usr_package) ;

	/* ###------------------------------------------------------### */
	/*    The first PRJBVL is for external driver.			*/
	/* ###------------------------------------------------------### */

  ptbussig->PRJBVL    = gss_addprjbvl (NULL, usr_package);

  ptbussig->TYPE      = ptlosig->USER->TYPE;

  ptbussig->NEXT      = lastbussig;


  return (ptbussig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frebussig						*/
/* description	: free a list of BUSSIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_frebussig (listbussig, usr_package)

struct bussig *listbussig;
struct funpack *usr_package;

  {
  struct bussig *ptbussig;

  while (listbussig != NULL)
    {
    gss_fredepsig (listbussig->DEPSIG);
    gss_freevthis (listbussig->HISTORY);
    gss_freevalue (listbussig->CURVAL, usr_package);
    gss_freevalue (listbussig->PRJVAL, usr_package);
    gss_freprjbvl (listbussig->PRJBVL, usr_package);

    ptbussig   = listbussig;
    listbussig = listbussig->NEXT;
    mbkfree (ptbussig);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addwrbsig						*/
/* description	: Used to add an structure wrbsig to an lkdfig		*/
/* called funct	: mbkalloc, gss_addprjbvl				*/
/* ###--------------------------------------------------------------### */

static struct wrbsig *gss_addwrbsig (ptlkdins, ptbebus, usr_package)

struct lkdins  *ptlkdins;
struct bebus   *ptbebus;
struct funpack *usr_package;

  {
  struct wrbsig *ptwrbsig;
  struct binode *ptbinode;

  ptwrbsig = (struct wrbsig *)mbkalloc (sizeof(struct wrbsig));

  ptwrbsig->PRJBVL = NULL;
  ptwrbsig->BUSSIG = NULL;
  ptwrbsig->BEBUS  = ptbebus;
  ptwrbsig->LKDINS = ptlkdins;
  ptwrbsig->EVAL      = 0;

  ptbinode = ptbebus->BINODE;
  while (ptbinode != NULL)
    {
    ptwrbsig->PRJBVL = gss_addprjbvl (ptwrbsig->PRJBVL, usr_package);
    ptbinode         = ptbinode->NEXT;
    }

  ptwrbsig->NEXT   = ptlkdins->WRBSIG;
  ptlkdins->WRBSIG = ptwrbsig;

  return (ptwrbsig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frewrbsig						*/
/* description	: free a list of SIMSIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_frewrbsig (listwrbsig, usr_package)

struct wrbsig *listwrbsig;
struct funpack *usr_package;

  {
  struct wrbsig *ptwrbsig;

  while (listwrbsig != NULL)
    {
    gss_freprjbvl (listwrbsig->PRJBVL, usr_package);

    ptwrbsig   = listwrbsig;
    listwrbsig = listwrbsig->NEXT;
    mbkfree (ptwrbsig);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addwrireg						*/
/* description	: Used to add an structure wrireg to an lkdfig		*/
/* called funct	: mbkalloc, gss_addprjbvl				*/
/* ###--------------------------------------------------------------### */

static struct wrireg *gss_addwrireg (ptlkdins, ptbereg, usr_package)

struct lkdins *ptlkdins;
struct bereg  *ptbereg;
struct funpack *usr_package;

  {
  struct wrireg *ptwrireg;
  struct binode *ptbinode;

  ptwrireg = (struct wrireg *)mbkalloc (sizeof(struct wrireg));

  ptwrireg->LKDINS    = ptlkdins;
  ptwrireg->BEREG     = ptbereg;

  ptwrireg->DEPSIG    = NULL;
  ptwrireg->EVAL      = 0;

	/* ###------------------------------------------------------### */
	/*    The first PRJBVL is for external driver.			*/
	/* ###------------------------------------------------------### */

  ptwrireg->PRJBVL    = gss_addprjbvl (NULL, usr_package);

  ptwrireg->HISTORY   = gss_addevthis (NULL, 0, 0);

  ptwrireg->CURVAL    = (void *) gss_addvalue (usr_package) ;

  ptwrireg->PRJVAL    = (void *) gss_addvalue (usr_package) ;

  ptbinode = ptbereg->BINODE;
  while (ptbinode != NULL)
    {
    ptwrireg->PRJBVL = gss_addprjbvl (ptwrireg->PRJBVL, usr_package);
    ptbinode         = ptbinode->NEXT;
    }

  ptwrireg->NEXT   = ptlkdins->WRIREG;
  ptlkdins->WRIREG = ptwrireg;

  return (ptwrireg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frewrireg						*/
/* description	: free a list of WRIREG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_frewrireg (listwrireg, usr_package)

struct wrireg *listwrireg;
struct funpack *usr_package;

  {
  struct wrireg *ptwrireg;

  while (listwrireg != NULL)
    {
    gss_freevthis (listwrireg->HISTORY);
    gss_freevalue (listwrireg->CURVAL, usr_package);
    gss_freprjbvl (listwrireg->PRJBVL, usr_package);

    gss_fredepsig (listwrireg->DEPSIG);

    ptwrireg   = listwrireg;
    listwrireg = listwrireg->NEXT;
    mbkfree (ptwrireg);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addwribux						*/
/* description	: Used to add an structure wribux to an lkdfig		*/
/* called funct	: mbkalloc, gss_addprjbvl				*/
/* ###--------------------------------------------------------------### */

static struct wribux *gss_addwribux (ptlkdins, ptbebux, usr_package)

struct lkdins *ptlkdins;
struct bebux  *ptbebux;
struct funpack *usr_package;

  {
  struct wribux *ptwribux;
  struct binode *ptbinode;

  ptwribux = (struct wribux *)mbkalloc (sizeof(struct wribux));

  ptwribux->LKDINS    = ptlkdins;
  ptwribux->BEBUX     = ptbebux;

  ptwribux->DEPSIG    = NULL;
  ptwribux->EVAL      = 0;

	/* ###------------------------------------------------------### */
	/*    The first PRJBVL is for external driver.			*/
	/* ###------------------------------------------------------### */

  ptwribux->PRJBVL    = gss_addprjbvl (NULL, usr_package);

  ptwribux->HISTORY   = gss_addevthis (NULL, 0, 0);

  ptwribux->CURVAL    = (void *) gss_addvalue (usr_package) ;

  ptwribux->PRJVAL    = (void *) gss_addvalue (usr_package) ;

  ptbinode = ptbebux->BINODE;
  while (ptbinode != NULL)
    {
    ptwribux->PRJBVL = gss_addprjbvl (ptwribux->PRJBVL, usr_package);
    ptbinode         = ptbinode->NEXT;
    }

  ptwribux->NEXT   = ptlkdins->WRIBUX;
  ptlkdins->WRIBUX = ptwribux;

  return (ptwribux);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frewribux						*/
/* description	: free a list of WRIBUX structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_frewribux (listwribux, usr_package)

struct wribux *listwribux;
struct funpack *usr_package;

  {
  struct wribux *ptwribux;

  while (listwribux != NULL)
    {
    gss_freevthis (listwribux->HISTORY);
    gss_freevalue (listwribux->CURVAL, usr_package);
    gss_freprjbvl (listwribux->PRJBVL, usr_package);

    gss_fredepsig (listwribux->DEPSIG);

    ptwribux   = listwribux;
    listwribux = listwribux->NEXT;
    mbkfree (ptwribux);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addwriaux						*/
/* description	: Used to add an structure wriaux to an lkdfig		*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct wriaux *gss_addwriaux (ptlkdins, ptbeaux, usr_package)

struct lkdins *ptlkdins;
struct beaux  *ptbeaux;
struct funpack *usr_package;

  {
  struct wriaux *ptwriaux;

  ptwriaux = (struct wriaux *)mbkalloc (sizeof(struct wriaux));

  ptwriaux->LKDINS    = ptlkdins;
  ptwriaux->BEAUX     = ptbeaux;

  ptwriaux->DEPSIG    = NULL;
  ptwriaux->EVAL      = 0;

  ptwriaux->HISTORY   = gss_addevthis (NULL, 0, 0);

  ptwriaux->CURVAL         = (void *) gss_addvalue (usr_package) ;

  ptwriaux->PRJVAL         = (void *) gss_addvalue (usr_package) ;

  ptwriaux->NEXT   = ptlkdins->WRIAUX;
  ptlkdins->WRIAUX = ptwriaux;

  return (ptwriaux);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frewriaux						*/
/* description	: free a list of SIMSIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_frewriaux (listwriaux, usr_package)

struct wriaux *listwriaux;
struct funpack *usr_package;

  {
  struct wriaux *ptwriaux;

  while (listwriaux != NULL)
    {
    gss_freevthis (listwriaux->HISTORY);
    gss_freevalue (listwriaux->CURVAL, usr_package);
    gss_freevalue (listwriaux->PRJVAL, usr_package);

    gss_fredepsig (listwriaux->DEPSIG);

    ptwriaux   = listwriaux;
    listwriaux = listwriaux->NEXT;
    mbkfree (ptwriaux);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addwrimsg						*/
/* description	: Used to add an structure wrimsg to an lkdfig		*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct wrimsg *gss_addwrimsg (ptlkdins, ptbemsg, usr_package)

struct lkdins *ptlkdins;
struct bemsg  *ptbemsg;
struct funpack *usr_package;

  {
  struct wrimsg *ptwrimsg;

  ptwrimsg = (struct wrimsg *)mbkalloc (sizeof(struct wrimsg));

  ptwrimsg->LKDINS    = ptlkdins;
  ptwrimsg->BEMSG     = ptbemsg;

  ptwrimsg->EVAL      = 0;

  ptwrimsg->HISTORY   = gss_addevthis (NULL, 0, 0);

  ptwrimsg->CURVAL         = (void *) gss_addvalue (usr_package) ;

  ptwrimsg->PRJVAL         = (void *) gss_addvalue (usr_package) ;

  ptwrimsg->NEXT   = ptlkdins->WRIMSG;
  ptlkdins->WRIMSG = ptwrimsg;

  return (ptwrimsg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frewrimsg						*/
/* description	: free a list of SIMSIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_frewrimsg (listwrimsg, usr_package)

struct wrimsg *listwrimsg;
struct funpack *usr_package;

  {
  struct wrimsg *ptwrimsg;

  while (listwrimsg != NULL)
    {
    gss_freevthis (listwrimsg->HISTORY);
    gss_freevalue (listwrimsg->CURVAL, usr_package);
    gss_freevalue (listwrimsg->PRJVAL, usr_package);

    ptwrimsg   = listwrimsg;
    listwrimsg = listwrimsg->NEXT;
    mbkfree (ptwrimsg);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addwrssig						*/
/* description	: Used to add an structure wrssig to an lkdfig		*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct wrssig *gss_addwrssig (ptlkdins, ptbeout)

struct lkdins *ptlkdins;
struct beout  *ptbeout;

  {
  struct wrssig *ptwrssig;

  ptwrssig = (struct wrssig *) mbkalloc (sizeof(struct wrssig));

  ptwrssig->SIMSIG = NULL;
  ptwrssig->BEOUT  = ptbeout;
  ptwrssig->LKDINS = ptlkdins;
  ptwrssig->EVAL      = 0;

  ptwrssig->NEXT   = ptlkdins->WRSSIG;
  ptlkdins->WRSSIG = ptwrssig;

  return (ptwrssig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frewrssig						*/
/* description	: free a list of SIMSIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_frewrssig (listwrssig)

struct wrssig *listwrssig;

  {
  struct wrssig *ptwrssig;

  while (listwrssig != NULL)
    {
    ptwrssig   = listwrssig;
    listwrssig = listwrssig->NEXT;
    mbkfree (ptwrssig);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addredlst						*/
/* description	: Used to add an redlst to an lkdins			*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct redlst *gss_addredlst (count)

int count;

  {
  struct redlst *ptredlst;
  struct redlst *pnt;
  int            i;
/* printf("addredlist with count = %d\n", count); */
	/* ###------------------------------------------------------### */
	/*    The redlst is allocated in one block. The NEXT field	*/
	/* is not really used (more like a table than like a list)	*/
	/* => We use pnt++ rather than pnt->NEXT.			*/
	/* ###------------------------------------------------------### */

  ptredlst = (struct redlst *)mbkalloc (count * sizeof(struct redlst));

  pnt = ptredlst;
  for (i=1 ; i<count ; i++)
    {
    pnt->VALU = NULL;
    pnt->NEXT = pnt + 1;
    pnt++;
    }
  pnt->VALU = NULL;
  pnt->NEXT = NULL;

  return (ptredlst);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_freredlst						*/
/* description	: free a list of SIMSIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_freredlst (listredlst, count)

struct redlst *listredlst;
int            count;

  {
  struct redlst *ptredlst = NULL;
  int            i=0;
/* printf("freredlist with count = %d\n", count); */

  while (i < count)
    {
    ptredlst = listredlst;
    listredlst = listredlst++;
    mbkfree (ptredlst);
    i++;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_adddepend						*/
/* description	: Create the dependance ( depsig list ) using the berin	*/
/*		  depend informations for a signal.			*/
/* ###--------------------------------------------------------------### */

static struct depsig *gss_adddepend (ptdepsig, ptberin, ptlkdins)

struct depsig     *ptdepsig;
struct berin      *ptberin;
struct lkdins     *ptlkdins;

  {
  struct chain      *ptberef;
  unsigned char      type;
  char              *name;
  void              *signal;

  ptberef = ptberin->OUT_REF;

  while (ptberef != NULL)
    {
    name     =
     concatname (ptlkdins->INSNAME, ((struct beout*)(ptberef->DATA))->NAME);
    signal   = (void *) gss_chktab (GSS_DICTIONNARY, name, NULL, GSS_PNT0FIELD);

    type     = GSS_SIGDEF;
    ptdepsig = gss_adddepsig (ptdepsig, signal, type);
    ptberef  = ptberef->NEXT;
    }

  ptberef = ptberin->BUS_REF;

  while (ptberef != NULL)
    {
    name     =
     concatname (ptlkdins->INSNAME, ((struct bebus*)(ptberef->DATA))->NAME);
    signal   = (void *) gss_chktab (GSS_DICTIONNARY, name, NULL, GSS_PNT0FIELD);

    type     = GSS_BUSDEF;
    ptdepsig = gss_adddepsig (ptdepsig, signal, type);
    ptberef  = ptberef->NEXT;
    }

  ptberef = ptberin->AUX_REF;

  while (ptberef != NULL)
    {
    name     =
     concatname (ptlkdins->INSNAME, ((struct beaux*)(ptberef->DATA))->NAME);
    signal   = (void *) gss_chktab (GSS_DICTIONNARY, name, NULL, GSS_PNT0FIELD);

    type     = GSS_AUXDEF;
    ptdepsig = gss_adddepsig (ptdepsig, signal, type);
    ptberef  = ptberef->NEXT;
    }

  ptberef = ptberin->REG_REF;

  while (ptberef != NULL)
    {
    name     =
     concatname (ptlkdins->INSNAME, ((struct bereg*)(ptberef->DATA))->NAME);
    signal   = (void *) gss_chktab (GSS_DICTIONNARY, name, NULL, GSS_PNT0FIELD);

    type     = GSS_REGDEF;
    ptdepsig = gss_adddepsig (ptdepsig, signal, type);
    ptberef  = ptberef->NEXT;
    }

  ptberef = ptberin->BUX_REF;

  while (ptberef != NULL)
    {
    name     =
     concatname (ptlkdins->INSNAME, ((struct bebux*)(ptberef->DATA))->NAME);
    signal   = (void *) gss_chktab (GSS_DICTIONNARY, name, NULL, GSS_PNT0FIELD);

    type     = GSS_BUXDEF;
    ptdepsig = gss_adddepsig (ptdepsig, signal, type);
    ptberef  = ptberef->NEXT;
    }

  ptberef = ptberin->MSG_REF;

  while (ptberef != NULL)
    {
    name     =
     concatname (ptlkdins->INSNAME, ((struct bemsg*)(ptberef->DATA))->LABEL);
    signal   = (void *) gss_chktab (GSS_DICTIONNARY, name, NULL, GSS_PNT0FIELD);

    type     = GSS_MSGDEF;
    ptdepsig = gss_adddepsig (ptdepsig, signal, type);
    ptberef  = ptberef->NEXT;
    }

  ptberef = ptberin->DLY_REF;

  while (ptberef != NULL)
    {
    name     =
     concatname (ptlkdins->INSNAME, ((struct beaux*)(ptberef->DATA))->NAME);
    signal   = (void *) gss_chktab (GSS_DICTIONNARY, name, NULL, GSS_PNT0FIELD);

    type     = GSS_AUXDEF;
    ptdepsig = gss_adddepsig (ptdepsig, signal, type);
    ptberef  = ptberef->NEXT;
    }

  return (ptdepsig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addlkdins						*/
/* description	: prepare an empty LKDINS structure and then fill its	*/
/*		  fields.						*/
/* called funct	: gss_addwrbsig, gss_addwrireg, gss_addredlst,		*/
/*		  gss_addwriaux, gss_addwribux, gss_toolbug  ,		*/
/*		  gss_chktab   , gss_addtab   , reverse      ,		*/
/*		  concatname   , mbkalloc     , addchain		*/
/* ###--------------------------------------------------------------### */

static struct lkdins *gss_addlkdins (ptlkdfig, ptloins, usr_package)

struct lkdfig     *ptlkdfig;
struct loins      *ptloins;
struct funpack *usr_package;

  {
  int            count;
  struct locon  *ptlocon;
  struct losig  *ptlosig;
  struct befig  *ptbefig;
  struct berin  *ptberin;
  struct beout  *ptbeout;
  struct bebus  *ptbebus;
  struct bebux  *ptbebux;
  struct bereg  *ptbereg;
  struct beaux  *ptbeaux;
  struct bemsg  *ptbemsg;
  struct lkdins *ptlkdins;
  struct redlst *ptredlst;
  struct simsig *ptsimsig;
  struct bussig *ptbussig;
  struct wrbsig *ptwrbsig;
  struct prjbvl *ptprjbvl;
  struct wrssig *ptwrssig;
  struct wribux *ptwribux;
  struct wriaux *ptwriaux;
  struct wrireg *ptwrireg;
  char          *name;

  ptlkdins          = (struct lkdins *) mbkalloc (sizeof(struct lkdins));
  ptlkdins->INSNAME = ptloins->INSNAME;

	/* ###------------------------------------------------------### */
	/*    Extract from the dictionnary the  pointer of the BEFIG	*/
	/* related to the instance					*/
	/* ###------------------------------------------------------### */

  ptbefig = (struct befig *) gss_chktab (GSS_DICTIONNARY,ptloins->FIGNAME,NULL,GSS_PNT0FIELD);

  if (ptbefig == NULL)
    gss_toolbug (1,"gss_addlkdins",ptloins->INSNAME,0);

	/* ###------------------------------------------------------### */
	/*    Link with the instance's model				*/
	/* ###------------------------------------------------------### */

  ptlkdins->BEFIG   = ptbefig;
  ptlkdins->WRSSIG  = NULL;
  ptlkdins->WRBSIG  = NULL;
  ptlkdins->WRIAUX  = NULL;
  ptlkdins->WRIBUX  = NULL;
  ptlkdins->WRIREG  = NULL;
  ptlkdins->WRIMSG  = NULL;
  ptlkdins->REDLST  = NULL;
  ptlkdins->CIRCUI  = ptbefig->CIRCUI;

  ptlkdins->NEXT    = ptlkdfig->LKDINS;
  ptlkdfig->LKDINS  = ptlkdins;

	/* ###------------------------------------------------------### */
	/*    For each BERIN create an empty REDLST (due to evaluation	*/
	/* constraints redlst is organised as a table - see gss_evalue)	*/
	/* ###------------------------------------------------------### */

  count   = 0;
  ptberin = ptbefig->BERIN;
  while (ptberin != NULL)
    {
    count++;
    ptberin = ptberin->NEXT;
    }

  if (count == 0)
    ptlkdins->REDLST = NULL;
  else
    ptlkdins->REDLST = gss_addredlst (count);

	/* ###------------------------------------------------------### */
	/*    For each BEOUT create an empty WRSSIG			*/
	/* ###------------------------------------------------------### */

  ptbeout = ptbefig->BEOUT;
  while (ptbeout != NULL)
    {
    ptlkdins->WRSSIG = gss_addwrssig (ptlkdins, ptbeout);
    name         = concatname    (ptlkdins->INSNAME,ptbeout->NAME);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD,ptlkdins->WRSSIG);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_SIGFIELD,GSS_SIGDEF);

    ptbeout          = ptbeout->NEXT;
    }
  ptlkdins->WRSSIG = (struct wrssig *)
                     reverse ((struct chain *)ptlkdins->WRSSIG);

	/* ###------------------------------------------------------### */
	/*    For each BEBUS create an WRBSIG. The WRBSIG list must be	*/
	/* reversed to obtain the same order of ports as in BEBUS list.	*/
	/* ###------------------------------------------------------### */

  ptbebus = ptbefig->BEBUS;
  while (ptbebus != NULL)
    {
    ptlkdins->WRBSIG = gss_addwrbsig (ptlkdins, ptbebus, usr_package);
    name         = concatname    (ptlkdins->INSNAME,ptbebus->NAME);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD,ptlkdins->WRBSIG);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_SIGFIELD,GSS_BUSDEF);

    ptbebus          = ptbebus->NEXT;
    }
  ptlkdins->WRBSIG = (struct wrbsig *)
                     reverse ((struct chain *)ptlkdins->WRBSIG);


	/* ###------------------------------------------------------### */
	/*    For each BEBUX create an WRIBUX. The WRIBUX list must be	*/
	/* reversed to obtain the same order of ports as in BEBUX list.	*/
	/*    Scan the BERIN list (and REDLST) to find the primary input*/
	/* that has the same name and make a link			*/
	/* ###------------------------------------------------------### */

  ptbebux = ptbefig->BEBUX;
  while (ptbebux != NULL)
    {
    ptlkdins->WRIBUX = gss_addwribux (ptlkdins, ptbebux, usr_package);
    name         = concatname    (ptlkdins->INSNAME,ptbebux->NAME);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD,ptlkdins->WRIBUX);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_SIGFIELD,GSS_BUXDEF);

    ptberin  = ptbefig->BERIN;
    ptredlst = ptlkdins->REDLST;
    while ((char *)ptberin->NAME != ptbebux->NAME)
      {
      ptberin = ptberin->NEXT;
      ptredlst++;
      }

    ptredlst->HIST   = ptlkdins->WRIBUX->HISTORY;
    ptredlst->VALU   = ptlkdins->WRIBUX->CURVAL;
    ptbebux          = ptbebux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    The WRIBUX list must be reversed to obtain the same order	*/
	/* of as in BEBUX list.						*/
	/* ###------------------------------------------------------### */

  ptlkdins->WRIBUX = (struct wribux *)
                     reverse ((struct chain *)ptlkdins->WRIBUX);

	/* ###------------------------------------------------------### */
	/*    For each BEREG create an WRIREG. The WRIREG list must be	*/
	/* reversed to obtain the same order of ports as in BEREG list.	*/
	/*    Scan the BERIN list (and REDLST) to find the primary input*/
	/* that has the same name and make a link			*/
	/* ###------------------------------------------------------### */

  ptbereg = ptbefig->BEREG;

  while (ptbereg != NULL)
    {
    ptlkdins->WRIREG = gss_addwrireg (ptlkdins, ptbereg, usr_package);
    name = concatname (ptlkdins->INSNAME,ptbereg->NAME);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD,ptlkdins->WRIREG);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_SIGFIELD,GSS_REGDEF);

    ptberin  = ptbefig->BERIN;
    ptredlst = ptlkdins->REDLST;
    while ((char *)ptberin->NAME != ptbereg->NAME)
      {
      ptberin = ptberin->NEXT;
      ptredlst++;
      }

    ptredlst->HIST   = ptlkdins->WRIREG->HISTORY;
    ptredlst->VALU   = ptlkdins->WRIREG->CURVAL;
    ptbereg          = ptbereg->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    The WRIREG list must be reversed to obtain the same order	*/
	/* of as in BEREG list.						*/
	/* ###------------------------------------------------------### */

  ptlkdins->WRIREG = (struct wrireg *)
                     reverse ((struct chain *)ptlkdins->WRIREG);

	/* ###------------------------------------------------------### */
	/*    For each BEDLY create an WRIAUX.				*/
	/* ###------------------------------------------------------### */

  ptbeaux = ptbefig->BEDLY;
  while (ptbeaux != NULL)
    {
    ptlkdins->WRIAUX = gss_addwriaux (ptlkdins, ptbeaux, usr_package);
    name         = concatname    (ptlkdins->INSNAME,ptbeaux->NAME);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD,ptlkdins->WRIAUX);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_SIGFIELD,GSS_AUXDEF);

    ptberin  = ptbefig->BERIN;
    ptredlst = ptlkdins->REDLST;
    while (ptberin->NAME != ptbeaux->NAME)
      {
      ptberin = ptberin->NEXT;
      ptredlst++;
      }
    ptredlst->HIST   = ptlkdins->WRIAUX->HISTORY;
    ptredlst->VALU   = ptlkdins->WRIAUX->CURVAL;
    ptbeaux          = ptbeaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    For each BEAUX create an WRIAUX.				*/
	/* ###------------------------------------------------------### */

  ptbeaux = ptbefig->BEAUX;
  while (ptbeaux != NULL)
    {
    ptlkdins->WRIAUX = gss_addwriaux (ptlkdins, ptbeaux, usr_package);
    name             = concatname    (ptlkdins->INSNAME,ptbeaux->NAME);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD,ptlkdins->WRIAUX);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_SIGFIELD,GSS_AUXDEF);

    ptberin  = ptbefig->BERIN;
    ptredlst = ptlkdins->REDLST;
    while (ptberin->NAME != ptbeaux->NAME)
      {
      ptberin = ptberin->NEXT;
      ptredlst++;
      }

    ptredlst->HIST   = ptlkdins->WRIAUX->HISTORY;
    ptredlst->VALU   = ptlkdins->WRIAUX->CURVAL;
    ptbeaux          = ptbeaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    The WRIAUX list must be reversed to obtain the same order	*/
	/* of as in BEAUX (top) and BEDLY (bottom) list.		*/
	/* ###------------------------------------------------------### */

  ptlkdins->WRIAUX = (struct wriaux *)
                     reverse ((struct chain *)ptlkdins->WRIAUX);

	/* ###------------------------------------------------------### */
	/*    For each BEMSG create an WRIMSG.				*/
	/* ###------------------------------------------------------### */

  ptbemsg = ptbefig->BEMSG;
  while (ptbemsg != NULL)
    {
    ptlkdins->WRIMSG = gss_addwrimsg (ptlkdins, ptbemsg, usr_package);
    name             = concatname    (ptlkdins->INSNAME,ptbemsg->LABEL);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD,ptlkdins->WRIMSG);

    gss_addtab (GSS_DICTIONNARY,name,NULL,GSS_SIGFIELD,GSS_MSGDEF);

    ptbemsg          = ptbemsg->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    The WRIMSG list must be reversed to obtain the same order	*/
	/* of as in BEMSG list.						*/
	/* ###------------------------------------------------------### */

  ptlkdins->WRIMSG = (struct wrimsg *)
                     reverse ((struct chain *)ptlkdins->WRIMSG);

	/* ###------------------------------------------------------### */
	/*    For each formal port of the instance find the actual	*/
	/* signal connected to the port and make the link depending on	*/
	/* the type of the port.					*/
	/* ###------------------------------------------------------### */

  ptlocon = ptloins->LOCON;
  while (ptlocon != NULL)
    {
    switch (ptlocon->DIRECTION)
      {

	/* ###------------------------------------------------------### */
	/* For a simple IN port :					*/
	/*   Find the correct REDLST and link it to the BUSSIG or SIMSIG*/
	/* ###------------------------------------------------------### */

      case ('I') :
        ptberin  = ptbefig->BERIN;
        ptredlst = ptlkdins->REDLST;
        while ((char *)ptberin->NAME != ptlocon->NAME)
          {
          ptberin  = ptberin->NEXT;
          ptredlst++;
          }
        ptlosig = ptlocon->SIG;

        if (gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_SIGFIELD) == GSS_BUSDEF)
          {
          ptbussig = (struct bussig *)
                     gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);
          ptredlst->HIST   = ptbussig->HISTORY;
          ptredlst->VALU   = ptbussig->CURVAL;
          }
        else
          {
          ptsimsig = (struct simsig *)
                     gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);
          ptredlst->HIST   = ptsimsig->HISTORY;
          ptredlst->VALU   = ptsimsig->CURVAL;
          }
        break;

	/* ###------------------------------------------------------### */
	/* For a simple OUT port :					*/
	/*    Find the correct WRSSIG and link it to the SIMSIG		*/
	/* ###------------------------------------------------------### */

      case ('O') :
        ptbeout  = ptbefig->BEOUT;
        ptwrssig = ptlkdins->WRSSIG;
        while (ptbeout->NAME != ptlocon->NAME)
          {
          ptbeout  = ptbeout->NEXT;
          ptwrssig = ptwrssig->NEXT;
          }
        ptlosig          = ptlocon->SIG;
        ptwrssig->SIMSIG = (struct simsig *)
                           gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);
        break;

	/* ###------------------------------------------------------### */
	/* For a bused OUT port :					*/
	/*    Find the correct WRBSIG and link it to the BUSSIG.	*/
	/*    complete the list of projected values of the BUSSIG.	*/
	/* ###------------------------------------------------------### */

      case ('Z') :
        ptbebus  = ptbefig->BEBUS;
        ptwrbsig = ptlkdins->WRBSIG;
        while (ptbebus->NAME != ptlocon->NAME)
          {
          ptbebus  = ptbebus->NEXT;
          ptwrbsig = ptwrbsig->NEXT;
          }
        ptlosig  = ptlocon->SIG;
        ptbussig = (struct bussig *)gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);

        ptwrbsig->BUSSIG = ptbussig;
        ptprjbvl         = ptbussig->PRJBVL;

        while (ptprjbvl->NEXT != NULL)
          ptprjbvl = ptprjbvl->NEXT;
        ptprjbvl->NEXT = ptwrbsig->PRJBVL;
        break;

	/* ###------------------------------------------------------### */
	/* For a simple INOUT port :					*/
	/*    Find the correct REDLST and link it to the SIMSIG.	*/
	/*    Find the correct WRSSIG an link it to the SIMSIG		*/
	/* ###------------------------------------------------------### */

      case ('B') :
        ptberin  = ptbefig->BERIN;
        ptredlst = ptlkdins->REDLST;
        while ((char *)ptberin->NAME != ptlocon->NAME)
          {
          ptberin  = ptberin->NEXT;
          ptredlst++;
          }
        ptlosig  = ptlocon->SIG;
        ptsimsig = (struct simsig *)gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);

        ptredlst->HIST   = ptsimsig->HISTORY;
        ptredlst->VALU   = ptsimsig->CURVAL;

        ptbeout  = ptbefig->BEOUT;
        ptwrssig = ptlkdins->WRSSIG;
        while (ptbeout->NAME != ptlocon->NAME)
          {
          ptbeout  = ptbeout->NEXT;
          ptwrssig = ptwrssig->NEXT;
          }
        ptwrssig->SIMSIG = ptsimsig;
        break;

	/* ###------------------------------------------------------### */
	/* For a bused INOUT port :					*/
	/*    Find the correct REDLST and link it to the BUSSIG.	*/
	/*    Find the correct WRBSIG an link it to the BUSSIG.		*/
	/*    complete the list of projected values of the BUSSIG.	*/
	/* ###------------------------------------------------------### */

      case ('T') :
        ptberin  = ptbefig->BERIN;
        ptredlst = ptlkdins->REDLST;
        while (ptberin->NAME != ptlocon->NAME)
          {
          ptberin  = ptberin->NEXT;
          ptredlst++;
          }
        ptlosig  = ptlocon->SIG;
        ptbussig = (struct bussig *)gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);

        ptredlst->HIST   = ptbussig->HISTORY;
        ptredlst->VALU   = ptbussig->CURVAL;

        ptbebus  = ptbefig->BEBUS;
        ptwrbsig = ptlkdins->WRBSIG;
        while (ptbebus->NAME != ptlocon->NAME)
          {
          ptbebus  = ptbebus->NEXT;
          ptwrbsig = ptwrbsig->NEXT;
          }
        ptwrbsig->BUSSIG = ptbussig;

        ptprjbvl         = ptbussig->PRJBVL;
        while (ptprjbvl->NEXT != NULL)
          ptprjbvl = ptprjbvl->NEXT;
        ptprjbvl->NEXT = ptwrbsig->PRJBVL;

        break;

      default :
        gss_toolbug (2,"gss_addlkdins",ptloins->INSNAME,ptlocon->DIRECTION);
      }

    ptlocon = ptlocon->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    For each BEBUX create the list of dependencies.		*/
	/* ###------------------------------------------------------### */

  ptbebux  = ptbefig->BEBUX;
  while (ptbebux != NULL)
    {
    name = concatname (ptlkdins->INSNAME,ptbebux->NAME);
    ptwribux = (struct wribux *) gss_chktab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD);

    ptberin  = ptbefig->BERIN;
    while ((char *)ptberin->NAME != ptbebux->NAME)
      {
      ptberin = ptberin->NEXT;
      }

    ptwribux->DEPSIG =
               gss_adddepend(ptwribux->DEPSIG, ptberin, ptlkdins);

    ptbebux          = ptbebux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    For each BEREG create the list of dependencies.		*/
	/* ###------------------------------------------------------### */

  ptbereg = ptbefig->BEREG;
  while (ptbereg != NULL)
    {
    name = concatname (ptlkdins->INSNAME,ptbereg->NAME);

    ptwrireg = (struct wrireg *) gss_chktab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD);

    ptberin  = ptbefig->BERIN;
    while ((char *)ptberin->NAME != ptbereg->NAME)
      {
      ptberin = ptberin->NEXT;
      }

    ptwrireg->DEPSIG =
               gss_adddepend(ptwrireg->DEPSIG, ptberin, ptlkdins);

    ptbereg = ptbereg->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    For each BEDLY create the list of dependencies.		*/
	/* ###------------------------------------------------------### */

  ptbeaux = ptbefig->BEDLY;
  while (ptbeaux != NULL)
    {
    name = concatname (ptlkdins->INSNAME,ptbeaux->NAME);

    ptwriaux = (struct wriaux *) gss_chktab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD);

    ptberin  = ptbefig->BERIN;
    while (ptberin->NAME != ptbeaux->NAME)
      {
      ptberin = ptberin->NEXT;
      }

    ptwriaux->DEPSIG =
               gss_adddepend(ptwriaux->DEPSIG, ptberin, ptlkdins);

    ptbeaux = ptbeaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    For each BEAUX create the list of dependencies.		*/
	/* ###------------------------------------------------------### */

  ptbeaux = ptbefig->BEAUX;
  while (ptbeaux != NULL)
    {
    name = concatname (ptlkdins->INSNAME,ptbeaux->NAME);

    ptwriaux = (struct wriaux *) gss_chktab (GSS_DICTIONNARY,name,NULL,GSS_PNT0FIELD);

    ptberin  = ptbefig->BERIN;
    while (ptberin->NAME != ptbeaux->NAME)
      {
      ptberin = ptberin->NEXT;
      }

    ptwriaux->DEPSIG =
               gss_adddepend(ptwriaux->DEPSIG, ptberin, ptlkdins);

    ptbeaux = ptbeaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    For each BEPOR create the list of dependencies.		*/
	/* ###------------------------------------------------------### */

  ptlocon = ptloins->LOCON;
  while (ptlocon != NULL)
    {
    switch (ptlocon->DIRECTION)
      {

      case ('I') :
        ptberin  = ptbefig->BERIN;
        while (ptberin->NAME != ptlocon->NAME)
          {
          ptberin  = ptberin->NEXT;
          }

        ptlosig = ptlocon->SIG;

        if (gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_SIGFIELD) == GSS_BUSDEF)
          {
          ptbussig = (struct bussig *)
                     gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);

          ptbussig->DEPSIG =
                     gss_adddepend(ptbussig->DEPSIG, ptberin, ptlkdins);
          }
        else
          {
          ptsimsig = (struct simsig *)
                     gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);

          ptsimsig->DEPSIG =
                     gss_adddepend(ptsimsig->DEPSIG, ptberin, ptlkdins);
          }
        break;

      case ('O') :
      case ('Z') :
        break;

      case ('B') :
        ptberin  = ptbefig->BERIN;
        while (ptberin->NAME != ptlocon->NAME)
          {
          ptberin  = ptberin->NEXT;
          }

        ptlosig = ptlocon->SIG;

        ptsimsig = (struct simsig *)
                   gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);

        ptsimsig->DEPSIG =
                   gss_adddepend(ptsimsig->DEPSIG, ptberin, ptlkdins);
        break;

      case ('T') :
        ptberin  = ptbefig->BERIN;
        while (ptberin->NAME != ptlocon->NAME)
          {
          ptberin  = ptberin->NEXT;
          }

        ptlosig = ptlocon->SIG;

        ptbussig = (struct bussig *)
                   gss_chktab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD);

        ptbussig->DEPSIG =
                   gss_adddepend(ptbussig->DEPSIG, ptberin, ptlkdins);

        break;

      default :
        gss_toolbug (2,"gss_addlkdins",ptloins->INSNAME,ptlocon->DIRECTION);
      }

    ptlocon = ptlocon->NEXT;
    }

  return (ptlkdins);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frelkdins						*/
/* description	: free a list of LKDFIG structure			*/
/* called funct	: mbkfree						*/
/* ###--------------------------------------------------------------### */

static void gss_frelkdins (listlkdins, usr_package)

struct lkdins *listlkdins;
struct funpack *usr_package;

  {
  int            count;
  struct lkdins *ptlkdins;
  struct berin  *ptberin;

  while (listlkdins != NULL)
    {
    gss_frewrssig (listlkdins->WRSSIG);
    gss_frewriaux (listlkdins->WRIAUX, usr_package);
    gss_frewrimsg (listlkdins->WRIMSG, usr_package);
    gss_frewribux (listlkdins->WRIBUX, usr_package);
    gss_frewrbsig (listlkdins->WRBSIG, usr_package);
    gss_frewrireg (listlkdins->WRIREG, usr_package);


    count   = 0;
    ptberin = listlkdins->BEFIG->BERIN;
    while (ptberin != NULL)
      {
      count++;
      ptberin = ptberin->NEXT;
      }
/*------- Il y a un probleme avec freredlst... apres la these ---------*/
/*-------
    if (count != 0)
      gss_freredlst (listlkdins->REDLST,count);
--------*/
    ptlkdins   = listlkdins;
    listlkdins = listlkdins->NEXT;
    mbkfree (ptlkdins);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_makgst						*/
/* description	: This function creates a ghost structure level to	*/
/*		  handle fully behavioural descriptions			*/
/* called funct	: addlofig, addlosig, addchain, mbkalloc		*/
/* ###--------------------------------------------------------------### */

struct lofig *gss_makgst (ptbefig)

struct befig *ptbefig;

  {
  struct chain *namelst;
  struct locon *ptlocon;
  struct bepor *ptbepor;
  struct lofig *ptlofig;
  struct loins *ptloins;
  struct losig *ptlosig;
  int           index=0;
  char          lofigname[80];

  sprintf (lofigname,"%s",ptbefig->NAME);
  ptlofig = addlofig (lofigname);

  ptloins = (struct loins *) mbkalloc (sizeof(struct loins));
  ptloins->INSNAME = ptbefig->NAME;
  ptloins->FIGNAME = ptbefig->NAME;
  ptloins->LOCON   = NULL;
  ptloins->USER    = NULL;
  ptloins->NEXT    = NULL;

  ptlofig->LOINS = ptloins;

  ptbepor = ptbefig->BEPOR;
  while (ptbepor != NULL)
    {
    namelst      = addchain (NULL   ,ptbepor->NAME);
    /* LUDO
    ptlosig      = addlosig (ptlofig,index,namelst,'E', 0.0);
    */
    ptlosig  = addlosig (ptlofig,index,namelst,'E');

    if (ptbepor->TYPE != 'B')
      ptlosig->USER = addptype(ptlosig->USER,ptbepor->TYPE,NULL);

    ptlocon      = (struct locon *) mbkalloc (2 * sizeof(struct locon));

    ptlocon->NAME      = ptbepor->NAME;
    ptlocon->SIG       = ptlosig;
    ptlocon->ROOT      = (void *)ptlofig;
    ptlocon->DIRECTION = ptbepor->DIRECTION;
    ptlocon->TYPE      = 'E';
    ptlocon->USER      = NULL;
    ptlocon->NEXT      = ptlofig->LOCON;
    ptlofig->LOCON     = ptlocon++;

    ptlocon->NAME      = ptbepor->NAME;
    ptlocon->SIG       = ptlosig;
    ptlocon->ROOT      = (void *)ptloins;
    ptlocon->DIRECTION = ptbepor->DIRECTION;
    ptlocon->TYPE      = 'I';
    ptlocon->USER      = NULL;
    ptlocon->NEXT      = ptloins->LOCON;
    ptloins->LOCON     = ptlocon;

    ptbepor = ptbepor->NEXT;
    index++;
    }
  ptlofig->LOCON = (struct locon *) reverse (ptlofig->LOCON);
  ptloins->LOCON = (struct locon *) reverse (ptloins->LOCON);

  return (ptlofig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_linker						*/
/* description	: This function generates the structure used by the	*/
/*		  simulator						*/
/* called funct	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct lkdfig *gss_linker (ptlofig, usr_package)

struct lofig      *ptlofig;
struct funpack    *usr_package;

  {
  struct losig  *ptlosig;
  struct loins  *ptloins;
  struct locon  *ptlocon;
  struct lkdfig *ptlkdfig;
  struct lkdins *ptlkdins;
  struct bussig *ptbussig;
  struct simsig *ptsimsig;

  if (ptlofig == NULL)
    gss_toolbug (1, "gss_linker", NULL, 0);

  ptlkdfig = gss_addlkdfig (ptlofig);

	/* ###------------------------------------------------------### */
	/*   Create a BUSSIG or a SIMSIG for each LOCON of root		*/
	/* ###------------------------------------------------------### */

  ptlocon = ptlofig->LOCON;
  while (ptlocon != NULL)
    {
    ptlosig = ptlocon->SIG;
    if (ptlosig->USER == NULL)
      {
      ptlkdfig->SIMSIG = gss_addsimsig (ptlkdfig->SIMSIG, ptlosig, usr_package);

      ptsimsig         = ptlkdfig->SIMSIG;

      gss_addtab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD,ptsimsig);
      gss_addtab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_SIGFIELD,GSS_SIGDEF);

      }
    else
      {
      ptlkdfig->BUSSIG = gss_addbussig (ptlkdfig->BUSSIG, ptlosig, usr_package);

      ptbussig         = ptlkdfig->BUSSIG;

      gss_addtab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_PNT0FIELD,ptbussig);
      gss_addtab (GSS_DICTIONNARY,getsigname(ptlosig),NULL,GSS_SIGFIELD,GSS_BUSDEF);

      }

    ptlocon = ptlocon->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*   Create a BUSSIG or a SIMSIG for each intern LOSIG of root	*/
	/* ###------------------------------------------------------### */

  ptlosig = ptlofig->LOSIG;
  while (ptlosig != NULL)
    {
    if (ptlosig->TYPE == 'I')
      {
      if (ptlosig->USER == NULL)
        {
        ptlkdfig->SIMSIG = gss_addsimsig (ptlkdfig->SIMSIG, ptlosig, usr_package);

        ptsimsig         = ptlkdfig->SIMSIG;

        gss_addtab (GSS_DICTIONNARY,getsigname (ptlosig) ,NULL,GSS_PNT0FIELD,ptsimsig);
        gss_addtab (GSS_DICTIONNARY,getsigname (ptlosig) ,NULL,GSS_SIGFIELD,GSS_SIGDEF);
        }
      else
        {
        ptlkdfig->BUSSIG = gss_addbussig (ptlkdfig->BUSSIG, ptlosig, usr_package);

        ptbussig         = ptlkdfig->BUSSIG;

        gss_addtab (GSS_DICTIONNARY,getsigname (ptlosig),NULL,GSS_PNT0FIELD,ptbussig);
        gss_addtab (GSS_DICTIONNARY,getsigname (ptlosig),NULL,GSS_SIGFIELD,GSS_BUSDEF);
        }
      }

    ptlosig = ptlosig->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*   Create a LKDINS for each LOINS of root			*/
	/* ###------------------------------------------------------### */

  ptloins = ptlofig->LOINS;
  while (ptloins != NULL)
    {
    ptlkdins  = gss_addlkdins (ptlkdfig, ptloins, usr_package);
    ptloins = ptloins->NEXT;
    }

  return (ptlkdfig);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_restor						*/
/* description	: restore the mode of external ports of the description	*/
/*		  when the declared mode is LINKAGE.			*/
/*									*/
/*		  First, it restores (if necessary) the informations on	*/
/*		  internal ports checking the consistency between the	*/
/*		  model and the instance. Then, it propagates restored	*/
/*		  informations through the signals til the external	*/
/*		  ports. In the final step it checks the consistency of	*/
/*		  connexions on the restored external ports.		*/ 
/* called funct	: gss_initab , gss_chktab, gss_addtab, gss_fretab,	*/
/*		  gss_toolbug, gss_error , mbkalloc  , addptype		*/
/* ###--------------------------------------------------------------### */

int gss_restor (ptlofig, head_befig)

struct lofig      *ptlofig;
struct befig      *head_befig;

  {
  struct loins      *ptloins;
  struct locon      *ptlocon;
  struct losig      *ptlosig;
  struct befig      *ptbefig;
  struct bepor      *ptbepor;
  int                update_flg;
  int                mode;
  int                new_mode;
  int                receiv_nbr;
  int                emettr_nbr;
  int                connec_nbr;
  int                errflg = 0;
  struct dct_entry **lcl_dic;

  if ((ptlofig != NULL) && (head_befig != NULL))
    {
    lcl_dic = gss_initab ();

	/* ###------------------------------------------------------### */
	/*    Fill the dictionnary with external LOCONs of mode linkage	*/
	/* ###------------------------------------------------------### */

    ptlocon = ptlofig->LOCON;
    while (ptlocon != NULL)
      {
      gss_addtab (lcl_dic,ptlocon->SIG,NULL,GSS_MODFIELD,GSS_UNKDFN);
      ptlocon = ptlocon->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    Fill the dictionnary with internal signals		*/
	/* ###------------------------------------------------------### */

    ptlosig = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      if (ptlosig->TYPE == 'I')
        gss_addtab (lcl_dic,ptlosig,NULL,GSS_MODFIELD,GSS_UNKDFN);
      ptlosig = ptlosig->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    For each LOINS ... :					*/
	/* ###------------------------------------------------------### */

    ptloins = ptlofig->LOINS;
    while (ptloins != NULL)
      {
	/* ###------------------------------------------------------### */
	/*    Scan the list of BEFIGs to find the model of the instance	*/
	/* ###------------------------------------------------------### */

      if ((ptbefig=(struct befig *)gss_chktab(GSS_DICTIONNARY,ptloins->FIGNAME,
                                              NULL,GSS_PNT0FIELD)) == NULL)
        {
        ptbefig = head_befig;
        while ((ptbefig != NULL) && (ptbefig->NAME != ptloins->FIGNAME))
          ptbefig = ptbefig->NEXT;

        if (ptbefig == NULL)
          gss_toolbug (1,"gss_restor",ptloins->INSNAME,0);
        else
          gss_addtab(GSS_DICTIONNARY,ptloins->FIGNAME,NULL,GSS_PNT0FIELD,ptbefig);
        }

	/* ###------------------------------------------------------### */
	/*    Check consistency of the instance and its model :		*/
	/* for each port names must be the same				*/
	/* ###------------------------------------------------------### */

      ptbepor = ptbefig->BEPOR;
      ptlocon = ptloins->LOCON;
      while ((ptbepor != NULL) && (ptlocon != NULL))
        {
        update_flg = 0;
        if (ptbepor->NAME != ptlocon->NAME)
          errflg = gss_error (102,ptbefig->NAME);
        else
          {

          if (ptlocon->DIRECTION != 'X')
            {

	/* ###------------------------------------------------------### */
	/*    if the LOCON's mode is not 'linkage', modes must be the	*/
	/* same. In such a case if the LOSIG has to be restored, read	*/
	/* the old RDW flags of the signal then update these flags.	*/
	/* ###------------------------------------------------------### */

            if (ptbepor->DIRECTION != ptlocon->DIRECTION)
              errflg = gss_error (102,ptbefig->NAME);
            else
              {
              if ((mode=gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_MODFIELD)) != 0)
                {
                update_flg = 1;
                receiv_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_RNBFIELD);
                emettr_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_WNBFIELD);
                connec_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_CNBFIELD);
                }
              }
            }
          else
            {

	/* ###------------------------------------------------------### */
	/*    if the LOCON's mode is 'linkage', restore the mode, read	*/
	/* the old RDW flags of the LOSIG then update these flags.	*/
	/* ###------------------------------------------------------### */

            ptlocon->DIRECTION = ptbepor->DIRECTION;
            update_flg = 1;
            mode       = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_MODFIELD);
            receiv_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_RNBFIELD);
            emettr_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_WNBFIELD);
            connec_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_CNBFIELD);
            }

	/* ###------------------------------------------------------### */
	/*    Update RDW flags						*/
	/* ###------------------------------------------------------### */

          if (update_flg == 1)
            {
            connec_nbr++;
            switch (ptbepor->DIRECTION)
              {
              case 'I' :
                new_mode = mode | GSS_REDDFN;
                receiv_nbr++;
                break;
              case 'O' :
                new_mode = mode | GSS_WRTDFN;
                emettr_nbr++;
                break;
              case 'Z' :
                new_mode = mode | GSS_DECDFN;
                if (ptlocon->SIG->USER == NULL)
                  ptlocon->SIG->USER = addptype (NULL,ptbepor->TYPE,NULL);
                else
                  {
                  if (ptlocon->SIG->USER->TYPE !=  ptbepor->TYPE)
                    errflg = gss_error (111,getsigname (ptlocon->SIG));
                  }
                emettr_nbr++;
                break;
              case 'B' :
                new_mode = mode | GSS_WRTDFN | GSS_REDDFN;
                receiv_nbr++;
                emettr_nbr++;
                break;
              case 'T' :
                new_mode = mode | GSS_DECDFN | GSS_REDDFN;
                if (ptlocon->SIG->USER == NULL)
                  ptlocon->SIG->USER = addptype (NULL,ptbepor->TYPE,NULL);
                else
                  {
                  if (ptlocon->SIG->USER->TYPE !=  ptbepor->TYPE)
                    errflg = gss_error (111, getsigname (ptlocon->SIG));
                  }
                receiv_nbr++;
                emettr_nbr++;
                break;
              }
            gss_addtab (lcl_dic,ptlocon->SIG,NULL,GSS_MODFIELD,new_mode);
            gss_addtab (lcl_dic,ptlocon->SIG,NULL,GSS_RNBFIELD,receiv_nbr);
            gss_addtab (lcl_dic,ptlocon->SIG,NULL,GSS_WNBFIELD,emettr_nbr);
            gss_addtab (lcl_dic,ptlocon->SIG,NULL,GSS_CNBFIELD,connec_nbr);
            }
          }

        ptbepor = ptbepor->NEXT;
        ptlocon = ptlocon->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*    It is an error if there is not the same number of BEPOR	*/
	/* and LOCON							*/
	/* ###------------------------------------------------------### */

      if ((ptbepor != NULL) || (ptlocon != NULL))
        errflg = gss_error (102,ptbefig->NAME);

      ptloins = ptloins->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    Scan the list of external LOCONs looking for linkage. If	*/
	/* found restore its mode					*/
	/* ###------------------------------------------------------### */

    ptlocon = ptlofig->LOCON;
    while (ptlocon != NULL)
      {
      mode = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_MODFIELD);
      if (mode != 0)
        {
        connec_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_CNBFIELD) + 1;
        receiv_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_RNBFIELD);
        emettr_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,GSS_WNBFIELD);
        mode      &= GSS_REDDFN | GSS_WRTDFN | GSS_DECDFN;

        switch (mode)
          {
          case (GSS_REDDFN):
            if ((ptlocon->DIRECTION != 'X') && (ptlocon->DIRECTION != 'I'))
              errflg = gss_error (111,ptlocon->NAME);
            ptlocon->DIRECTION = 'I';
            emettr_nbr++;
            break;
          case (GSS_WRTDFN):
            if ((ptlocon->DIRECTION != 'X') && (ptlocon->DIRECTION != 'O'))
              errflg = gss_error (111,ptlocon->NAME);
            ptlocon->DIRECTION = 'O';
            receiv_nbr++;
            break;
          case (GSS_DECDFN):
            if ((ptlocon->DIRECTION != 'X') && (ptlocon->DIRECTION != 'Z'))
              errflg = gss_error (111,ptlocon->NAME);
            ptlocon->DIRECTION = 'Z';
            emettr_nbr++;
            receiv_nbr++;
            break;
          case (GSS_REDDFN | GSS_WRTDFN):
            if ((ptlocon->DIRECTION != 'X') && (ptlocon->DIRECTION != 'B'))
              errflg = gss_error (111,ptlocon->NAME);
            ptlocon->DIRECTION = 'B';
            receiv_nbr++;
            break;
          case (GSS_REDDFN | GSS_DECDFN):
            if ((ptlocon->DIRECTION != 'X') && (ptlocon->DIRECTION != 'T'))
              errflg = gss_error (111,ptlocon->NAME);
            ptlocon->DIRECTION = 'T';
            emettr_nbr++;
            receiv_nbr++;
            break;
          }
        gss_addtab (lcl_dic,ptlocon->SIG,NULL,GSS_RNBFIELD,receiv_nbr);
        gss_addtab (lcl_dic,ptlocon->SIG,NULL,GSS_WNBFIELD,emettr_nbr);
        gss_addtab (lcl_dic,ptlocon->SIG,NULL,GSS_CNBFIELD,connec_nbr);
        }
      ptlocon = ptlocon->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    For each LOSIG added into the dictionnary check the	*/
	/* coherence of connections					*/
	/* ###------------------------------------------------------### */

    ptlosig = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      if ((mode=gss_chktab(lcl_dic,ptlosig,NULL,GSS_MODFIELD)) != 0)
        {
        receiv_nbr = gss_chktab (lcl_dic,ptlosig,NULL,GSS_RNBFIELD);
        emettr_nbr = gss_chktab (lcl_dic,ptlosig,NULL,GSS_WNBFIELD);
        connec_nbr = gss_chktab (lcl_dic,ptlosig,NULL,GSS_CNBFIELD);

        if (receiv_nbr == 0)
          gss_warning (1, getsigname (ptlosig));

        if (emettr_nbr == 0)
          errflg = gss_error (109, getsigname (ptlosig));
        if ((emettr_nbr > 1) && ((mode & GSS_WRTDFN) == GSS_WRTDFN))
          errflg = gss_error (114, getsigname (ptlosig));
        if (connec_nbr == 0)
          errflg = gss_error (110, getsigname (ptlosig));

        if (((mode & GSS_WRTDFN) == GSS_WRTDFN) &&
            ((mode & GSS_DECDFN) == GSS_DECDFN))
          errflg = gss_error (111, getsigname (ptlosig));
        }
      ptlosig = ptlosig->NEXT;
      }

    gss_fretab (lcl_dic);
    }

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_resord						*/
/* description	: put the LOCONs of each instance  in the same order as	*/
/*		  the BEPOR.						*/
/*		  WARNING : This function is used to repare a bug in	*/
/*		  ap2alx tools. So it is to be removed when a correct	*/
/*		  version of ap2alx is installed			*/
/* ###--------------------------------------------------------------### */

void gss_resord (ptlofig, head_befig)

struct lofig      *ptlofig;
struct befig      *head_befig;

  {
  struct loins      *ptloins;
  struct locon      *ptlocon;
  struct befig      *ptbefig;
  struct bepor      *ptbepor;
  struct dct_entry **lcl_dic;

  if (ptlofig != NULL)
    {
    lcl_dic = gss_initab ();

	/* ###------------------------------------------------------### */
	/*    Find the root model of the description by scanning the	*/
	/* list of LOFIGs til the end. (The last LOFIG must be the root)*/
	/* ###------------------------------------------------------### */

/*   while (ptlofig->NEXT != NULL) */
/*     ptlofig = ptlofig->NEXT; */

	/* ###------------------------------------------------------### */
	/*    For each LOINS ... :					*/
	/* ###------------------------------------------------------### */

    ptloins = ptlofig->LOINS;
    while (ptloins != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    Fill the local dictionnary with the internal LOCONs' SIG	*/
	/* and DIRECTION fields. Then, delete the LOCON's list		*/
	/* ###------------------------------------------------------### */

      ptlocon = ptloins->LOCON;
      while (ptlocon != NULL)
        {
        gss_addtab (lcl_dic,ptlocon->NAME,ptloins->FIGNAME,GSS_DIRFIELD,
                    ptlocon->DIRECTION);
        gss_addtab (lcl_dic,ptlocon->NAME,ptloins->FIGNAME,GSS_PNT0FIELD,
                    ptlocon->SIG);
        ptlocon = ptlocon->NEXT;
        }
      ptloins->LOCON = NULL;

	/* ###------------------------------------------------------### */
	/*    Scan the list of BEFIGs to find the model of the instance	*/
	/* ###------------------------------------------------------### */

      if ((ptbefig=(struct befig *)gss_chktab(GSS_DICTIONNARY,ptloins->FIGNAME,
                                              NULL,GSS_PNT0FIELD)) == NULL)
        {
        ptbefig = head_befig;
        while ((ptbefig != NULL) && (ptbefig->NAME != ptloins->FIGNAME))
          ptbefig = ptbefig->NEXT;

        if (ptbefig == NULL)
          gss_toolbug (1,"gss_restor",ptloins->INSNAME,0);
        else
          gss_addtab(GSS_DICTIONNARY,ptloins->FIGNAME,NULL,GSS_PNT0FIELD,ptbefig);
        }

	/* ###------------------------------------------------------### */
	/*    Make a copy of list of BEPORs (create a list of LOCONs)	*/
	/* ###------------------------------------------------------### */

      ptbepor = ptbefig->BEPOR;
      while (ptbepor != NULL)
        {
        ptlocon = (struct locon *) mbkalloc (sizeof(struct locon));

        ptlocon->NAME      = ptbepor->NAME;
        ptlocon->ROOT      = ptloins;
        ptlocon->TYPE      = 'I';
        ptlocon->USER      = NULL;
        ptlocon->NEXT      = ptloins->LOCON;
        ptlocon->DIRECTION = gss_chktab (lcl_dic,ptlocon->NAME,ptloins->FIGNAME,
                                         GSS_DIRFIELD);
        ptlocon->SIG       = (struct losig *)
                             gss_chktab (lcl_dic,ptlocon->NAME,ptloins->FIGNAME,
                                         GSS_PNT0FIELD);

        ptloins->LOCON = ptlocon;
        ptbepor = ptbepor->NEXT;
        }
      ptloins->LOCON = (struct locon *)
                       reverse ((struct chain *) ptloins->LOCON);

      ptloins = ptloins->NEXT;
      }

    gss_fretab (lcl_dic);
    }
  }
