
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: gss_package.c						*/
/* date		: Jun  1 1994						*/
/* version	: v1.00							*/
/* author	: DUNOYER Julien					*/
/* description	: This file contains the set of generic functions	*/
/*                doing the interface with custom client functions. 	*/
/*		  gss_addvalue , gss_freevalue, gss_cpyvalue,		*/
/*		  gss_cmpvalue , gss_evalvalue, gss_resolvalue,		*/
/*		  gss_evaldelay, gss_addfunpack, gss_frefunpack,	*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <sys/time.h>

#include MUT_H
#include "beh.h"
#include GSS_H
#include SCG_H

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */



/* ###--------------------------------------------------------------### */
/* function	: gss_addfunpack					*/
/* description	: Used to add a function package 			*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

struct funpack *gss_addfunpack ( fun_addvalue,
                                 fun_freevalue,
                                 fun_cmpvalue,
                                 fun_cpyvalue,
                                 fun_updvalue,
                                 fun_evalvalue,
                                 fun_resolvalue,
                                 fun_evaldelay,
                                 fun_evalinert,
                                 fun_colvalue)

  void *         (*fun_addvalue)();
  void           (*fun_freevalue)();
  int            (*fun_cmpvalue)();
  int            (*fun_cpyvalue)();
  int            (*fun_updvalue)();
  int            (*fun_evalvalue)();
  int            (*fun_resolvalue)();
  unsigned long  (*fun_evaldelay)();
  unsigned long  (*fun_evalinert)();
  int            (*fun_colvalue)();

  {
  struct funpack  *fun_package;

  fun_package = (struct funpack*) mbkalloc(sizeof(struct funpack));

  fun_package->ADDVALUE   = fun_addvalue;
  fun_package->FREEVALUE  = fun_freevalue;
  fun_package->CMPVALUE   = fun_cmpvalue;
  fun_package->CPYVALUE   = fun_cpyvalue;
  fun_package->UPDVALUE   = fun_updvalue;
  fun_package->EVALVALUE  = fun_evalvalue;
  fun_package->RESOLVALUE = fun_resolvalue;
  fun_package->EVALDELAY  = fun_evaldelay;
  fun_package->EVALINERT  = fun_evalinert;
  fun_package->COLVALUE   = fun_colvalue;

  return (fun_package);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_frefunpack					*/
/* description	: Used to free a function package			*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

void gss_frefunpack (fun_package)

struct funpack *fun_package;

  {
  mbkfree(fun_package);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addvalue						*/
/* description	: Used to add a value 					*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

void *gss_addvalue (usr_package)

struct funpack *usr_package;

  {
  void *value = NULL;

  if ((usr_package != NULL) && ((usr_package->ADDVALUE) != NULL))
    {
    value = (*(usr_package->ADDVALUE)) ();
    }
  else
    {
    gss_error (0,"gss_addvalue");
    }
  return (value);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_freevalue						*/
/* description	: Used to free a value 					*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

int gss_freevalue (value, usr_package)

void	       *value;
struct funpack *usr_package;

  {
  int errflg = 0;

  if ((usr_package != NULL) && ((usr_package->FREEVALUE) != NULL))
    {
    (*(usr_package->FREEVALUE)) (value);
    }
  else
    {
    errflg = gss_error (0,"gss_freevalue");
    }

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_cmpvalue						*/
/* description	: Used to compare two values				*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

int gss_cmpvalue (value1, value2, usr_package)

void	       *value1;
void	       *value2;
struct funpack *usr_package;

  {
  int result = 1;

  if ((usr_package != NULL) && ((usr_package->CMPVALUE) != NULL))
    {
    result = (*(usr_package->CMPVALUE)) (value1, value2);
    }
  else
    {
    gss_error (0, "gss_cmpvalue");
    }

  return (result);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_cpyvalue						*/
/* description	: Used to copy a value					*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

int gss_cpyvalue (value1, value2, usr_package)

void	       *value1;
void	       *value2;
struct funpack *usr_package;

  {
  int errflg = 0;

  if ((usr_package != NULL) && ((usr_package->CPYVALUE) != NULL))
    {
    errflg = (*(usr_package->CPYVALUE)) (value1, value2);
    }
  else
    {
    errflg = gss_error (0, "gss_cpyvalue");
    }

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_updvalue						*/
/* description	: Used to update a value				*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

int gss_updvalue (value1, value2, usr_package)

void	       *value1;
void	       *value2;
struct funpack *usr_package;

  {
  int errflg = 0;

  if ((usr_package != NULL) && ((usr_package->CPYVALUE) != NULL))
    {
    errflg = (*(usr_package->UPDVALUE)) (value1, value2);
    }
  else
    {
    errflg = gss_error (0, "gss_updvalue");
    }

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_resolvalue					*/
/* description	: Used to resolve an bivalue list into a single value	*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

int gss_resolvalue (projected_bivalue,
                    resolution_function_type,
                    current_value,
                    result_value,
                    usr_package)

struct prjbvl  *projected_bivalue;
char	        resolution_function_type;
void	       *current_value;
void	       *result_value;
struct funpack *usr_package;

  {
  int errflg = 0;

  if ((usr_package != NULL) && ((usr_package->RESOLVALUE) != NULL))
    {
    errflg = (*(usr_package->RESOLVALUE)) (projected_bivalue,
                                           resolution_function_type,
                                           current_value,
                                           result_value);
    }
  else
    {
    errflg = gss_error (0, "gss_resolvalue");
    }

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_evalvalue						*/
/* description	: Used to evaluate an expression into a value		*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

int gss_evalvalue (instance,
                   expression,
                   target_type,
                   result_value,
                   usr_package)

struct lkdins  *instance;
void           *expression;
unsigned char   target_type;
void	       *result_value;
struct funpack *usr_package;

  {
  int errflg = 0;

  if ((usr_package != NULL) && ((usr_package->EVALVALUE) != NULL))
    {
    errflg = (*(usr_package->EVALVALUE)) (instance,
                                          expression,
                                          target_type,
                                          result_value);

    }
  else
    {
    errflg = gss_error (0, "gss_evalvalue");
    }

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_evaldelay						*/
/* description	: Used to evaluate a delay				*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

unsigned long gss_evaldelay (instance, target, target_type, usr_package)

struct lkdins  *instance;
void           *target;
unsigned char   target_type;
struct funpack *usr_package;

  {
  unsigned long   result = 0;

  switch (GSS_DELAY_MODE)
    {
    case GSS_ZERO_DELAY :
      result = 0;
      break;

    case  GSS_INERTIAL_DELAY :
    case GSS_TRANSPORT_DELAY :
      switch (target_type)
        {
        case GSS_SIGDEF :
          result = ((struct beout *)target)->TIME;
          break;

        case GSS_BUSDEF :
          result = ((struct binode *)target)->TIME;
          break;

        case GSS_BUXDEF :
          result = ((struct binode *)target)->TIME;
          break;

        case GSS_REGDEF :
          result = ((struct binode *)target)->TIME;
          break;

        case GSS_AUXDEF :
          result = ((struct beaux *)target)->TIME;
          break;

        case GSS_MSGDEF :
          result = 0;
          break;

        default :
          gss_error (0, "gss_evaldelay");
          break;
        }
      break;

    case GSS_CUSTOM_DELAY :
      if ((usr_package != NULL) && ((usr_package->EVALDELAY) != NULL))
        {
        result = (*(usr_package->EVALDELAY)) (instance, target, target_type);
        }
      else
        {
        gss_error (0, "gss_evaldelay");
        }
      break;

    default :
      gss_error (0, "gss_evaldelay");
      break;
    }

  return (result);
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_evalinert						*/
/* description	: Used to evaluate the Pulse Rejection Limit for trans.	*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

unsigned long gss_evalinert (instance, target, target_type, usr_package)

struct lkdins  *instance;
void           *target;
unsigned char   target_type;
struct funpack *usr_package;

  {
  unsigned long   result = 0;

  switch (GSS_DELAY_MODE)
    {
    case GSS_ZERO_DELAY :
    case GSS_TRANSPORT_DELAY :
      result = 0;
      break;

        /* ###------------------------------------------------------### */
        /*    gss_evalinert renvoie la Pulse Rejection Limit au sens    */
        /*    de la norme VHDL'93. L'echeancier supprime alors les      */
        /*    glitches dont la duree est inferieure a cette limite.     */
        /* ###------------------------------------------------------### */

    case GSS_INERTIAL_DELAY :
      switch (target_type)
        {
        case GSS_SIGDEF :
          result = ((struct beout *)target)->TIME;
          break;

        case GSS_BUSDEF :
          result = ((struct binode *)target)->TIME;
          break;

        case GSS_BUXDEF :
          result = ((struct binode *)target)->TIME;
          break;

        case GSS_REGDEF :
          result = ((struct binode *)target)->TIME;
          break;

        case GSS_AUXDEF :
          result = ((struct beaux *)target)->TIME;
          break;

        case GSS_MSGDEF :
          result = 0;
          break;

        default :
          gss_error (0, "gss_evalinert");
          break;
        }
      break;

    case GSS_CUSTOM_DELAY :
      if ((usr_package != NULL) && ((usr_package->EVALINERT) != NULL))
        {
        result = (*(usr_package->EVALINERT)) (instance, target, target_type);
        }
      else
        {
        gss_error (0, "gss_evalinert");
        }
      break;

    default :
      gss_error (0, "gss_evalinert");
      break;
    }

  return (result);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_colvalue						*/
/* description	: Used to collapse 2 values in scheduler		*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

int gss_colvalue (value1, value2, usr_package)

void	       *value1;
void	       *value2;
struct funpack *usr_package;

  {
  int errflg = 0;

  if ((usr_package != NULL) && ((usr_package->COLVALUE) != NULL))
    {
    errflg = (*(usr_package->COLVALUE)) (value1, value2);
    }
  else
    {
    errflg = gss_error (0, "gss_colvalue");
    }

  return (errflg);
  }
