
/* ###--------------------------------------------------------------### */
/* file		: gss_scheduler.c					*/
/* date		: Aug  5 1997						*/
/* version	: v107							*/
/* author	: Nizar ABDALLAH, Pirouz BARAGAN SABET, Julien DUNOYER	*/
/* description	: interface with scheduler functions			*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include SCG_H
#include GSS_H


/* ###--------------------------------------------------------------### */
/* function	: gss_inswave						*/
/* description	: insert a new transaction into the scheduler (main	*/
/*		  identifier access)					*/
/* called func.	: scg_rmventry						*/
/* ###--------------------------------------------------------------### */

/* ###--------------------------------------------------------------### */
/* modification	:- Un parametre inert permet de passer le Pulse Reject	*/
/*   22/01/98      Limit et de se preparer ainsi en douceur a VHDL'93.	*/
/*               - Le parametre flags a ete supprime. Il etait inutile	*/
/*                 car on peut lire ces flags dans pt_shtra->FLAGS.	*/
/*               - La comparaison des valeurs est faites a l'aide d'une	*/
/*                 fonction utilisateur pour la genericite due a GSS.	*/
/*               - Ajout d'un mode collector pour autorise l'addition  	*/
/*                 des transactions d'une wave (pour les densites)      */
/* ###--------------------------------------------------------------### */

void gss_collecwave (pt_shtra, pt_afttra, usr_package)

struct shtra   *pt_shtra;
struct shtra   *pt_afttra;
struct funpack *usr_package;

  {
  if ((pt_shtra->VALU != NULL) && (pt_afttra->VALU != NULL))
    gss_colvalue (pt_shtra->VALU, pt_afttra->VALU, usr_package);
  if ((pt_shtra->COND != NULL) && (pt_afttra->COND != NULL))
    gss_colvalue (pt_shtra->COND, pt_afttra->COND, usr_package);
  }


void gss_inswave (pt_shdul, pt_shtra, inert, usr_package)

struct shdul   *pt_shdul;
struct shtra   *pt_shtra;
unsigned int    inert   ;
struct funpack *usr_package;

  {
  struct shtra *pt_beftra;
  struct shtra *pt_afttra;
  struct shtra *pt_deltra;
  struct shtra *pt_lsttra;
  struct shtra *pt_tmptra;
  struct shwav *pt_shwav ;
  unsigned int  index    ;

  index    = scg_hash (pt_shtra->IDENT);
  pt_shwav = pt_shdul->ITABLE [index];
  
  while ((pt_shwav != NULL) && (pt_shwav->IDENT != pt_shtra->IDENT))
    pt_shwav = pt_shwav->NEXT;

  if (pt_shwav == NULL)
    {
    pt_shwav                 = scg_addshwav (pt_shdul->ITABLE [index],
                                             pt_shtra->IDENT          );
    pt_shdul->ITABLE [index] = pt_shwav;
    }

  if ((pt_shtra->FLAGS & SCG_TRA_COLLECTOR) != 0)
    {
    pt_shtra->WAVE = pt_shwav->TRANS;
    pt_shwav->TRANS = pt_shtra;

    pt_afttra = pt_shtra->WAVE;
    while (pt_afttra != NULL)
      {
      gss_collecwave( pt_shtra, pt_afttra, usr_package );
      pt_afttra = scg_rmventry (pt_afttra);
      }
    pt_shtra->WAVE = NULL;
    }

  else
    {
    pt_lsttra = NULL;
    pt_afttra = pt_shwav->TRANS;

    while ((pt_afttra != NULL) && (pt_afttra->DATE < (pt_shtra->DATE - inert)))
      {
      pt_lsttra = pt_afttra;
      pt_afttra = pt_afttra->WAVE;
      }

    pt_deltra = NULL;

    while ((pt_afttra != NULL) && (pt_afttra->DATE < pt_shtra->DATE))
      {
      pt_lsttra = pt_afttra;

      if (gss_cmpvalue (pt_afttra->VALU, pt_shtra->VALU, usr_package) != 0)
        {
        if ((pt_afttra->COND == NULL) && (pt_shtra->COND == NULL))
          {
          pt_deltra = pt_afttra;
          }
        else if ((pt_afttra->COND != NULL) && (pt_shtra->COND != NULL))
          {
          if (gss_cmpvalue (pt_afttra->COND, pt_shtra->COND, usr_package) != 0)
            pt_deltra = pt_afttra;
          }
        }

      pt_afttra = pt_afttra->WAVE;
      }

    if ((pt_shtra->FLAGS & SCG_TRA_DELAFTER) != 0)
      {
      while (pt_afttra != NULL)
        pt_afttra = scg_rmventry (pt_afttra);
      }
    
    if (pt_lsttra != NULL)
      pt_lsttra->WAVE = pt_shtra;
    else
      pt_shwav->TRANS = pt_shtra;

    pt_shtra->WAVE = pt_afttra;

    if ((pt_deltra != NULL) && ((pt_shtra->FLAGS & SCG_TRA_DELBEFORE) != 0))
      {
      pt_tmptra = pt_shwav->TRANS;
      while (pt_tmptra != pt_deltra)
        pt_tmptra = scg_rmventry (pt_tmptra);

      pt_shwav->TRANS = scg_rmventry (pt_tmptra);
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addtransaction					*/
/* description	: insert a new transaction into the GSS scheduler	*/
/* called func.	: scg_insentry, scg_addshtra, gss_inswave		*/
/* ###--------------------------------------------------------------### */

void gss_addtransaction (ident, scdid, date, inert, cond, valu, uflag, flags, usr_package)

void           *ident   ;
void           *scdid   ;
unsigned int    date    ;
unsigned int    inert   ;
void           *cond    ;
void           *valu    ;
unsigned int    uflag   ;
unsigned short  flags   ;
struct funpack *usr_package;

  {
  struct shtra  *pt_shtra;

  date     += GSS_SCHEDULER->CURTIME;
  pt_shtra  = scg_addshtra (ident, scdid, date, cond, valu, uflag, flags);

  scg_insentry (GSS_SCHEDULER, pt_shtra);
  gss_inswave  (GSS_SCHEDULER, pt_shtra, inert, usr_package);
  }

