
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: gss_util.c						*/
/* date		: Feb 13 1995						*/
/* authors	: Julien DUNOYER					*/
/* version	: v1.00							*/
/* description	: This file contains some utility functions :		*/
/*		  gss_addtab , gss_chktab , gss_fretab , gss_error  ,	*/
/*		  gss_addent , gss_addrcd , gss_toolbug, 		*/
/*		  gss_warning, gss_initab , gss_avers			*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <stdlib.h>
#include MUT_H
#include GSS_H
#include SCG_H

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */


#define GSS_ALODFN     600		/* minimal size of allocation	 */
					/* for dct_entry and dct_recrd	 */
#define GSS_HSZDFN   16001		/* number of entry points in the */
					/* dictionnary			 */

static struct dct_entry  *GSS_DCEHED;   /* free dct_entry's head        */
static struct dct_recrd  *GSS_DCRHED;   /* free dct_recrd's head        */

/* ###--------------------------------------------------------------### */
/* function	: gss_avers						*/
/* description	: return the current version				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

char *gss_avers ()
  {
  return ("v1.00");
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addent						*/
/* description	: add a new entry in a dictionary			*/
/* called func.	: mbkaloc						*/
/* ###--------------------------------------------------------------### */

static struct dct_entry *gss_addent (head, key)

struct dct_entry *head;
char             *key;

  {
  struct dct_entry *entry;
  int               i;

  if (GSS_DCEHED == NULL)
    {
    GSS_DCEHED = (struct dct_entry *)
                 mbkalloc (sizeof(struct dct_entry) * GSS_ALODFN);

    entry = GSS_DCEHED;
    for (i=1 ; i<GSS_ALODFN ; i++)
      {
      entry->next = entry + 1;
      entry++;
      }
    entry->next = NULL;
    }

  entry       = GSS_DCEHED;
  GSS_DCEHED  = GSS_DCEHED->next;

  entry->next = head;
  entry->data = NULL;
  entry->key  = key;

  return (entry);
  }

static struct dct_recrd *gss_addrcd (head , key)

struct dct_recrd *head;
char             *key;

  {
  struct dct_recrd *recrd;
  int               i;

  if (GSS_DCRHED == NULL)
    {
    GSS_DCRHED = (struct dct_recrd *)
                 mbkalloc (sizeof(struct dct_recrd) * GSS_ALODFN);

    recrd = GSS_DCRHED;
    for (i=1 ; i<GSS_ALODFN ; i++)
      {
      recrd->next = recrd + 1;
      recrd++;
      }
    recrd->next = NULL;
    }

  recrd           = GSS_DCRHED;
  GSS_DCRHED      = GSS_DCRHED->next;

  recrd->next     = head;
  recrd->fd0_val  = 0;
  recrd->fd1_val  = 0;
  recrd->fd2_val  = 0;
  recrd->fd3_val  = 0;
  recrd->fd4_val  = 0;
  recrd->pnt0_val = 0;
  recrd->pnt1_val = 0;
  recrd->key      = key;

  return (recrd);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_initab						*/
/* description	: create a new dictionary				*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct dct_entry **gss_initab ()

  {
  struct dct_entry **head;
  int                i;

  head = (struct dct_entry **)
         mbkalloc (sizeof(struct dct_entry *) * GSS_HSZDFN);

  for (i=0 ; i<GSS_HSZDFN ; i++)
    head[i] = NULL;

  return (head);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_warning						*/
/* description	: print a warning message corresponding to the argument	*/
/*		  code							*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void gss_warning (code,str1)

int   code;
char *str1;

  {
  fprintf (stderr,"Warning %d : ",code);
  switch (code)
    {
    case 1 :
      fprintf (stderr,"signal `%s` never used\n",str1);
      break;
    default:
      fprintf (stderr,"gss_warning : code %d unknown.\n",code);
      break;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_toolbug						*/
/* description	: print a fatal error message corresponding to the	*/
/*		  argument code and exit				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void gss_toolbug (code,str1,str2,nbr1)

int   code;
char *str1;
char *str2;
int   nbr1;

  {
  (void) fprintf (stdout,"Fatal error %d executing `%s`: ", code,str1);
  switch (code)
    {
    case 1 :
      (void) fprintf (stderr,"cannot find model of `%s`\n",str2);
      break;
    case 2 :
      (void) fprintf (stderr,"unknown mode `%c` for port `%s`\n",nbr1,str2);
      break;
    default:
      (void) fprintf (stderr,"gss_toolbug : code %d unknown.\n",code);
      break;
    }
  EXIT (1);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_addtab						*/
/* description	: add a new information into a dictionary. If the	*/
/*		  key or the context doesn't exist create a new record	*/
/* called func.	: gss_addrcd, gss_addent				*/
/* ###--------------------------------------------------------------### */

void gss_addtab (head,key_str,ctx_str,field,valu)

struct dct_entry **head;
char              *key_str;
char              *ctx_str;
int                field;
int                valu;

  {
  int               found = 0;
  int               index;
  struct dct_entry *entry_pnt;
  struct dct_recrd *recrd_pnt;

  index     = (int) key_str % GSS_HSZDFN;
  entry_pnt = head[index];

  while (entry_pnt != NULL)
    {
    if (entry_pnt->key == key_str)
      {
      found = 1;
      break;
      }
    entry_pnt = entry_pnt->next;
    }

  if (found == 0)
    {
    head[index] = gss_addent (head[index],key_str); 
    entry_pnt = head[index]; 
    }

  found = 0;
  recrd_pnt = entry_pnt->data;
  while (recrd_pnt != NULL)
    {
    if (recrd_pnt->key == ctx_str)
      {
      found = 1;
      break;
      }
    recrd_pnt = recrd_pnt->next;
    }

  if (found == 0)
    {
    entry_pnt->data = gss_addrcd (entry_pnt->data,ctx_str); 
    recrd_pnt       = entry_pnt->data ;
    }

  switch (field)
    {
    case 0 :
      recrd_pnt->fd0_val = valu; break;
    case 1 :
      recrd_pnt->fd1_val = valu; break;
    case 2 :
      recrd_pnt->fd2_val = valu; break;
    case 3 :
      recrd_pnt->fd3_val = valu; break;
    case 4 :
      recrd_pnt->fd4_val = valu; break;
    case 5 :
      recrd_pnt->pnt0_val = valu; break;
    case 6 :
      recrd_pnt->pnt1_val = valu; break;
    }

  }

/* ###--------------------------------------------------------------### */
/* function	: gss_chktab						*/
/* description	: extract an information from a dictionary. Return	*/
/*		  0 if the information has not been found.		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int  gss_chktab (head,key_str,ctx_str,field)

struct dct_entry **head;
char              *key_str;
char              *ctx_str;
int                field;

  {
  int               found = 0;
  int               valu = 0;
  struct dct_entry *entry_pnt;
  struct dct_recrd *recrd_pnt;

  entry_pnt = head [(int)key_str % GSS_HSZDFN];

  while (entry_pnt != NULL)
    {
    if (entry_pnt->key == key_str)
      {
      found = 1;
      break;
      }
    entry_pnt = entry_pnt->next;
    }

  if (found == 1)
    {
    found = 0;
    recrd_pnt = entry_pnt->data;
    while (recrd_pnt != NULL)
      {
      if (recrd_pnt->key == ctx_str)
        {
        found = 1;
        break;
        }
      recrd_pnt = recrd_pnt->next;
      }
    if (found == 1)
      {
      switch (field)
        {
        case 0 :
          valu = recrd_pnt->fd0_val; break;
        case 1 :
          valu = recrd_pnt->fd1_val; break;
        case 2 :
          valu = recrd_pnt->fd2_val; break;
        case 3 :
          valu = recrd_pnt->fd3_val; break;
        case 4 :
          valu = recrd_pnt->fd4_val; break;
        case 5 :
          valu = recrd_pnt->pnt0_val; break;
        case 6 :
          valu = recrd_pnt->pnt1_val; break;
        }
      }
    }

  return (valu);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_fretab						*/
/* description	: remove entirely a dictionary				*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void gss_fretab (pt_hash)

struct dct_entry **pt_hash;
  {
  struct dct_entry *pt_entry;
  struct dct_entry *pt_nxtentry;
  struct dct_recrd *pt_record;
  int               i;

  if (pt_hash != NULL)
    {
    for (i=0 ; i<GSS_HSZDFN ; i++)
      {
      if ((pt_entry = pt_hash[i]) != NULL)
        {
        while (pt_entry != NULL)
          {
          pt_record = pt_entry->data;

          while (pt_record->next != NULL)
            pt_record = pt_record->next;

          pt_record->next = GSS_DCRHED;
          GSS_DCRHED      = pt_entry->data;

          pt_nxtentry     = pt_entry->next;
          pt_entry->next  = GSS_DCEHED;
          GSS_DCEHED      = pt_entry;
          pt_entry        = pt_nxtentry;
          }
        }
      }
    mbkfree (pt_hash);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_error						*/
/* description	: print an error message depending on the argument code	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int gss_error (code,str1)

int   code;
char *str1;

  {
  if (code < 200)
    (void)fprintf (stderr,"Error %d :",code);

  switch (code)
    {


    case 0:
      (void) fprintf (stderr, "unknown error in function : %s\n", str1);
      break;
    case 1:
      (void) fprintf (stderr, "combinatory loop : `%s`\n", str1);
      break;
    case 3:
      (void) fprintf (stderr, "cannot find terminal `%s`\n", str1);
      break;
    case 6:
      (void) fprintf (stderr, "cannot make abl of empty expression\n");
      break;
    case 7:
      (void) fprintf (stderr, "call on empty list\n");
      break;
    case 8:
      (void) fprintf (stderr, "missing extraction lists\n");
      break;
    case 9:
      (void) fprintf (stderr, "missing propagation lists\n");
      break;
    case 10:
      (void) fprintf (stderr, "missing propagation table\n");
      break;
    case 11:
      (void) fprintf (stderr, "missing extraction table\n");
      break;
    case 12:
      (void) fprintf (stderr, "missing propagation tree\n");
      break;
    case 13:
      (void) fprintf (stderr, "missing extraction list\n");
      break;
    case 14:
      (void) fprintf (stderr, "Signal %s doesn't exist: This event is ignored\n",str1);
      break;
    case 15:
      (void) fprintf (stderr, "Assert Violation\n");
      break;
    case 16:
      (void) fprintf (stderr, "Bus Resolution Error\n");
      break;
    case 17:
      (void) fprintf (stderr, "Register Resolution Error\n");
      break;
    case 18:
      (void) fprintf (stderr, "Signal %s bad type: Events ignored\n",str1);
      break;


    case 100:
      (void) fprintf (stderr,"cannot find `%s`\n", str1);
      break;
    case 102:
      (void) fprintf (stderr,"instance and model of `%s` mismatch\n",str1);
      break;
    case 103:
      (void) fprintf (stderr,"incompatible input/output mode for `%s`\n",str1);
      break;
    case 104:
      (void) fprintf (stderr,"input/output `%s` doesn't exist\n",str1);
      break;
    case 105:
      (void) fprintf (stderr,"Possible combinatory loop. Simulation aborted\n");
      break;
    case 106:
      (void) fprintf (stderr,"Too many errors. Simulation interrupted\n");
      break;
    case 107:
      (void) fprintf (stderr,"Cannot open result file\n");
      break;
    case 109:
      (void) fprintf (stderr,"signal `%s` never assigned\n",str1);
      break;
    case 110:
      (void) fprintf (stderr,"signal `%s` not connected to any port\n",str1);
      break;
    case 111:
      (void) fprintf (stderr,"illegal connections on signal `%s`\n",str1);
      break;
    case 112:
      (void) fprintf (stderr,"multiple driver on guarded signal `%s`\n",str1);
      break;
    case 113:
      (void) fprintf (stderr,"expected value and simulation's result mismatch on `%s`\n",str1);
      break;
    case 114:
      (void) fprintf (stderr,"multiple driver on unguarded signal `%s`\n",str1);
      break;
    case 115:
      (void) fprintf (stderr,"save file doesn't match with the description\n");
      break;
    case 116:
      (void) fprintf (stderr,"cannot open file `%s`\n", str1);
      break;

    case 200:
      (void) fprintf (stderr,"\n	cannot continue further more.\n");
      (void) fprintf (stderr,"\n		Have a nice day...\n");
      break;

    default:
      (void) fprintf (stderr,"Unknown Gss Error\n");
      break;
    }

  return (1);
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_init						*/
/* description	: initialize different global variables and so on...	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void gss_init ()
{
char *str;

	str = mbkgetenv("GSS_MAXDELTA");
	if (str != NULL)
		GSS_MAXDELTA = (int)atoi(str);

        GSS_SCHEDULER   = scg_CreateScheduler( NULL, NULL);
        GSS_DICTIONNARY = gss_initab();
}
