
/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: gss_value.c						*/
/* date		: Jan 28 1998						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien					*/
/* content	: high-level gss function				*/
/*		  gss_getvalue, gss_getcurval, gss_getprjval		*/
/*		  gss_setvalue, gss_setcurval, gss_setprjval		*/
/* ###--------------------------------------------------------------### */

#include MUT_H
#include GSS_H
#include "beh.h"

/* ###--------------------------------------------------------------### */
/* function	: gss_getvalue						*/
/* description	: get current value by signal index and instance	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void *gss_getvalue (instance, indexbdd)

struct lkdins *instance;
short          indexbdd;

  {
  void        *value   = NULL;

      /* ###--------------------------------------------------------### */
      /*    if the GSS REDLST Table is expanded, use a right shift      */
      /* ###--------------------------------------------------------### */

  if ( BEH_CIRCUI & BEH_CIRCUI_EXPANDED)
    {
    if (indexbdd > 1)
        value = (instance->REDLST + ((indexbdd - 2) >> 1))->VALU;
    else
        gss_error (0,"gss_getvalue");
    }
  else
    {
    if (indexbdd > 1)
        value = (instance->REDLST + (indexbdd - 2))->VALU;
    else
        gss_error (0,"gss_getvalue");
    }
  return (value); 
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_setvalue						*/
/* description	: set projected value by signal index and instance	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void gss_setvalue (instance, indexbdd, value)

struct lkdins *instance;
short          indexbdd;
void          *value;

  {
  void             *signal  = NULL;


      /* ###--------------------------------------------------------### */
      /*    if the GSS REDLST Table is expanded, use a right shift      */
      /* ###--------------------------------------------------------### */

  if ( BEH_CIRCUI & BEH_CIRCUI_EXPANDED)
    {
    if (indexbdd > 1)
        (instance->REDLST + ((indexbdd - 2) >> 1))->VALU = value;
    else
        gss_error (0,"gss_setvalue");
    }
  else
    {
    if (indexbdd > 1)
        (instance->REDLST + (indexbdd - 2))->VALU = value;
    else
        gss_error (0,"gss_setvalue");
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_getcurval						*/
/* description	: get current value by signal name			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void *gss_getcurval (signame)

char              *signame;

  {
  void             *signal  = NULL;
  void             *value   = NULL;
  unsigned char     sigtype = GSS_UNKDEF;

  sigtype = (unsigned char) gss_chktab (GSS_DICTIONNARY,signame,NULL,GSS_SIGFIELD);
  signal  = (void *)        gss_chktab (GSS_DICTIONNARY,signame,NULL,GSS_PNT0FIELD);

  switch (sigtype)
    {
    case GSS_SIGDEF :
      value = ((struct simsig *)(signal))->CURVAL;
      break; 
    case GSS_BUSDEF :
      value = ((struct bussig *)(signal))->CURVAL;
      break; 
    case GSS_BUXDEF :
      value = ((struct wribux *)(signal))->CURVAL;
      break; 
    case GSS_REGDEF :
      value = ((struct wrireg *)(signal))->CURVAL;
      break; 
    case GSS_AUXDEF :
      value = ((struct wriaux *)(signal))->CURVAL;
      break; 
    case GSS_MSGDEF :
      value = ((struct wrimsg *)(signal))->CURVAL;
      break; 
    default :
      gss_error (0,"gss_getcurval");
      break; 
    }
  return (value); 
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_setcurval						*/
/* description	: set current value by signal name			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void gss_setcurval (signame, value)

char              *signame;
void              *value;

  {
  void             *signal  = NULL;
  unsigned char     sigtype = GSS_UNKDEF;

  sigtype = (unsigned char) gss_chktab (GSS_DICTIONNARY,signame,NULL,GSS_SIGFIELD);

  signal  = (void *)        gss_chktab (GSS_DICTIONNARY,signame,NULL,GSS_PNT0FIELD);

  switch (sigtype)
    {
    case GSS_SIGDEF :
      ((struct simsig *)(signal))->CURVAL = value;
      break;
    case GSS_AUXDEF :
      ((struct wriaux *)(signal))->CURVAL = value;
      break;
    case GSS_MSGDEF :
      ((struct wrimsg *)(signal))->CURVAL = value;
      break;
    case GSS_BUSDEF :
      ((struct bussig *)(signal))->CURVAL = value;
      break;
    case GSS_BUXDEF :
      ((struct wribux *)(signal))->CURVAL = value;
      break;
    case GSS_REGDEF :
      ((struct wrireg *)(signal))->CURVAL = value;
      break;
    default :
      gss_error (0,"gss_setcurval");
      break; 
    }
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_getprjval						*/
/* description	: get projected value by signal name and type		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void *gss_getprjval (signame, sigtype)

char              *signame;
unsigned char      sigtype;

  {
  void             *signal  = NULL;
  void             *value   = NULL;

  if ( (sigtype == GSS_UNKDEF) ||
       (sigtype == GSS_CNDDEF) || (sigtype == GSS_DRVDEF) )
    {
    sigtype |= (unsigned char) gss_chktab
                                (GSS_DICTIONNARY,signame,NULL,GSS_SIGFIELD);
    }

  signal  = (void *) gss_chktab (GSS_DICTIONNARY,signame,NULL,GSS_PNT0FIELD);

  switch (sigtype)
    {
    case GSS_SIGDEF :
      value = ((struct simsig *)(signal))->PRJVAL;
      break; 
    case GSS_AUXDEF :
      value = ((struct wriaux *)(signal))->PRJVAL;
      break; 
    case GSS_MSGDEF :
      value = ((struct wrimsg *)(signal))->PRJVAL;
      break; 
    case GSS_BUSDEF :
      value = ((struct bussig *)(signal))->PRJVAL;
      break; 
    case ( GSS_BUSDEF | GSS_CNDDEF ) :
      value = ((struct bussig *)(signal))->PRJBVL->CNDVAL;
      break;
    case ( GSS_BUSDEF | GSS_DRVDEF ) :
      value = ((struct bussig *)(signal))->PRJBVL->DRVVAL;
      break;
    case GSS_BUXDEF :
      value = ((struct wribux *)(signal))->PRJVAL;
      break; 
    case ( GSS_BUXDEF | GSS_CNDDEF ) :
      value = ((struct wribux *)(signal))->PRJBVL->CNDVAL;
      break;
    case ( GSS_BUXDEF | GSS_DRVDEF ) :
      value = ((struct wribux *)(signal))->PRJBVL->DRVVAL;
      break;
    case GSS_REGDEF :
      value = ((struct wrireg *)(signal))->PRJVAL;
      break; 
    case ( GSS_REGDEF | GSS_CNDDEF ) :
      value = ((struct wrireg *)(signal))->PRJBVL->CNDVAL;
      break;
    case ( GSS_REGDEF | GSS_DRVDEF ) :
      value = ((struct wrireg *)(signal))->PRJBVL->DRVVAL;
      break;
    default :
      gss_error (0,"gss_getprjval");
      break; 
    }
  return (value); 
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_setprjval						*/
/* description	: set projected value by signal name and type		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void gss_setprjval (signame, sigtype, value)

char              *signame;
unsigned char      sigtype;
void              *value;

  {
  void             *signal  = NULL;

  if ( (sigtype == GSS_UNKDEF) ||
       (sigtype == GSS_CNDDEF) || (sigtype == GSS_DRVDEF) )
    {
    sigtype |= (unsigned char) gss_chktab
                                (GSS_DICTIONNARY,signame,NULL,GSS_SIGFIELD);
    }

  signal  = (void *) gss_chktab (GSS_DICTIONNARY,signame,NULL,GSS_PNT0FIELD);

  switch (sigtype)
    {
    case GSS_SIGDEF :
      ((struct simsig *)(signal))->PRJVAL = value;
      break;
    case GSS_AUXDEF :
      ((struct wriaux *)(signal))->PRJVAL = value;
      break;
    case GSS_MSGDEF :
      ((struct wrimsg *)(signal))->PRJVAL = value;
      break;
    case GSS_BUSDEF :
      ((struct bussig *)(signal))->PRJVAL = value;
      break;
    case ( GSS_BUSDEF | GSS_CNDDEF ) :
      ((struct bussig *)(signal))->PRJBVL->CNDVAL = value;
      break;
    case ( GSS_BUSDEF | GSS_DRVDEF ) :
      ((struct bussig *)(signal))->PRJBVL->DRVVAL = value;
      break;
    case GSS_BUXDEF :
      ((struct wribux *)(signal))->PRJVAL = value;
      break;
    case ( GSS_BUXDEF | GSS_CNDDEF ) :
      ((struct wribux *)(signal))->PRJBVL->CNDVAL = value;
      break;
    case ( GSS_BUXDEF | GSS_DRVDEF ) :
      ((struct wribux *)(signal))->PRJBVL->DRVVAL = value;
      break;
    case GSS_REGDEF :
      ((struct wrireg *)(signal))->PRJVAL = value;
      break;
    case ( GSS_REGDEF | GSS_CNDDEF ) :
      ((struct wrireg *)(signal))->PRJBVL->CNDVAL = value;
      break;
    case ( GSS_REGDEF | GSS_DRVDEF ) :
      ((struct wrireg *)(signal))->PRJBVL->DRVVAL = value;
      break;
    default :
      gss_error (0,"gss_setprjval");
      break; 
    }
  }

