/* ###--------------------------------------------------------------### */
/*                                                                      */
/* file         : mkpat.c                                               */
/* date         : May. 1995                                             */
/* version      : v1.00                                                 */
/* author       : Chun CUI                                              */
/* description  : VHDL pattern generation tool.                         */
/*                                                                      */
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>
#include MUT_H
#include PAT_H
#include PHL_H
#include PPT_H
#include MLO_H
#include "beh.h"
#include "bvl.h"
#include LOG_H
#include PRB_H

int      Max_count         = 10000;        /* # of iteration                  */
int      base              = 2147483647;   /* random() function base (2**31)-1*/
float    Densi_eps         = 1.0;          /* min density error constraint    */

/******************************************************************************/
/******************************************************************************/


int mkp_error(code, str1, str2, str3)
int 	code;
char   *str1;
char   *str2;
char   *str3;
  {
  (void)fprintf (stderr,"Mkpat Error %d :",code);

  switch (code)
    {
    case 0 :
      (void) fprintf (stderr,"\n        cannot continue further more.\n");
      (void) fprintf (stderr,"\n                Have a nice day...\n");
      break;
    case 1  :
      (void) fprintf (stderr,"Cannot find the vhdl file %s.vbe\n\n", str1);
      break;
    case 2  :
      (void) fprintf (stderr,"Cannot find the vhdl file %s.vst\n\n", str1);
      break;
    case 3  :
      (void) fprintf (stderr,"Cannot open the result File %s.pat\n\n", str1);
      break;
    case 4  :
      (void) fprintf (stderr,"Cannot open the result File %s.%s\n\n", str1, str2);
      break;
    case 5  :
      (void) fprintf (stderr,"The connector %s has a UNKNOWN direction.\n", str1);
      break;

    default :
      (void) fprintf (stderr,"Unknown Mkpat Error\n\n");
      break;
    }

  exit(1);
  }



/* ###--------------------------------------------------------------### */
/* function	: randint	      				        */
/* description	: This function creates randomly an integer between  	*/
/*		  startp and endp                     	                */
/* ###--------------------------------------------------------------### */
int   randint(startp, endp)
int   startp;
int   endp;
{
  int      t;
  float    s;

  s = (float)random()/base;
  t = startp + s * (endp - startp);
  return(t);

 }

/* ###--------------------------------------------------------------### */
/* function	: inttostr	      				        */
/* description	: This function changes an integer to a character  	*/
/* ###--------------------------------------------------------------### */
char  inttostr(entier)
int   entier;
{
  char *str;

  str = (char *) mbkalloc(32 * sizeof(char));
  sprintf(str, "%d", entier);
  return(*str);
}

/* ###--------------------------------------------------------------### */
/* function	: change	      				        */
/* description	: This function calculates the number of transition  	*/
/*		  in a 0/1 array                     	                */
/* ###--------------------------------------------------------------### */
int     change(num, a)
int     num;
char   *a;
{
  int   i;
  char  last;
  int   num_chg = 0;

  last = *(a+0);
  for(i=1; i<num; i++)
    {
    if(*(a+i) != last) 
      {
      num_chg++;
      last = *(a+i);
      }
    }
  return(num_chg);
}

/* ###--------------------------------------------------------------### */
/* function	: prb		 				        */
/* description	: This function calculates the percentage of 1 in a  	*/
/*		  0/1 array                     	                */
/* ###--------------------------------------------------------------### */
float   prb(num, x)
int     num; 
char   *x;
{
  int   i;
  int   p = 0;
  float proba;

  for(i=0; i<num; i++)
    {
    if(*(x+i) == '1') p++;
    }
  proba = (float)p / num;
  return(proba);
}


/* ###--------------------------------------------------------------### */
/* function	: randpat						*/
/* description	: This function creates an array of pattern with 	*/
/*		  a certain probability of 1		                */
/* called funct	: mbkalloc, rand		                        */
/* ###--------------------------------------------------------------### */
void    randpat(startp, endp, sig_prb, num_pat, pat)
int     startp, endp;
float   sig_prb;
int     num_pat;
char   *pat;
{
  int      i, j, k, t, found, number;
  float    s;
  int     *tmp;

  tmp = (int *)mbkalloc(sizeof(int) * num_pat);

/*--------------------------------------------------------------*/
/* Calculate the number of 1 in the array                       */
/*--------------------------------------------------------------*/

  number = (int)(sig_prb * num_pat + 0.5);

  if (abs(number-num_pat) > 1)
    {
    for(k=0; k<number; k++)
      {
      found = 1;
      while(found)
        {
        found = 0;
        s = (float)random()/base;
        t = startp + s * (endp - startp);
/*--------------------------------------------------------------*/
/* For verifying t isn't equal to what have been chosen         */
/*--------------------------------------------------------------*/
        for (j=0; j<k; j++)
          {
          if(*(tmp + j) == t)
            {
            found = 1;
            break;
            }
          }
        *(tmp + k) = t;
        }
      }

    for(k=0; k<number; k++)
      {
      *(pat + *(tmp+k)) = '1';
      }
    }
/*--------------------------------------------------------------*/
/* For sig_prb = 1, all of the element of array equal to 1      */
/*--------------------------------------------------------------*/
  else                  
    {
    for(k=0; k<number; k++)
      {
      *(pat + k) = '1';
      }
    }
}

/* ###--------------------------------------------------------------### */
/* function	: proba_density						*/
/* description	: This function creates a pattern for a certain signal 	*/
/*		  with a predetermined probability and density		*/
/* called funct	: mbkalloc, randpat		                        */
/* ###--------------------------------------------------------------### */

void    proba_density(num_pat, proba, density, eps, HEAD_PAT)
int     num_pat;
float   proba, density, eps;
char   *HEAD_PAT;
  {
  int             index1, index2;
  float           diff1;
  int             i, tmp;
  int             count = 0;
  int             rev_mode;
  int             first_index1, last_index1;
  int             first_index2;
  char           *pat;
  char           *COPY_PAT;
  char            flag, exchg;

  if(eps < 1.0) eps = 1.0;
/*--------------------------------------------------------------*/
/* Allocate the places for temporarily use                      */
/*--------------------------------------------------------------*/
  COPY_PAT = (char *)mbkalloc(sizeof(char) * num_pat);
  pat = HEAD_PAT;

  randpat(0, num_pat-1, proba, num_pat, HEAD_PAT);  

  tmp = change(num_pat, HEAD_PAT);
  diff1 = (float)tmp - density;
  
  while ((fabs(diff1) > eps) && (count < Max_count))
    {
g1: count++;
    if (count >= Max_count) break; 
/*--------------------------------------------------------------*/
/* Randomly choose two inequal index between 0 and num_pat-1    */
/*--------------------------------------------------------------*/
    do {
       index1 = randint(0, num_pat-1);
       index2 = randint(0, num_pat-1);
       } while(abs(index1-index2) < 3);
    if (index1 > index2) 
      {
      tmp = index1;
      index1 = index2;
      index2 = tmp;
      }

    if (diff1 < 0) flag = 'U';
    else           flag = 'D';

    if (flag == 'D')
/*--------------------------------------------------------------*/
/* For decrease the density                                     */
/*--------------------------------------------------------------*/
      {
      first_index1 = index1;
      while (*(HEAD_PAT + first_index1) == *(HEAD_PAT + first_index1 + 1))
        {
        first_index1++;
        }

      last_index1 = first_index1 + 1;
      while (*(HEAD_PAT + last_index1) == *(HEAD_PAT + last_index1 + 1)) 
        {
        last_index1++;
        }
      
      if (last_index1 + 1 > index2) goto g1;

      rev_mode = 0;
      first_index2 = index2;

      while (*(HEAD_PAT + first_index2 ) == *(HEAD_PAT + first_index1))
        {
        first_index2++;
        if((rev_mode == 1) && (first_index2 >= first_index1)) goto g1;
        if (first_index2 > num_pat-1 ) 
          {
/*--------------------------------------------------------------*/
/* Go to the beginning of the array[0]                          */
/*--------------------------------------------------------------*/
          first_index2 -= num_pat;
          rev_mode = 1;
	  }
        }

      pat = COPY_PAT;

/*--------------------------------------------------------------*/
/* Replace the elements in the array                            */
/*--------------------------------------------------------------*/
      if (rev_mode == 0)
        {
        for (i=0; i<=first_index1; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
      
        for(i=last_index1+1; i<first_index2; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
 
        for(i=first_index1+1; i<=last_index1; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
      
        for(i=first_index2; i<num_pat; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
   
        for(i=0; i<num_pat; i++)
          {
          *(HEAD_PAT + i) = *(COPY_PAT + i);
          }
	diff1 = diff1 - 2;
        }
      else
        {
        for (i=0; i<first_index2; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
 
        for(i=first_index1+1; i<=last_index1; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }

        for(i=first_index2; i<=first_index1; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
       
        for(i=last_index1+1; i<num_pat; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }

        for(i=0; i<num_pat; i++)
          {
          *(HEAD_PAT + i) = *(COPY_PAT + i);
          }
	diff1 = diff1 - 2;
        }
      }
    else                           
/*--------------------------------------------------------------*/
/* For increase the density, flag == 'U'                        */
/*--------------------------------------------------------------*/
      {
      first_index1 = index1;
      while (*(HEAD_PAT + first_index1) != *(HEAD_PAT + first_index1 + 1))
        {
        first_index1++;
        }

      if (first_index1 + 1 > index2) goto g1;

      rev_mode = 0;
      first_index2 = index2;

      while ((*(HEAD_PAT + first_index2 ) == *(HEAD_PAT + first_index1)) 
             || (*(HEAD_PAT + first_index2 ) != *(HEAD_PAT + first_index2 + 1)))
        {
        first_index2++;
        if((rev_mode == 1) && (first_index2 >= first_index1)) goto g1;
        if (first_index2 > num_pat-1 ) 
          {
/*--------------------------------------------------------------*/
/* Go to the beginning of the array[0]                          */
/*--------------------------------------------------------------*/
          first_index2 -= num_pat;
          rev_mode = 1;
          }
        }

      pat = COPY_PAT;
/*--------------------------------------------------------------*/
/* Replace the elements in the array                            */
/*--------------------------------------------------------------*/
      if (rev_mode == 0)
        {
        for (i=0; i<=first_index1; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }

        *pat = *(HEAD_PAT + first_index2);
         pat++;

        for(i=first_index1+1; i<first_index2; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
      
        for(i=first_index2+1; i<num_pat; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }

        for(i=0; i<num_pat; i++)
          {
          *(HEAD_PAT + i) = *(COPY_PAT + i);
          }
	diff1 = diff1 + 2;
        }
      else
        {
        for (i=0; i<first_index2; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
 
        for(i=first_index2+1; i<=first_index1; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }

        *pat = *(HEAD_PAT + first_index2);
         pat++;

        for(i=first_index1+1; i<num_pat; i++, pat++)
          {
          *pat = *(HEAD_PAT + i);
          }
       
        for(i=0; i<num_pat; i++)
          {
          *(HEAD_PAT + i) = *(COPY_PAT + i);
          }
	diff1 = diff1 + 2;
        }
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: markov_density					*/
/* description	: This function creates a pattern for a certain signal 	*/
/*		  with a predetermined probability and density		*/
/* called funct	: mbkalloc, randpat		                        */
/* ###--------------------------------------------------------------### */

void    markov_density(num_pat, proba, density, eps, HEAD_PAT)
int     num_pat;
float   proba, density, eps;
char   *HEAD_PAT;
  {
  int             i;
  long            threshold;
  char           *pat;

  if(eps < 1.0) eps = 1.0;

  pat = HEAD_PAT;

  if ((pat != NULL) && (num_pat != 0))
    {

/*--------------------------------------------------------------*/
/* choose a value for the first pattern                         */
/*--------------------------------------------------------------*/

    if ( proba > 0.5)
      {
      *(pat) = '1';
      }
    else
      {
      *(pat) = '0';
      }

/*--------------------------------------------------------------*/
/* Then use markov chains (conditionnal probabilities)          */
/* for successive remainding patterns                           */
/*--------------------------------------------------------------*/

    for(i=1; i<num_pat; i++)
      {
      if (*(pat + i -1) == '0')
        {
        threshold = (long)((base * density) / ( 2 * (1 - proba)));
        }
      else
        {
        threshold = base - (long)((base * density) / ( 2 * proba));
        }

      if (random() < threshold)
        {
        *(pat + i) = '1';
        }
      else
        {
        *(pat + i) = '0';
        }
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: clock_density					*/
/* description	: This function creates a pattern for a certain signal 	*/
/*		  with a predetermined probability and density		*/
/* called funct	: mbkalloc, randpat		                        */
/* ###--------------------------------------------------------------### */

void    clock_density(num_pat, motif, HEAD_PAT)
int     num_pat;
char   *motif;
char   *HEAD_PAT;
  {
  int             i, j;
  char           *pat;

  pat = HEAD_PAT;

  if ((pat != NULL) && (num_pat != 0) && (motif != NULL))
    {

/*--------------------------------------------------------------*/
/* Use clock motif to generate patterns                         */
/*--------------------------------------------------------------*/

    for(i=0; i<num_pat; i++)
      {
      j = i % (strlen(motif));
      *(pat + i) = *(motif + j);
      printf (" I = %d, J = %d, motif = %c\n", i, j, *(motif + j));
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: Usage						        */
/* description	: This function gives a help information for mkpat	*/
/* ###--------------------------------------------------------------### */

void mkp_usage()
{
  puts("\nUsage : mkpat [options] <input_vhdl_file> <input_prb_file> <output_pat_file>\n");

  puts("               options : -b : use behavioural vhdl input file (.vbe)");
  puts("                         -i : generate pattern file with internal signals");
  puts("                         -p : generate corresponding probability file (.prb)");
  puts("                         -m : use markov-chain algorithm");
  puts("                         -n nbr_patterns  : generated patterns number");
  puts("                         -r random_point  : random starting point");
  puts("                         -d delay         : delay between 2 patterns");
  puts("                         -c clock pattern : generated clock name & pattern");
}

/* ###--------------------------------------------------------------### */
/* function	: makeghost						*/
/* description	: This function creates a ghost structure level to	*/
/*		  handle fully behavioural descriptions			*/
/* called funct	: addlofig, addlosig, addchain, mbkalloc		*/
/* ###--------------------------------------------------------------### */

struct lofig *makeghost (ptbefig)
struct befig *ptbefig;
{
  struct chain *namelst;
  struct locon *ptlocon;
  struct bepor *ptbepor;
  struct lofig *ptlofig;
  struct loins *ptloins;
  struct losig *ptlosig;
  int           index=0;
  char          lofigname[80];

  sprintf (lofigname,"root_of_%s",ptbefig->NAME);
  ptlofig = addlofig (lofigname);

  ptloins = (struct loins *) mbkalloc (sizeof(struct loins));
  ptloins->INSNAME = ptbefig->NAME;
  ptloins->FIGNAME = ptbefig->NAME;
  ptloins->LOCON   = NULL;
  ptloins->USER    = NULL;
  ptloins->NEXT    = NULL;

  ptlofig->LOINS = ptloins;

  ptbepor = ptbefig->BEPOR;
  while (ptbepor != NULL)
    {
    namelst      = addchain (NULL   ,ptbepor->NAME);
    ptlosig      = addlosig (ptlofig,index,namelst,'E',(float) 0);

    if (ptbepor->TYPE != 'B')
      ptlosig->USER = addptype(ptlosig->USER,ptbepor->TYPE,NULL);

    ptlocon      = (struct locon *) mbkalloc (2 * sizeof(struct locon));

    ptlocon->NAME      = ptbepor->NAME;
    ptlocon->SIG       = ptlosig;
    ptlocon->ROOT      = (void *)ptlofig;
    ptlocon->DIRECTION = ptbepor->DIRECTION;
    ptlocon->TYPE      = 'E';
    ptlocon->USER      = NULL;
    ptlocon->NEXT      = ptlofig->LOCON;
    ptlofig->LOCON     = ptlocon++;

    ptlocon->NAME      = ptbepor->NAME;
    ptlocon->SIG       = ptlosig;
    ptlocon->ROOT      = (void *)ptloins;
    ptlocon->DIRECTION = ptbepor->DIRECTION;
    ptlocon->TYPE      = 'I';
    ptlocon->USER      = NULL;
    ptlocon->NEXT      = ptloins->LOCON;
    ptloins->LOCON     = ptlocon;

    ptbepor = ptbepor->NEXT;
    index++;
    }
  ptlofig->LOCON = (struct locon *) reverse (ptlofig->LOCON);
  ptloins->LOCON = (struct locon *) reverse (ptloins->LOCON);

  return (ptlofig);
  }

/* ###--------------------------------------------------------------### */
/* function     : prb_loadprbseq                                        */
/* description  : load a probability sequence from a file               */
/*                add it at the top of a sequence list                  */
/* called func. : mbkfopen  , prb_error , prb_parser                    */
/* ###--------------------------------------------------------------### */

struct prbseq *prb_loadprbseq (name, ptseq)

char          *name  ;          /* name of the file to be parsed        */
struct prbseq *ptseq ;          /* list of existing sequences           */

{
  FILE          *fp      = NULL;
  struct prbseq *tmpseq ;

  fp = mbkfopen (name, PRB_IN_SUFFIX, READ_TEXT);
  if (fp == NULL)
  {
     prb_error (100, name, ' ', 0);
     exit (1);
   }

  tmpseq = prb_parser (fp, name);
  fclose (fp);

  if ((tmpseq == NULL) || (tmpseq->ERRFLG != 0))
  {
     prb_error (101, name, ' ', 0);
     exit (1);
  }

  tmpseq->NEXT = ptseq;

  return (tmpseq);
}

/* ###--------------------------------------------------------------### */
/* function     : pat_probab                                            */
/* description  : This function calculates the percentage of 1 for      */
/*                a certain variable *name in the pattern data          */
/*                or in a pattern file                                  */
/* called func. : namealloc                                             */
/* ###--------------------------------------------------------------### */
float  pat_probab(pt_paseq, name)
struct paseq  *pt_paseq;
char          *name;
{
  int       count_1 = 0;
  int       count_0 = 0; 
  char      value;
  float     result;
  unsigned  short  index = 0;
  struct    paiol  *pt_paiol = NULL;
  struct    papat  *pt_papat = NULL;
  struct    paevt  *pt_paevt = NULL;

  pt_paiol = pt_paseq->PAIOL;
/*--------------------------------------------------------------*/
/* Get the index in PAIOL of the parameter *name                */
/*--------------------------------------------------------------*/
  name = namealloc(name);
  while( name != pt_paiol->NAME)
    {
      index++;
      pt_paiol = pt_paiol->NEXT;
    }

/*--------------------------------------------------------------*/
/* If the event happened, find the initial value of *name and   */
/* do the counting as long as there happened an event for *name */
/*--------------------------------------------------------------*/
  pt_papat = pt_paseq->CURPAT;
  while(pt_papat != NULL)
    { 
      pt_paevt = pt_papat->PAEVT;
      while(pt_paevt != NULL)
	{
	  if(pt_paevt->INDEX == index)
	    {
	      value = pt_paevt->USRVAL;
	      if((value == '1') || (value == '+')) 
		count_1++;
	      else 
		if((value == '0') || (value == '-')) 
		  count_0++;
	      break;
	    }
	  pt_paevt = pt_paevt->NEXT;
	}
/*--------------------------------------------------------------*/
/* Keep the counting if there is no events for *name            */
/*--------------------------------------------------------------*/
      if(pt_paevt == NULL)
	{
	  if((value == '1') || (value == '+')) 
	    count_1++;
	  else 
	    if((value == '0') || (value == '-')) 
	      count_0++;
	}
      pt_papat = pt_papat->NEXT;
    }

  result = (float) count_1 / pt_paseq->PATNBR;
  return(result);
}

/* ###--------------------------------------------------------------### */
/* function     : pat_cntevt                                            */
/* description  : This function calculates the number of events for     */
/*                a certain variable *name in the pattern data          */
/*                or in a pattern file                                  */
/* called func. : namealloc                                             */
/* ###--------------------------------------------------------------### */
int    pat_cntevt(pt_paseq, name)
struct paseq  *pt_paseq;
char          *name;
{
  int       count_evt = 0;
  unsigned  short  index = 0;
  struct    paiol  *pt_paiol = NULL;
  struct    papat  *pt_papat = NULL;
  struct    paevt  *pt_paevt = NULL;

  pt_paiol = pt_paseq->PAIOL;
/*--------------------------------------------------------------*/
/* Get the index in PAIOL of the parameter *name                */
/*--------------------------------------------------------------*/
  name = namealloc(name);
  while(name !=  pt_paiol->NAME)
    {
      index++;
      pt_paiol = pt_paiol->NEXT;
    }

/*--------------------------------------------------------------*/
/* Do the counting as long as there happened an event for *name */
/*--------------------------------------------------------------*/
  pt_papat = pt_paseq->CURPAT;
  while(pt_papat != NULL)
    { 
    pt_paevt = pt_papat->PAEVT;
      while(pt_paevt != NULL)
	{
	  if(pt_paevt->INDEX == index)
	    {
	      count_evt++;
	      break;
	    }
	  pt_paevt = pt_paevt->NEXT;
	}
      pt_papat = pt_papat->NEXT;
    }
  count_evt--;   
  return(count_evt);
}

/* ###--------------------------------------------------------------### */
/* program      : mkpat                                                 */
/* description  : This program generates the pattern file according     */
/*                to the predetermined needs of the probability and     */
/*                density of each signal                                */
/* ###--------------------------------------------------------------### */
main (argc,argv)
int       argc;
char    **argv;
{
  struct lofig   *ptlofig;
  struct befig   *ptbefig;
  struct locon   *ptlocon;
  struct beaux   *ptbeaux;
  struct losig   *ptlosig;
  struct paseq   *pt_paseq  = NULL;
  struct paiol   *pt_paiol  = NULL;
  struct prbseq  *pt_prbseq = NULL;
  struct prbevt  *pt_prbevt = NULL;
  FILE	         *prbfile;
  FILE	         *patfile;
  char	         *mode;
  char	         *name;
  char 	         *vecteur;
  char            beh_flg       = 0;
  char            dly_flg       = 0;
  char            markov_flg    = 0;
  char            internal_flg  = 0;
  char            usage_flg     = 0;
  char            prbout_flg    = 0;
  char            clock_flg     = 0;
  char           *clock_pattern;
  char           *arg_flg   = NULL;
  int             last_opt  = 0;
  int             frst_fil  = 0;
  char	          s[3];
  time_t          exec_time;
  int	          left,right;
  int	          i, j, k, jth_sig;
  int             cur_pat;
  int             length_head = 0;
  int             length_delay = 0;
  int             found = 0;
  int             Num_pat   = 100;
  int             Rand_point=1121;
  int             p_evt;
  int             num_input_sig;
  int             num_internal_sig;
  float           p_1; 
  char           *pat;
  char           *delay = NULL;
  char           *HEAD_MATRIX;
  char            pat_tmp;

        /* ###------------------------------------------------------### */
        /*    analyse the command line, set option flags and find the   */
        /* first file in the argument list                              */
        /* ###------------------------------------------------------### */

  name          = (char *)mbkalloc(256);
  delay         = (char *)mbkalloc(256);
  mode          = (char *)mbkalloc(256);
  clock_pattern = (char *)mbkalloc(256);
  vecteur       = (char *)mbkalloc(4096);

  arg_flg = (char *) mbkalloc (argc * sizeof (char));
  arg_flg [0] = 1;
  for (i=1 ; i<argc ; i++)
    arg_flg [i] = 0;

  for (i=1 ; i<argc ; i++)
    {
    if (arg_flg [i] == 0)
      {
      if (!strcmp (argv[i], "-b"))
        {
        beh_flg     = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-m"))
        {
        markov_flg = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-p"))
        {
        prbout_flg = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-i"))
        {
        internal_flg     = 1;
        arg_flg [i] = 1;
        }

      if (!strcmp (argv[i], "-n"))
        {
        sscanf (argv[i+1], "%u", &Num_pat);
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        }

      if (!strcmp (argv[i], "-r"))
        {
        sscanf (argv[i+1], "%u", &Rand_point);
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        }

      if (!strcmp (argv[i], "-c"))
        {
        clock_flg     = i + 1;
        sscanf (argv[i+2], "%s", clock_pattern);
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        arg_flg [i+2] = 1;
        }

      if (!strcmp (argv[i], "-d"))
        {
        dly_flg     = 1;
        sscanf (argv[i+1], "%s", delay);
        arg_flg [i  ] = 1;
        arg_flg [i+1] = 1;
        }
      }
    }

  for (i=0 ; i<argc ; i++)
    {
    if (arg_flg [i] == 1)
      last_opt = i;
    }


  for (frst_fil=0; frst_fil<argc; frst_fil++)
    {
    if (arg_flg [frst_fil] == 0)
      break;
    }

  if ((argc == 1) || (last_opt != (frst_fil - 1)))
    usage_flg = 1;

  if (argc != (frst_fil + 3))
    usage_flg = 1;

  if (usage_flg == 1)
    {
    mkp_usage ();
    exit (1);
    }


  alliancebanner ("MKPAT", "v1.00", "A Pseudo-Random Pattern Generation Tool",
                  "1995..1996", "1.0");
 
        /* ###------------------------------------------------------### */
        /*    Initialize :                                              */
        /*      - BDD hash tables                                       */
        /*      - get MBK environment variables                         */
        /* ###------------------------------------------------------### */

  fprintf(stdout,"\nInitialize BDD system ...\n");
  initializeBdd (0);
  mbkenv();

  PRB_IN_SUFFIX  = namealloc("prb");
  PRB_OUT_SUFFIX = namealloc("prb");
        /*--------------------------------------------------------------*/
        /* Load the comportemental or structural file, and the          */
        /* probability description file                                 */
        /*--------------------------------------------------------------*/

  if (beh_flg == 1)
    {
    fprintf(stdout,"\nLoading comportement file ...\n");
    ptbefig = vhdlloadbefig (ptbefig, argv[frst_fil], 3);

    if (!ptbefig)
      mkp_error (1,argv[frst_fil], NULL, NULL);

    ptlofig = makeghost (ptbefig);
    }
  else
    {
    fprintf(stdout,"\nLoading structure file ...\n");
    ptlofig = getlofig(argv[frst_fil],'A');

    if (!ptlofig) 
      mkp_error (2,argv[frst_fil], NULL, NULL);
    }

  fprintf (stdout,"\nLoading probability file ...\n");
  pt_prbseq = prb_loadprbseq (argv[frst_fil+1], pt_prbseq);

        /*--------------------------------------------------------------*/
        /* Write the title in pattern file.                             */
        /*--------------------------------------------------------------*/

  fprintf(stdout,"\nOpening Result file ...\n");
  patfile = mbkfopen(argv[frst_fil+2], "pat", WRITE_TEXT);

  if (!patfile)
    mkp_error (3,argv[frst_fil+2], NULL, NULL);

  time(&exec_time);
  fprintf(patfile, "-- Description generated by Mkpat driver\n");
  fprintf(patfile, "-- \t\t date : %s\n\n", ctime(&exec_time));
  fprintf(patfile, "-- input / output list :\n\n");


  strcpy(vecteur, "");

  if (dly_flg == 1)
    {
    length_delay = strlen(delay) + 4;
    for (i=1; i<length_delay+1; i++)
      strcat(vecteur, " ");
    strcat(vecteur, " ");
    }

  strcat(vecteur, ptlofig->NAME);
  length_head = strlen(vecteur);
  strcat(vecteur, " : ");
 
  num_input_sig = 0;
  ptlofig->LOCON = (struct locon *)mvl_reverse((chain_list *)ptlofig->LOCON);

  for (ptlocon = ptlofig->LOCON ; ptlocon ; ptlocon = ptlocon->NEXT)
    {
    switch (ptlocon->DIRECTION) 
      {
      case IN       : 
        strcpy(mode,"IN");    
        strcpy(s,"0"); 
        num_input_sig++;
        break;
      case OUT      : 
        strcpy(mode,"OUT");   
        strcpy(s,"*"); 
        break;
      case INOUT    : 
        strcpy(mode,"INOUT"); 
        strcpy(s,"*"); 
        break;
      case TRISTATE : 
        strcpy(mode,"OUT");   
        strcpy(s,"*"); 
        break;
      case TRANSCV  : 
        strcpy(mode,"INOUT"); 
        strcpy(s,"*"); 
        break;
      default       :
        mkp_error (5, ptlocon->NAME, NULL, NULL);
      }

    ptlocon = (locon_list *)mvl_vectnam(ptlocon, &left, &right, &name, 1);
    if (left != -1)  
      {
      fprintf(patfile, "%s	%s 	(%d %s %d) 	B;;\n",
	      mode, name, left, (left >= right)? "DOWNTO":"TO", right);
      if(s[0] == '0') num_input_sig += ((left >= right)? left-right : right-left); 
      }
    else 
      {
      fprintf(patfile, "%s 	%s 			B;;\n", mode, name);
      }
    }

  fprintf(patfile,"\n\n");

/*--------------------------------------------------------------*/
/* Count the number of the internal signals                     */
/*--------------------------------------------------------------*/

  if (internal_flg == 1)
    {
    if (beh_flg == 1)
      {
      num_internal_sig = 0;
      ptbeaux = ptbefig->BEAUX;
      while (ptbeaux)
        {
        strcpy(mode, ptbefig->NAME);
        strcat(mode, ".");
        strcat(mode, ptbeaux->NAME);
        fprintf(patfile, "SIGNAL %s\tB;;\n", mode);
        num_internal_sig++;
        ptbeaux = ptbeaux->NEXT;
        }
      }
    else
      {
      num_internal_sig = 0;
      ptlosig = ptlofig->LOSIG;
      while((ptlosig) && (ptlosig->TYPE == 'I'))
        {
        strcpy(mode, getsigname(ptlosig));
        fprintf(patfile, "SIGNAL %s\tB;;\n", mode);
        num_internal_sig++;
        ptlosig = ptlosig->NEXT;
        }
      }
    }

  fprintf(patfile,"\nbegin\n\n");

/*--------------------------------------------------------------*/
/* Initialize the random generator 				*/
/*--------------------------------------------------------------*/

  srandom(Rand_point);

/*--------------------------------------------------------------*/
/* Write the patterns for each signal in pattern file           */
/*--------------------------------------------------------------*/

  pat = (char *)mbkalloc(sizeof(char) * Num_pat); 

/*--------------------------------------------------------------*/
/* Store the patterns for all of the input signals              */
/*--------------------------------------------------------------*/
  HEAD_MATRIX = (char *)mbkalloc(sizeof(char) * Num_pat * num_input_sig);

  jth_sig = 0;
  for (ptlocon = ptlofig->LOCON; ptlocon ; ptlocon = ptlocon->NEXT)
    {
    if (ptlocon->DIRECTION == 'I')
/*--------------------------------------------------------------*/
/* For each input signal, generate the patterns satisfying the  */
/* probability and density needs                                */
/*--------------------------------------------------------------*/
      {
      for (i=0; i<Num_pat; i++)
        {
        *(pat+i) = '0';
        }

      if ((ptlocon->NAME == namealloc(argv[clock_flg])) && (clock_flg != 0))
        {
        fprintf(stdout,"\nGenerating clock pattern (%dth input signal)...\n", jth_sig+1); 
        clock_density(Num_pat, clock_pattern, pat);
        }
      else
        {
        fprintf(stdout,"\nGenerating pattern of the %dth input signal ...\n", jth_sig+1); 
        pt_prbevt = pt_prbseq->PRBEVT;

        while(pt_prbevt)
          {
          if (ptlocon->NAME == pt_prbevt->NAME) 
            {
            found = 1;
            break;
            }
          pt_prbevt = pt_prbevt->NEXT;
          }

        if (found == 1)
          {
          if (markov_flg == 1)
            {
            markov_density(Num_pat, pt_prbevt->PROBVAL, pt_prbevt->DENSVAL,
                           Densi_eps, pat);
            }
          else
            {
            proba_density(Num_pat, pt_prbevt->PROBVAL, pt_prbevt->DENSVAL,
                          Densi_eps, pat);
            }
          }
        else
          {
          if (markov_flg == 1)
            {
            markov_density(Num_pat, pt_prbseq->PROBVAL, pt_prbseq->DENSVAL,
                           Densi_eps, pat);
            }
          else
            {
            proba_density(Num_pat, pt_prbseq->PROBVAL, pt_prbseq->DENSVAL,
                          Densi_eps, pat);
            }
          }
        found = 0;
        }
/*--------------------------------------------------------------*/
/* Store into the allocated place HEAD_MATRIX by column         */
/*--------------------------------------------------------------*/
      for (i=0; i<Num_pat; i++)
        {
        *(HEAD_MATRIX + i*num_input_sig + jth_sig) = *(pat + i);
        }  
      jth_sig++;
      }
    }

  fprintf(stdout,"\nWritting pattern file ...\n");

/*--------------------------------------------------------------*/
/* Output the patterns in its defined form                      */
/*--------------------------------------------------------------*/
  for (cur_pat=0; cur_pat<Num_pat; cur_pat++)
    {
    if (cur_pat != 0)
      {
      strcpy(vecteur, "");

      if (dly_flg == 1)
        {
        strcat(vecteur, "<+ ");
        strcat(vecteur, delay);
        strcat(vecteur, ">");
        }

      for (i=1; i<length_head-length_delay+1; i++)
        strcat(vecteur, " ");

      strcat(vecteur, " : "); 
     
/*--------------------------------------------------------------*/
/* The variable "vecteur" is used to present a pattern at a     */
/* certain instant                                              */
/*--------------------------------------------------------------*/
      }

    jth_sig = -1;
    for (ptlocon = ptlofig->LOCON; ptlocon ; ptlocon = ptlocon->NEXT)
      {
      switch (ptlocon->DIRECTION) 
	{
	case IN       : 
	  strcpy(s, "0"); 
          jth_sig++;
	  break;
        case OUT      : 
	  strcpy(s, "*"); 
	  break;
	case INOUT    : 
          strcpy(s, "*"); 
          break;
        case TRISTATE : 
          strcpy(s, "*"); 
          break;
        case TRANSCV  : 
          strcpy(s, "*"); 
          break;
        default       :
          mkp_error (5, ptlocon->NAME, NULL, NULL);
	}

/*--------------------------------------------------------------*/
/* Output the signals with the same name, but different index,  */
/* in a special form in the pattern file                        */
/*--------------------------------------------------------------*/

      ptlocon = (locon_list *)mvl_vectnam(ptlocon, &left, &right, &name, 1);
      if (s[0] == '*') strcat(vecteur,"?");

      if (left != -1)  
	{
        for (i = 0 ; i <= ((left >= right)? left-right : right-left) ; i++)
	  {
	  if (s[0] == '*')
	    {
	    strcat(vecteur, "*");
	    }
	  else
	    {
	    pat_tmp = *(HEAD_MATRIX + cur_pat*num_input_sig + jth_sig+i);
            strncat(vecteur, &pat_tmp, 1);
	    }
	  }
        if(s[0] == '0') jth_sig += ((left >= right)? left-right : right-left);
        }
      else 
        {
	if (s[0] == '*')
	  {
          strcat(vecteur, "*");
          }
        else
	  {
	  pat_tmp = *(HEAD_MATRIX + cur_pat*num_input_sig + jth_sig);  
	  strncat(vecteur, &pat_tmp, 1);
	  }
        }

      strcat(vecteur,"  ");
      }

    if (internal_flg == 1)
      {
      for(j=0; j<num_internal_sig; j++)
        {
	strcat(vecteur, "?*  ");
        }
      }

    fprintf(patfile,"%s ; \n",vecteur);
    }

  fprintf(patfile,"end;\n");
  fclose(patfile);

/*--------------------------------------------------------------*/
/* Create a .prb file as the verification of pattern generation */
/*--------------------------------------------------------------*/

  if (prbout_flg == 1)
    {
    fprintf(stdout,"\nLoading pattern file ...\n");
    pt_paseq = pat_lodpaseq (argv[frst_fil+2], NULL, 0);

    prbfile = mbkfopen(argv[frst_fil+2], PRB_IN_SUFFIX, WRITE_TEXT);

    if (!prbfile)
      mkp_error (4,argv[frst_fil+2], PRB_IN_SUFFIX, NULL);
 
    fprintf(stdout,"\nWriting probability file ...\n\n");
    pt_paiol = pt_paseq->PAIOL;
    fprintf(prbfile, "\n\n# Core file generated from %s\n", pt_paseq->NAME);
    fprintf(prbfile, "#\t\t date : %s\n\n", ctime(&exec_time));
    fprintf(prbfile, "\t# Figure : %s ==\n\n", pt_paseq->NAME);
    fprintf(prbfile, "\t# -- Simple Signals --\n\n");
    while(pt_paiol != NULL)
      {
      switch(pt_paiol->MODE)
        {
        case 'I' :
          p_evt = pat_cntevt(pt_paseq, pt_paiol->NAME);
          p_1 = pat_probab(pt_paseq, pt_paiol->NAME);
          fprintf(prbfile, "in   %-15s  : p('1') = %-6.4f | density = %-6.4f;\n",
                  pt_paiol->NAME, p_1, (float)(p_evt / (float)Num_pat) );
          break;
        case 'O' :
          break;
        case 'B' :
          break;
        }
      pt_paiol = pt_paiol->NEXT;
      }
    fclose(prbfile);
    exit(0);
  }
  return( 0 );
}

