
/* ###--------------------------------------------------------------### */
/* file		: prb_maksupvarlist.c					*/
/* date		: Jan 20 1998						*/
/* version	: v107							*/
/* authors	: DUNOYER Julien.					*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include "beh.h"
#include "bhl.h"
#include PRB_H


/* ###--------------------------------------------------------------### */
/* function	: prb_mvlsttot						*/
/* description	: make list of variables for all primary signals	*/
/*		  simple and bussed outputs,				*/
/*		  simple and bussed internal signals, registers.	*/
/* called func.	: prb_abl2varlist					*/
/* ###--------------------------------------------------------------### */

static void prb_mvlsttot (pt_befig)

struct befig *pt_befig;		/* pointer on current BEFIG	*/

  {
  struct beaux  *pt_beaux;
  struct bemsg  *pt_bemsg;
  struct beout  *pt_beout;
  struct bebus  *pt_bebus;
  struct bebux  *pt_bebux;
  struct bereg  *pt_bereg;
  struct binode *pt_binode;


	/* ###------------------------------------------------------### */
	/*    make a varlist for each simple internal signal		*/
	/* ###------------------------------------------------------### */

  pt_beaux = pt_befig->BEAUX;

  while (pt_beaux != NULL)
    {
    if (PRB_SUPERVARLIST(pt_beaux->NODE) == NULL)
      {
      if (PRB_SUPERFLAG(pt_beaux->NODE) == DEF_PRB_SUPERGATE)
        {
        PRB_SUPERVARLIST(pt_beaux->NODE) =
            beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_beaux->NODE));
        }
      else
        {
        PRB_SUPERVARLIST(pt_beaux->NODE) = PRB_VARLIST(pt_beaux->NODE);
        }
      }

    pt_beaux = pt_beaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a varlist for each delayed internal signal		*/
	/* ###------------------------------------------------------### */

  pt_beaux = pt_befig->BEDLY;
  while (pt_beaux != NULL)
    {
    if (PRB_SUPERVARLIST(pt_beaux->NODE) == NULL)
      {
      if (PRB_SUPERFLAG(pt_beaux->NODE) == DEF_PRB_SUPERGATE)
        {
        PRB_SUPERVARLIST(pt_beaux->NODE) =
            beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_beaux->NODE));
        }
      else
        {
        PRB_SUPERVARLIST(pt_beaux->NODE) = PRB_VARLIST(pt_beaux->NODE);
        }
      }

    pt_beaux = pt_beaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a varlist for each assertion				*/
	/* ###------------------------------------------------------### */

  pt_bemsg = pt_befig->BEMSG;
  while (pt_bemsg != NULL)
    {
    if (PRB_SUPERVARLIST(pt_bemsg->NODE) == NULL)
      {
      if (PRB_SUPERFLAG(pt_bemsg->NODE) == DEF_PRB_SUPERGATE)
        {
        PRB_SUPERVARLIST(pt_bemsg->NODE) =
            beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_bemsg->NODE));
        }
      else
        {
        PRB_SUPERVARLIST(pt_bemsg->NODE) = PRB_VARLIST(pt_bemsg->NODE);
        }
      }

    pt_bemsg = pt_bemsg->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a varlist for each simple output port		*/
	/* ###------------------------------------------------------### */

  pt_beout = pt_befig->BEOUT;
  while (pt_beout != NULL)
    {
    if (PRB_SUPERVARLIST(pt_beout->NODE) == NULL)
      {
      if (PRB_SUPERFLAG(pt_beout->NODE) == DEF_PRB_SUPERGATE)
        {
        PRB_SUPERVARLIST(pt_beout->NODE) =
            beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_beout->NODE));
        }
      else
        {
        PRB_SUPERVARLIST(pt_beout->NODE) = PRB_VARLIST(pt_beout->NODE);
        }
      }

    pt_beout = pt_beout->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a varlist for each bussed output port		*/
	/* ###------------------------------------------------------### */

  pt_bebus = pt_befig->BEBUS;
  while (pt_bebus != NULL)
    {
    pt_binode = pt_bebus->BINODE;

    while (pt_binode != NULL)
      {
      if (PRB_SUPERVARLIST(pt_binode->CNDNODE) == NULL)
        {
        if (PRB_SUPERFLAG(pt_binode->CNDNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERVARLIST(pt_binode->CNDNODE) =
              beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_binode->CNDNODE));
          }
        else
          {
          PRB_SUPERVARLIST(pt_binode->CNDNODE) = PRB_VARLIST(pt_binode->CNDNODE);
          }
        }

      if (PRB_SUPERVARLIST(pt_binode->VALNODE) == NULL)
        {
        if (PRB_SUPERFLAG(pt_binode->VALNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERVARLIST(pt_binode->VALNODE) =
              beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_binode->VALNODE));
          }
        else
          {
          PRB_SUPERVARLIST(pt_binode->VALNODE) = PRB_VARLIST(pt_binode->VALNODE);
          }
        }

      pt_binode = pt_binode->NEXT;
      }

    pt_bebus = pt_bebus->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a varlist for each bussed internal signal		*/
	/* ###------------------------------------------------------### */

  pt_bebux = pt_befig->BEBUX;
  while (pt_bebux != NULL)
    {
    pt_binode = pt_bebux->BINODE;

    while (pt_binode != NULL)
      {
      if (PRB_SUPERVARLIST(pt_binode->CNDNODE) == NULL)
        {
        if (PRB_SUPERFLAG(pt_binode->CNDNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERVARLIST(pt_binode->CNDNODE) =
              beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_binode->CNDNODE));
          }
        else
          {
          PRB_SUPERVARLIST(pt_binode->CNDNODE) = PRB_VARLIST(pt_binode->CNDNODE);
          }
        }

      if (PRB_SUPERVARLIST(pt_binode->VALNODE) == NULL)
        {
        if (PRB_SUPERFLAG(pt_binode->VALNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERVARLIST(pt_binode->VALNODE) =
              beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_binode->VALNODE));
          }
        else
          {
          PRB_SUPERVARLIST(pt_binode->VALNODE) = PRB_VARLIST(pt_binode->VALNODE);
          }
        }

      pt_binode = pt_binode->NEXT;
      }
    pt_bebux = pt_bebux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a varlist for each internal register			*/
	/* ###------------------------------------------------------### */

  pt_bereg = pt_befig->BEREG;
  while (pt_bereg != NULL)
    {
    pt_binode = pt_bereg->BINODE;

    while (pt_binode != NULL)
      {
      if (PRB_SUPERVARLIST(pt_binode->CNDNODE) == NULL)
        {
        if (PRB_SUPERFLAG(pt_binode->CNDNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERVARLIST(pt_binode->CNDNODE) =
              beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_binode->CNDNODE));
          }
        else
          {
          PRB_SUPERVARLIST(pt_binode->CNDNODE) = PRB_VARLIST(pt_binode->CNDNODE);
          }
        }

      if (PRB_SUPERVARLIST(pt_binode->VALNODE) == NULL)
        {
        if (PRB_SUPERFLAG(pt_binode->VALNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERVARLIST(pt_binode->VALNODE) =
              beh_abl2varlist (pt_befig, PRB_SUPERABL(pt_binode->VALNODE));
          }
        else
          {
          PRB_SUPERVARLIST(pt_binode->VALNODE) = PRB_VARLIST(pt_binode->VALNODE);
          }
        }

      pt_binode = pt_binode->NEXT;
      }
    pt_bereg = pt_bereg->NEXT;
    }

  }

/* ###--------------------------------------------------------------### */
/* function	: prb_maksupvarlist					*/
/* description	: make var list for each signal of a befig structure 	*/
/* called func.	: initializeCct, addInputCct, prb_mvlsttot		*/
/* ###--------------------------------------------------------------### */

void prb_maksupvarlist (pt_befig)

struct befig *pt_befig;

  {
  int         err_flg   = 0;

  if ((pt_befig->TYPE & BEH_NODE_VARLIST )
  &&  (pt_befig->TYPE & BEH_NODE_QUAD    )
  &&  (pt_befig->TYPE & BEH_NODE_USER    ))
    {
    prb_mvlsttot (pt_befig);
    }

        /* ###------------------------------------------------------### */
        /*    Then get a new list of variables for all expressions	*/
        /* ###------------------------------------------------------### */
  else
    {
    prb_toolbug (4, "prb_maksupbdd", ' ');
    }

  if (err_flg != 0)
    pt_befig->ERRFLG = 1;
  }

