%{

/* ###--------------------------------------------------------------### */
/*									*/
/* file		: prb_parser.lex					*/
/* date		: May 24 1994						*/
/* version	: v100							*/
/* author	: Diane PITON 						*/
/* content	: parser-driver function				*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <stdlib.h>
#include MUT_H
#include "prb_defs.h"
#include "prb_parser.h"


int	PRB_LINNUM = 1;


%}

%s LEFT_PART
%s RIGHT_PART

letter		[a-zA-Z]
number		[0-9]

%%

  /*###------------------------------------------------------###*/
  /*			separators				*/
  /*###------------------------------------------------------###*/

<INITIAL>\,				{ return (Comma);} 

<INITIAL>\.				{ return (Dot);}


<INITIAL>\:				{ 
			BEGIN LEFT_PART; 
			return (Colon);
					}

<LEFT_PART>\=				{ 
			BEGIN RIGHT_PART; 
			return (Equal); 
					}

<RIGHT_PART>\|				{ 
			BEGIN LEFT_PART; 
			return (Pipe); 
					}

<RIGHT_PART>\;				{ 
			BEGIN INITIAL; 
			return (Semicolon); 
					}

  /*###------------------------------------------------------###*/
  /*			LEFT PART				*/
  /*###------------------------------------------------------###*/

<LEFT_PART>[pP]\(\'[01X]\'\)		{ 
			if (yytext[3] == '0')
			   return (Pof0);
			else if (yytext[3] == '1')
			   return (Pof1);
			else
			   return (PofX);
					}

<LEFT_PART>{letter}{letter}+		{  
			int         code;
			static int  lcl_search ();

			yylval.text = namealloc (yytext);
			code = lcl_search (yylval.text); 
			switch (code)
			  {
			  case DENSITY   : return (DENSITY) ; break;
			  case REGACCESS : return (REGACCESS) ; break;
			  default        : return (UnknownChar);
			  }
					}

  /*###------------------------------------------------------###*/
  /*			RIGHT PART and INITIAL PART			*/
  /*###------------------------------------------------------###*/

<INITIAL>\(			{  return (LeftParen); }
<INITIAL>\)			{  return (RightParen); }

<INITIAL,RIGHT_PART>{letter}(_?({letter}|{number}))*	{
			int         code;
			static int  lcl_search ();

			yylval.text = namealloc (yytext);
			if ((code = lcl_search (yylval.text)) == -1)
			  return (Identifier);
			else
			{
			  if (code == DEFAULT)
			     BEGIN LEFT_PART;
			  return (code);
			}
					}
<INITIAL>{number}+		{
			yylval.valu = atoi (yytext);
			return (AbstractLit);
					}
<RIGHT_PART>{number}+\.{number}*	{
			yylval.fvalu = atof (yytext);
			return (Float);
					}

\#.*$					{ }
[ \t]					{ }

\n					{ PRB_LINNUM++; }

.					{ return (UnknownChar); }
%%

/* ###--------------------------------------------------------------### */
/* function	: yywrap						*/
/* description	: what to do at the end of a file 1 => finish		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static int yywrap()
{
    return (1);
}

/* ###--------------------------------------------------------------### */
/* function	: lcl_search						*/
/* description	: check that an identifier is a reserved word or not	*/
/* called func.	: addht, addhtitem, gethtitem, namealloc		*/
/* ###--------------------------------------------------------------### */

static int lcl_search (key)

char  *key;

  {
  static ht *pt_hash = NULL;

  if (pt_hash == NULL)
    {
    pt_hash = addht (20);

    addhtitem (pt_hash, namealloc("default")	, DEFAULT);
    addhtitem (pt_hash, namealloc("in")		, _IN);
    addhtitem (pt_hash, namealloc("out")	, _OUT);
    addhtitem (pt_hash, namealloc("inout")	, _INOUT);
    addhtitem (pt_hash, namealloc("tristate")	, _TRISTATE);
    addhtitem (pt_hash, namealloc("transc")	, _TRANSC);
    addhtitem (pt_hash, namealloc("port")	, _PORT);
    addhtitem (pt_hash, namealloc("signal")	, _SIGNAL);
    addhtitem (pt_hash, namealloc("assert")	, _ASSERT);
    addhtitem (pt_hash, namealloc("register")	, _REGISTER);
    addhtitem (pt_hash, namealloc("_clock")	, _CLOCK);
    addhtitem (pt_hash, namealloc("_regaccess")	, REGACCESS);
    addhtitem (pt_hash, namealloc("_scannable")	, TOSCAN);
    addhtitem (pt_hash, namealloc("_unscannable"), UNSCAN);
    addhtitem (pt_hash, namealloc("density")	, DENSITY);
    addhtitem (pt_hash, namealloc("to")		, TO);
    addhtitem (pt_hash, namealloc("downto")	, DOWNTO);
    }
  return (gethtitem (pt_hash, namealloc(key)));
  }
