/* ###--------------------------------------------------------------### */
/*									*/
/* file		: prb_simulator.c					*/
/* date		: Mar  1 1994						*/
/* version	: v1.00							*/
/* author	: DUNOYER Julien					*/
/* description	: This file contains the high level function :		*/
/*		  prb_simulator						*/
/*		  this  function do a symbolic simulation to calculate	*/
/*		  static probabilities in a circuit.			*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <sys/time.h>

#include MUT_H
#include LOG_H
#include SCG_H
#include GSS_H
#include PRB_H

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: prb_simulator						*/
/* description	: symbolic simulation function for correlation		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */


struct dct_entry **prb_simulator (pt_lofig,
                                  pt_befig,
                                  input_filename)

  struct lofig      *pt_lofig;
  struct befig      *pt_befig;
  char              *input_filename;

  {
  struct dct_entry **dictionnary;
  struct gssseq     *pt_gssseq;
  struct lkdfig     *pt_lkdfig;
  struct funpack    *prb_package;
  char               prb_errflg = 0;
  int                cur_error;
  int                i = 0;

  gss_init ();

  GSS_DELAY_MODE = GSS_ZERO_DELAY;
  GSS_SCG_MODE   = SCG_TRA_DEFAULT;
  GSS_MAXDELTA   = 1000;

  dictionnary = gss_initab ();

  prb_package = gss_addfunpack( prb_addvalue,
                                prb_freevalue,
                                prb_cmpvalue,
                                prb_cpyvalue,
                                prb_updvalue,
                                prb_evalvalue,
                                prb_resolvalue,
                                NULL,
                                NULL,
                                NULL);

  /* put the LOCONs of each instance  in the same order as the BEPOR.*/

  gss_resord (pt_lofig, pt_befig);

  /* restore the mode of ext. ports when the declared mode is LINKAGE*/

  gss_restor (pt_lofig, pt_befig);

  if (prb_errflg != 0)
    {
    prb_errflg = prb_error (0, NULL, 0.0, ' ', 0);
    exit (1);
    }

        /* ###------------------------------------------------------### */
        /*    Link previous structures to create the simulator's data   */
        /* base structures                                              */
        /* ###------------------------------------------------------### */

  prb_message (9, NULL, 0);

  pt_lkdfig   = gss_linker (pt_lofig, prb_package);

/*   gss_debug ( pt_lkdfig, "lkdfig"); */
  
        /* ###------------------------------------------------------### */
        /*    Running the sequence description loader                   */
        /* ###------------------------------------------------------### */

  prb_message (8, input_filename, 0);

  pt_gssseq  = prb_loadsequence (input_filename);

  prb_message (9, NULL, 0);

  gss_lnkgssseq (pt_gssseq);

  /* ###------------------------------------------------------------### */
  /*    Simulating the current figure                                   */
  /* ###------------------------------------------------------------### */

  prb_message (10, NULL, 0);

  gss_circuitinit(pt_lkdfig, PRB_INPUT_VALUE, prb_package);

        /* ###------------------------------------------------------### */
        /*    Perform the static simulation.                            */
        /* ###------------------------------------------------------### */

  gss_forcesequence (pt_gssseq, prb_package);

  gss_circuitstable (pt_lkdfig, prb_package);

  fprintf (stdout,"\n End Of Static Simulation \n");
  fflush (stdout);

        /* ###------------------------------------------------------### */
        /*    Perform the dynamic simulation.                           */
        /* ###------------------------------------------------------### */

  if ( PRB_MODE & PRB_DYNAMIC_MODE )
    {
    prb_package = gss_addfunpack( prb_addvalue,
                                  prb_freevalue,
                                  prb_dyncmpvalue,
                                  prb_dyncpyvalue,
                                  prb_dynupdvalue,
                                  prb_dynevalvalue,
                                  prb_dynresolvalue,
                                  prb_dynevaldelay,
                                  prb_dynevalinert,
                                  prb_dyncolvalue);

    GSS_DELAY_MODE = GSS_CUSTOM_DELAY;
    GSS_SCG_MODE   = SCG_TRA_COLLECTOR;
/*     GSS_SCG_MODE   = SCG_TRA_DEFAULT; */


    gss_circuitinit(pt_lkdfig, PRB_INIT_VALUE, prb_package);
    gss_forcesequence (pt_gssseq, prb_package);

/*     gss_circuitstable (pt_lkdfig, prb_package); */

    gss_deltaloop (prb_package);
	
    while ( GSS_NEXT_TIME != GSS_CURRENT_TIME )
      {
      GSS_CURRENT_TIME = GSS_NEXT_TIME;

      if ( i == 20 )
        {
        fprintf (stdout,"%10d ps\b\b\b\b\b\b\b\b\b\b\b\b\b", GSS_CURRENT_TIME);
        fflush (stdout);
        i = 1;
        }
      else
        {
        i++;
        }
      gss_deltaloop (prb_package);
      }

    fprintf (stdout,"%10d ps\n", GSS_CURRENT_TIME);
    fprintf (stdout,"\n End Of Dynamic Simulation \n");
    fflush (stdout);
    }

  prb_storesequence (dictionnary, pt_lofig, pt_befig);

  return(dictionnary);
  }
