
/* ###--------------------------------------------------------------### */
/* file		: prb_supdepend.c					*/
/* date		: Jan 21 1998						*/
/* version	: v111							*/
/* authors	: Pirouz BAZARGAN SABET, Julien DUNOYER			*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include "beh.h"
#include "bhl.h"
#include PRB_H

	/* ###------------------------------------------------------### */
	/*    A hash table containing inputs name and their pointers	*/
	/*    This is not ideal since this table is already constructed	*/
	/*    in beh_depend (the direct dependencies lists)...but...	*/
	/* ###------------------------------------------------------### */

static ht *rin_list = NULL;

/* ###--------------------------------------------------------------### */
/* function	: prb_expdepend						*/
/* description	: find dependencies for an expression			*/
/* called func.	: beh_namelist, gethtitem				*/
/* ###--------------------------------------------------------------### */

static struct chain *prb_expdepend (pt_fig, pt_exp)

struct befig *pt_fig;
struct chain *pt_exp;

  {
  struct chain *res_chain = NULL;
  struct chain *del_chain = NULL;
  struct chain *tmp_chain = NULL;
  struct chain *mem_chain = NULL;
  int           value           ;

  if (pt_fig->ERRFLG == 0)
    {
	/* ###------------------------------------------------------### */
	/*   if there is no error in the current figure, find the list	*/
	/* of terminals for the expression				*/
	/* ###------------------------------------------------------### */

    res_chain = beh_namelist (pt_exp);

	/* ###------------------------------------------------------### */
	/*   scan the list of terminals to replace terminal names by	*/
	/* the pointer on the corresponding primary input (berin).	*/
	/*   check the unicity of terminals ...				*/
	/* ###------------------------------------------------------### */

    tmp_chain = res_chain;

    while (tmp_chain != NULL)
      {
      del_chain = tmp_chain->NEXT;
      while (del_chain != NULL)
        {
        mem_chain = del_chain->NEXT;

        if (del_chain->DATA == tmp_chain->DATA)
          tmp_chain = delchain (tmp_chain, del_chain);

        del_chain = mem_chain;
        }

      value = gethtitem (rin_list, tmp_chain->DATA);

      if (value == EMPTYHT)
        pt_fig->ERRFLG = beh_error (3, tmp_chain->DATA);
      else
        tmp_chain->DATA = (void *) value;

      tmp_chain = tmp_chain->NEXT;
      }
    }

  return (res_chain);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_supdepend						*/
/* description	: find dependencies between signals for a behavioural	*/
/*		  description.						*/
/* called func.	: prb_expdepend, addht    , addhtitem, addchain,	*/
/*		  freechain    , append   , delhtitem 			*/
/* ###--------------------------------------------------------------### */

void prb_supdepend (pt_fig)

struct befig *pt_fig;

  {
  struct chain  *res_chain = NULL;
  struct beaux  *pt_beaux    ;
  struct beaux  *pt_bedly    ;
  struct bebux  *pt_bebux    ;
  struct bereg  *pt_bereg    ;
  struct bemsg  *pt_bemsg    ;
  struct berin  *pt_berin    ;
  struct beout  *pt_beout    ;
  struct bebus  *pt_bebus    ;
  struct binode *pt_binode  ;
  unsigned int   count     = 0;

	/* ###------------------------------------------------------### */
	/*   if there is no error in the current figure ...		*/
	/* ###------------------------------------------------------### */

  if ((pt_fig != NULL) && (pt_fig->ERRFLG == 0))
    {
	/* ###------------------------------------------------------### */
	/*   initialize a hash table with inputs			*/
	/* ###------------------------------------------------------### */

    pt_berin = pt_fig->BERIN;
    while (pt_berin != NULL)
      {
      count++;
      pt_berin = pt_berin->NEXT;
      }

    pt_berin = pt_fig->BERIN;

    if (pt_berin != NULL)
      rin_list = addht (count);

    while (pt_berin != NULL)
      {
      addhtitem (rin_list, pt_berin->NAME, pt_berin);
      pt_berin = pt_berin->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   Delete previously recorded _REF lists			*/
	/* ###------------------------------------------------------### */

    pt_berin = pt_fig->BERIN;
    while (pt_berin != NULL)
      {
      if (pt_berin->AUX_REF != NULL)
        {
        freechain (pt_berin->AUX_REF);
        pt_berin->AUX_REF = NULL;
        }

      if (pt_berin->DLY_REF != NULL)
        {
        freechain (pt_berin->DLY_REF);
        pt_berin->DLY_REF = NULL;
        }

      if (pt_berin->OUT_REF != NULL)
        {
        freechain (pt_berin->OUT_REF);
        pt_berin->OUT_REF = NULL;
        }

      if (pt_berin->BUS_REF != NULL)
        {
        freechain (pt_berin->BUS_REF);
        pt_berin->BUS_REF = NULL;
        }

      if (pt_berin->BUX_REF != NULL)
        {
        freechain (pt_berin->BUX_REF);
        pt_berin->BUX_REF = NULL;
        }

      if (pt_berin->REG_REF != NULL)
        {
        freechain (pt_berin->REG_REF);
        pt_berin->REG_REF = NULL;
        }

      if (pt_berin->MSG_REF != NULL)
        {
        freechain (pt_berin->MSG_REF);
        pt_berin->MSG_REF = NULL;
        }

      pt_berin = pt_berin->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process simple internal signals.				*/
	/*   check that the signal does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    pt_beaux = pt_fig->BEAUX;
    while (pt_beaux != NULL)
      {
      res_chain = prb_expdepend (pt_fig, PRB_SUPERABL(pt_beaux->NODE));
      while (res_chain != NULL)
        {
        pt_berin = (struct berin *) res_chain->DATA;

        if ((pt_berin->AUX_REF == NULL) || (pt_berin->AUX_REF->DATA != pt_beaux))
          pt_berin->AUX_REF = addchain (pt_berin->AUX_REF, pt_beaux);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_beaux = pt_beaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process delayed internal signals				*/
	/*     check that the signal does not already belong to the	*/
	/*     dependency list before adding it to the list		*/
	/* ###------------------------------------------------------### */

    pt_bedly = pt_fig->BEDLY;
    while (pt_bedly != NULL)
      {
      res_chain = prb_expdepend (pt_fig, PRB_SUPERABL(pt_bedly->NODE));
      while (res_chain != NULL)
        {
        pt_berin = (struct berin *) res_chain->DATA;

        if ((pt_berin->DLY_REF == NULL) || (pt_berin->DLY_REF->DATA != pt_bedly))
          pt_berin->DLY_REF = addchain (pt_berin->DLY_REF, pt_bedly);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_bedly = pt_bedly->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process simple ouputs					*/
	/*     check that the signal does not already belong to the	*/
	/*     dependency list before adding it to the list		*/
	/* ###------------------------------------------------------### */

    pt_beout = pt_fig->BEOUT;
    while (pt_beout != NULL)
      {
      res_chain = prb_expdepend (pt_fig, PRB_SUPERABL(pt_beout->NODE));
      while (res_chain != NULL)
        {
        pt_berin = (struct berin *) res_chain->DATA;

        if ((pt_berin->OUT_REF == NULL) || (pt_berin->OUT_REF->DATA != pt_beout))
          pt_berin->OUT_REF = addchain (pt_berin->OUT_REF, pt_beout);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_beout = pt_beout->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process bussed ouputs.					*/
	/*   check that the signal does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    res_chain = NULL;
    pt_bebus    = pt_fig->BEBUS;
    while (pt_bebus != NULL)
      {
      pt_binode = pt_bebus->BINODE;
      while (pt_binode != NULL)
        {
        res_chain = append (res_chain, prb_expdepend (pt_fig, PRB_SUPERABL(pt_binode->VALNODE)));
        res_chain = append (res_chain, prb_expdepend (pt_fig, PRB_SUPERABL(pt_binode->CNDNODE)));
        pt_binode  = pt_binode->NEXT;
        }

      while (res_chain != NULL)
        {
        pt_berin = (struct berin *) res_chain->DATA;

        if ((pt_berin->BUS_REF == NULL) || (pt_berin->BUS_REF->DATA != pt_bebus))
          pt_berin->BUS_REF = addchain (pt_berin->BUS_REF, pt_bebus);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_bebus = pt_bebus->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process bussed internal signals.				*/
	/*   check that the signal does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    res_chain = NULL;
    pt_bebux    = pt_fig->BEBUX;
    while (pt_bebux != NULL)
      {
      pt_binode = pt_bebux->BINODE;
      while (pt_binode != NULL)
        {
        res_chain = append (res_chain, prb_expdepend (pt_fig, PRB_SUPERABL(pt_binode->VALNODE)));
        res_chain = append (res_chain, prb_expdepend (pt_fig, PRB_SUPERABL(pt_binode->CNDNODE)));
        pt_binode  = pt_binode->NEXT;
        }

      while (res_chain != NULL)
        {
        pt_berin = (struct berin *) res_chain->DATA;

        if ((pt_berin->BUX_REF == NULL) || (pt_berin->BUX_REF->DATA != pt_bebux))
          pt_berin->BUX_REF = addchain (pt_berin->BUX_REF, pt_bebux);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_bebux = pt_bebux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process internal registers.				*/
	/*   check that the signal does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    res_chain = NULL;
    pt_bereg    = pt_fig->BEREG;
    while (pt_bereg != NULL)
      {
      pt_binode = pt_bereg->BINODE;
      while (pt_binode != NULL)
        {
        res_chain = append (res_chain, prb_expdepend (pt_fig, PRB_SUPERABL(pt_binode->VALNODE)));
        res_chain = append (res_chain, prb_expdepend (pt_fig, PRB_SUPERABL(pt_binode->CNDNODE)));
        pt_binode  = pt_binode->NEXT;
        }

      while (res_chain != NULL)
        {
        pt_berin = (struct berin *) res_chain->DATA;

        if ((pt_berin->REG_REF == NULL) || (pt_berin->REG_REF->DATA != pt_bereg))
          pt_berin->REG_REF = addchain (pt_berin->REG_REF, pt_bereg);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_bereg = pt_bereg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process assert statements.					*/
	/*   check that the statement does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    res_chain = NULL;
    pt_bemsg    = pt_fig->BEMSG;
    while (pt_bemsg != NULL)
      {
      res_chain = prb_expdepend (pt_fig, PRB_SUPERABL(pt_beaux->NODE));
      while (res_chain != NULL)
        {
        pt_berin = (struct berin *) res_chain->DATA;

        if ((pt_berin->MSG_REF == NULL) || (pt_berin->MSG_REF->DATA != pt_bemsg))
          pt_berin->MSG_REF = addchain (pt_berin->MSG_REF, pt_bemsg);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_bemsg = pt_bemsg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   release hash tables and lists that they contain		*/
	/* ###------------------------------------------------------### */

    if (pt_fig->BERIN != NULL)
      delht (rin_list);

    }
  }
