#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include MUT_H
#include MLO_H
#include MLU_H
#include RCN_H

#define ZAN 129929

lofig_list *load_densities();

static char buf[256];
void usage()
{
   fprintf(stderr, "Usage: sicc swing cout input output\n" );
   fprintf(stderr, "       swing is the power supply swing in Volt\n");
   fprintf(stderr, "       cout  is the outputs loading capacity in ff\n");
   fprintf(stderr, "       input is a netlist file\n");
   fprintf(stderr, "       output is a .nrj file\n");
   exit(1);
}

void compute_nrj(lf, file_name, swing, cout)
lofig_list *lf;
char       *file_name;
float       swing;
float       cout;
  {
  losig_list *si;
  losig_list *Signal;
  locon_list *co;
  float       capa;
  chain_list *ch;
  float       nrj;
  FILE       *output_file;

   if ((output_file = mbkfopen(file_name, "nrj", "w")) == NULL)
     {
     fprintf(stderr, "Cannot open %s.nrj file", file_name);
     exit(2);
     }

/*  lofigchain (lf); */

 for (co = lf->LOCON; co; co = co->NEXT)
     {
     if (co->DIRECTION != IN)
       {
   /* sig->CAPA is the routing capacity in pf  */
   /* cout is the output capacity in ff        */
         Signal = co->SIG;
         if ( Signal->PRCN != NULL )
         {
           Signal->PRCN->CAPA += 0.001 * cout;
         }
       }
     else
       {
   /* In the preceding version, no input power consumption was accounted. */
   /* Now, just keep sig->CAPA (the routing capacity in pf).              */
       }
     }

   /* Traversal:                                                       */
   /* for all signals, do the magical $\frac{1}{2} C v^2 computation.  */

   for (si = lf->LOSIG; si; si = si->NEXT)
     {

/*
     for(ch = (chain_list *)(getptype(si->USER, (long)LOFIGCHAIN)->DATA); ch; ch = ch->NEXT)
       {
       co = (locon_list *)(ch->DATA);
       if (co->TYPE == INTERNAL)
         {
         if (co->DIRECTION != IN)
           {
           if (co->DIRECTION != OUT)
             {
             fprintf(stderr,"WARNING : signal %s connected ", getsigname(si));
             fprintf(stderr,"to instance %s via a non output ", insname);
             fprintf(stderr," connector %s\n", co->NAME);
             }
           insname = ((loins_list *)(co->ROOT))->INSNAME;
           }
         }
       }
     if ( insname == NULL )
       insname = namealloc("external");
*/
     if ( si->PRCN != NULL )
     {
       capa = si->PRCN->CAPA;
     }
     else
     {
       fprintf( stdout, "ICI !\n" );
       capa = 0.0;
     }

     nrj = capa * swing * swing * 0.5;

     fprintf(output_file,
             "%-15s : %-6.4f 10e-12 J [ %-6.4f 10e-12 F]\n",
             getsigname(si), nrj, capa );
     }
  fclose(output_file);
  }

int main(argc, argv)
int argc;
char **argv;
{
lofig_list *lf;

   mbkenv();

  alliancebanner_with_authors( 
      "SiCC", "1.00", "SIgnal Consumption Computation", "1999",
      ALLIANCE_VERSION, "Julien Dunoyer" );

   if (argc != 5)
      usage(argv[0]);

   lf = getlofig(argv[3],'A');

   compute_nrj(lf, argv[4], atof(argv[1]), atof(argv[2]));
   exit(0);
}

