/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - CMOS Cells Compiler
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */




#ident "$Id: c4_debug.c,v 1.2 2000/11/07 16:15:16 czo Exp $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include BEH_H
#include MLO_H
#include LOG_H

#include "c4_local.h"

/******************************************************************************
 ***  Version 1.0b01                                            14/06/1991  ***
 ******************************************************************************

 ******************************************************************************
 ******************************************************************************
 ***                                                                        ***
 ***                            c.c.c.c tools                               ***
 ***                    Complex Cmos Cells Compiler                         ***
 ***                                                                        ***
 ******************************************************************************
 ******************************************************************************

 ******************************************************************************
 ***                            DEBUG MODULE                                ***
 *****************************************************************************/

/*::::::::::::::::::::::::: start of debug func. ::::::::::::::::::::::::::::: */

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*
  =-=-=-=- low-level func =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-[]=-
  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  SYNTAX : void c4_dbg_printInfo(MemFig *)
  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  EXEMPLE : c4_dbg_printInfo(cell)
  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  FONCTION : 
  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  ERROR : none
 *=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

void
c4_dbg_printInfo (cell)
     MemFig *cell;
{
  extern chain_list *expr2Model ();
  chain_list *ex;

  if (cell)
    {
      printf ("\n=======> Infos <=========================================\n\n");
      printf ("Exe. checkExpr:   %s\n", (C4_CHECKok (cell) ? "SUCCES" : "ECHEC"));
      printf ("Exe. placeInput:  %s\n", (C4_PLACEok (cell) ? "SUCCES" : "ECHEC"));
      printf ("Exe. createInt:   %s\n", (C4_CONNECTok (cell) ? "SUCCES" : "ECHEC"));
      printf ("Exe. rouTrans:    %s\n", (C4_VROUTok (cell) ? "SUCCES" : "ECHEC"));
      printf ("Compilation OK?:  %s\n", (C4_COMPok (cell) ? "OUI" : "NON"));
      printf ("Routage OK?:      %s\n", (C4_ROUTok (cell) ? "OUI" : "NON"));
      printf ("ABL d'entree:    ");
      displayExpr (C4_INEXPR (cell));
      if (cell->status.checkExprOK)
	{
	  printf ("\nABL de sortie:   ");
	  displayExpr (C4_TREXPR (cell));
	  printf ("\n");
	}
      else
	printf ("\nABL de sortie non disponible.\n");
      if (cell->status.placeInputsOK)
	{
	  printf ("Modele:           sortie =");
	  if (C4_POLARITY (cell) == 1)
	    displayExpr (ex = expr2Model (C4_TREXPR (cell)));
	  else
	    displayExpr (ex = notExpr (expr2Model (C4_TREXPR (cell))));
	  printf ("\n");
	  freeExpr (ex);
	}
      else
	printf ("Modele non disponible.\n");
      if (cell->status.placeInputsOK)
	{
	  printf ("Liste assoc.:    ");
	  ex = C4_ASSOCLST (cell);
	  while (ex)
	    {
	      printf ("%3d ", (int) ex->DATA);
	      ex = ex->NEXT;
	    }
	  printf ("\n");
	}
      else
	printf ("Liste assoc. non disponible.\n");
      printf ("Polarite:         ");
      if (C4_POLARITY (cell) == 1)
	printf ("POSITIVE\n");
      else if (C4_POLARITY (cell) == 0)
	printf ("NEGATIVE\n");
      else
	printf ("ni NEGATIVE, ni POSITIVE !\n");
      printf ("Tr N serie:       %d\n", C4_NBNTRAN (cell));
      printf ("Tr P serie:       %d\n", C4_NBPTRAN (cell));
      printf ("\n");
      printf ("Nb Entrees:       %d\n", C4_NBINP (cell));
      printf ("Nb Coupures:      %d\n", C4_NBCUT (cell));
      printf ("Nb Pitchs:        %d\n", C4_NBPITCH (cell));
      printf ("Nb Ch N voulus:   %d\n", C4_NBNCHAN0 (cell));
      printf ("Nb Ch P voulus:   %d\n", C4_NBPCHAN0 (cell));
      printf ("Nb Ch N obtenus:  %d\n", C4_NBNCHAN (cell));
      printf ("Nb Ch P obtenus:  %d\n", C4_NBPCHAN (cell));
    }
  else
    {
      printf ("\n   === Erreur ===\n");
      printf ("Impossible d'afficher les infos!\n");
    }
}



/*:::::::::::::::::::::::::: end of debug func. :::::::::::::::::::::::::::::: */
