/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident "$Id: sl_prepar_bib.c,v 1.2 2000/11/07 16:15:27 czo Exp $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include BEH_H
#include LOG_H

#include "lax_param.h"
#include "sl_local.h"


/*------------------------------------------------------------------
  IDENT : levelExpr
  ------------------------------------------------------------------
  FUNCTION : calcule la profondeur en ET et OU d'une expression.
  ------------------------------------------------------------------*/

int
levelExpr (expr)
     chain_list *expr;
{
  if (ATOM (expr))
    return (0);
  else
    {
      int max = 0, profCar;

      if (OPER (expr) == NOT)
	return levelExpr (CADR (expr));

      while ((expr = CDR (expr)))
	{
	  profCar = levelExpr (CAR (expr));
	  if (profCar > max)
	    max = profCar;
	}
      return (++max);
    }
}


/*-------------------------------------------------------------------------
calculTNTPExpr	: calcule le nbre de transistors N et P en serie 
		  dans une fonction positive. 
---------------------------------------------------------------------------
retour		: tp et tn par variable.
---------------------------------------------------------------------------*/
void
calculTPTNExpr (expr, tp, tn)
     chain_list *expr;
     int *tp, *tn;
{
  if (ATOM (expr))
    {
      *tn = 1;
      *tp = 1;
    }
  else
    {
      if (OPER (expr) == OR)
	{
	  int max = 0;
	  int som = 0;

	  while ((expr = CDR (expr)))
	    {
	      int auxn = 0, auxp = 0;

	      calculTPTNExpr (CAR (expr), &auxp, &auxn);
	      som += auxp;
	      if (auxn > max)
		max = auxn;
	    }
	  *tn = max;
	  *tp = som;
	  return;
	}
      if (OPER (expr) == AND)
	{
	  int som = 0;
	  int max = 0;

	  while ((expr = CDR (expr)))
	    {
	      int auxn = 0, auxp = 0;

	      calculTPTNExpr (CAR (expr), &auxp, &auxn);
	      som += auxn;
	      if (auxp > max)
		max = auxp;
	    }
	  *tn = som;
	  *tp = max;
	  return;
	}
      if (OPER (expr) == NOT)
	{
	  *tn = 1;
	  *tp = 1;
	  return;
	}
    }
}

/*-------------------------------------------------------------------------
polarityExpr	: calcule la  polarite d'une expression
---------------------------------------------------------------------------
retour		: un int = 0 pour negatif, 1 pour positif et 2 sinon
---------------------------------------------------------------------------*/
int
polarityExpr (expr)
     chain_list *expr;
{
  if (ATOM (expr))
    return 1;
  else
    {
      int polar;

      if (OPER (expr) == NOT)
	{
	  if (ATOM (CADR (expr)))
	    return 0;
	  else
	    {
	      printf ("polarityExpr : error - bad expression\n");
	      displayExpr (expr);
	      exit (-1);
	    }
	}
      if (OPER (expr) != AND && (OPER (expr) != OR))
	{
	  printf ("polarityExpr : error - bad operator\n");
	  exit (-1);
	}
      expr = CDR (expr);
      polar = polarityExpr (CAR (expr));
      if (polar == 2)
	return 2;
      while ((expr = CDR (expr)))
	{
	  int val = polarityExpr (CAR (expr));

	  if ((val == 2) || (val ^ polar))
	    return 2;
	}
      return polar;
    }
}

/*-------------------------------------------------------------------------
respectTrExpr : verifie qu'une expression respecte la contrainte 
	        d'empilement niveau par niveau.
---------------------------------------------------------------------------
retour		: un int = 0 pour KO, 1 pour OK
---------------------------------------------------------------------------*/
int
respectTrExpr (expr, tp, tn, optim)
     chain_list *expr;
     int tp, tn, optim;
{

/* 18/01/96 */
/*
 */
  if (optim == 0)
    return 1;
  else
    {
      /* optimisation en delai */
      if (ATOM (expr))
	return 1;
      if (OPER (expr) == NOT)
	{
	  if (ATOM (CADR (expr)))
	    return 1;
	  else
	    {
	      printf ("respectTrExpr : error - bad expression\n");
	      exit (-1);
	    }
	}
      if (OPER (expr) != AND && (OPER (expr) != OR))
	{
	  printf ("respectTrExpr : error - bad operator\n");
	  exit (-1);
	}
      if (OPER (expr) == OR)
	{
	  if (lengthExpr (expr) > tp)
	    return 0;
	}
      if (OPER (expr) == AND)
	{
	  if (lengthExpr (expr) > tn)
	    return 0;
	}
      while ((expr = CDR (expr)))
	if (!respectTrExpr (CAR (expr), tp - 1, tn - 1, optim))
	  return 0;
      return 1;
    }
}

/*-------------------------------------------------------------------------
empileOKExpr : verifie qu'une expression respecte la contrainte 
	        d'empilement.
---------------------------------------------------------------------------
retour		: un int = 0 pour KO, 1 pour OK
---------------------------------------------------------------------------*/
int
empileOKExpr (expr, tp, tn, polarite, optim)
     chain_list *expr;
     int tp, tn, polarite, optim;
{
  int tpint = 0, tnint = 0;
  chain_list *inter;
  int result;

  if (levelExpr (expr) > MAXLEVEL)
    return (0);

  inter = copyExpr (expr);
  flatArityExpr (inter);

  if (polarite == 1)
    {
      calculTPTNExpr (expr, &tpint, &tnint);

      if (tpint > tp || tnint > tn || !respectTrExpr (inter, tp, tn, optim))
	result = 0;
      else
	result = 1;
    }
  else if (polarite == 0)
    {
      calculTPTNExpr (expr, &tnint, &tpint);
      if (tpint > tp || tnint > tn || !respectTrExpr (inter, tn, tp, optim))
	result = 0;
      else
	result = 1;
    }
  freeExpr (inter);

/*
   printf("  Sn = %d, Sp = %d\n", tnint, tpint);
 */

  return result;
}

/*-------------------------------------------------------------------------
maxTH : calcule la valeur max d'une table de hachage. 
---------------------------------------------------------------------------
retour		: un int.
---------------------------------------------------------------------------*/
int
maxTH (pTable)
     pTH pTable;
{
  int i;
  pElemTH pEl;
  int max = 0;

  pEl = pTable->pElem;
  for (i = 0; i < pTable->length; i++)
    {
      if (pEl->value != EMPTYTH && pEl->value != DELETETH)
	{
	  if (pEl->value > max)
	    max = pEl->value;
	}
      pEl++;
    }
  return max;
}
/*-------------------------------------------------------------------------
delaisCarac : calcul le delais au dessus duquel le cone doit etre
              optimise en delais. 
---------------------------------------------------------------------------
retour		: un int.
---------------------------------------------------------------------------*/
int
delaisCarac (pTable, optim)
     pTH pTable;
     int optim;
{
  int max = maxTH (pTable);

  switch (optim)
    {
    case 0:
      return max;
    case 1:
      return (9 * max) / 10;
    case 2:
      return (8 * max) / 10;
    case 3:
      return (8 * max) / 10;
    case 4:
      return (7 * max) / 10;
    }

 fprintf(stderr,"## ERROR %s.%d## optim : '%d'\n",__FILE__,__LINE__, optim );
 exit(23);

}
