/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident "$Id: sl_systemp.c,v 1.3 2001/05/21 15:29:29 czo Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include MLO_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"



chain_list *compileVHDL ();

/*----------------------------------------------------------------------------
initializeParam : initialise les parametres de la synthese.
------------------------------------------------------------------------------
retour          : un pointeur sur une structure param.
------------------------------------------------------------------------------*/
param *
initializeParam (optim, aux, level, delayPI, intermediate, optimPO, numTransN, numTransP)
     int optim;
     int aux;
     int level;
     ptype_list *delayPI;
     chain_list *intermediate;
     chain_list *optimPO;
     int numTransN, numTransP;
{
  param *par;

  par = (param *) mbkalloc (sizeof (param));
  par->mode = optim;
  par->aux = aux;
  par->level = level;
  par->delayPI = delayPI;
  par->intermediate = intermediate;
  par->earlyPO = optimPO;
  par->numTransN = numTransN;
  par->numTransP = numTransP;

  par->maxSlopeTime = 1000;
  par->impedancePI = NULL;
  par->capaPO = NULL;
  par->capaPI = NULL;

  return (par);
}

/*----------------------------------------------------------------------------
initializeSLSystem:  alloue sys 
------------------------------------------------------------------------------
retour		: un pointeur sur une structure systeme. 
------------------------------------------------------------------------------*/

sl_system *
initializeSLSystem ()
{
  sl_system *sl;

  sl = (sl_system *) mbkalloc (sizeof (sl_system));

  sl->orderPI = NULL;
  sl->order = 0;
  sl->version = (char *) mbkalloc (4096);
  sl->fileVhdl = (char *) mbkalloc (4096);
  sl->fileOut = (char *) mbkalloc (4096);
  sl->file = (char *) mbkalloc (4096);
  sl->fileParam = (char *) mbkalloc (4096);
  sl->filePar = (char *) mbkalloc (4096);

  sl->file_lib = (char *) mbkalloc (4096);
  sl->cata_lib = (char *) mbkalloc (4096);
  sl->cata_save = (char *) mbkalloc (4096);
  sl->c4_lib = (char *) mbkalloc (4096);
  sl->in_lo = (char *) mbkalloc (4096);
  sl->out_lo = (char *) mbkalloc (4096);
  sl->out_ph = (char *) mbkalloc (4096);

  return (sl);
}


/* ================================================================================
   mappingSC : fonction de mapping SC
   ---------------------------------------------------------------------------------------------
   retour : une lofig
   =================================================================================== */

lofig_list *
mappingSC ()
{
  chain_list *fileLib;
  int lit, maxLit, litRed, litReg, profMax, delaisMax, operBin;
  char *sauve_cata_lib;
  lofig_list *fig;
  char *sauve_work_lib;

  printf ("Running Standard Cell Mapping\n");

  /* compilation de la bibliotheque */
  fileLib = (chain_list *) SLgetFile (SL_CATA_LIB, "vbe");
  if (!fileLib)
    {
      printf ("Mapping SC : Error\n--> There is no cell in directory '%s'\n", SL_CATA_LIB);
      exit (-1);
      return (NULL);
    }

  VHB_HEDFIG = NULL;

  coutAblBeh (SL_BEH, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
	      &litReg, 0, SL_DELAYPI);

  printf ("=============================  INITIAL COST  ==================================\n");


  printf ("Total number of literals    = %d\n", lit);
  printf ("Number of reduced literals  = %d\n", litRed);
  printf ("Number of latches           = %d\n", litReg);
  printf ("Maximum logical depth       = %d\n", profMax);
  printf ("Maximum delay               = %.3f\n", ((float) delaisMax) / 1000);


  printf ("===============================================================================\n");

  if (SL_TRACE)
    printf ("Removing the auxiliary signals...\n");

  mapCarExprBeh (SL_BEH, simplif10Expr);

  if (SL_AUX == 0)
    compactExprBeh (SL_BEH, SL_OPTIM, SL_DELAYPI, SL_OPTIMPO, 1, SL_GUARDEDAUX);

  mapCarExprBeh (SL_BEH, simplif10Expr);

  if (SL_TRACE)
    {
      coutAblBeh (SL_BEH, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
		  &litReg, 0, SL_DELAYPI);
      printf ("==============================  INITIAL COST  ==================================\n");

      printf ("Total number of literals    = %d\n", lit);
      printf ("Number of reduced literals  = %d\n", litRed);
      printf ("Number of latches           = %d\n", litReg);
      printf ("Maximum logical depth       = %d\n", profMax);
      printf ("Maximum delay               = %.3f\n", ((float) delaisMax) / 1000);

      printf ("===============================================================================\n");
    }

  if (SL_TRACE && SL_AUX == 0)
    {
      printf ("apres decompactage\n");
      displayBeh (SL_BEH, 1);
    }

  /* on sauvegarde et on positionne MBK_CATA_LIB a SL_CATA_LIB */

  sauve_cata_lib = (char *) mbkalloc (4096);
  sauve_work_lib = WORK_LIB;
  WORK_LIB = *CATA_LIB;

  if (!mbkgetenv ("MBK_CATA_LIB"))
    strcpy (sauve_cata_lib, ".");
  else
    sauve_cata_lib = (char *) mbkgetenv ("MBK_CATA_LIB");

  put_CATA_LIB ((char *) genNAME ("%s", SL_CATA_LIB));

  printf ("Compiling library '%s'\n", SL_CATA_LIB);

  while (fileLib)
    {
      if (SL_TRACE)
	printf ("compile %s\n", (char *)fileLib->DATA);
      if (compileVHDL (fileLib->DATA, NULL, 0, NULL, 1, 0, 1, 1, 0, renameSig, NULL) != NULL)
	{
	  removeAuxBeh (VHB_HEDFIG, NULL);
	  bddToAblBeh (VHB_HEDFIG, 0, 0);
	}
      fileLib = fileLib->NEXT;
    }

  WORK_LIB = sauve_work_lib;

  if (SL_TRACE)
    printf ("Init Mapping \n");

  mappSys = initMapping ();

  printf ("Generating Expert System ...\n");
  if (compileLibrary () == 0)
    return (NULL);

  mapper ();

  printf ("\n");

  /* Generation de la netlist en memoire */

  fig = genMBKBeh (SL_BEH, 1, 1, SL_FILEOUT);

  changeAblBeh (SL_BEH);

  endMapping ();

  if (SL_TRACE)
    printf ("End of mappingSC\n");
  else
    printf ("\n");

  /* on restore MBK_CATA_LIB */

  put_CATA_LIB (sauve_cata_lib);
/* mbkfree(sauve_cata_lib); */

  return fig;
}



/* ==========================================================================
   mappingC4    : fonction de mapping avec C4
   -----------------------------------------------------------------------------
   retour               : une lofig
   ============================================================================= */
lofig_list *
mappingC4 ()
{
  lofig_list *fig;
  int lit, maxLit, litRed, litReg, profMax, delaisMax, operBin;
  char *save_WORK_LIB;
  char *sauve_cata_lib;
  chain_list *lstvbe;
  FILE *fich;
  chain_list *filevbe;

/* ATTENTION : 27/01/95
   if (SL_BEH->TYPE & BEH_STABLE)
   {
   printf("Mapping C4 : Error \n");
   displayBehMap(SL_BEH, 1);
   return(NULL); 
   }
 */

  printf ("Running Mapping with C4\n");

  coutAblBeh (SL_BEH, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
	      &litReg, 0, SL_DELAYPI);

  printf ("=============================  INITIAL COST  ==================================\n");

  printf ("Total number of literals    = %d\n", lit);
  printf ("Number of reduced literals  = %d\n", litRed);
  printf ("Number of latches           = %d\n", litReg);
  printf ("Maximum logical depth       = %d\n", profMax);
  printf ("Maximum delay               = %.3f\n", ((float) delaisMax) / 1000);

  printf ("===============================================================================\n");
  printf ("Decomposing in feasible functions...\n");

  printf (".");
  fflush (stdout);

  if (SL_TRACE == 3)
    printf ("-------------------- simplif10Expr --------------------\n");
  mapCarExprBeh (SL_BEH, simplif10Expr);
  if (SL_TRACE == 3)
    {
      printf ("-------------------- FIN simplif10Expr --------------------\n");
      displayBeh (SL_BEH, 1);
    }


  if (SL_TRACE == 3)
    printf ("-------------------- devXorBeh --------------------\n");
  devXorBeh (SL_BEH, SL_OPTIM, SL_DELAYPI, SL_OPTIMPO);
  if (SL_TRACE == 3)
    {
      printf ("-------------------- FIN devXorBeh --------------------\n");
      displayBeh (SL_BEH, 1);
    }

  printf (".");
  fflush (stdout);

  if (SL_TRACE)
    printf ("\nRemoving the auxiliary signals...\n");

  mapCarExprBeh (SL_BEH, simplif10Expr);

  if (SL_TRACE == 3)
    printf ("-------------------- compactExprBeh --------------------\n");
/*
   printf("-------------------- SANS compactExprBeh --------------------\n");
 */
  if (SL_AUX == 0)
    compactExprBeh (SL_BEH, SL_OPTIM, SL_DELAYPI, SL_OPTIMPO, 1, SL_GUARDEDAUX);
  if (SL_TRACE == 3)
    {
      printf ("-------------------- FIN compactExprBeh --------------------\n");
      displayBeh (SL_BEH, 1);
    }

/*******************/
/*******************/

  mapCarExprBeh (SL_BEH, simplif10Expr);

  if (SL_TRACE == 2 && SL_AUX == 0)
    {
      printf ("apres decompactage\n");
      displayBeh (SL_BEH, 1);
    }

  printf (".");
  fflush (stdout);

  if (SL_TRACE == 3)
    printf ("-------------------- flatPolarityBeh --------------------\n");
  flatPolarityBeh (SL_BEH);
  if (SL_TRACE == 3)
    {
      printf ("-------------------- FIN flatPolarityBeh --------------------\n");
      displayBeh (SL_BEH, 1);
    }

  printf (".");
  fflush (stdout);

  if (SL_TRACE == 3)
    printf ("-------------------- decompBoolBeh --------------------\n");
  decompBoolBeh (SL_BEH, SL_OPTIM, SL_DELAYPI, SL_OPTIMPO, SL_NUMTRANSP, SL_NUMTRANSN);
  if (SL_TRACE == 3)
    {
      printf ("-------------------- FIN decompBoolBeh --------------------\n");
      displayBeh (SL_BEH, 1);
    }

  /* on appelle une deuxieme fois (les temps ont change) */
  printf (".");
  fflush (stdout);

  if (SL_TRACE == 3)
    printf ("-------------------- decompBoolBeh --------------------\n");
  decompBoolBeh (SL_BEH, SL_OPTIM, SL_DELAYPI, SL_OPTIMPO, SL_NUMTRANSP, SL_NUMTRANSN);
  if (SL_TRACE == 3)
    {
      printf ("-------------------- FIN decompBoolBeh --------------------\n");
      displayBeh (SL_BEH, 1);
    }

  printf (".");
  fflush (stdout);

  if (SL_TRACE == 3)
    printf ("-------------------- mapBeh --------------------\n");
  mapBeh (SL_BEH, SL_OPTIM, SL_DELAYPI, SL_OPTIMPO, SL_NUMTRANSP, SL_NUMTRANSN);
  if (SL_TRACE == 3)
    {
      printf ("-------------------- FIN mapBeh --------------------\n");
      displayBeh (SL_BEH, 1);
    }

  printf (".");
  fflush (stdout);

  if (SL_TRACE == 3)
    printf ("-------------------- postBeh --------------------\n");
  postBeh (SL_BEH, SL_AUX);
  if (SL_TRACE == 3)
    {
      printf ("-------------------- FIN postBeh --------------------\n");
      displayBeh (SL_BEH, 1);
    }

  printf (".");
  fflush (stdout);

  printf ("\n");

  /* ==== appel a C4 ==== */

  printf ("Generating CMOS Cells in '%s'...\n", SL_CCCC_LIB);

  /* on modifie WORK_LIB a SL_CCCC_LIB pour sauver les fichiers de C4 */

  save_WORK_LIB = WORK_LIB;

  WORK_LIB = (char *) genNAME ("%s", SL_CCCC_LIB);

  /* on fait un putenv() a cause du mbkenv() de GENLIB */

  putenv ((char *) genNAME ("MBK_WORK_LIB=%s", SL_CCCC_LIB));

/*
   {
   beout_list *o;
   beaux_list *a;
   int cptCell = 0;

   for(o = SL_BEH->BEOUT; o; o = o->NEXT)
   cptCell++;
   for(a = SL_BEH->BEAUX; a; a = a->NEXT)
   cptCell++;

   printf("NOMBRE DE CELLULES = %d\n", cptCell);
   exit(-1);
   }
 */

  c4MappBeh (SL_BEH, SL_TRACE, SL_DELAYPI);


  /* on remet WORK_LIB a sa valeur initiale */

  WORK_LIB = save_WORK_LIB;

  /* on sauvegarde et on positionne MBK_CATA_LIB a SL_CCCC_LIB */

  sauve_cata_lib = (char *) mbkalloc (4096);
  if (!mbkgetenv ("MBK_CATA_LIB"))
    strcpy (sauve_cata_lib, ".");
  else
    sauve_cata_lib = (char *) mbkgetenv ("MBK_CATA_LIB");
  put_CATA_LIB ((char *) genNAME ("%s", SL_CCCC_LIB));

  /* on cree un fichier CATAL dans SL_CCCC_LIB */

  filevbe = (chain_list *) SLgetFile (SL_CCCC_LIB, "vbe");

  /* system (genNAME ("rm %s/CATAL", SL_CCCC_LIB)); */
  /* voir incatalog */
  fich = fopen ((char *) genNAME ("%s/CATAL", SL_CCCC_LIB), "w");

  if (fich)
    {
      lstvbe = filevbe;
      while (lstvbe)
	{
	  char *nameCell = (char *) getIdent (lstvbe->DATA);

	  /* 16/02/95 = pour scr : tie   F */
	  if (!strcmp (nameCell, namealloc ("tie")))
	    fprintf (fich, "%s\tF\n", nameCell);

	  fprintf (fich, "%s\tC\n", nameCell);
	  lstvbe = lstvbe->NEXT;
	}
      fclose (fich);
    }
  else
    {
      printf ("mappingC4 : Error \n---> Can't open file 'CATAL'\n");
      return (NULL);
    }
  freechain (filevbe);

  fig = genMBKBeh (SL_BEH, 0, 0, SL_FILEOUT);

  changeAblBeh (SL_BEH);

  /* on restore MBK_CATA_LIB */

  put_CATA_LIB (sauve_cata_lib);
/* mbkfree(sauve_cata_lib); */
  return fig;
}
