/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include MUT_H
#include MLO_H
#include "environment.h"
#include "genscan.h"
#include "utile.h"

#define GATE 0
#define SIGNAL 1

locon_list *scan_in_sig    = NULL;/* scin connector */
locon_list *scan_out_sig   = NULL;/* scout connector */
locon_list *clock_sig      = NULL;/* clock connector */
locon_list *test_sig       = NULL;/* test connector */
locon_list *vdd_sig        = NULL;
locon_list *vss_sig        = NULL;
locon_list *bist_mode1_sig = NULL;

int test_mode   = 1;      /* test signal polarity  1= test_mode 0= fonc_mode */
int clock_mode   = 1;     /* clock signal polarity 0= falling edge 1= rising edge */
int scan_ff_mode = 1;     /* scan select in scan_dff cells */

char *scan_in   = NULL;
char *scan_out  = NULL;
char *clock_    = NULL;
char *test      = NULL;

int ff_only     = 1;      /* means netlist contains only non scannable flipflops */
int t_state     = 0;      /* means netlist contains tristate gates */


/* declaration of gate models needed by genbist */ 

char *SCAN_FF     = NULL; 
char *SCAN_FF_N   = NULL; 
 
char * D_FF       = NULL; 
char * D_FF_N     = NULL;    

char * LATCH_1    = NULL; 
char * LATCH_1_N  = NULL; 
char * LATCH_2    = NULL; 
char * LATCH_2_N  = NULL; 
char * LATCH_3    = NULL; 
char * LATCH_3_N  = NULL; 
char * LATCH_4    = NULL; 
char * LATCH_4_N   = NULL; 
char * LATCH_5    = NULL; 
char * LATCH_5_N  = NULL;
char *LATCH_6     = NULL; 
char * LATCH_6_N  = NULL;

char * TSTATE1    = NULL; 
char *TSTATE1_N   = NULL; 
char * TSTATE     = NULL; 

char * AND_GATE   = NULL;
char * NAND_GATE  = NULL;
char * NOR_GATE   = NULL; 
char * XOR_GATE   = NULL; 
char * MUX_GATE   = NULL; 
char * INV_GATE   = NULL; 

/* scan cell connectors names */

char * SCAN_FF_SCIN    = NULL; 
char * SCAN_FF_OUT     = NULL; 
char * SCAN_FF_IN      = NULL; 
char * SCAN_FF_SE      = NULL;
char * SCAN_FF_CLOCK   = NULL;

/* flip-flop cell connectors names */

char * FF_IN           = NULL;
char * FF_OUT          = NULL;
char * FF_CLOCK        = NULL;


/*******************************************************************************************************************/
/*                                                                                                                 */
/*       These fuctions get useful informations from environment variables                                          */
/*                                                                                                                 */
/*******************************************************************************************************************/
char *get_string(st,index)
char *st;
int index;
{
   char tmp[20];
   int j,i = 0;
   
   if (st == NULL) return (NULL);
   
   if (index!=0) 
   {
	   for (j=0;j<index;j++)
	   { 
		   if (st[i] == ':') i++;
		   while((st[i] != ':') && (st[i]!='\0'))
		   {
			   i++;
						   
		   }
		   		   
		}
   }
   
   if (st[i] == ':') i++;
   j=0;
   
   while ((st[i] != '\0') && (st[i] != ':')) 	
   {
	   tmp[j++]=st[i++];
    /*printf("%c\n",st[i]);*/
   }	  
	
   tmp[j] =  '\0';
   if (tmp == '\0') return(NULL);   
   return (namealloc(tmp));   
}

char * get_variable(variable,nbr,type)
char *variable;
int nbr,type;
{

  	char *st ;

	if (mbkgetenv(variable) == NULL) 
	{
        if (!strcmp(variable,"MUX")) return(NULL);
        else 
		{
			printf("error in variable %s !\n",variable);
			exit(1);
		}
	}
	
	st = namealloc(mbkgetenv(variable));
    
	if (!get_string(st,nbr) ) 
	{
		printf("variable %s not defined!\n",variable);
		exit(1);
	}
    
    if ((type == GATE) & (nbr == 0))
		if (getlofig(get_string(st,nbr),'A') == NULL) printf("Lofig %s not found\n",get_string(st,nbr));
	
    return(get_string(st,nbr));
	
}
   
void get_environment_variables()
{
   /* special signals */
	
 
    scan_in = get_variable("SCIN",0,SIGNAL);
 
    scan_out= get_variable("SCOUT",0,SIGNAL);
 
    test    =get_variable("TEST",0,SIGNAL);
   /* test_mode = atoi(get_variable("TEST",1,SIGNAL));*/
	
	clock_ = get_variable("CLOCK",0,SIGNAL);
    /*clock_mode = atoi(get_variable("CLOCK",1,SIGNAL));rising or falling edge */ 
    
/*  gates */

   
/*    NAND_GATE = get_variable("NAND",0,GATE);*/
	
/*    AND_GATE = get_variable("AND",0);*/
    
/*    NOR_GATE = get_variable("NOR",0);*/
	
/*    XOR_GATE = get_variable("XOR",0,GATE);*/
	
/*    MUX_GATE = get_variable("MUX",0,GATE);*/
	
/*    INV_GATE = get_variable("INV",0,GATE);*/
		

/* scannable flip-flop connectors names */


SCAN_FF    = get_variable("SCAN_DFF",0,GATE); 
SCAN_FF_SCIN    = get_variable("SCAN_DFF",2,GATE); 
 get_con_fig(getlofig(SCAN_FF,'A'),SCAN_FF_SCIN);
SCAN_FF_OUT     = get_variable("SCAN_DFF",5,GATE); 
 get_con_fig(getlofig(SCAN_FF,'A'),SCAN_FF_OUT);
SCAN_FF_IN      = get_variable("SCAN_DFF",1,GATE); 
 get_con_fig(getlofig(SCAN_FF,'A'),SCAN_FF_IN);
SCAN_FF_SE      = get_variable("SCAN_DFF",3,GATE);
 get_con_fig(getlofig(SCAN_FF,'A'),SCAN_FF_SE);
SCAN_FF_CLOCK   =get_variable("SCAN_DFF",4,GATE);
 get_con_fig(getlofig(SCAN_FF,'A'),SCAN_FF_CLOCK);


/* flip-flop cell connectors names */


D_FF           = get_variable("DFF",0,GATE);
FF_IN           = get_variable("DFF",1,GATE);
 get_con_fig(getlofig(D_FF,'A'),FF_IN);
FF_OUT          = get_variable("DFF",3,GATE);
 get_con_fig(getlofig(D_FF,'A'),FF_OUT);
FF_CLOCK        = get_variable("DFF",2,GATE);
 get_con_fig(getlofig(D_FF,'A'),FF_CLOCK);


}

/*******************************************************************************************************************/
/*                                                                                                                 */
/*          gets the poiters to the special connectors clock, test, scan in, scan out ,vdd and vss                 */
/*                                                                                                                 */
/*******************************************************************************************************************/          

void get_special_signals(fig)
lofig_list *fig;
{
	int scin_   = 0;
	int scout_  = 0;
	int test_   = 0;
	int Clock  = 0;
	
    locon_list *ptc;
	for (ptc = fig->LOCON ; ptc ; ptc=ptc->NEXT)
	{
		if (isvdd(ptc->NAME)) vdd_sig = ptc;
		if (isvss(ptc->NAME)) vss_sig = ptc;
        if (strncmp((char *)ptc->NAME, scan_in,strlen(scan_in))==0) 
		{ 
			scin_ = 1;
			scan_in = ptc->NAME;
		}
	if (ptc->NAME == namealloc(scan_out)) scout_ = 1;
	if (ptc->NAME == namealloc(test)) test_ = 1;
        if (ptc->NAME == namealloc(clock_)) Clock = 1;

    }
	if (!vdd_sig || !vss_sig)
	{
		printf(" no vdd or vss signals in the netlist ");
		exit(1);
	}
	
	if (scin_) scan_in_sig  = getlocon(fig,scan_in);/* SCIN connector */
	
	if (scout_) scan_out_sig = getlocon(fig,scan_out);/* SCOUT connector */
	
	if (Clock) clock_sig = getlocon(fig,clock_);/* CK connector */

	if (test_) test_sig     = getlocon(fig,test);/* TEST connector */
	
} 



/*******************************************************************************************************************/
/*                                                                                                                 */
/*   this function returns true if a connector is a special signal like ck, scin, scout and test                   */
/*                                                                                                                 */
/*******************************************************************************************************************/


int special_signal(ptc)
locon_list *ptc;
{   
    char *st =  namealloc(ptc->NAME);
	int scin_   = 0;
	int scout_  = 0;
	int test_   = 0;
	int Clock   = 0;
    int vdd_    =0 ;
	int vss_    =0;
	
	
	if (scan_in_sig)  scin_ =(scan_in_sig->NAME == st);
	if (scan_out_sig) scout_ = (scan_out_sig->NAME == st );
        if (clock_sig)    Clock = (clock_sig->NAME == st );
        if (test_sig)     test_ = (test_sig->NAME  ==st );
	if (vdd_sig)      vdd_  = (vdd_sig->NAME  == st );
	if (vss_sig)      vss_  = (vss_sig->NAME  == st );
	return(scin_ || scout_ || Clock || test_ || vdd_ || vss_); 
}


/*******************************************************************************************************************/
/*                                                                                                                 */
/*                               this function identifys the flip-flop instances                                   */
/*                                                                                                                 */
/*******************************************************************************************************************/

int is_tstate(inst)
loins_list *inst;
{
    char *name = inst->FIGNAME;
	
	return((namealloc(TSTATE1) == name) ||
           (namealloc(TSTATE1_N) == name) || 
           (namealloc(TSTATE) == name));
}

int is_mem(inst)
loins_list *inst;
{  char *model = inst->FIGNAME;
   int  non_scan_ff   = 0;
   int  latch         = 0;
   int  scan_ff       =0;
   
   non_scan_ff  =  ((namealloc(D_FF) == model) ||
                    (namealloc(D_FF_N) == model));
   
   scan_ff =  ((namealloc(SCAN_FF) == model) ||
			   (namealloc(SCAN_FF_N) == model) );
   
   latch = ((namealloc(LATCH_1) == model) ||
            (namealloc(LATCH_1_N) == model) ||
			(namealloc(LATCH_2) == model) ||
			(namealloc(LATCH_2_N) == model) ||
			(namealloc(LATCH_3) == model) ||
			(namealloc(LATCH_3_N) == model) ||
			(namealloc(LATCH_4) == model) ||
			(namealloc(LATCH_4_N) == model) ||
			(namealloc(LATCH_5) == model) ||
			(namealloc(LATCH_5_N) == model) ||
			(namealloc(LATCH_6) == model) ||
			(namealloc(LATCH_6_N) == model));
   
   
   if (latch )
   {
	   printf("This cell model %s is not allowed in this version\n",model);
	   exit(1);
   }

   if (scan_ff)
   {
	   ff_only = 0;
   }
   
   return(scan_ff || non_scan_ff);
   
}


/*******************************************************************************************************************/
/*                                                                                                                 */
/*    This function gives a pointer to the input connectors and output connectiors lists                           */
/*                                                                                                                 */
/*******************************************************************************************************************/



